#ifndef __MBUILTINCLASSES_H
#define __MBUILTINCLASSES_H

#include "mString.h"


// Classes which are built into the language operate directly via instructions which
// are declared herein.


//////////////////////////////////////////////////////////////////////////
// Integer
//////////////////////////////////////////////////////////////////////////
class ClassInteger{

  // this is where Integer instructions start
  static const short INTEGER_START = 0;
  
 public:


  // name of integer
  //static char* integerName() { return "Integer"; }

  // opcodes for Integer methods

  // create new integer with default value 0
  // l0.new
  static const short NEW = INTEGER_START + 0;

  // create new integer from literal in constant pool
  // l0.new lit_address
  static const short NEW_FROM_LITERAL = INTEGER_START + 1;

  // create new integer from integer local variable
  // l0.new l_i
  static const short NEW_COPY = INTEGER_START + 2;

  // create new integer from double local variable
  // l0.new l_d
  static const short NEW_FROM_DOUBLE = INTEGER_START + 3;

  // create new integer from float local variable
  // l0.new l_f
  static const short NEW_FROM_FLOAT = INTEGER_START + 4;

  // create new integer from short local variable
  // l0.new l_s
  static const short NEW_FROM_SHORT = INTEGER_START + 5;

  // create new integer from byte local variable
  // l0.new l_b
  static const short NEW_FROM_BYTE = INTEGER_START + 6;

  // create new integer from char local variable
  // l0.new l_c
  static const short NEW_FROM_CHAR = INTEGER_START + 7;

  // create new integer from string local variable
  // l0.new l_c
  static const short NEW_FROM_STRING = INTEGER_START + 8;


  // add another integer to this integer
  // l0.add l_i
  static const short ADD = INTEGER_START + 9;

  // sub another integer from this integer
  // l0.sub l_i
  static const short SUB = INTEGER_START + 10;

  // mul another integer with this integer
  // l0.mul l_i
  static const short MUL = INTEGER_START + 11;

  // divide this integer with another integer
  // l0.div l_i
  static const short DIV = INTEGER_START + 12;

  // modulo
  // l0.mod l_i
  static const short MOD = INTEGER_START + 13;

  // compares this integer with another one and sets boolean variable accordingly
  // greater
  // l0.ifgt l_i, l_b
  static const short IFGT = INTEGER_START + 14;

  // greater or equals
  // l0.ifgteq l_i, l_b
  static const short IFGTEQ = INTEGER_START + 15;

  // less
  // l0.iflw l_i, l_b
  static const short IFLS = INTEGER_START + 16;

  // less or equal
  // l0.iflweq l_i, l_b
  static const short IFLSEQ = INTEGER_START + 17;

  // equals
  // l0.ifeq l_i, l_b
  static const short IFEQ = INTEGER_START + 18;

  // not equals
  // l0.ifneq l_i, l_b
  static const short IFNEQ = INTEGER_START + 19;

  // sets value
  // l0.set l_i
  static const short SET = INTEGER_START + 20;


  // prints Integer value to stdout
  // l0.print l_i
  static const short PRINT = INTEGER_START + 21;


  // increments Integer value
  // l0.inc
  static const short INC = INTEGER_START + 22;

  // decrements Integer value
  // l0.dec
  static const short DEC = INTEGER_START + 23;

  // this is where Integer instructions end
  static const short INTEGER_END = DEC;

 

};


//////////////////////////////////////////////////////////////////////////
// Double
//////////////////////////////////////////////////////////////////////////
class ClassDouble{

  // this is where Double instructions start
  static const short DOUBLE_START = ClassInteger::INTEGER_END + 1;
  
 public:
  // opcodes for Double methods

  // create new double with default value 0.0
  // l0.new
  static const short NEW = DOUBLE_START + 0;

  // create new double from literal in constant pool
  // l0.new lit_address
  static const short NEW_FROM_LITERAL = DOUBLE_START + 1;

  // create new double from double local variable
  // l0.new l_i
  static const short NEW_COPY = DOUBLE_START + 2;

  // create new double from integer local variable
  // l0.new l_i
  static const short NEW_FROM_INTEGER = DOUBLE_START + 3;

  // create new double from float local variable
  // l0.new l_f
  static const short NEW_FROM_FLOAT = DOUBLE_START + 4;

  // create new double from short local variable
  // l0.new l_s
  static const short NEW_FROM_SHORT = DOUBLE_START + 5;

  // create new double from byte local variable
  // l0.new l_b
  static const short NEW_FROM_BYTE = DOUBLE_START + 6;

  // create new double from char variable
  // l0.new l_c
  static const short NEW_FROM_CHAR = DOUBLE_START + 7;

  // create new double from string variable
  // l0.new l_c
  static const short NEW_FROM_STRING = DOUBLE_START + 8;
  

  // add another double to this double
  // l0.add l_d
  static const short ADD = DOUBLE_START + 9;

  // sub another double from this double
  // l0.sub l_d
  static const short SUB = DOUBLE_START + 10;

  // mul another double with this double
  // l0.mul l_d
  static const short MUL = DOUBLE_START + 11;

  // divide this double with another double
  // l0.div l_d
  static const short DIV = DOUBLE_START + 12;

  // compares this double with another one and sets boolean variable accordingly
  // greater
  // l0.ifgt l_d, l_b
  static const short IFGT = DOUBLE_START + 13;

  // greater or equals
  // l0.ifgteq l_d, l_b
  static const short IFGTEQ = DOUBLE_START + 14;

  // less
  // l0.iflw l_d, l_b
  static const short IFLS = DOUBLE_START + 15;

  // less or equal
  // l0.iflweq l_d, l_b
  static const short IFLSEQ = DOUBLE_START + 16;

  // equals
  // l0.ifeq l_d, l_b
  static const short IFEQ = DOUBLE_START + 17;

  // not equals
  // l0.ifneq l_d, l_b
  static const short IFNEQ = DOUBLE_START + 18;

  // sets value
  // l0.set l_d
  static const short SET = DOUBLE_START + 19;

  // prints double value to stdout
  // l0.print l_d
  static const short PRINT = DOUBLE_START + 20;


  // increments Double value
  // l0.inc
  static const short INC = DOUBLE_START + 21;

  // decrements Double value
  // l0.dec
  static const short DEC = DOUBLE_START + 22;

  // this is where Double instructions end
  static const short DOUBLE_END = DEC;
};


//////////////////////////////////////////////////////////////////////////
// String
//////////////////////////////////////////////////////////////////////////
class ClassString{

  // this is where String instructions start
  static const short STRING_START = ClassDouble::DOUBLE_END + 1;
  
 public:
  // opcodes for String methods

  // create new String with default value ""
  // l0.new
  static const short NEW = STRING_START + 0;

  // create new string from literal in constant pool
  // l0.new lit_address, length
  static const short NEW_FROM_LITERAL = STRING_START + 1;

  // create new string from string local variable
  // l0.new l_i
  static const short NEW_COPY = STRING_START + 2;

  // create new string from double local variable
  // l0.new l_i
  static const short NEW_FROM_DOUBLE = STRING_START + 3;

  // create new string from integer local variable
  // l0.new l_i
  static const short NEW_FROM_INTEGER = STRING_START + 4;

  // create new string from float local variable
  // l0.new l_f
  static const short NEW_FROM_FLOAT = STRING_START + 5;

  // create new string from short local variable
  // l0.new l_s
  static const short NEW_FROM_SHORT = STRING_START + 6;

  // create new string from byte local variable
  // l0.new l_b
  static const short NEW_FROM_BYTE = STRING_START + 7;

  // create new string from char variable
  // l0.new l_c
  static const short NEW_FROM_CHAR = STRING_START + 8;

  // create new string from boolean variable
  // l0.new l_b
  static const short NEW_FROM_BOOLEAN = STRING_START + 9;
  
  // add another string to this string
  // l0.add l_d
  static const short ADD = STRING_START + 10;

  // sets value
  // l0.set l_s
  static const short SET = STRING_START + 11;
  

  // equals
  // l0.ifeq l_s, l_b
  static const short IFEQ = STRING_START + 12;

  // not equals
  // l0.ifneq l_s, l_b
  static const short IFNEQ = STRING_START + 13;


  // prints string value to stdout
  // l0.print l_d
  static const short PRINT = STRING_START + 14;

  // this is where String instructions end
  static const short STRING_END = PRINT;
};



//////////////////////////////////////////////////////////////////////////
// Boolean
//////////////////////////////////////////////////////////////////////////
class ClassBoolean{

  // this is where Boolean instructions start
  static const short BOOLEAN_START = ClassString::STRING_END + 1;
  
 public:
  // opcodes for Boolean methods

  // create new Boolean with default value true
  // l0.new
  static const short NEW = BOOLEAN_START + 0;

  // create new Boolean from literal in constant pool
  // l0.new lit_address
  static const short NEW_FROM_LITERAL = BOOLEAN_START + 1;

  // create new Boolean from Boolean local variable
  // l0.new l_i
  static const short NEW_COPY = BOOLEAN_START + 2;

  // create new string from double local variable
  // l0.new l_i
  //static const short NEW_FROM_DOUBLE = STRING_START + 3;

  // sets value
  // l0.set l_b
  static const short SET = BOOLEAN_START + 3;

  // performs and on this boolean and another one and sets result boolean accordingly
  // l0.and l_b, l_br
  static const short AND = BOOLEAN_START + 4;

  // performs or on this boolean and another one and sets result boolean accordingly
  // l0.or l_b, l_br
  static const short OR = BOOLEAN_START + 5;

  // performs not on this boolean and sets result boolean accordingly
  // l0.not l_br
  static const short NOT = BOOLEAN_START + 6;

  // equals
  // l0.ifeq l_b, l_b
  static const short IFEQ = BOOLEAN_START + 7;

  // not equals
  // l0.ifneq l_b, l_b
  static const short IFNEQ = BOOLEAN_START + 8;
  

  // prints boolean value to stdout
  // l0.print
  static const short PRINT = BOOLEAN_START + 9;

  // this is where Boolean instructions end
  static const short BOOLEAN_END = PRINT;
};



//////////////////////////////////////////////////////////////////////////
// Character
//////////////////////////////////////////////////////////////////////////
class ClassCharacter{

  // this is where Character instructions start
  static const short CHARACTER_START = ClassBoolean::BOOLEAN_END + 1;
  
 public:
  // opcodes for Character methods

  // create new Character with default value true
  // l0.new
  static const short NEW = CHARACTER_START + 0;

  // create new Boolean from literal in constant pool
  // l0.new lit_address
  static const short NEW_FROM_LITERAL = CHARACTER_START + 1;

  // create new Boolean from Boolean local variable
  // l0.new l_i
  static const short NEW_COPY = CHARACTER_START + 2;  

  // sets value
  // l0.set l_b
  static const short SET = CHARACTER_START + 3;

  
  

  // prints character value to stdout
  // l0.print
  static const short PRINT = CHARACTER_START + 4;

  // this is where Character instructions end
  static const short CHARACTER_END = PRINT;
};



//////////////////////////////////////////////////////////////////////////
// Object
//////////////////////////////////////////////////////////////////////////
class ClassObject{

  // this is where Object instructions start
  static const short OBJECT_START = ClassCharacter::CHARACTER_END + 1;
  
 public:
  // opcodes for Object methods

  // create new Object
  // l0.new
  static const short NEW = OBJECT_START + 0;

  // create new Object from Object local variable
  // l0.new l_o
  static const short NEW_COPY = OBJECT_START + 2;  
  

  // prints object value to stdout
  // l0.print
  static const short PRINT = OBJECT_START + 4;

  // returns instance of Class
  // l0.getclass
  static const short GETCLASS = OBJECT_START + 5;

  // this is where Object instructions end
  static const short OBJECT_END = GETCLASS;
};


//////////////////////////////////////////////////////////////////////////
// Core
//////////////////////////////////////////////////////////////////////////
class ClassCore{

  // this is where Core instructions start
  static const short CORE_START = ClassObject::OBJECT_END + 1;
  
 public:
  // opcodes for Core methods

  // terminates program
  // Core.terminate
  static const short TERMINATE = CORE_START + 0;

  // allocates new frame
  // Core.new_frame #items
  static const short NEW_FRAME = CORE_START + 1;

  // allocates new frame and copies N items into it
  // Core.new_frame #items, #copy
  static const short NEW_FRAME_COPY = CORE_START + 2;

  // frees frame
  // Core.free_frame #items  
  static const short FREE_FRAME = CORE_START + 3;

  // pushes local by expanding frame
  // Core.push l_src
  static const short PUSH = CORE_START + 4;

  // pushes stack address of local by expanding frame
  // Core.pushref l_src
  static const short PUSHREF = CORE_START + 5;

  // pops local from top of frame into dest local
  // Core.pop l_dest
  static const short POP = CORE_START + 6;

  // copies reference from one local variable to another
  // Core.copy l_dest, l_src  
  static const short COPY = CORE_START + 7;

  // copies reference from one local variable to another which is itself a reference
  // Core.copy l_dest, l_src  
  static const short COPYREF = CORE_START + 8;

  // dereferences local from stack address
  // Core.deref l_dest, l_src  (l_src is local containing stack address, l_dest is dest. for reference)  
  static const short DEREF = CORE_START + 9;

  // sets local variable reference to nil
  // Core.nilify l_dest  
  static const short NILIFY = CORE_START + 10;

  // creates new object
  // Core.new_object dest_ref, address of class data, index in vtable
  static const short NEW_OBJECT = CORE_START + 11;

  // returns from method
  // Core.return_1
  static const short RETURN_1 = CORE_START + 12;

  // returns local variable from method
  // Core.return_1 local
  static const short RETURN_1_LOCAL = CORE_START + 13;

  // returns from method and removes arguments
  // Core.return_2 #args
  static const short RETURN_2 = CORE_START + 14;

  // returns local variable from method and removes arguments
  // Core.return_2 local, #args
  static const short RETURN_2_LOCAL = CORE_START + 15;

  // invokes procedure method 
  // Core.invoke_1_p ref, index  (index in vtable, ref in local variable)  
  static const short INVOKE_1_P = CORE_START + 16;

  // invokes function method 
  // Core.invoke_1_f ref, index, result  (index in vtable, ref in local variable, result is dest. local)  
  static const short INVOKE_1_F = CORE_START + 17;

  // invokes super constructor
  // Core.invoke_s ref, index  (index in mtable, ref in local variable)  
  static const short INVOKE_S = CORE_START + 18;

  // skips n bytes if specified bit in flags is set 
  // Core.skip_if bit, #bytes  
  static const short SKIP_IFSET = CORE_START + 19;

  // skips n bytes if specified bit in flags is cleared 
  // Core.skip_nif bit, #bytes  
  static const short SKIP_IFCLEAR = CORE_START + 20;

  // sets method  without parent environment
  // Core.setmethod_1 l2, l3, index, address  (l2-object,l3-source self,index-method index, address-closure)
  static const short SETMETHOD_1 = CORE_START + 21;

  // sets method  with parent environment
  // Core.setmethod_2 l2, l3, nP, index, address  (l2-object,l3-source self, nP- number of parent vars,
  //                                               index-method index, address-closure)
  static const short SETMETHOD_2 = CORE_START + 22;

  // upcasts object
  // Core.upcast l2, n (l2-object to be casted, n-offset to which selfish is set
  static const short UPCAST = CORE_START + 23;

  // downcasts object
  // Core.downcast l2, off (l2-object to be casted, off-offset of object into which we're casting
  static const short DOWNCAST = CORE_START + 24;

  // compare if references are equal
  // Core.ifrefeq l1, l2, lb  (l1-first reference, l2-second reference, lb-boolean variable)
  static const short IFREFEQ = CORE_START + 25;

  // compare if references are not equal
  // Core.ifrefneq l1, l2, lb  (l1-first reference, l2-second reference, lb-boolean variable)
  static const short IFREFNEQ = CORE_START + 26;

  // load nil into local variable
  // Core.new_nil l1
  static const short NEW_NIL = CORE_START + 27;

  // invokes native procedure method 
  // Core.invoke_n lit_address, frame_size  (method name literal address, size of frame)  
  static const short INVOKE_N_P = CORE_START + 28;

  // invokes native function method 
  // Core.invoke_n lit_address, frame_size  (method name literal address, size of frame)  
  static const short INVOKE_N_F = CORE_START + 29;

  // invokes native procedure method directly
  // Core.invoke_n lit_address, frame_size  (method name literal address, size of frame)  
  static const short INVOKE_N_P_1 = CORE_START + 30;

  // invokes native function method directly
  // Core.invoke_n lit_address, frame_size  (method name literal address, size of frame)  
  static const short INVOKE_N_F_1 = CORE_START + 31;


  // creates anonymous method  without parent environment
  // Core.createmethod_1 l1, l2, nParams, address  (l2-actual self (owner), nParams-# of parameters,
  //                                           address-closure)
  static const short CREATEMETHOD_1 = CORE_START + 32;

  // creates anonymous method with parent environment
  // Core.createmethod_2 l1, l2, nParams, address, nP  (l2-actual self (owner), nParams-# of parameters, 
  //                                               address-closure, nP- number of 
  //                                               parent vars)
  static const short CREATEMETHOD_2 = CORE_START + 33;


  // pushes direct value by expanding frame
  // Core.pushd 1000
  static const short PUSHD = CORE_START + 34;

  // checks method type
  // Core.mcheck l1, nParams
  static const short MCHECK = CORE_START + 35;


  // this is where Core instructions end
  static const short CORE_END = MCHECK;



};



//////////////////////////////////////////////////////////////////////////
// Block
//////////////////////////////////////////////////////////////////////////
class ClassBlock{

  // this is where Block instructions start
  static const short BLOCK_START = ClassCore::CORE_END + 1;
  
 public:
  // opcodes for Block methods

  // creates new block
  // Block.new l5, closure_address, num_parent_locals
  static const short NEW = BLOCK_START + 0;

  // executes block
  // Block.exec
  static const short EXEC = BLOCK_START + 1;

  // exits from block
  // Block.exit
  static const short EXIT_1 = BLOCK_START + 2;

  // sets block exit status according to Boolean variable
  // sets to i_status1 if Boolean is true, to i_status2 otherwise
  // Block.stat l_b, i_status1, i_status2
  static const short STATUS = BLOCK_START + 3;

  // returns local from block
  // Block.return_1 l5, numtoclean
  static const short RETURN_1_LOCAL = BLOCK_START + 4;

  // returns nothing from block
  // Block.return_1 l5
  static const short RETURN_1 = BLOCK_START + 5;

  // returns true if this block is top level
  // Block.isTopLevel implemented in NATIVE
  //static const short ISTOPLEVEL = BLOCK_START + 6;

  // this is where Block instructions end
  static const short BLOCK_END = RETURN_1;

};



//////////////////////////////////////////////////////////////////////////
// While
//////////////////////////////////////////////////////////////////////////
class ClassWhile{

  // this is where While instructions start
  static const short WHILE_START = ClassBlock::BLOCK_END + 1;
  
 public:
  // opcodes for While methods

  // creates new while
  // While.new l5, l6 (l5-conditional block, l6-execution block)
  static const short NEW = WHILE_START + 0;

  // executes while
  // While.exec
  static const short EXEC = WHILE_START + 1;

  // exits from block
  // Block.exit
  //static const short EXIT_1 = BLOCK_START + 2;

  // returns local from block
  // Block.return_1 l5, numtoclean
  //static const short RETURN_1_LOCAL = BLOCK_START + 3;

   // returns nothing from block
  // Block.return_1 l5
  //static const short RETURN_1 = BLOCK_START + 4;

  // this is where While instructions end
  static const short WHILE_END = EXEC;


};


//////////////////////////////////////////////////////////////////////////
// DoWhile
//////////////////////////////////////////////////////////////////////////
class ClassDoWhile{

  // this is where While instructions start
  static const short DOWHILE_START = ClassWhile::WHILE_END + 1;
  
 public:
  // opcodes for DoWhile methods

  // creates new dowhile
  // DoWhile.new l5, l6 (l5-conditional block, l6-execution block)
  static const short NEW = DOWHILE_START + 0;

  // executes dowhile
  // DoWhile.exec
  static const short EXEC = DOWHILE_START + 1;

  // exits from block
  // Block.exit
  //static const short EXIT_1 = BLOCK_START + 2;

  // returns local from block
  // Block.return_1 l5, numtoclean
  //static const short RETURN_1_LOCAL = BLOCK_START + 3;

   // returns nothing from block
  // Block.return_1 l5
  //static const short RETURN_1 = BLOCK_START + 4;

  // this is where DoWhile instructions end
  static const short DOWHILE_END = EXEC;

};



//////////////////////////////////////////////////////////////////////////
// If
//////////////////////////////////////////////////////////////////////////
class ClassIf{

  // this is where If instructions start
  static const short IF_START = ClassDoWhile::DOWHILE_END + 1;
  
 public:
  // opcodes for If methods

  // creates new if (one condition/block only)
  // If.new l5, l6 (l5-conditional block, l6-execution block)
  static const short NEW = IF_START + 0;

  // creates new if
  // If.new i_immed (number of condition/block pairs)
  static const short NEW_1 = IF_START + 1;

  // adds condition/block pair
  // If.addpair l5, l6  (l5-condition block, l6-execution block)
  static const short ADDPAIR = IF_START + 2;

  // adds block only (condition is empty)
  // If.addblock l5  (l5-execution block)
  static const short ADDBLOCK = IF_START + 3;



  // executes if
  // If.exec
  static const short EXEC = IF_START + 4;

  // exits from block
  // Block.exit
  //static const short EXIT_1 = BLOCK_START + 2;

  // returns local from block
  // Block.return_1 l5, numtoclean
  //static const short RETURN_1_LOCAL = BLOCK_START + 3;

   // returns nothing from block
  // Block.return_1 l5
  //static const short RETURN_1 = BLOCK_START + 4;

  // this is where DoWhile instructions end
  static const short IF_END = EXEC;

};



//////////////////////////////////////////////////////////////////////////
// For
//////////////////////////////////////////////////////////////////////////
class ClassFor{

  // this is where For instructions start
  static const short FOR_START = ClassIf::IF_END + 1;
  
 public:
  // opcodes for For methods

  // creates new For
  // For.new l5, l6 (l5-initialisation block, l6-condition block)
  static const short NEW = FOR_START + 0;

  // appends next expression block and main execution block
  // For.addblocks l5, l6 (l5-next block, l6-execution block)
  static const short ADDBLOCKS = FOR_START + 1;

  // executes For
  // For.exec
  static const short EXEC = FOR_START + 2;

  // exits from block
  // Block.exit
  //static const short EXIT_1 = BLOCK_START + 2;

  // returns local from block
  // Block.return_1 l5, numtoclean
  //static const short RETURN_1_LOCAL = BLOCK_START + 3;

   // returns nothing from block
  // Block.return_1 l5
  //static const short RETURN_1 = BLOCK_START + 4;

  // this is where DoWhile instructions end
  static const short FOR_END = EXEC;

};


#endif
