#include "mClass.h"
#include "mLiteral.h"
#include "mBuiltinClasses.h"
#include "mClosure.h"
#include "mMethodClass.h"

using namespace std;

/*********************************************************************************************/
mClass::mClass()
{
 access = ACCESS_PUBLIC;
 symbolType = SYMBOL_CLASS;
 
 methodList = new mList<mMethod>();
 methodList->setAutoDelete( true );

 relocTable = 0;
 relocEntries = 0;
 parsed = false;
 isMethodType = false;

 // create method table
 //methodTable = new mMethodTable();

 // and add this class to it
 //methodTable->addClass( this );


 // create linkage table
 //linkTable = new mList<linkTableEntry>();
 //linkTable->setAutoDelete( true );


 superClasses = new mList<mClass>();
 superClasses->setAutoDelete( false );


 //objectTable = new mObjectTable();

}


/*********************************************************************************************/

mClass::mClass(mString *Name, short Access)
{
 name = Name;
 access = Access;
 symbolType = SYMBOL_CLASS;
 
 methodList = new mList<mMethod>();
 methodList->setAutoDelete( true );

 relocTable = 0;
 relocEntries = 0;
 parsed = false;
 isMethodType = false;


 // create method table
 methodTable = new mMethodTable();

 // and add this class to it
 methodTable->addClass( this );


 // create linkage table
 linkTable = new mList<linkTableEntry>();
 linkTable->setAutoDelete( true );


 superClasses = new mList<mClass>();
 superClasses->setAutoDelete( false );


 objectTable = new mObjectTable();

}


/*********************************************************************************************/


mClass::~mClass()
{
  
 //  !!!!!!!!!!environment will delete them!!!!!!!!
 delete methodList;
 delete methodTable;
 delete linkTable;
 delete objectTable;

 delete superClasses;
 
 if(relocTable) delete[] relocTable; 
}

/*********************************************************************************************/


bool mClass::addMethod(mMethod *m, mClass*& error)
{  

  // first set method's owner class - this is this class
  m->theClass = this;


  // check if this method is compatible with all superclasses
  for(mClass *c = superClasses->first(); c != 0; c = superClasses->next() ){

    mClass* clas = c->isCompatibleWith(m);
    if(clas != 0){

      error = clas;
      return false;
    }
  }

 
  addMethod(m);

  return true;
}


bool mClass::addMethod(mMethod *m)
{  

  // add it to class...  
  methodList->append( m );

  // first set method's owner class - this is this class
  m->theClass = this;

  // ...and to method table
  // check if this is builtin method of a primitive class
  // we don't add primitive methods to mtable because they don't need index if they're
  // invoked directly with instruction




  if(!m->isBuiltin())
    methodTable->addMethod( m );

  return true;
}




/*********************************************************************************************/
void mClass::addSuperClass(mClass* cl)
{
  superClasses->append( cl );

  if(!isMethodType)  // for method types we have no own methodtable

    // append method table of superclass to ours
    methodTable->addSuper( cl->methodTable );

}




/*********************************************************************************************/



void mClass::addVariable(mVariable *v)
{
 //variableList->append( v );
}


/*********************************************************************************************/


mString mClass::toStringFull()
{
 mString str("class ");
 str += *name;
 str += "\n";
 str += "- methods\n";

 for(mMethod *m =  methodList->first(); m != 0; m = methodList->next()){
     str += "--- ";
     str += m->toString();
     str += "\n";
 }

 str += "- table\n";
 str += methodTable->toString();

 
 
 return str; 
}


/*********************************************************************************************/


mString mClass::toString()
{
 mString str("class '");
 str += *name;
 str += "'";
 return str; 
}


/*********************************************************************************************/


mClass* mClass::isCompatibleWith(mMethod* meth)
{

  // first we check if there is a method in this class that is not supertype of meth
  // because overriding rules are simple, we only check methods with the same name and same number
  // of parameters
  for(mMethod *m =  methodList->first(); m != 0; m = methodList->next())
    if( meth->mayOverride(m) )

      // meth must be a subtype of m
      if(!meth->isSubtypeOf( m ))
	
	return this;
	     
    


  // check all superclasses
  for(mClass *c = superClasses->first(); c != 0; c = superClasses->next() ){

    mClass *clas = c->isCompatibleWith(meth);
    if(clas != 0) return clas;

  }


  return 0;

}





/*********************************************************************************************/

mMethod* mClass::getMethod(mString *name)
{

  bool global= true; bool ambig=false;   // ......... crap


  ambiguous = ambig = false;

  // first check this class
  mList<candidateMethod>* list = new mList<candidateMethod>();
  list->setAutoDelete( true );

  mMethod* result = 0;

  // get all appropriate methods
  if( getPossibleMethods( name, list, result, global ) ){


    // if there is nothing in the list then no method was found
    if( list->size() == 0 ){
      return 0;
    }
    
    return result;
    

  }else{
    // we have an ambiguity error
    ambiguous = ambig = true;
  }


  return 0;

  /*
  // we're looking for method without parameters
  for(mMethod *m =  methodList->first(); m != 0; m = methodList->next())
    if( *name == *m->name && m->parameterList->size() == 0 ) return m;


  // check superclasses
  for(mClass *c=superClasses->first(); c != 0; c = superClasses->next()){

    // tukaj je potreben enak pristop kot pri metodah s parametri, ker lahko ustrezno 
    // metodo najdemo po vec poteh PRI VECKRATNEM!!!

    mMethod *meth = c->getMethod(name);
    if(meth != 0) return meth;

  }

  return 0;*/
}


/*********************************************************************************************/

mMethod* mClass::getMethod(mString *name, mParseTree* args, bool global, bool& ambig)
{


  ambiguous = ambig = false;

  // first check this class
  mList<candidateMethod>* list = new mList<candidateMethod>();
  list->setAutoDelete( true );

  mMethod* result = 0;

  // get all appropriate methods
  if( getPossibleMethods( name, args, list, result, global ) ){


    // if there is nothing in the list then no method was found
    if( list->size() == 0 ){
      return 0;
    }
    
    
    return result;
    

  }else{
    // we have an ambiguity error
    ambiguous = ambig = true;
  }


  return 0;
}

/*********************************************************************************************/


bool mClass::methodDefinedHere(mString* name, mList<mParameter>* params)
{

  // create params if they don't exist
  bool paramsExist = params != 0;
  if(!paramsExist) params= new  mList<mParameter>();

  for(mMethod *m =  methodList->first(); m != 0; m = methodList->next()){


    // names and number of parameters must match
    if( *name != *m->name || (int)params->size() != m->numParameters() ) continue;

    // parameters too
    bool match=true;
    for( unsigned i=0; i < params->size(); i++ ){

      if( *params->at( i )->theClass->name != 
	   *m->parameterList->at( i )->theClass->name ){ match = false; break; }
    }

    if(match){ if(!paramsExist) delete params; return true; }

  }

  if(!paramsExist) delete params;
  return false;
}

/*********************************************************************************************/

bool mClass::getPossibleMethods(mString* name, mParseTree* args, mList<candidateMethod>* list, 
				mMethod*& res, bool global)
{

  int start = list->size();   // this is where methods for this class begin

  // find all methods in this class that meet the requirement
  for(mMethod *m =  methodList->first(); m != 0; m = methodList->next()){


    // if name or number of arguments differ then this is definitely not the method
    // we're looking for
    if( *name != *m->name || m->parameterList->size() != args->children->size() ) continue;

    bool match = true;  // assume we have a match
    int total_error = 0;

    // name and the number of args match so this is a potential candidate.
    // check if type of every parameter matches the type of actual argument
    for( unsigned i=0; i < m->parameterList->size(); i++ ){

      mClass* paramType = m->parameterList->at( i )->theClass;       // parameter type
      mClass* argType = args->children->at( i )->returnType;         // argument type
            
      // -1 is returned is types are disjunctive, 0 if types match exactly, and > 0 if
      // argType is subtype of paramType or the two types can be converted via primitive casts
      int dist = paramType->distance( argType );
      total_error += dist;
      // if at least one argument doesn't match then we have an error
      if( dist == -1 ){ match = false; break; }
    }

    // if all parameters matched then we add this method to the list
    if(match)      
      list->append( new candidateMethod( m, total_error ) );

  }

  // check whether there are two or more methods in the list with the same error value
  
  
  // find minimum error value
  int minError = 999999999;
  mMethod* meth = 0;
  for( unsigned i = start; i < list->size(); i++ )
    if( list->at(i)->error < minError ){
      minError = list->at(i)->error;
      meth = list->at(i)->method;
    }
    
  // if there is more than 1 method with this minimum value then we have ambiguity.
  int numBest=0;
  for( unsigned i = start; i < list->size(); i++ )
    if( list->at(i)->error == minError ) numBest++;
  
  if( numBest > 1 ) 

    return false;  // error

  else if( numBest == 1 ){
    // exactly 1 match means we're done
    res = meth;  // the found method
    return true;
  }
  
  // now check superclasses only if list is empty (we haven't found ANY appropriate method)

  
  if( list->size() - start == 0 && global){

    for(mClass *c=superClasses->first(); c != 0; c = superClasses->next()){

      if(!c->getPossibleMethods( name, args, list, res, global )) return false;

    }


  }
  // otherwise we have a method from this class


  return true;
}


/*********************************************************************************************/

bool mClass::getPossibleMethods(mString* name, mList<candidateMethod>* list, 
				mMethod*& res, bool global)
{

  int start = list->size();   // this is where methods for this class begin

  // find all methods in this class that meet the requirement
  for(mMethod *m =  methodList->first(); m != 0; m = methodList->next()){


    // if name or number of arguments differ then this is definitely not the method
    // we're looking for
    if( *name != *m->name || m->parameterList->size() > 0 ) continue;

    int total_error = 0;

    list->append( new candidateMethod( m, total_error ) );

  }

  // check whether there are two or more methods in the list with the same error value
  
  
  // find minimum error value
  int minError = 999999999;
  mMethod* meth = 0;
  for( unsigned i = start; i < list->size(); i++ )
    if( list->at(i)->error < minError ){
      minError = list->at(i)->error;
      meth = list->at(i)->method;
    }
    
  // if there is more than 1 method with this minimum value then we have ambiguity.
  int numBest=0;
  for( unsigned i = start; i < list->size(); i++ )
    if( list->at(i)->error == minError ) numBest++;
  
  if( numBest > 1 ) 

    return false;  // error

  else if( numBest == 1 ){
    // exactly 1 match means we're done
    res = meth;  // the found method
    return true;
  }
  
  // now check superclasses only if list is empty (we haven't found ANY appropriate method)

  
  if( list->size() - start == 0 && global){

    for(mClass *c=superClasses->first(); c != 0; c = superClasses->next()){

      if(!c->getPossibleMethods( name, list, res, global )) return false;

    }


  }
  // otherwise we have a method from this class


  return true;
}


/*********************************************************************************************/


mMethod* mClass::getDefaultConstructor()
{
  for(mMethod *m =  methodList->first(); m != 0; m = methodList->next())
     if( m->isDefaultConstructor() ) return m;

  return 0;  // no default constructor
}



/*********************************************************************************************/


mMethod* mClass::getDeclaredMethod(mString* name, mList<mParameter>* params)
{

  // create params if they don't exist
  bool paramsExist = params != 0;
  if(!paramsExist) params= new  mList<mParameter>();

  for(mMethod *m =  methodList->first(); m != 0; m = methodList->next()){


    // names and number of parameters must match
    if( *name != *m->name || (int)params->size() != m->numParameters() ) continue;

    // parameters too
    bool match=true;
    for( unsigned i=0; i < params->size(); i++ ){

      if( *params->at( i )->theClass->name != 
	   *m->parameterList->at( i )->theClass->name ){ match = false; break; }
    }

    if(match){ if(!paramsExist) delete params; return m; }

  }

  if(!paramsExist) delete params;
  return 0;
}


/*********************************************************************************************/

int mClass::distance(mClass* other)
{

  // 1st check if types are equal
  if( *name == *other->name ) return 0;


  // now check primitive types. primitive types can be converted among each other so mismatch here
  // is still valid. we set distance to 1
  if( isPrimitive() && other->isPrimitive() ) return 1;


  // now check if other is a subclass of this class and calculate its distance
  if(this->isSubclass(other) || other->isSubclass(this)) return 1;
  
  // otherwise the classes are not related
  return -1;
}

/*********************************************************************************************/

bool mClass::isSubclass(mClass* other)
{

  
    // check if this class is a special method type
    if(isMethodType){

      // we have method type which means that other class must either be Object or some other
      // method supertype

      // if other class is Method then this class is subclass of Method because
      // all special method types are subtypes of Method and Object
      if( *other->name == "Method" || *other->name == "Object" ) return true;

      // now check if other is also special method type, then we have two functional subtypes
      if( other->isMethodType ){

	// this method class must be subclass of other method class
	// TUKAJ
	//cout<<"**************** checking two functional types: "<<*name <<" with "<< *other->name<<endl;

      }
      


    }


  // check if other is in our superClasses list
  for(mClass* super = superClasses->first(); super != 0; super = superClasses->next()){

    if(super == other) return true;
    if(super->isSubclass(other)) return true;
    
  }

  return false;
}


/*********************************************************************************************/

bool mClass::isSubtype(mClass* other)
{

  // if classes are equal
  if(this == other) return true;

  // if subclass then it must be subtype
  if(isSubclass(other)) return true;

  
  // if Self, then both must be Selves
  if(*this->name == "Self" && *other->name == "Self") return true;


  return false;
}


/*********************************************************************************************/

bool mClass::isPrimitive()
{
  // the following are primitive types
  return *name== "Integer" || *name== "Double" || *name== "String" || *name== "Float"
      || *name== "Long" || *name== "Byte" || *name== "Character" || *name== "Short" || *name== "Object";
}


/*********************************************************************************************/

bool mClass::isPrimitiveButNoObject()
{
  return isPrimitive() && *name != "Object";
}

/*********************************************************************************************/

int mClass::getSize()
{
 return 0; //variableList->size() * 4;
}


/*********************************************************************************************/


void mClass::updateSelfSuperTypes()
{

  for(mClass* super = superClasses->first(); super != 0; super = superClasses->next()){

    for(mMethod *m =  super->methodList->first(); m != 0; m = super->methodList->next()){

      for(mParameter *p = m->parameterList->first(); p != 0; p = m->parameterList->next()){

	if(p->isSelfType) p->theClass = this;

      }

      // and return type as well
      if(m->returnsSelf)
	m->returnType = this;

    }
    
  }

}


void mClass::evaluate()
{
  
  int poolSize = 0;

  // calculate size of literal pool
  for(mSymbol *l =  literalList->first(); l != 0; l = literalList->next()){
    poolSize += ((mLiteral*)l)->length();

  }

  // class data starts right after constant pool and meta data
  thisClassAddress = poolSize + getMetaSize();

  

  // code of 1st method starts after literal pool, meta data, class info and method table
  // classinfo is 16 bytes
  currentCodeOffset = poolSize + getInfoSize() + methodTable->size() + getMetaSize();


  //cout<<"-------> methodtable size="<<methodTable->size()<<endl;

  

  ///////////////////////////////////////////////////////////////////////

  // evaluate every method environment
  
  for(mMethod *m =  methodList->first(); m != 0; m = methodList->next()){
      cout<<"evaluating "<<m->toString()<<":"<<endl;

      // only now can we set address of this method in method table
      methodTable->setMethodAddress( m , currentCodeOffset );


      // add types of parameters for this method to linktable
      // we do this here, because if a parameter is not used in method body, it will not be added
      // to linktable in evaluate(), but we still need it for meta data
      for(mParameter* p = m->parameterList->first(); p != 0; p = m->parameterList->next()){

	addLinkEntry( p->theClass, -1 );

      }

      // and do the same with return type
      if(m->isFunction()){
	addLinkEntry( m->returnType, -1 );

      }



      // method environments always have 1 child - the block (closure)
      mEnvironment *block = m->environment->children->at(0);
      

      // create frame for this method (copies all named locals into frame)
      FrameRecord frame(  block, m );

      // to 0
      updatingMethod = 0;
      
      // we must reserve space for prologue which will be inserted here AFTER the
      // code has been generated. this is to ensure consistency of relocation 
      // addresses which are set up during the code generation

      // 6 bytes for instruction Core.new_frame size, or 10 bytes for Core.new_frame_copy
      // change to static later
      currentCodeOffset += *m->name != "main" ? 10 : 6;
      

      // create stack for nested closures within this method
      nestedClosures = new mStack<mClosure>();

      closures = new mList<mClosure>();
      closures->setAutoDelete( true );
       
      
      evaluateNode( block->closure, &frame, block );

      mCode *code = block->code;   // generated code


      // length of prologue code
      int prologueLength = 0;


      // non-static class method
      // if( !m->isStatic() )
      if( *m->name != "main" ){ 
	

	// prepend method prologue (frame allocation)

	// we must provide space for parameters, 3 implicitly pushed values (stack_ptr, ret address, self)
	// and all local variables
	// if this is function, we must also provide one additional location for result destination
	int itemsToCopy = m->parameterList->size() + 3;


	if( m->isFunction() && !m->isConstructor() ) itemsToCopy++;  // for functions

	int frame_size = frame.calculateSize();
	

	mString frame_instruction="Core.new_frame_copy ";
	frame_instruction += frame_size;    // total frame size
	frame_instruction += ", ";
	frame_instruction += itemsToCopy;  // number of params + return addr. + self
      
	frame_instruction += '\n';

	
	prologueLength = code->prependInstruction( frame_instruction, ClassCore::NEW_FRAME_COPY, 
						   frame_size, itemsToCopy );


	// method epilogue (return with frame disposal) is taken care of in evaluation
	// at NODE_RETURN

	
      }else{

	// main method

	// prepend method prologue (frame allocation)
	int frame_size = frame.calculateSize();
	mString frame_instruction="Core.new_frame ";
	frame_instruction += frame_size;    // total frame size
	frame_instruction += '\n';

      
	prologueLength = code->prependInstruction( frame_instruction, ClassCore::NEW_FRAME, 
						   frame_size );


	// append method epilogue (frame disposal and return)
	frame_instruction="Core.free_frame ";
	frame_instruction += frame_size;
	frame_instruction += '\n';      
	
	currentCodeOffset += code->appendInstruction( frame_instruction, 
						      ClassCore::FREE_FRAME, 
						      frame_size );      

	// add termination instruction at the end of main
	mString term_instruction="Core.terminate\n";
	currentCodeOffset += code->appendInstruction( term_instruction, 
						      ClassCore::TERMINATE );
      }

      // remove nested closures
      delete nestedClosures;
            

      //cout<<"offset after method = "<<currentCodeOffset<<endl;

      // relocate all nested closures to address after this method (currentCodeOffset)
      for(mClosure *closure = closures->first(); closure != 0; closure = closures->next()){

	
	closure->updateRelocationTable( relocTable, currentCodeOffset );

	closure->updateLinkTable( currentCodeOffset );
	
	closure->fixParentCode( currentCodeOffset, prologueLength );
	
	currentCodeOffset += closure->getLength();
	
      }


      // remove them completely
      delete closures;

      cout<<"------------------- code ---------------------"<<endl;
      dumpClosures( block );
           
  }   //  for(mMethod *m =  methodList->first(); m != 0; m = methodList->next())



  // object table contains every superclass as well as this class. every class in object table
  // contains class_address, offset_in_vtable, and offset_in_mtable
  // in order for class_address to be properly setup at run-time we must include it in link table

  // object table starts after literal pool, meta data and class info
  int currentAddress = poolSize + getMetaSize() + getInfoSize();

  linkSuperClasses(this, this->methodTable, currentAddress );


  // now we calculate selfish offsets into virtual table for every class in this hierarchy
  // offset of vtable for first superclass is after class_info and object_table 
  methodTable->calculateVirtualTableOffsets( getInfoSize() + methodTable->objectTableSize() );
  
  // and the same for method table which starts after class_info, object_table AND virtual_table
  

  methodTable->calculateMethodTableOffsets( getInfoSize() + methodTable->objectTableSize() + 
					    methodTable->virtualTableEntries()*4 );
    


  /*
  cout<<"---------------- reloc table -----------------"<<endl;
  for(int i=0;i<relocEntries;i++){
    cout<<"code offset "<<relocTable[i]<<endl;
  }
  cout<<"----------------------------------------------"<<endl;
  */
  
}



// this is where we evaluate the parsetree and generate code
// in addition to that we also generate entries for relocation and link tables
void mClass::evaluateNode(mParseTree *tree, FrameRecord *frame, mEnvironment* env)
{
  mList<mParseTree> *children = tree->children;
  mSymbol* symbol = tree->symbol;  // symbol in this node
  mClass* returnType = tree->returnType;


  // add class of returnType to link table without address (-1). this is just to ensure that
  // this class gets linked at runtime

  // every type except Void...which is not actually a type
  if( returnType != 0 && *returnType->name != "Void")
    addLinkEntry( returnType, -1 );




  mCode& out = *env->code;

  mString code;  // textual code representation

  mString newline = "\n";

  int fvar = 0;  // free variable
  int leftr = 0,rightr = 0;  // result

  bool partial = false;   // partial code is part of an instruction
            
  mString temp1;
  mString result;

  mParseTree *first=0, *second=0;  // 1st & 2nd child are used extensively
  mParseTree *third=0;

  if( children->size() > 0 ) first = children->at(0);
  if( children->size() > 1 ) second = children->at(1);
  if( children->size() > 2 ) third = children->at(2);


  frame->constant = false;

  switch( tree->type ){


    ////////////////////////////////////////////////////////////////////
    // closure
    ////////////////////////////////////////////////////////////////////
    case NODE_CLOSURE: {              

      mEnvironment* closureEnv = env;  // closure environment
      FrameRecord* theFrame = frame;   // frame


      // we only create nested closures (those that are within method blocks)

      // nested closure always has parent
      mClosure *closure = 0;
      if( tree->parent ){
	
	//cout<<"++++++++++++ compiling nested closure "<<endl;

	// the following closure is evaluated in its own environment
	closureEnv = env->getChildOfClosure( tree );

	

	// this is where we store our closure
	int fvar = frame->getFreeVariable();
	//cout<<" ++++++++++++++++ got variable for closure - l"<<fvar<<endl;

	// push all variables from parent which are used in this closure
	int stackDelta = 0;

	// 1st is current self

	int currentSelf = env->getSelfLocation();


	// if this is static scope, self doesn't exist so we just set it to dummy local l1
	// eventually this should be changed so that no self will be passed at all

	if(currentSelf == -1) currentSelf = 1;


	code = "Core.push l"; code += currentSelf;
	mString dbg = "push current self";
	code = debug(code, dbg);
	code += '\n';
	currentCodeOffset += out.appendInstruction( code, ClassCore::PUSH,  currentSelf + stackDelta );
	stackDelta++;

	
	int nParentVariables = 1;   // at least 1 variable - self

	// update locations of parent variables in this closure
	closureEnv->updateParentVariables();	
	

	// for all from parent....
	int frameEntry=5;   // 1st parent var is at location 5
	for(mSymbol *s = closureEnv->parentSymbols->first(); s != 0; s = closureEnv->parentSymbols->next()){


	  if( s->used ){

	    int entry = 0;
	    if( s->symbolType == SYMBOL_VARIABLE ){ 

	      entry = ((mVariable*)s)->entry;

	      ((mVariable*)s)->entry = frameEntry;

	      //cout<<"pushing environment variable "<<*s->name<<" at loc "<<entry<<endl;

	    }else if( s->symbolType == SYMBOL_PARAMETER ){

	      entry = ((mParameter*)s)->entry;

	      //cout<<"pushing parameter "<<*s->name<<" at loc "<<entry<<endl;

	      ((mParameter*)s)->entry = frameEntry;
	    }

	   
 	    


	    // primitive values are pushed as references
	    if( s->theClass->isPrimitive() || !closureEnv->declaredInParent(s->name) ){
	      
	      code = "Core.push l"; code += (entry + stackDelta);
	      mString dbg = "push environment var ";  dbg += *s->name;
	      code = debug(code, dbg);
	      code += '\n';
	      currentCodeOffset += out.appendInstruction( code, ClassCore::PUSH, entry + stackDelta  );

	    }else{

	      // all other values are pushed as stack addresses

	      // if variable is declared in this environment then we push it as address
	      // however, if it was declared in one of the parent environments we push it as
	      // plain local because its address was already pushed

	      code = "Core.pushref l"; code += (entry + stackDelta);
	      mString dbg = "push environment var address ";  dbg += *s->name;
	      code = debug(code, dbg);
	      code += '\n';
	      currentCodeOffset += out.appendInstruction( code, ClassCore::PUSHREF, entry + stackDelta  );
	      
	    }

	    stackDelta++;
	    nParentVariables++;
	    frameEntry++;

	  }  // if(s->used)

	}   // for( parentSymbols....
       

	
	// now shift all variables
	closureEnv->shiftVariables( nParentVariables-1 );

	// make room for parent locals
	closureEnv->currentVariableEntry += nParentVariables-1;

	// generate code for closure instantiation
	code="l"; code+=(stackDelta + fvar); code+= ".new Block "; code+= "??, ";
	code += nParentVariables;
	dbg = "new block";

	code = debug(code, dbg);
	code +='\n';	

	// block will hold only its original self and parental variables. return address,stack_ptr
	// and block reference will be pushed when it gets executed!

	// closure address is at this point not known so we just set it to 0.
	// it will be updated after method has been compiled
	currentCodeOffset += out.appendInstruction( code, ClassBlock::NEW, stackDelta+fvar, 0, 
						    nParentVariables );	

	// relocate closure address	
	addRelocation( currentCodeOffset-8 );

	// we need 'direct' reference to closure , that is without literals, etc.
	int closureRef = env->code->binaryLength - 8;



	// if this is a standalone closure, we execute it immediately. otherwise we don't (i.e.
	// if block is assigned to a variable)
	if( tree->parent->type == NODE_CLOSURE ){

	  // direct parent is closure, so this is a standalone closure

	  // generate exec instruction
	  code="l"; code+=fvar; code+= ".exec";
	  dbg = "execute block";

	  code = debug(code, dbg);
	  code +='\n';
	  currentCodeOffset += out.appendInstruction( code, ClassBlock::EXEC,  fvar );

	  // variable no longer needed because block has already been executed
	  frame->free( fvar );

	  // closure may have returned something which we check here. bit 1 is set if return did
	  // occur within block and result is on top of stack. if block just returned (from a procedure)
	  // no value is passed on stack but bit will be set anyway

	  // we don't know the number of bytes to skip until we have generated actual return instruction
	  code="Core.skip_ifclear 0, ??";
	  dbg = "test if block returned";

	  code = debug(code, dbg);
	  code +='\n';
	  // make it 6 bytes
	  currentCodeOffset += out.appendInstruction( code, ClassCore::SKIP_IFCLEAR,  0, 6 );

	  // this is where number of bytes to skip is stored
	  int sizeOffset = out.binaryLength - 4;
	  

	  // now we return value that was returned by block. if this is method level we do 
	  // plain method return. if block is around us we do block.return

	  if( nestedClosures->size() == 0 || updatingMethod ){   // method level

	    int numParameters = frame->ownerMethod->numParameters();

	    // we may return from function or procedure. in functions we actually return
	    // some value, whereas in procedures nothing is returned

	    if( numParameters == 0 ){   // no parameters means no arguments cleanup
	
	      
	      if(frame->ownerMethod->isFunction()){  // return from function
		// return top of stack	
		code = "Core.return_1 l1"; 
		mString dbgmsg = "return block result on TOS";
		code = debug(code, dbgmsg);
		code+='\n';
		
		currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_1_LOCAL, 1 );
		// skip_ifclear will skip 6 bytes which is the length of this instruction


	      }else{   // return from procedure
		
		code = "Core.return_1"; 
		mString dbgmsg = "return block result (nothing)";
		code = debug(code, dbgmsg);
		code+='\n';
		
		currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_1 );
		// we should skip 2 bytes
		out.setInstructionValue( sizeOffset, 2 );

	      }

	    }else{   // return with arguments cleanup
	      
	      // otherwise we call return_2 which also cleans pushed arguments	


	      
	      if(frame->ownerMethod->isFunction()){   // return from function

		code = "Core.return_2 l1, "; code += numParameters;
		mString dbgmsg = "return block result on TOS and cleanup"; 
		code = debug(code, dbgmsg);
		code+='\n';
		currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_2_LOCAL, 1,
							    numParameters );
		// we should skip 10 bytes
		out.setInstructionValue( sizeOffset, 10 );

	      }else{   // return from procedure
		
		code = "Core.return_2, "; code += numParameters; 
		mString dbgmsg = "return block result (nothing) and cleanup";
		code = debug(code, dbgmsg);
		code+='\n';
		
		currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_2, numParameters );
		// we should skip 6 bytes which is already set!

	      }
	    }

	  }else{  // if( nestedClosures->size() == 0 )

	    // still within closure (this is nested block). we pass return result to our parent via 
	    // Block.return

	    int envSize = nestedClosures->top()->nParentVariables;

	    if(frame->ownerMethod->isFunction()){   // return from function

	      // return local variable from block
	      code = "Block.return l1"; 
	      code +=", "; code += envSize;
	      mString dbgmsg = "return block result on TOS and cleanup (in block)"; 
	      code = debug(code, dbgmsg);
	      code +='\n';
	      currentCodeOffset += out.appendInstruction( code, ClassBlock::RETURN_1_LOCAL, 1,
							  envSize );
	      // we should skip 10 bytes
	      out.setInstructionValue( sizeOffset, 10 );

	    }else{   // return from procedure
		
	      code = "Block.return "; code += envSize; 
	      mString dbgmsg = "return block result (nothing) and cleanup (in block)";
	      code = debug(code, dbgmsg);
	      code+='\n';
	      
	      currentCodeOffset += out.appendInstruction( code, ClassBlock::RETURN_1, envSize );
	      // we should skip 6 bytes which is already set!

	    }
	    
	  }
	  
	
	}else{  // if( tree->parent->type == NODE_CLOSURE )
	  
	  // closure must be preserved because it will be stored rather than executed
	  //cout<<"saving block reference for later"<<endl;
	  frame->result = fvar;

	 
	}

	// closure will store current code offset and relocation entry for closure address in
	// ClassBlock::NEW instruction
	closure = new mClosure( currentCodeOffset, closureRef, env->code );

	// number of parent variables without self
	closure->nParentVariables = nParentVariables - 1 ;


	// create frame for this closure. owner method remains the same
	FrameRecord* closureFrame = new FrameRecord( closureEnv, frame->ownerMethod );

	
	//cout<<"closureEnv->currentVariableEntry="<<closureEnv->currentVariableEntry<<endl;


	currentCodeOffset += 10;   // reserve for new_frame_copy instruction
	
	theFrame = closureFrame;


	nestedClosures->push( closure );   // all relocation and link entries will be recorded in this closure

	closures->append( closure );  // also add it to list	
		
      }
            

      // create code for this closure
      mCode *code = new mCode();
      closureEnv->code = code;

      partial = true;

      // evaluate all children (statements) of this closure
      for(mParseTree *c = children->first(); c != 0; c = children->next()){
	evaluateNode( c, theFrame, closureEnv );
      }
      
      
      if( tree->parent ){

	//////////////////////////////////////
	// remove pushed closure from stack
	//////////////////////////////////////
	
	mClosure* cl = nestedClosures->top();

	// calculate size
	cl->calculateSize( currentCodeOffset );
	


	// remove closure and add it to list of closures
	nestedClosures->pop();


	//////////////////////////////////////////////////
	// now add block prologue and epilogue
	//////////////////////////////////////////////////
	mCode* code = closureEnv->code;


	// we must provide space for  3 implicitly pushed values (stack_ptr, ret address, self),
	// all variables from parent closures and all local variables
	int itemsToCopy = 3;  // at least 3 values will always be copied (stack, ret, self)

	//if( m->returnType != 0 && !m->isConstructor() ) itemsToCopy++;  // for functions

	itemsToCopy += cl->nParentVariables;

	int frame_size = theFrame->calculateSize();
	//frame_size += cl->nParentVariables;

	mString frame_instruction="Core.new_frame ";
	frame_instruction += frame_size;    // total frame size
	frame_instruction += ", ";
	frame_instruction += itemsToCopy;  // number of pushed variables + stack + return addr. + self
      
	frame_instruction += '\n';

	
	// closure frame no longer needed
	delete theFrame;

	code->prependInstruction( frame_instruction, ClassCore::NEW_FRAME_COPY, 
				  frame_size, itemsToCopy );

	

	

	// restore old offset
	currentCodeOffset = cl->codeOffset;
	
	// restore old frame
	//frame = cl->parentFrame;
	

      }


      break;
    }
    



    ////////////////////////////////////////////////////////////////////
    // new object
    ////////////////////////////////////////////////////////////////////
    case NODE_NEW: {  // new object
                      // object may be created without initialisation in
                      // which case there is no right-hand side or with
                      // explicit initialisation which is represented by
                      // the right side of the expression

      // RIGHT HAND SIDE - the expression

      // this is used for primitive types
      mString fromType = "";


      // if initialiser is present we must evaluate it first
      if( second != 0 ){

	// if this is literal we don't need to load it because all builtin
        // types can be constructed directly from constant pool literals
        if( second->type == NODE_LITERAL ){  // only literal
	  
          // result is simply constant pool offset of this literal
          frame->result = ((mLiteral*)second->symbol)->offset;

	  frame->constant = true;
	  
	// explicit construction
	}else if( second->type == NODE_CONSTRUCTION ){
	  // new object with explicit construction: new someType(...)
	  // is a special case because we don't actually create
	  // object here (which is the case for primitive types with
	  // implicit construction), but in evaluation of right hand side,
	  // where we also store reference to this variable

	  // set destination variable where reference should be stored
	  //frame->dest_variable = ((mVariable*)first->symbol)->entry;
	  evaluateNode( second, frame, /*out,*/ env );

        }else{

	  // otherwise we have a plain expression, which must be evaluated and
	  // its result used to create an object


	  // OPTIMISATION: for primitive types check if second is a primitive cast node and
	  // ignore it. evaluate just second->first instead because all primitive types can be
	  // constructed directly from all other primitive types
	  if( second->type == NODE_PRIMITIVE_CAST ){

	    evaluateNode( second->children->at(0), frame, /*out,*/ env );
	    fromType = *second->children->at(0)->returnType->name;  

	  }else{

	    evaluateNode( second, frame, /*out,*/ env );
	    fromType = *second->returnType->name;  	    

	  }
	    // frame->result holds the location of this result
	}
      }  // if( second != 0 )


      // LEFT HAND SIDE is the variable where we store the reference of a
      // newly created object

      // new variable
      if( first->type == NODE_VARIABLE ){	

        mVariable *v = (mVariable*)first->symbol; 

	int varEntry =  v->entry;  // location of our variable

        code += "l"; code += varEntry; code += ".new ";


	// now we check for builtin types. if right side (initialiser) is present
	// we construct new object with its value, otherwise we construct new
	// object with default value

	////////////////////////////////////////////////////
	// Integer 
	////////////////////////////////////////////////////
        if( *v->theClass->name == "Integer" ){
	   code += "Integer ";

	   if(!second){

	     // Integer object with default value
	     code = debug(code, "new Integer object with default value");
	     currentCodeOffset += out.appendInstruction( code, ClassInteger::NEW,
							 varEntry );

	   }else{

	     // Integer object with specified value
	     code += frame->strResult();  // result can be const or local var
	   	   
	     if( frame->constant ){  // constructing from literal
	       mString dbg = "new Integer object from literal ";
	       dbg += *second->symbol->name;
	       code = debug(code, dbg);
	       currentCodeOffset += out.appendInstruction( code, 
				    ClassInteger::NEW_FROM_LITERAL, varEntry,
				    frame->result );
	       // relocate literal
	       addRelocation( currentCodeOffset-4 );

	     }else{                   // constructing from local variable

	       code = debug(code, "new Integer object from local");
	       short instruction = -1;
	       if( fromType == "Integer" ) instruction = ClassInteger::NEW_COPY;
	       else if( fromType == "String" ) instruction = ClassInteger::NEW_FROM_STRING;
	       else if( fromType == "Double" ) instruction = ClassInteger::NEW_FROM_DOUBLE;
          	       
	       currentCodeOffset += out.appendInstruction( code, instruction, varEntry, frame->result );

	       frame->free( frame->result );

	     }

	   } // !second

	////////////////////////////////////////////////////   
	// Double
	////////////////////////////////////////////////////
	}else if( *v->theClass->name == "Double" ){         
	   code += "Double ";

	   if(!second){

	     // Double object with default value
	     code = debug(code, "new Double object with default value");
	     currentCodeOffset += out.appendInstruction( code, ClassDouble::NEW,
							 varEntry );	     

	   }else{

	     // Double object with specified value
	     code += frame->strResult();  // result can be const or local var
	   
	     if( frame->constant ){  // constructing from literal
	       mString dbg = "new Double object from literal ";
	       dbg += *second->symbol->name;
	       code = debug(code, dbg);
	       currentCodeOffset += out.appendInstruction( code, 
			      ClassDouble::NEW_FROM_LITERAL, varEntry, frame->result );
	       // relocate literal
	       addRelocation( currentCodeOffset-4 );
	   
	     }else{                   // constructing from local variable
	       code = debug(code, "new Double object from local");
	       short instruction = -1;
	       if( fromType == "Integer" ) instruction = ClassDouble::NEW_FROM_INTEGER;
	       else if( fromType == "String" ) instruction = ClassDouble::NEW_FROM_STRING;
	       else if( fromType == "Double" ) instruction = ClassDouble::NEW_COPY;
          
	       currentCodeOffset += out.appendInstruction( code, instruction, varEntry, frame->result );
	       frame->free( frame->result );
	     }
	   } // !second

	////////////////////////////////////////////////////
	// String
	////////////////////////////////////////////////////
	}else if( *v->theClass->name == "String" ){         
	   code += "String ";
	   
	   if(!second){

	     // String object with default value
	     code = debug(code, "new String object with default value");
	     currentCodeOffset += out.appendInstruction( code, ClassString::NEW,
							 varEntry );

	   }else{

	     // String object with specified value
	     code += frame->strResult();  // result can be const or local var
	   	     
	     if( frame->constant ){  // constructing from literal
	       mString dbg = "new String object from literal ";
	       dbg += *second->symbol->name;
	       code = debug(code, dbg);	       
	       // get literal length from right side of parent (this is only needed for string literals)
	       int litLength = ((mLiteral*)second->symbol)->length();
	       currentCodeOffset += out.appendInstruction( code,
			      ClassString::NEW_FROM_LITERAL, varEntry, frame->result, litLength );
	       // relocate literal
	       addRelocation( currentCodeOffset-8 );
	   
	     }else{                   // constructing from local variable
	       code = debug(code, "new String object from local");
	       short instruction = -1;
	       if( fromType == "Integer" ) instruction = ClassString::NEW_FROM_INTEGER;
	       else if( fromType == "Double" ) instruction = ClassString::NEW_FROM_DOUBLE;
	       else if( fromType == "String" ) instruction = ClassString::NEW_COPY;
	       else if( fromType == "Boolean" ) instruction = ClassString::NEW_FROM_BOOLEAN;
          
	       currentCodeOffset += out.appendInstruction( code, instruction, varEntry, frame->result );
	       frame->free( frame->result );
	     }
	   } // !second

	   ////////////////////////////////////////////////////   
	   // Boolean
	   ////////////////////////////////////////////////////
	}else if( *v->theClass->name == "Boolean" ){         
	   code += "Boolean ";

	   if(!second){

	     // Boolean object with default value
	     code = debug(code, "new Boolean object with default value");
	     currentCodeOffset += out.appendInstruction( code, ClassBoolean::NEW,
							 varEntry );	     

	   }else{

	     // Boolean object with specified value
	     code += frame->strResult();  // result can be const or local var
	   
	     if( frame->constant ){  // constructing from literal
	       mString dbg = "new Boolean object from literal ";
	       dbg += *second->symbol->name;
	       code = debug(code, dbg);
	       currentCodeOffset += out.appendInstruction( code, 
			      ClassBoolean::NEW_FROM_LITERAL, varEntry, frame->result );
	       // relocate literal
	       addRelocation( currentCodeOffset-4 );
	   
	     }else{                   // constructing from local variable
	       code = debug(code, "new Boolean object from local");
	       // booleans can be constructed only from booleans
	       short instruction = ClassBoolean::NEW_COPY;
          
	       currentCodeOffset += out.appendInstruction( code, instruction, varEntry, frame->result );
	       frame->free( frame->result );
	     }
	   } // !second

           ////////////////////////////////////////////////////   
	   // Object
	   ////////////////////////////////////////////////////
	}else if( *v->theClass->name == "Object" ){         
	   code += "Object ";

	   if(!second){

	     // just Object
	     code = debug(code, "new Objectsa");
	     currentCodeOffset += out.appendInstruction( code, ClassObject::NEW,
							 varEntry );	     

	   }else{

	     // Object with specified value. because Object cannot be constructed with 
	     // any particular value it is always casted into which means we're not actually creating
	     // an object but assigning to it
	     code = "l"; code += varEntry; code += ".copy l"; code += frame->result;
	   	     
	     code = debug(code, "assigning to Object");
	     short instruction = ClassCore::COPY;
          
	     currentCodeOffset += out.appendInstruction( code, instruction, varEntry, frame->result );
	     frame->free( frame->result );
	     
	   } // !second
	
	}else{   // end of any builtin type

	////////////////////////////////////////////////////
	// otherwise we have a user type
	////////////////////////////////////////////////////


	  // if no initialiser is present we set reference in local variable to nil
	  if(!second){

	    // reference with value nil
	    code = "l"; code += varEntry; code += ".nilify";
	    code = debug(code, "new nil reference");	    
	    currentCodeOffset += out.appendInstruction( code, ClassCore::NILIFY,
							varEntry );
	  }else{

	    // if right side is present it can be a literal, new expression or ordinary 
	    // expression

	    // if it's literal it can only be nil, so we do the same as above:
	    // set reference to nil	    
	    if( second->type == NODE_LITERAL ){

	      // reference with value nil
	      code = "l"; code += varEntry; code += ".nilify";
	      code = debug(code, "new nil reference");	    
	      currentCodeOffset += out.appendInstruction( code, ClassCore::NILIFY,
							  varEntry );

	      // construction with 'new' expression
	    }else if( second->type == NODE_CONSTRUCTION ){

	      // right side has already been evaluated (and reference stored) above, 
	      // so we don't do anything here
	     
	      // ordinary expression means we just copy reference from frame->result
	      // to this variable
	    }else{

	      code = "l"; code += varEntry; code += ".copy l"; code += frame->result;
	      code = debug(code, "copy reference");	    
	      currentCodeOffset += out.appendInstruction( code, ClassCore::COPY,
							  varEntry, frame->result );

	      frame->free( frame->result );

	    }
	  }  // end of right side present

	}  // end of user type

      }  // if( first->type == NODE_VARIABLE )
    }
      break;


    ////////////////////////////////////////////////////////////////////
    // literal access
    ////////////////////////////////////////////////////////////////////
    case NODE_LITERAL: {

      // literal is loaded from constant pool into a local variable

      // first we check if literal with this name has already been loaded so
      // that we may save us a new object
      
      mLiteral *l = (mLiteral*)symbol;
      receiver = l->theClass;

      FrameVariable* lvar = frame->findLiteral( l );



      // if this literal will be passed as argument we create a copy of it, because it
      // may be modified within method. we must keep it as it is!

      if( lvar && !evaluatingArguments){  // only permanent literal can already be loaded
	
	// we simply set location
	frame->result = lvar->location;
	partial = true;

      }else{


	// otherwise we must load it into frame
	
	// 1st we find a free variable
	fvar = frame->getFreeVariable();
	code += 'l'; code += fvar; code += ".";  // location
	frame->result = fvar;

	//if it's permanent we add it to loaded literal list
	frame->bindPermanentLiteral( l, fvar );	  


	// now we check which object we create

	// Integer literal

	if( *l->theClass->name == "Integer" ){         
	  code += "new Integer ";
	  code += " ";
	  code += l->offset;
	       
          mString dbgmsg = "new object from literal "; dbgmsg += *l->name;
	  code = debug(code, (const char*)dbgmsg);
	  code += '\n';

	  currentCodeOffset += out.appendInstruction( code, 
			    ClassInteger::NEW_FROM_LITERAL, frame->result, l->offset );

	  // relocate literal
	  addRelocation( currentCodeOffset-4 );

	// Double literal
	}else if( *l->theClass->name == "Double" ){
	  code += "new Double ";
	  code += " ";
	  code += l->offset;
	       
          mString dbgmsg = "new object from literal "; dbgmsg += *l->name;
	  code = debug(code, (const char*)dbgmsg);
	  code += '\n';

	  currentCodeOffset += out.appendInstruction( code, 
			    ClassDouble::NEW_FROM_LITERAL, frame->result, l->offset );

	  // relocate literal
	  addRelocation( currentCodeOffset-4 );
	

	// String literal
	}else if( *l->theClass->name == "String" ){
	  code += "new String ";
	  code += " ";
	  code += l->offset;
	       
          mString dbgmsg = "new object from literal "; dbgmsg += *l->name;
	  code = debug(code, (const char*)dbgmsg);
	  code += '\n';

	  currentCodeOffset += out.appendInstruction( code, ClassString::NEW_FROM_LITERAL, frame->result, 
						      l->offset,l->stringValue->length()  );
	  // relocate literal
	  addRelocation( currentCodeOffset-8 );

	// Boolean literal
	}else if( *l->theClass->name == "Boolean" ){
	  code += "new Boolean ";
	  code += " ";
	  code += l->offset;
	       
          mString dbgmsg = "new object from literal "; dbgmsg += *l->name;
	  code = debug(code, (const char*)dbgmsg);
	  code += '\n';

	  currentCodeOffset += out.appendInstruction( code, ClassBoolean::NEW_FROM_LITERAL, frame->result, 
						      l->offset  );
	  // relocate literal
	  addRelocation( currentCodeOffset-4 );


	  // check if this is standalone expression inside block
	  if(tree->parent->type == NODE_CLOSURE){

	    // SE SE VEDNO NASTAVI STATUS, CE IMAMO NPR. INVOKACIJO NAD BOOLEAN LITERALOM??

	    // may require status setting
	    if(requireStatus){
	    
	      code = "l2.status l"; code += frame->result; code += ", 1, 2";
	      code = debug(code, "set block status");
	      
	      currentCodeOffset += out.appendInstruction( code, ClassBlock::STATUS, frame->result, 1, 2 );
	    
	    }

	  }


	// nil literal
	}else if( *l->theClass->name == "Object" ){   // the only literal with Object type is NIL
	  code += "new_nil ";
	       
          mString dbgmsg = "new nil literal";
	  code = debug(code, (const char*)dbgmsg);
	  code += '\n';

	  currentCodeOffset += out.appendInstruction( code, ClassCore::NEW_NIL, frame->result );
	  // relocate literal
	  //addRelocation( currentCodeOffset-4 );

	}

	

      }

      // additional invocation on expression
      mParseTree* invocation = tree->getInvocationChild();
      if( invocation )	evaluateNode( invocation, frame, env );

    }  // case crap
      break;


    ////////////////////////////////////////////////////////////////////
    // primitive type cast
    ////////////////////////////////////////////////////////////////////
    case NODE_PRIMITIVE_CAST: {

      // cast is simulated by creating new object of desired type from old one
      // i.e. casting Integer to Double is achieved by creating new Double 
      // object from Integer object


      // evaluate expression that is to be casted
      evaluateNode( first, frame, /*out,*/ env );

      // 1st we find free variable
      fvar = frame->getFreeVariable();
      code += 'l'; code += fvar; code += ".new ";  // location
      

      // object of which type do we create?
      if( *returnType->name == "Integer" ){
	//////////////////////////////////////////////
	// we're creating Integer
	//////////////////////////////////////////////

	// from what?
	if( *first->returnType->name == "Double" ){
	  // create new Integer from Double
	  code += "Integer l";
	  code += frame->result;
	  mString dbgmsg = "new Integer object casted from Double";
	  code = debug(code, (const char*)dbgmsg);
	  currentCodeOffset += out.appendInstruction( code, 
						      ClassInteger::NEW_FROM_DOUBLE, 
						      fvar,
						      frame->result );

	}else if( *first->returnType->name == "String" ){
	  // create new Integer from String
	  code += "Integer l";
	  code += frame->result;
	  mString dbgmsg = "new Integer object casted from String";
	  code = debug(code, (const char*)dbgmsg);
	  currentCodeOffset += out.appendInstruction( code, 
						      ClassInteger::NEW_FROM_STRING, 
						      fvar,
						      frame->result );
	}


      }else if( *returnType->name == "Double" ){	
	//////////////////////////////////////////////
	// we're creating Double
	//////////////////////////////////////////////

	// from what?
	if( *first->returnType->name == "Integer" ){
	  // create new Double from Integer
	  code += "Double l";
	  code += frame->result;
	  mString dbgmsg = "new Double object casted from Integer";
	  code = debug(code, (const char*)dbgmsg);

	  currentCodeOffset += out.appendInstruction( code, 
						      ClassDouble::NEW_FROM_INTEGER, 
						      fvar,
						      frame->result );

	}else if( *first->returnType->name == "String" ){
	  // create new Double from String
	  code += "Double l";
	  code += frame->result;
	  mString dbgmsg = "new Double object casted from String";
	  code = debug(code, (const char*)dbgmsg);

	  currentCodeOffset += out.appendInstruction( code, 
						      ClassDouble::NEW_FROM_STRING, 
						      fvar,
						      frame->result );
	}


      }else if( *returnType->name == "String" ){	
	//////////////////////////////////////////////
	// we're creating String
	//////////////////////////////////////////////

	// from what?
	if( *first->returnType->name == "Integer" ){
	  // create new String from Integer
	  code += "String l";
	  code += frame->result;
	  mString dbgmsg = "new String object casted from Integer";
	  code = debug(code, (const char*)dbgmsg);

	  currentCodeOffset += out.appendInstruction( code, 
						      ClassString::NEW_FROM_INTEGER, 
						      fvar,
						      frame->result );

	}else if( *first->returnType->name == "Double" ){
	  // create new String from Double
	  code += "String l";
	  code += frame->result;
	  mString dbgmsg = "new String object casted from Double";
	  code = debug(code, (const char*)dbgmsg);

	  currentCodeOffset += out.appendInstruction( code, 
						      ClassString::NEW_FROM_DOUBLE, 
						      fvar,
						      frame->result );

	}else if( *first->returnType->name == "Boolean" ){
	  // create new String from Boolean
	  code += "String l";
	  code += frame->result;
	  mString dbgmsg = "new String object casted from Boolean";
	  code = debug(code, (const char*)dbgmsg);

	  currentCodeOffset += out.appendInstruction( code, 
						      ClassString::NEW_FROM_BOOLEAN, 
						      fvar,
						      frame->result );
	}

      }
      
      frame->free( frame->result );

      frame->result = fvar;  // this is where our final (casted) result is
      
      

      
    }  // case crap
      break;  




    ////////////////////////////////////////////////////////////////////
    // complex type cast
    ////////////////////////////////////////////////////////////////////
    case NODE_COMPLEX_CAST: {
      


      // evaluate expression that is to be casted
      evaluateNode( first, frame, env );


      // first we check if we're upcasting (from derived class to base class). this means that
      // returnType of first must be subclass of second of this tree

      if(first->returnType->isMethodType || second->returnType->isMethodType){


	if(first->returnType->isSubclass( second->returnType )){

	  // casting method to its supertype which means we don't need
	  // to check anything

	}else /*if(second->returnType->isSubclass( first->returnType )) */  {

	  // checking for method type of second->returnType
	  
	  
	  int stackDelta = 0;

	  // first we push return type class
	  int classAddress = 0;  // assume 0 (for procedure)
	  
	  // and also return type which in case of procedure is 0
	  if( second->returnType != 0 )
	    
	    if( second->returnType == this ){
	      
	      classAddress = thisClassAddress;
	      // relocate offset
	      addRelocation( currentCodeOffset+2 );
	      
	    }else{
	      
	      // otherwise we don't know the address so we have to add this offset to link table
	      addLinkEntry( second->returnType, currentCodeOffset + 2 );
	    }
	  
	  
	  code = "Core.pushd "; code += classAddress;
	  mString dbg = "push return class address ";
	  code = debug(code, dbg);
	  code += '\n';
	  currentCodeOffset += out.appendInstruction( code, ClassCore::PUSHD, classAddress  );
	  
	  stackDelta++;
	  
	  
	  
	  // and now parameter class addresses
	  
	  classAddress = 0;
	  
	  mMethodClass* mclass = (mMethodClass*)second->returnType;
	  int nParameters = mclass->parameters->size();

	  for(mClass* param = mclass->parameters->last(); param != 0; param = mclass->parameters->prev()){
	    
	    classAddress = 0;
	    
	    if(param == this){
	      
	      classAddress = thisClassAddress;
	      // relocate offset
	      addRelocation( currentCodeOffset+2 );
	      
	    }else{
	      
	      // otherwise we don't know the address so we have to add this offset to link table
	      addLinkEntry( param, currentCodeOffset + 2 );
	    }
	    
	    
	    code = "Core.pushd "; code += classAddress;
	    mString dbg = "push parameter class address ";  dbg += *param->name;
	    code = debug(code, dbg);
	    code += '\n';
	    currentCodeOffset += out.appendInstruction( code, ClassCore::PUSHD, classAddress  );
	    
	    stackDelta++;
	    
	  }
	  
	  

	  // 1st we find free variable for casted object
	  //fvar = frame->getFreeVariable();
	  code ="Core.mcheck l";  code += (frame->result + stackDelta);
	  code += ", "; code += nParameters;
	  
	  mString dbgmsg = "type checking method";
	  code = debug(code, (const char*)dbgmsg);
	  
	  currentCodeOffset += out.appendInstruction( code, 
						      ClassCore::MCHECK, 
						      frame->result + stackDelta, nParameters );
	}


      }else{   // not method type
      
	if(first->returnType->isSubclass( second->returnType )){



	  // 1st we find free variable for casted object
	  fvar = frame->getFreeVariable();
	  code ="Core.upcast l";  code += frame->result;
	  code += ", l"; code += fvar; 
	  code += ", "; code += "??";
	  
	  mString dbgmsg = "upcasting";
	  code = debug(code, (const char*)dbgmsg);
	  
	  // class address is set to 0 because it is not known at this point
	  
	  // if we're casting into our own class then we know it's address, we just have to
	  // relocate it
	  int classAddress = 0;
	  
	  if(second->returnType == this){
	    
	    classAddress = thisClassAddress;
	    // relocate offset
	    addRelocation( currentCodeOffset+10 );
	    
	  }else{
	    
	    // otherwise we don't know the address so we have to add this offset to link table
	    addLinkEntry( second->returnType, currentCodeOffset + 10 );
	  }
	  
	  
	  
	  currentCodeOffset += out.appendInstruction( code, 
						      ClassCore::UPCAST, 
						      frame->result, fvar, classAddress );
	  
	  frame->free( frame->result );
	  frame->result = fvar;    // casted object becomes result
	  	
	}else{
	  
	  
	  // otherwise we're downcasting
	  // 1st we find free variable for casted object
	  fvar = frame->getFreeVariable();
	  code ="Core.downcast l";  code += frame->result;
	  code += ", l"; code += fvar; 
	  code += ", "; code += "??";
	  
	  mString dbgmsg = "downcasting";
	  code = debug(code, (const char*)dbgmsg);
	  
	  // class address is set to 0 because it is not known at this point
	  
	  // if we're castion into our own class then we know it's address, we just have to
	  // relocate it
	  int classAddress = 0;
	  
	  if(second->returnType == this){
	    
	    classAddress = thisClassAddress;
	    // relocate offset
	    addRelocation( currentCodeOffset+10 );
	    
	  }else{
	    
	    // otherwise we don't know the address so we have to add this offset to link table
	    addLinkEntry( second->returnType, currentCodeOffset + 10 );
	    
	  }
	  
	  
	  
	  // UPCAST takes all complex casts for now
	  currentCodeOffset += out.appendInstruction( code, 
						      ClassCore::UPCAST, 
						      frame->result, fvar, classAddress );
	  
	  frame->free( frame->result );
	  frame->result = fvar;    // casted object becomes result
	  
	}

      }    
	
      receiver = second->returnType;
      // additional invocation on cast expression
      mParseTree* invocation = tree->getInvocationChild( 2 );
      if( invocation ){ 
	out.appendText ( newline ); 
	evaluateNode( invocation, frame, env );
      }
      


    } // case crap
      break;



    ////////////////////////////////////////////////////////////////////
    // local variable/parameter access
    ////////////////////////////////////////////////////////////////////
    case NODE_VARIABLE: {

      // we return this variable's location in frame

      // variable may reside in local closure (actual local) or in one of
      // the parent closures
      if( symbol->symbolType == SYMBOL_VARIABLE ){
	mVariable *v = (mVariable*)symbol;
	receiver = v->theClass;

	// is this local symbol or inherited from parent closure?	

	if( env->getLocalSymbol(v->name, SYMBOL_VARIABLE ) != 0 || v->theClass->isPrimitive() 
	    || env->methodClosure ){

	  // this is a local symbol or primitive so we reference it directly
	  frame->result =  v->entry;

	}else{

	  // symbol was declared in one of the parent closures which means we have its address here
	  // first we must dereference it

	  // this is where we store actual reference
	  fvar = frame->getFreeVariable();
	  
	  code = "Core.deref l"; code += fvar;  code += ", l"; code += v->entry;
	  mString dbgmsg = "dereference object";
	  code = debug(code, (const char*)dbgmsg);
	  code += '\n';
	  currentCodeOffset += out.appendInstruction( code, 
						      ClassCore::DEREF, 
						      fvar,
						      v->entry );

	  frame->result =  fvar;
	  
	  //frame->result =  v->entry;

	}
	

      }else if( symbol->symbolType == SYMBOL_PARAMETER ){
	mParameter *p = (mParameter*)symbol;
	receiver = p->theClass;

	
	frame->result = p->entry;


	

      }
      
	// in addition to standalone reference a variable can also serve as a reference 
	// for method invocation such as:  myVar.method1.method2...methodN

	// if it has a child then we have an invocation
      if( first != 0){

         // evaluate method call on this variable
         //evaluateNode( first, frame, out );


      }else
         partial = true;

      // additional invocation on expression
      mParseTree* invocation = tree->getInvocationChild();
      if( invocation )

	evaluateNode( invocation, frame, /*out,*/ env );

      else{

	// check if this is standalone expression inside block
	if(tree->parent->type == NODE_CLOSURE){

	  // otherwise we check if we need to set block status, if this is a Boolean variable
	  if(requireStatus){
	    
	    if(*symbol->theClass->name == "Boolean"){
	      code = "l2.status l"; code += frame->result; code += ", 1, 2";
	      code = debug(code, "set block status");
	      
	      currentCodeOffset += out.appendInstruction( code, ClassBlock::STATUS, frame->result, 1, 2 );
	    }
	    
	  }
	  
	}

      }

    }  // case crap
      break;
    



    ////////////////////////////////////////////////////////////////////
    // +, -, *, /, %
    // parser ensures that we're always operating on two values of equal
    // primitive types
    ////////////////////////////////////////////////////////////////////
    case NODE_PLUS:     // adding two primitive values
    case NODE_MINUS:    // subtracting two primitive values
    case NODE_MULTIPLY: // multiplying two primitive values
    case NODE_DIVIDE:   // dividing two primitive values
    case NODE_MODULE:   // mod
      
    {
      

      mString operation;  // actual operation
      short opcode = 0;   // actual opcode for operation
      short new_opc = 0;  // opcode for cloning (when creating a temporary object)

      switch(tree->type){
        case NODE_PLUS:

	  ///////////////////////////////////////////
	  // adding two values
	  ///////////////////////////////////////////

	  // because both values are of the same type, we simply take first to obtain type
	  if( *first->returnType->name == "Integer" ){
	    opcode=ClassInteger::ADD;
	    new_opc = ClassInteger::NEW_COPY;
	  }else if( *first->returnType->name == "Double" ){
	    opcode=ClassDouble::ADD;
	    new_opc = ClassDouble::NEW_COPY;
	  }else if( *first->returnType->name == "String" ){
	    opcode=ClassString::ADD;
	    new_opc = ClassString::NEW_COPY;
	  }

	  operation =".add ";

	  break;

        case NODE_MINUS:
 
	  ///////////////////////////////////////////
	  // subtracting two values
	  ///////////////////////////////////////////

	  if( *first->returnType->name == "Integer" ){
	    opcode=ClassInteger::SUB;
	    new_opc = ClassInteger::NEW_COPY;
	  }else if( *first->returnType->name == "Double" ){
	    opcode=ClassDouble::SUB;
	    new_opc = ClassDouble::NEW_COPY;
	  }

	  operation =".sub ";

	  break;

        case NODE_MULTIPLY: 

	  ///////////////////////////////////////////
	  // multiplying two values
	  ///////////////////////////////////////////
	  

	  if( *first->returnType->name == "Integer" ){
	    opcode=ClassInteger::MUL;
	    new_opc = ClassInteger::NEW_COPY;
	  }else if( *first->returnType->name == "Double" ){
	    opcode=ClassDouble::MUL;
	    new_opc = ClassDouble::NEW_COPY;
	  }

	  operation =".mul "; 
	  
	  break;

        case NODE_DIVIDE: 
	  
	  ///////////////////////////////////////////
	  // dividing two values
	  ///////////////////////////////////////////
	 

	  if( *first->returnType->name == "Integer" ){
	    opcode=ClassInteger::DIV;
	    new_opc = ClassInteger::NEW_COPY;
	  }else if( *first->returnType->name == "Double" ){
	    opcode=ClassDouble::DIV;
	    new_opc = ClassDouble::NEW_COPY;
	  }
	  

	  operation =".div "; 

	  break;


	case NODE_MODULE: 
	  
	  ///////////////////////////////////////////
	  // modulo
	  ///////////////////////////////////////////
	 

	  if( *first->returnType->name == "Integer" ){
	    opcode=ClassInteger::MOD;	  
	    new_opc = ClassInteger::NEW_COPY;
	  }

	  operation =".mod "; 

	  break;


      }  // switch

      
      // 1st we evaluate left side expression
      evaluateNode( first, frame, /*out,*/ env );
      leftr = frame->result;  // result of left side

      // now we evaluate right side
      evaluateNode( second, frame, /*out,*/ env );
      rightr = frame->result;  // result of right side
      

      // if left side was temporary we don't need to preserve it so we just 
      // a/s/m/d right side to it
      if( frame->isTemp(leftr) ){ 
	code += 'l'; code += leftr; code += operation; 
	code += frame->strResult();	

	currentCodeOffset += out.appendInstruction( code, opcode, leftr, rightr );

	// result may have been pushed: add any cleanup code if necessary
	//currentCodeOffset += frame->free( rightr, out, code );
	frame->free(rightr);
	frame->result = leftr;

      }else{
	// otherwise we must create intermediate temporary to store left result
	// because left may be a local variable access which we must preserve
	fvar = frame->getFreeVariable();
	// create object set to left result
	code += 'l'; code += fvar; code += ".new ";
	code += *tree->returnType->name;
	code += " ";
	code += 'l'; code += leftr;	
	code = debug(code, "new temporary object");
	code +='\n';

	currentCodeOffset += out.appendInstruction( code, new_opc, fvar, leftr );
	

        //... and now a/s/m/d right side to it
	code = 'l'; code += fvar; code += operation; code += frame->strResult();
	code +='\n';
	frame->result = fvar;

	currentCodeOffset += out.appendInstruction( code, opcode, fvar, rightr );

	// free left & right
	frame->free( leftr );  frame->free( rightr );

      }
      
      partial = false;

      // additional invocation on expression
      receiver = first->returnType;
      mParseTree* invocation = tree->getInvocationChild( 2 );
      if( invocation ){ 
	out.appendText ( newline ); 
	evaluateNode( invocation, frame, /*out,*/ env );
      }

    }
      break;



    ////////////////////////////////////////////////////////////////////
    // relational operators
    // >, >=, <, <=, ==, !=
    // parser ensures that we're always operating on two values of equal
    // primitive types
    ////////////////////////////////////////////////////////////////////
    case NODE_LESS:
    case NODE_GREATER:
    case NODE_EQUALS: 
    case NODE_NOT_EQUALS:  
    case NODE_GREATER_OR_EQUALS:
    case NODE_LESS_OR_EQUALS:
      
    {
      
      mString operation;  // actual operation
      short opcode = 0;   // actual opcode for operation

      bool comparingReferences = false;  // true when we're comparing two references

      switch(tree->type){
        case NODE_LESS:

	  ///////////////////////////////////////////
	  // less than
	  ///////////////////////////////////////////

	  if( *first->returnType->name == "Integer" ){
	    opcode=ClassInteger::IFLS;
	  }else if( *first->returnType->name == "Double" ){
	    opcode=ClassDouble::IFLS;
	  }

	  operation =".ifls ";

	  break;

        case NODE_GREATER:
 
	  ///////////////////////////////////////////
	  // greater than
	  ///////////////////////////////////////////

	  if( *first->returnType->name == "Integer" ){
	    opcode=ClassInteger::IFGT;
	  }else if( *first->returnType->name == "Double" ){
	    opcode=ClassDouble::IFGT;
	  }

	  
	  operation =".ifgt ";

	  break;

	  case NODE_GREATER_OR_EQUALS:
 
	  ///////////////////////////////////////////
	  // greater or equals than
	  ///////////////////////////////////////////

	  if( *first->returnType->name == "Integer" ){
	    opcode=ClassInteger::IFGTEQ;
	  }else if( *first->returnType->name == "Double" ){
	    opcode=ClassDouble::IFGTEQ;
	  }

	  
	  operation =".ifgteq ";

	  break;


	  case NODE_LESS_OR_EQUALS:
 
	  ///////////////////////////////////////////
	  // less or equals than
	  ///////////////////////////////////////////

	  if( *first->returnType->name == "Integer" ){
	    opcode=ClassInteger::IFLSEQ;
	  }else if( *first->returnType->name == "Double" ){
	    opcode=ClassDouble::IFLSEQ;
	  }

	  
	  operation =".iflseq ";

	  break;


	  case NODE_EQUALS:
 
	  ///////////////////////////////////////////
	  // equals 
	  ///////////////////////////////////////////	    

	    operation =".ifeq ";

	    if( *first->returnType->name == "Integer" ){
	      opcode=ClassInteger::IFEQ;
	    }else if( *first->returnType->name == "Double" ){
	      opcode=ClassDouble::IFEQ;
	    }else if( *first->returnType->name == "Boolean" ){
	      opcode=ClassBoolean::IFEQ;
	    }else if( *first->returnType->name == "String" ){
	      opcode=ClassString::IFEQ;
	    }else{

	      // otherwise we are issuing == on references 

	      opcode=ClassCore::IFREFEQ;
	      operation ="Core.ifrefeq ";
	      comparingReferences = true;
	    }
	    
	  	  
	    
	    break;
	    
          case NODE_NOT_EQUALS:
 
	  ///////////////////////////////////////////
	  // not equals 
	  ///////////////////////////////////////////

	  if( *first->returnType->name == "Integer" ){
	    opcode=ClassInteger::IFNEQ;
	  }else if( *first->returnType->name == "Double" ){
	    opcode=ClassDouble::IFNEQ;
	  }else if( *first->returnType->name == "Boolean" ){
	    opcode=ClassBoolean::IFNEQ;
	  }else if( *first->returnType->name == "String" ){
	    opcode=ClassString::IFNEQ;
	  }else{
	      opcode=ClassCore::IFREFNEQ;
	      operation ="Core.ifrefneq ";
	      comparingReferences = true;
	    }

	  
	  operation =".ifneq ";

	  break;


      }  // switch


      // 1st we evaluate left side expression
      evaluateNode( first, frame,  env );
      leftr = frame->result;  // result of left side

      // now we evaluate right side
      evaluateNode( second, frame, env );
      rightr = frame->result;  // result of right side

      // now we execute compare

      // find free location for boolean and create it
      fvar = frame->getFreeVariable();

      code = 'l'; code += fvar; code+=".new Boolean";
      code = debug(code, "new Boolean for comparison");	
      code += '\n';
      currentCodeOffset += out.appendInstruction( code, ClassBoolean::NEW, fvar );
      
      if(comparingReferences){

	code = operation; code += " l"; code += leftr;
	code += ", l"; code += rightr;
	code += ", l"; code += fvar;
	code = debug(code, "compare references and set boolean");
	code += '\n';

      }else{

	code = 'l'; code += leftr; code += operation;
	code += ", l"; code += rightr;
	code += ", l"; code += fvar;
	code = debug(code, "compare and set boolean");
	code += '\n';
      }

      currentCodeOffset += out.appendInstruction( code, opcode, leftr, rightr, fvar );

      // free left & right
      frame->free( leftr );  frame->free( rightr );

      frame->result = fvar;

      // check if this is standalone expression
      if(tree->parent->type == NODE_CLOSURE){

	
	// if this is a conditional expression in a loop we set block status accordingly


	if(requireStatus){

	  code = "l2.status l"; code += fvar; code += ", 1, 2";
	  code = debug(code, "set block status");
	  
	  currentCodeOffset += out.appendInstruction( code, ClassBlock::STATUS, fvar, 1, 2 );
	  
	}

	frame->free( frame->result );

      }

      partial = false;

      // additional invocation on expression
      receiver = first->returnType;
      mParseTree* invocation = tree->getInvocationChild( 2 );
      if( invocation ){ 
	out.appendText ( newline ); 
	evaluateNode( invocation, frame, /*out,*/ env );
      }

    }
    break;



    ////////////////////////////////////////////////////////////////////
    // logical operators
    // &&, ||, !
    // parser ensures that we're always operating on two boolean values
    ////////////////////////////////////////////////////////////////////
    case NODE_AND:
    case NODE_OR:
    case NODE_NOT:
    
      
    {
      
      mString operation;  // actual operation
      short opcode = 0;   // actual opcode for operation


      switch(tree->type){
        case NODE_AND:

	  ///////////////////////////////////////////
	  // and
	  ///////////////////////////////////////////

	  opcode=ClassBoolean::AND;
	  operation =".and ";

	  break;

        case NODE_OR:
 
	  ///////////////////////////////////////////
	  // or
	  ///////////////////////////////////////////

	  opcode=ClassBoolean::OR;
	  operation =".or ";
	  
	  break;

        case NODE_NOT:
 
	  ///////////////////////////////////////////
	  // not
	  ///////////////////////////////////////////

	  opcode=ClassBoolean::NOT;
	  operation =".not ";
	  
	  break;

	  


      }  // switch


      // 1st we evaluate left side expression
      evaluateNode( first, frame,  env );
      leftr = frame->result;  // result of left side


      if( tree->type != NODE_NOT ){   // NOT is unary and has only one expression

	// now we evaluate right side
	evaluateNode( second, frame, env );
	rightr = frame->result;  // result of right side

      }

      // now we execute operation

      // find free location for boolean and create it
      fvar = frame->getFreeVariable();

      code = 'l'; code += fvar; code+=".new Boolean";
      code = debug(code, "new Boolean for logical expression");	
      code += '\n';
      currentCodeOffset += out.appendInstruction( code, ClassBoolean::NEW, fvar );
            
      
      code = 'l'; code += leftr; code += operation;

      if( tree->type != NODE_NOT ){   // no second variable for unary !
	code += "l"; code += rightr;
	code += ", l"; code += fvar;
      }

      code = debug(code, "logical expression and set boolean");
      code += '\n';
      
      if( tree->type != NODE_NOT ){
	currentCodeOffset += out.appendInstruction( code, opcode, leftr, rightr, fvar );
	frame->free( rightr );   // free right

      }else
	currentCodeOffset += out.appendInstruction( code, opcode, leftr, fvar );
      

      // free left
      frame->free( leftr );  

      frame->result = fvar;

      // check if this is standalone expression
      if(tree->parent->type == NODE_CLOSURE){

	
	// if this is a conditional expression in a loop we set block status accordingly


	if(requireStatus){

	  code = "l2.status l"; code += fvar; code += ", 1, 2";
	  code = debug(code, "set block status");
	  
	  currentCodeOffset += out.appendInstruction( code, ClassBlock::STATUS, fvar, 1, 2 );
	  
	}

	frame->free( frame->result );

      }



      partial = false;

      // additional invocation on expression
      receiver = first->returnType;
      mParseTree* invocation = tree->getInvocationChild( 2 );
      if( invocation ){ 
	out.appendText ( newline ); 
	evaluateNode( invocation, frame, /*out,*/ env );
      }

    }
    break;


    ////////////////////////////////////////////////////////////////////
    // ++, --,
    ////////////////////////////////////////////////////////////////////
    case NODE_INCREMENT:
    case NODE_DECREMENT:    
      
    {
      

      mString operation;  // actual operation
      short opcode = 0;   // actual opcode for operation
      short new_opc = 0;  // opcode for cloning (when creating a temporary object)

      switch(tree->type){
        case NODE_INCREMENT:

	  ///////////////////////////////////////////
	  // incrementing
	  ///////////////////////////////////////////

	  if( *first->returnType->name == "Integer" ){
	    opcode=ClassInteger::INC;
	    new_opc = ClassInteger::NEW_COPY;
	  }else if( *first->returnType->name == "Double" ){
	    opcode=ClassDouble::INC;
	    new_opc = ClassDouble::NEW_COPY;
	  }

	  operation =".inc ";

	  break;

        case NODE_DECREMENT:
 
	  ///////////////////////////////////////////
	  // decrementing
	  ///////////////////////////////////////////

	  if( *first->returnType->name == "Integer" ){
	    opcode=ClassInteger::DEC;
	    new_opc = ClassInteger::NEW_COPY;
	  }else if( *first->returnType->name == "Double" ){
	    opcode=ClassDouble::DEC;
	    new_opc = ClassDouble::NEW_COPY;
	  }

	  operation =".dec ";

	  break;        

      }  // switch

      
      // 1st we evaluate left side expression
      evaluateNode( first, frame, env );
      leftr = frame->result;  // result of left side

      // now we evaluate right side
      //evaluateNode( second, frame, /*out,*/ env );
      //rightr = frame->result;  // result of right side
      

      // if left side was temporary we don't need to preserve it so we just 
      // i/d it
      //if( frame->isTemp(leftr) ){ 
	code += 'l'; code += leftr; code += operation; 
	//code += frame->strResult();	

	currentCodeOffset += out.appendInstruction( code, opcode, leftr );

	// result may have been pushed: add any cleanup code if necessary
	//currentCodeOffset += frame->free( rightr, out, code );
	//frame->free(rightr);
	frame->result = leftr;

	/*
      }else{
	// otherwise we must create intermediate temporary to store left result
	// because left may be a local variable access which we must preserve
	fvar = frame->getFreeVariable();
	// create object set to left result
	code += 'l'; code += fvar; code += ".new ";
	code += *tree->returnType->name;
	code += " ";
	code += 'l'; code += leftr;	
	code = debug(code, "new temporary object");
	code +='\n';

	currentCodeOffset += out.appendInstruction( code, new_opc, fvar, leftr );
	

        //... and now i/d it
	code = 'l'; code += fvar; code += operation; code += frame->strResult();
	code +='\n';
	frame->result = fvar;

	currentCodeOffset += out.appendInstruction( code, opcode, fvar );

	// free left
	frame->free( leftr );

	}*/
      
      partial = false;

      // additional invocation on expression
      receiver = first->returnType;
      mParseTree* invocation = tree->getInvocationChild( 1 );
      if( invocation ){ 
	out.appendText ( newline ); 
	evaluateNode( invocation, frame, env );
      }

    }
      break;



    ////////////////////////////////////////////////////////////////////
    // assignment
    ////////////////////////////////////////////////////////////////////
    case NODE_ASSIGN: {      
      
      // first is the variable, second is value to be set to

      //mVariable *v = 0;
      //mParameter *p = 0;

      int entry = 0;

      if( first->symbol->symbolType == SYMBOL_VARIABLE )
	entry = ((mVariable*)first->symbol)->entry;

      else if( first->symbol->symbolType == SYMBOL_PARAMETER )
	entry = ((mParameter*)first->symbol)->entry;
	
      

      

      mString fromType = "";
      
      evaluateNode( second, frame,  env );
      fromType = *second->returnType->name;  	    


      short instruction = -1;
      if( *tree->returnType->name == "Integer" ){

	// assigning to Integer variable

	code += 'l'; code += entry;
	 	
	instruction = ClassInteger::SET;
	code += ".set "; code += 'l'; code += frame->result;
	code = debug(code, "set Integer value");
     

      }else if( *tree->returnType->name == "Double" ){

	// assigning to Double variable

	code += 'l'; code += entry;
	 	
	instruction = ClassDouble::SET;
	code += ".set "; code += 'l'; code += frame->result;
	code = debug(code, "set Double value");


      }else if( *tree->returnType->name == "String" ){

	// assigning to String variable
	code += 'l'; code += entry;
	 	
	instruction = ClassString::SET;
	code += ".set "; code += 'l'; code += frame->result;
	code = debug(code, "set String value");
	

      }else if( *tree->returnType->name == "Boolean" ){

	// assigning to Boolean variable

	code += 'l'; code += entry;
	 	
	instruction = ClassBoolean::SET;
	code += ".set "; code += 'l'; code += frame->result;
	code = debug(code, "set Boolean value");
		
      }else{

	// else we have some user type in which case we just copy the reference from src to our variable


	if(env->getLocalSymbol(first->symbol->name, SYMBOL_VARIABLE) != 0){

	  // if this is locally declared variable we just copy to it

	  instruction = ClassCore::COPY;
	  code += 'l'; code += entry;
	  code += ".copy "; code += 'l'; code += frame->result;
	  code = debug(code, "copy reference");

	}else{

	  // otherwise we must copy to its reference because that's what we have in ins local variable

	  instruction = ClassCore::COPYREF;
	  code += 'l'; code += entry;
	  code += ".copyref "; code += 'l'; code += frame->result;
	  code = debug(code, "copy reference to referenced object");


	}
      }

      currentCodeOffset += out.appendInstruction( code, instruction, entry, 
						  frame->result );
      

      frame->free( frame->result );
  

    }  // case crap
      break;


    ////////////////////////////////////////////////////////////////////
    // parenthesised invocation:  (expression).invocation
    ////////////////////////////////////////////////////////////////////
    case NODE_PARENTHESES: {      
    
      // method in second is called on expression in first
      
      evaluateNode( first, frame,  env );      

      //int reference = frame->result;   // reference of receiver object

      evaluateNode( second, frame,  env ); 
      
      // additional invocation on expression
      receiver = second->returnType;
      /*mParseTree* invocation = tree->getInvocationChild( 2 );
      if( invocation ){ 
	out.appendText ( newline ); 
	evaluateNode( invocation, frame, env );
      }*/


    }  // case crap
      break;



    ////////////////////////////////////////////////////////////////////
    // method invocation
    ////////////////////////////////////////////////////////////////////
    case NODE_METHODCALL: {      
    
      // method can be called with implicitly or explicitly given reference
      // implicit referencing occurs when class method is called directly:
      //   method1.method2...methodN(...);

      // explicit referencing is done through a variable like this:
      //   myvar.method(...);


      mParseTree* parent = tree->parent;

      bool standalone = false;


      mMethod* meth = (mMethod*)symbol;


      int reference = frame->result;   // reference of receiver object

      mClass* receiver_temp = receiver;

      // get arguments and additional call
      mParseTree* arguments = 0;  // assume no arguments
      mParseTree* additional = 0;  // assume no additional methodcall
      
      // check first
      if( first && first->type == NODE_ARGUMENTS ) arguments = first;
      else if( first && first->type == NODE_METHODCALL ) additional = first;
      // check second (which can only be additional method call)
      if( second ) additional = second;
           

      
      int nArguments = 0;
      int stackDelta = 0;


      // if arguments are present we evaluate them first
      if( arguments ){

	nArguments = arguments->children->size();

	// results
	int* results = new int[ nArguments ];

	// evaluate arguments
	int count=0;
	for(mParseTree* arg = arguments->children->last(); arg != 0; arg = arguments->children->prev()){

	  evaluatingArguments = true;
	  evaluateNode( arg, frame, env );
	  evaluatingArguments = false;

	  results[ count++ ] = frame->result;
	}

	// push results on stack and free variables
	for(int i=0;i<nArguments;i++){
	  code = "Core.push l"; code += results[i] + stackDelta;
	  mString dbg = "push argument result";
	  code = debug(code, dbg);
	  code += '\n';
	  currentCodeOffset += out.appendInstruction( code, ClassCore::PUSH, results[i] + stackDelta );
	  frame->free( results[i] );

	  // inc stackDelta for every push (stackDelta is used later in 'reference' which moves downwards)
	  stackDelta++;
	}


      }  // arguments

      
      // restore receiver that may have been modified in arguments evaluation
      receiver = receiver_temp;


      // method invocation via expression - self is given explicitly
      if ( tree->implicitReference == false ){
      
	// check if parent of this variable is closure. if it is, it's a standalone invocation


	// tukaj je treba preverit, ce je standalone (samo na drugacen nacin)!!!!!
	if( parent->parent->type == NODE_CLOSURE ) standalone = true;


	// reference is taken from frame->result

	// first we check if this is a builtin method of a primitive type
	if( meth->isBuiltin() ){

	  if( !meth->isFunction() ){   // builtin procedure

	    code = 'l'; code += (reference + stackDelta); code +=".";
	    code += *meth->name;
	    mString dbg = "builtin method invocation ";
	    dbg += *meth->theClass->name; dbg += '.';
	    dbg += *meth->name;
	    code = debug(code, dbg);
	    currentCodeOffset += out.appendInstruction( code, meth->builtinOpcode, reference+stackDelta );
	    
	  }else{    // builtin function
	    
	    frame->result = frame->getFreeVariable();  // this is where function will store its result

	    code = 'l'; code += (reference + stackDelta); code +=".";
	    code += *meth->name;
	    code += ", l"; code += frame->result;
	    mString dbg = "builtin method invocation ";
	    dbg += *meth->theClass->name; dbg += '.';
	    dbg += *meth->name;
	    code = debug(code, dbg);
	    currentCodeOffset += out.appendInstruction( code, meth->builtinOpcode, reference+stackDelta,
							frame->result );
	    
	    
	  }
	  
	}else{	 // if( meth->isBuiltin() ) 	  
	  
	  // 'normal' method invocation (user defined, not built-in)
	  // in this case we have two options: procedure invocation and function invocation
	  

	  if( !meth->isFunction() ){   // procedure

	    // procedures invocations don't take destination variable to store result

	    // invoke with reference in frame->result
	    code = 'l'; code += (reference + stackDelta); code +=".invoke_1_p #";
	    
	    //int vindex = receiver->methodTable->getMethodIndex(meth);
	    int vindex = meth->theClass->methodTable->getMethodIndex(meth);
	    int descriptor = 0;

	    // always use correct descriptor, even in our own class
	    //if(meth->theClass != receiver->theClass)
	      descriptor = receiver->methodTable->getDescriptor(meth->theClass, 8 );
	    code += vindex; code += ", "; code += descriptor;
	    mString dbg = "virtual procedure method invocation ";
	    code = debug(code, dbg);
	    code += '\n';	
	    currentCodeOffset += out.appendInstruction( code, ClassCore::INVOKE_1_P,
							reference+stackDelta,
							vindex, descriptor );
	  }else{   // function

	    // function invocations take destination variable to store result
	    frame->result = frame->getFreeVariable();  // this is where function will store its result


	    // invoke with reference in frame->result
	    code = 'l'; code += (reference + stackDelta); code +=".invoke_1_f #";
	    //int vindex = receiver->methodTable->getMethodIndex(meth);
	    int vindex = meth->theClass->methodTable->getMethodIndex(meth);
	    int descriptor = 0;
	    // same here ...
	    //if(meth->theClass != receiver->theClass)
	      descriptor = receiver->methodTable->getDescriptor(meth->theClass, 8 );
	    code += vindex; 
	    code += ", l"; code += frame->result; code += ", "; code += descriptor;
	    mString dbg = "virtual function method invocation ";
	    code = debug(code, dbg);
	    code += '\n';	
	    currentCodeOffset += out.appendInstruction( code, ClassCore::INVOKE_1_F,
							reference+stackDelta,
							vindex, 
							frame->result, descriptor );
	  }  // end of function invocation
	}    // end of user method invocation

	// reference no longer needed (if it was temporary)
	frame->free( reference );

      }else{  // if (tree->implicitReference == false)

	// otherwise we have invocation with implicit reference being self

	// if direct parent of this invocation is closure then it's a standalone invocation
	if( parent->type == NODE_CLOSURE ) standalone = true;

	
	// get reference to self from local variable. self may be on different locations depending on
	// whether method returns or not
        
	reference = env->getSelfLocation();

	// first we check if this is a builtin method of a primitive type
	if( meth->isBuiltin() ){

	  if( !meth->isFunction() ){   // builtin procedure

	    code = 'l'; code += (reference + stackDelta); code +=".";
	    code += *meth->name;
	    mString dbg = "builtin method invocation ";
	    dbg += *meth->theClass->name; dbg += '.';
	    dbg += *meth->name;
	    code = debug(code, dbg);
	    currentCodeOffset += out.appendInstruction( code, meth->builtinOpcode, reference+stackDelta );
	    
	  }else{    // builtin function
	    
	    frame->result = frame->getFreeVariable();  // this is where function will store its result

	    code = 'l'; code += (reference + stackDelta); code +=".";
	    code += *meth->name;
	    code += ", l"; code += frame->result;
	    mString dbg = "builtin method invocation ";
	    dbg += *meth->theClass->name; dbg += '.';
	    dbg += *meth->name;
	    code = debug(code, dbg);
	    currentCodeOffset += out.appendInstruction( code, meth->builtinOpcode, reference+stackDelta,
							frame->result );	    	    
	  }
	  
	}else{   // if( meth->isBuiltin() ){

	  // user defined method

	  if( !meth->isFunction() ){   // procedure

	    // procedures invocations don't take destination variable to store result
	    
	    // invoke with reference in frame->result
	    code = 'l'; code += (reference + stackDelta); code +=".invoke_1_p #";
	    int vindex = meth->theClass->methodTable->getMethodIndex(meth);
	    int descriptor = 0;
	    //if(meth->theClass != this)
	      descriptor = this->methodTable->getDescriptor(meth->theClass, 8 );
	    //int vindex = this->methodTable->getMethodIndex(meth);
	    code += vindex; code += ", "; code += descriptor;
	    mString dbg = "virtual procedure method invocation ";
	    code = debug(code, dbg);
	    code += '\n';	
	    currentCodeOffset += out.appendInstruction( code, ClassCore::INVOKE_1_P,
							reference+stackDelta,
							vindex, descriptor );
	  }else{   // function
	    
	    // function invocations take destination variable to store result
	    frame->result = frame->getFreeVariable();  // this is where function will store its result
	    
	    
	    // invoke with reference in frame->result
	    code = 'l'; code += (reference + stackDelta); code +=".invoke_1_f #";
	    int vindex = meth->theClass->methodTable->getMethodIndex(meth);
	    int descriptor = 0;
	    //if(meth->theClass != this)
	      descriptor = this->methodTable->getDescriptor(meth->theClass, 8 );
	    //int vindex = this->methodTable->getMethodIndex(meth);
	    code += vindex; 
	    code += ", l"; code += frame->result ; code += ", "; code += descriptor;
	    mString dbg = "virtual function method invocation ";
	    code = debug(code, dbg);
	    code += '\n';	
	    currentCodeOffset += out.appendInstruction( code, ClassCore::INVOKE_1_F,
							reference+stackDelta,
							vindex, 
							frame->result, descriptor );
	  }

	}
      
      }
      receiver = meth->returnType;
							 

      // do we have compound method call?
      if( additional ){

	receiver = meth->returnType;
	// evaluate additional
	// frame->result will only be used as reference for succeeding call
	evaluateNode( additional, frame, env );


	// free reference after succeeding call has been evaluated
	frame->free( frame->result );

      }else{

	// no additional call! this means that this call was the last. if this was a function
	// that returns Boolean and if it is called inside conditional block we must set
	// block status accordingly. But we only set status if 
	
	// check if this is standalone expression inside block. note however this:
	// if we have invocation like if { a.b.boolMethod; } then ...
	// then we should set status accordingly, but tree->parent->type of last boolMethod
	// will not be NODE_CLOSURE but rather NODE_METHODCALL of previous invocation
	// so this is what we do: we move up the parse tree until we come to a NODE_CLOSURE node type
	// and all nodes in between must be either NODE_METHODCALL or NODE_VARIABLE (which may only be
	// first anyway)
	
	if(requireStatus && tree->isStandaloneClosureMethodCall( tree ) ){

	    code = "l2.status l"; code += frame->result; code += ", 1, 2";
	    code = debug(code, "set block status");
	  
	    currentCodeOffset += out.appendInstruction( code, ClassBlock::STATUS, frame->result, 1, 2 );
	  	 
	}
	
      }


      // if this method returns something then it's on top of stack. 
      // but we don't necessarily need to push the result, because it may be discarded. so check whether
      // the result is being assigned to something - if not we have to dispose it from frame
      // ourselves
      
      if( meth->returnType != 0 ){

	// this is just formal push
	//frame->pushResult();

	if( standalone ){

	  // free destination variable
	  frame->free( frame->result );

	  // generate code for result disposal
	  /*code = "\nCore.free_frame 1";
	  code += "  ( unused result disposal )";
	  currentCodeOffset = out.appendInstruction( code, ClassCore::FREE_FRAME, 1 );
	  */
	}

      }
	
	

      

    }  // case crap
      break;



    ////////////////////////////////////////////////////////////////////
    // native method invocation
    ////////////////////////////////////////////////////////////////////
    case NODE_NATIVECALL: {      
          
      
      mMethod* meth = (mMethod*)symbol; // the method


      // get literal for this method's name
      int literalAddress = ((mLiteral*)first->symbol)->offset;


      // pushed values are: all parameters, 3 (self, return_addr, old_stack)
      int nArguments = meth->numParameters() + 3;

      // + 1 (result local) for functions
      if( meth->isFunction() ) nArguments++;
      	
	
      // generate invocation instruction
      mString instruction = "Core.invoke_n_p";   // assume procedure
      short opcode = ClassCore::INVOKE_N_P;
      if( meth->isFunction() ){ instruction = "Core.invoke_n_f"; opcode = ClassCore::INVOKE_N_F; }
      


      code = instruction; code += literalAddress; code += ", "; code += nArguments;
      mString dbg = "native method invocation ";
      code = debug(code, dbg);
      code += '\n';	
      currentCodeOffset += out.appendInstruction( code, opcode,
						  literalAddress, nArguments );

      // relocate literal
      addRelocation( currentCodeOffset-8 );

    }  // case crap
      break;



    ////////////////////////////////////////////////////////////////////
    // super constructor invocation
    ////////////////////////////////////////////////////////////////////
    case NODE_SUPERCONSTRUCTOR: {      
        

      mParseTree* parent = tree->parent;

      bool standalone = false;


      mMethod* meth = (mMethod*)symbol;


      int reference = env->getSelfLocation();  // reference of receiver object

      

      // get arguments and additional call
      mParseTree* arguments = 0;  // assume no arguments
      mParseTree* additional = 0;  // assume no additional methodcall
      
      // check first
      if( first && first->type == NODE_ARGUMENTS ) arguments = first;
      else if( first && first->type == NODE_METHODCALL ) additional = first;
      // check second (which can only be additional method call)
      if( second ) additional = second;
           

      
      int nArguments = 0;
      int stackDelta = 0;
      

      // if arguments are present we evaluate them first
      if( arguments ){

	nArguments = arguments->children->size();

	// results
	int* results = new int[ nArguments ];

	// evaluate arguments
	int count=0;
	for(mParseTree* arg = arguments->children->last(); arg != 0; arg = arguments->children->prev()){

	  evaluatingArguments = true;
	  evaluateNode( arg, frame, env );
	  evaluatingArguments = false;

	  results[ count++ ] = frame->result;
	}

	// push results on stack and free variables
	for(int i=0;i<nArguments;i++){
	  code = "Core.push l"; code += results[i] + stackDelta;
	  mString dbg = "push argument result";
	  code = debug(code, dbg);
	  code += '\n';
	  currentCodeOffset += out.appendInstruction( code, ClassCore::PUSH, results[i] + stackDelta );
	  frame->free( results[i] );

	  // inc stackDelta for every push (stackDelta is used later in 'reference' which moves downwards)
	  stackDelta++;
	}


      }  // arguments

                	

      // if direct parent of this invocation is closure then it's a standalone invocation
      if( parent->type == NODE_CLOSURE ) standalone = true;


      // first perform upcast because selfish may be modified for multiple inheritance


      code = 'l'; code += (reference + stackDelta); code +=".invoke_s #";
      int start=0;
      int vindex = this->methodTable->getStaticMethodIndex(meth, start);
      int descriptor = 0;
      if(meth->theClass != this)
	descriptor = this->methodTable->getDescriptor(meth->theClass, 8 );
      code += vindex; code += ", "; code += descriptor;
      mString dbg = "superconstructor invocation ";
      code = debug(code, dbg);
      code += '\n';	
      currentCodeOffset += out.appendInstruction( code, ClassCore::INVOKE_S,
						  reference+stackDelta,
						  vindex, descriptor );
   

      
      
							 

      // do we have compound method call?
      if( additional ){

	receiver = meth->returnType;
	// evaluate additional
	// frame->result will be only used as reference for succeding call
	evaluateNode( additional, frame, /*out,*/ env );


	// free reference after succeeding call has been evaluated
	frame->free( frame->result );

      }

      

    }  // case crap
      break;

    ////////////////////////////////////////////////////////////////////
    // construction
    ////////////////////////////////////////////////////////////////////
    case NODE_CONSTRUCTION: {

      // constructing an object with new



      // first & second of method_call node (this tree's first)
      mParseTree *first_of_call = 0;
      mParseTree *second_of_call = 0;
      
      if( first->children->size() > 0 ) first_of_call = first->children->at(0);
      if( first->children->size() > 1 ) second_of_call = first->children->at(1);

          
      // first is call to constructor
      mMethod* meth = (mMethod*)first->symbol;


      // get arguments and additional call
      mParseTree* arguments = 0;  // assume no arguments
      mParseTree* additional = 0;  // assume no additional methodcall
      
      // check first
      if( first_of_call && first_of_call->type == NODE_ARGUMENTS ) arguments = first_of_call;
      else if( first_of_call && first_of_call->type == NODE_METHODCALL ) additional = first_of_call;
      // check second (which can only be additional method call)
      if( second_of_call ) additional = second_of_call;



      int nArguments = 0;
      int stackDelta = 0;

      mClass* receiver_temp = receiver;


      // if arguments are present we evaluate them first
      if( arguments ){

	nArguments = arguments->children->size();

	// results
	int* results = new int[ nArguments ];

	// evaluate arguments
	int count=0;
	for(mParseTree* arg = arguments->children->last(); arg != 0; arg = arguments->children->prev()){

	  evaluatingArguments = true;
	  evaluateNode( arg, frame, env );
	  evaluatingArguments = false;

	  results[ count++ ] = frame->result;
	}

	// push results on stack and free variables
	for(int i=0;i<nArguments;i++){
	  code = "Core.push l"; code += results[i] + stackDelta;
	  mString dbg = "push argument result";
	  code = debug(code, dbg);
	  code += '\n';
	  currentCodeOffset += out.appendInstruction( code, ClassCore::PUSH, results[i] + stackDelta );
	  frame->free( results[i] );

	  // inc stackDelta for every push (stackDelta is used later in 'reference' which moves downwards)
	  stackDelta++;
	}


      }  // arguments
      

      receiver = receiver_temp;


      int reference = 0;

      // if parent is NODE_NEW we're storing into a variable which resides
      // in parent's first
      if( tree->parent->type == NODE_NEW )
	// !!!!!!!!!!!!!!!!!!!!! KAJ CE PRIREJAMO PARAMETRU?  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
          reference = ((mVariable*)tree->parent->children->at(0)->symbol)->entry;
      else
          // otherwise we store our reference to a temporary variable
          reference = frame->getFreeVariable();


      //cout<<"++++++++++ storing new object to l"<<reference<<endl;

      // if class of this method is our (this) class we don't need to
      // make a link entry for it because we already know the adress of
      // its data; it's thisClassAddress. however, we must relocate it
      
      if( meth->theClass == this ){
      
          // method belongs to our own class          

          // we only need to relocate
          addRelocation( currentCodeOffset + 6 );

          code = 'l'; code += (reference + stackDelta); code +=".new_object ";
          code += thisClassAddress; code +=" #"; 
	  int vindex = meth->theClass->methodTable->getMethodIndex(meth);
	  code += vindex;
          code = debug(code, "new own class instance");
          currentCodeOffset += out.appendInstruction( code, ClassCore::NEW_OBJECT,
						      (reference + stackDelta), thisClassAddress, 
						      vindex );

      }else{
      
          // however if this method belongs to an external class we must
          // store this offset into link table for that class
          addLinkEntry( meth->theClass, currentCodeOffset + 6 );

          // because the address is not known we just set it to 0
	  // (it will be set during the linkage process when the class is loaded)
          code = 'l'; code += (reference + stackDelta); code +=".new_object ";
          code += "0 #"; 
	  int vindex = meth->theClass->methodTable->getMethodIndex(meth);
	  code += vindex;
          code = debug(code, "new foreign class instance");
          currentCodeOffset += out.appendInstruction( code, ClassCore::NEW_OBJECT,
						      (reference + stackDelta), 0, 
						      vindex );

      }      


      if( !second ){
	
	if( tree->parent->type == NODE_CLOSURE )
	  frame->free( reference );
	frame->result = reference;

      }else{

	// additional methodcall
	// (1st child is methodcall for constructor 2nd is methodcall
	// on new expression
	receiver = tree->returnType;
	frame->result = reference;      
	out.appendText ( newline ); 
	evaluateNode( second, frame, env );
      }


    }  // case crap
      break;



    ////////////////////////////////////////////////////////////////////
    // return statement
    ////////////////////////////////////////////////////////////////////
    case NODE_RETURN: {

      // number of parameters of this method
      int numParameters = frame->ownerMethod->numParameters();

      
      if( first ){

	//////////////////////////////////////////////////////////////////////7
	// return local variable
	//////////////////////////////////////////////////////////////////////7


	evaluateNode( first, frame, /*out,*/ env );

	// if returning from a method without arguments we call return_1
	
	if( nestedClosures->size() == 0 || updatingMethod ){

	  // return local variable from method

	  if( numParameters == 0 ){
	    
	    // return frame->result	
	    code = "Core.return_1 l"; code += frame->result; 
	    mString dbgmsg = "return l"; dbgmsg += frame->result;
	    code = debug(code, dbgmsg);
	    
	    currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_1_LOCAL, frame->result );
	    frame->free( frame->result );
	    
	  }else{
	    
	    // otherwise we call return_2 which also cleans pushed arguments	
	    code = "Core.return_2 l"; code += frame->result; 
	    code +=", "; code += numParameters;
	    mString dbgmsg = "return l"; dbgmsg += frame->result; dbgmsg += ", cleanup "; 
	    dbgmsg += numParameters;
	    code = debug(code, dbgmsg);
	    
	    currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_2_LOCAL, frame->result,
							numParameters );
	    frame->free( frame->result );
	    
	  }
	}else{  // if( nestedClosure->size() == 0 )


	  // return local from block

	  //int envSize = nestedClosures->top()->nParentVariables;

	  // return local variable from block
	  code = "Block.return l"; code += frame->result; 
	  //code +=", "; code += envSize;
	  mString dbgmsg = "return l"; dbgmsg += frame->result; dbgmsg +=" from block and cleanup";
	  code = debug(code, dbgmsg);
	  
	  currentCodeOffset += out.appendInstruction( code, ClassBlock::RETURN_1_LOCAL, frame->result );
	  frame->free( frame->result );
	  
	}

      }else{   // end of if(first)

	//////////////////////////////////////////////////////////////////////7
	// returning nothing
	//////////////////////////////////////////////////////////////////////7


	if( nestedClosures->size() == 0 || updatingMethod ){   // returning nothing from method

	  if( numParameters == 0){
	    
	    // just return
	    
	    code = "Core.return_1";
	    code = debug(code, "return nothing");
	    currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_1 );
	    
	  }else{
	    
	    // return and cleanup pushed args
	    
	    code = "Core.return_2";
	    mString dbgmsg = "return nothing, cleanup "; dbgmsg += numParameters;
	    code = debug(code, dbgmsg);	  
	    currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_2, numParameters );
	    
	  }

	}else{  // if( nestedClosure->size() == 0 )


	  // return nothing from block

	  //int envSize = nestedClosures->top()->nParentVariables;

	  // return local variable from block
	  code += "Block.return "; //code += envSize;
	  mString dbgmsg = "return nothing from block and cleanup"; 
	  code = debug(code, dbgmsg);
	  
	  currentCodeOffset += out.appendInstruction( code, ClassBlock::RETURN_1 );
	  
	}


      }




    }  // case crap
      break;


    ////////////////////////////////////////////////////////////////////
    // exit from block
    ////////////////////////////////////////////////////////////////////
    case NODE_BLOCK_EXIT: {

      // number of parameters of this method
      //int numParameters = frame->ownerMethod->numParameters();

     
      // plain exit with cleanup

      code = "Block.exit_1";
      code = debug(code, "exiting from block");
      currentCodeOffset += out.appendInstruction( code, ClassBlock::EXIT_1 );

    }  // case crap
    break;


    ////////////////////////////////////////////////////////////////////
    // while
    ////////////////////////////////////////////////////////////////////
    case NODE_WHILE: {

      // first is condition block, second is execution block

      // evaluate them both
      requireStatus = true;
      evaluateNode( first, frame, env );
      requireStatus = false;

      leftr = frame->result;

      evaluateNode( second, frame, env );
      rightr = frame->result;


      // find free variable to store loop
      fvar = frame->getFreeVariable();      

      code = 'l'; code+=fvar; code +=".new While ";
      code += "l"; code += leftr;  code += ", l"; code += rightr;
      code = debug(code, "new while loop");
      code +='\n';

      currentCodeOffset += out.appendInstruction( code, ClassWhile::NEW, fvar, leftr, rightr );

      // free 
      frame->free( leftr );  frame->free( rightr );



      // if this is standalone while loop we execute it
      if( tree->parent->type == NODE_CLOSURE ){

	  // direct parent is closure, so this is a standalone while

	  // generate exec instruction
	  code="l"; code+=fvar; code+= ".exec";
	  mString dbg = "execute while";

	  code = debug(code, dbg);
	  code +='\n';
	  currentCodeOffset += out.appendInstruction( code, ClassWhile::EXEC,  fvar );

	  // variable no longer needed because block has already been executed
	  frame->free( fvar );

	  
	  /////////////////////////////////////////////////////////////////////////////////
	  // test for return from within while
	  /////////////////////////////////////////////////////////////////////////////////


	  // execution closure of while loop may have returned something which we check here. bit 1 is set 
	  // if return did
	  // occur within block and result is on top of stack. if block just returned (from a procedure)
	  // no value is passed on stack but bit will be set anyway

	  // we don't know the number of bytes to skip until we have generated actual return instruction
	  code="Core.skip_ifclear 0, ??";
	  dbg = "test if block returned";

	  code = debug(code, dbg);
	  code +='\n';
	  // make it 6 bytes
	  currentCodeOffset += out.appendInstruction( code, ClassCore::SKIP_IFCLEAR,  0, 6 );

	  // this is where number of bytes to skip is stored
	  int sizeOffset = out.binaryLength - 4;
	  

	  // now we return value that was returned by block. if this is method level we do 
	  // plain method return. if block is around us we do block.return

	  if( nestedClosures->size() == 0 || updatingMethod ){   // method level

	    int numParameters = frame->ownerMethod->numParameters();

	    // we may return from function or procedure. in functions we actually return
	    // some value, whereas in procedures nothing is returned

	    if( numParameters == 0 ){   // no parameters means no arguments cleanup
	
	      
	      if(frame->ownerMethod->isFunction()){  // return from function
		// return top of stack	
		code = "Core.return_1 l1"; 
		mString dbgmsg = "return block result on TOS";
		code = debug(code, dbgmsg);
		code+='\n';
		
		currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_1_LOCAL, 1 );
		// skip_ifclear will skip 6 bytes which is the length of this instruction


	      }else{   // return from procedure
		
		code = "Core.return_1"; 
		mString dbgmsg = "return block result (nothing)";
		code = debug(code, dbgmsg);
		code+='\n';
		
		currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_1 );
		// we should skip 2 bytes
		out.setInstructionValue( sizeOffset, 2 );

	      }

	    }else{   // return with arguments cleanup
	      
	      // otherwise we call return_2 which also cleans pushed arguments	
	      
	      if(frame->ownerMethod->isFunction()){   // return from function

		code = "Core.return_2 l1, "; code += numParameters;
		mString dbgmsg = "return block result on TOS and cleanup"; 
		code = debug(code, dbgmsg);
		code+='\n';
		currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_2_LOCAL, 1,
							    numParameters );
		// we should skip 10 bytes
		out.setInstructionValue( sizeOffset, 10 );

	      }else{   // return from procedure
		
		code = "Core.return_2, "; code += numParameters; 
		mString dbgmsg = "return block result (nothing) and cleanup";
		code = debug(code, dbgmsg);
		code+='\n';
		
		currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_2, numParameters );
		// we should skip 6 bytes which is already set!

	      }
	    }   // if( numParameters != 0 )

	  }else{  // if( nestedClosures->size() == 0 )

	    // still within closure (this is nested block). we pass return result to our parent via 
	    // Block.return

	    int envSize = nestedClosures->top()->nParentVariables;

	    if(frame->ownerMethod->isFunction()){   // return from function

	      // return local variable from block
	      code = "Block.return l1"; 
	      code +=", "; code += envSize;
	      mString dbgmsg = "return block result on TOS and cleanup (in block)"; 
	      code = debug(code, dbgmsg);
	      code +='\n';
	      currentCodeOffset += out.appendInstruction( code, ClassBlock::RETURN_1_LOCAL, 1,
							  envSize );
	      // we should skip 10 bytes
	      out.setInstructionValue( sizeOffset, 10 );

	    }else{   // return from procedure
		
	      code = "Block.return "; code += envSize; 
	      mString dbgmsg = "return block result (nothing) and cleanup (in block)";
	      code = debug(code, dbgmsg);
	      code+='\n';
	      
	      currentCodeOffset += out.appendInstruction( code, ClassBlock::RETURN_1, envSize );
	      // we should skip 6 bytes which is already set!

	    }
	    
	  }   // if( nestedClosures->size() != 0 )
	  
	
      

      /////////////////////////////////////////////////////////////////////////////////
      // end of test for return from within while
      /////////////////////////////////////////////////////////////////////////////////


      // otherwise this loop will be stored somewhere, so save its reference

      }else{

	// save reference
	frame->result = fvar;

      }

      
      
    }  // case crap
    break;



    ////////////////////////////////////////////////////////////////////
    // dowhile
    ////////////////////////////////////////////////////////////////////
    case NODE_DOWHILE: {

      // first is execution block, second is condition block

      // evaluate them both
      
      evaluateNode( first, frame, env );

      leftr = frame->result;

      requireStatus = true;
      evaluateNode( second, frame, env );
      requireStatus = false;

      rightr = frame->result;


      // find free variable to store loop
      fvar = frame->getFreeVariable();

      code = 'l'; code+=fvar; code +=".new DoWhile ";
      code += "l"; code += leftr;  code += ", l"; code += rightr;
      code = debug(code, "new dowhile loop");
      code +='\n';

      currentCodeOffset += out.appendInstruction( code, ClassDoWhile::NEW, fvar, leftr, rightr );

      // free
      frame->free( leftr );  frame->free( rightr );



      // if this is standalone dowhile loop we execute it
      if( tree->parent->type == NODE_CLOSURE ){

          // direct parent is closure, so this is a standalone while

          // generate exec instruction
          code="l"; code+=fvar; code+= ".exec";
          mString dbg = "execute dowhile";

          code = debug(code, dbg);
          code +='\n';
          currentCodeOffset += out.appendInstruction( code, ClassDoWhile::EXEC,  fvar );

          // variable no longer needed because block has already been executed
          frame->free( fvar );


          /////////////////////////////////////////////////////////////////////////////////
          // test for return from within dowhile
          /////////////////////////////////////////////////////////////////////////////////


	  // execution closure of while loop may have returned something which we check here. bit 1 is set
          // if return did
          // occur within block and result is on top of stack. if block just returned (from a procedure)
          // no value is passed on stack but bit will be set anyway

          // we don't know the number of bytes to skip until we have generated actual return instruction
          code="Core.skip_ifclear 0, ??";
          dbg = "test if block returned";

          code = debug(code, dbg);
          code +='\n';
          // make it 6 bytes
          currentCodeOffset += out.appendInstruction( code, ClassCore::SKIP_IFCLEAR,  0, 6 );

          // this is where number of bytes to skip is stored
          int sizeOffset = out.binaryLength - 4;


          // now we return value that was returned by block. if this is method level we do
          // plain method return. if block is around us we do block.return


	  if( nestedClosures->size() == 0 || updatingMethod ){   // method level
 
            int numParameters = frame->ownerMethod->numParameters();
 
            // we may return from function or procedure. in functions we actually return
            // some value, whereas in procedures nothing is returned
 
            if( numParameters == 0 ){   // no parameters means no arguments cleanup
         
               
              if(frame->ownerMethod->isFunction()){  // return from function
                // return top of stack
                code = "Core.return_1 l1";
                mString dbgmsg = "return block result on TOS";
                code = debug(code, dbgmsg);
                code+='\n';
                 
                currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_1_LOCAL, 1 );
                // skip_ifclear will skip 6 bytes which is the length of this instruction
 
 
              }else{   // return from procedure
                 
                code = "Core.return_1";
                mString dbgmsg = "return block result (nothing)";
                code = debug(code, dbgmsg);
                code+='\n';
                 
                currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_1 );
                // we should skip 2 bytes
                out.setInstructionValue( sizeOffset, 2 );
 
              }
 
            }else{   // return with arguments cleanup

              // otherwise we call return_2 which also cleans pushed arguments

              if(frame->ownerMethod->isFunction()){   // return from function

                code = "Core.return_2 l1, "; code += numParameters;
                mString dbgmsg = "return block result on TOS and cleanup";
                code = debug(code, dbgmsg);
                code+='\n';
                currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_2_LOCAL, 1,
                                                            numParameters );
                // we should skip 10 bytes
                out.setInstructionValue( sizeOffset, 10 );

              }else{   // return from procedure
                code = "Core.return_2, "; code += numParameters;
                mString dbgmsg = "return block result (nothing) and cleanup";
                code = debug(code, dbgmsg);
                code+='\n';

                currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_2, numParameters );
                // we should skip 6 bytes which is already set!

              }
            }   // if( numParameters != 0 )

          }else{  // if( nestedClosures->size() == 0 )

            // still within closure (this is nested block). we pass return result to our parent via
            // Block.return

            int envSize = nestedClosures->top()->nParentVariables;

            if(frame->ownerMethod->isFunction()){   // return from function

              // return local variable from block
              code = "Block.return l1";
              code +=", "; code += envSize;
              mString dbgmsg = "return block result on TOS and cleanup (in block)";
              code = debug(code, dbgmsg);
              code +='\n';
              currentCodeOffset += out.appendInstruction( code, ClassBlock::RETURN_1_LOCAL, 1,
                                                          envSize );
              // we should skip 10 bytes
              out.setInstructionValue( sizeOffset, 10 );

            }else{   // return from procedure

              code = "Block.return "; code += envSize;
              mString dbgmsg = "return block result (nothing) and cleanup (in block)";
              code = debug(code, dbgmsg);
              code+='\n';

              currentCodeOffset += out.appendInstruction( code, ClassBlock::RETURN_1, envSize );
              // we should skip 6 bytes which is already set!

            }

          }   // if( nestedClosures->size() != 0 )




      /////////////////////////////////////////////////////////////////////////////////
      // end of test for return from within dowhile
      /////////////////////////////////////////////////////////////////////////////////


      // otherwise this loop will be stored somewhere, so save its reference

      }else{

        // save reference
        frame->result = fvar;

      }


      }  // case crap
    break;





    ////////////////////////////////////////////////////////////////////
    // if then else
    ////////////////////////////////////////////////////////////////////
    case NODE_IF: {

      // first is condition block, second is execution block, third is optional else block or
      // another if statement

      // evaluate them both
      requireStatus = true;
      evaluateNode( first, frame, env );
      requireStatus = false;

      leftr = frame->result;

      evaluateNode( second, frame, env );
      rightr = frame->result;


      /*if(third){
	
	evaluateNode( third, frame, env );
	thirdres = frame->result;

	} */


      // find free variable to store if
      fvar = frame->getFreeVariable();      
      int ifResult = fvar;

      // if we have only 1 condition/block we use 1st constructor

      if( !third ){

	code = 'l'; code+=fvar; code +=".new If ";
	code += "l"; code += leftr;  code += ", l"; code += rightr;
	code = debug(code, "new if");
	code +='\n';

	currentCodeOffset += out.appendInstruction( code, ClassIf::NEW, fvar, leftr, rightr );
     
	// free 
	frame->free( leftr );  frame->free( rightr );

      }else {

	// we have additional condition/block so we add them separately

	int nPairs = tree->numIfConditions();

	code = 'l'; code+=fvar; code +=".new If ";
	code += nPairs;
	code = debug(code, "new if");
	code +='\n';

	currentCodeOffset += out.appendInstruction( code, ClassIf::NEW_1, fvar, nPairs );
     

	// add first pair
	code = 'l'; code+=fvar; code +=".addpair ";
	code += "l"; code += leftr;  code += ", l"; code += rightr;
	code = debug(code, "add condition/block pair");
	code +='\n';

	currentCodeOffset += out.appendInstruction( code, ClassIf::ADDPAIR, fvar, leftr, rightr );
	// free 
	frame->free( leftr );  frame->free( rightr );

	mParseTree *node = third;

	while( true ){

	  // closure means we have else block without condition
	  if(node->type == NODE_CLOSURE ){
	    
	    evaluateNode( node, frame, env );

	    code = 'l'; code+=ifResult; code +=".addblock ";
	    code += "l"; code += frame->result;
	    code = debug(code, "add block only");
	    code +='\n';
	  
	    currentCodeOffset += out.appendInstruction( code, ClassIf::ADDBLOCK, ifResult, frame->result );
	    frame->free( frame->result );

	    // no further if statements can follow else clause
	    break;

	  }else{

	    // we have additional if statement

	    // evaluate them both
	    requireStatus = true;
	    evaluateNode( node->children->at(0), frame, env );
	    requireStatus = false;

	    leftr = frame->result;

	    evaluateNode( node->children->at(1), frame, env );
	    rightr = frame->result;

	    // add pair
	    code = 'l'; code+=ifResult; code +=".addpair ";
	    code += "l"; code += leftr;  code += ", l"; code += rightr;
	    code = debug(code, "add cond/block pair");
	    code +='\n';

	    currentCodeOffset += out.appendInstruction( code, ClassIf::ADDPAIR, ifResult, leftr, rightr );

	    // free 
	    frame->free( leftr );  frame->free( rightr );

	    // next node may exist ot not
	    if(node->children->size() == 3)
	      node = node->children->at(2);
	    else
	      break;


	  }   // node->type == NODE_IF

	}    // while(true)
       	      	

      }  // if( !third )

    



      // if this is standalone we execute it
      if( tree->parent->type == NODE_CLOSURE ){

	  // direct parent is closure, so this is a standalone if

	  // generate exec instruction
	  code="l"; code+=fvar; code+= ".exec";
	  mString dbg = "execute if";

	  code = debug(code, dbg);
	  code +='\n';
	  currentCodeOffset += out.appendInstruction( code, ClassIf::EXEC,  fvar );

	  // variable no longer needed because block has already been executed
	  frame->free( fvar );

	  
	  /////////////////////////////////////////////////////////////////////////////////
	  // test for return from within if
	  /////////////////////////////////////////////////////////////////////////////////


	  // execution closure of if may have returned something which we check here. bit 1 is set 
	  // if return did
	  // occur within block and result is on top of stack. if block just returned (from a procedure)
	  // no value is passed on stack but bit will be set anyway

	  // we don't know the number of bytes to skip until we have generated actual return instruction
	  code="Core.skip_ifclear 0, ??";
	  dbg = "test if block returned";

	  code = debug(code, dbg);
	  code +='\n';
	  // make it 6 bytes
	  currentCodeOffset += out.appendInstruction( code, ClassCore::SKIP_IFCLEAR,  0, 6 );

	  // this is where number of bytes to skip is stored
	  int sizeOffset = out.binaryLength - 4;
	  

	  // now we return value that was returned by block. if this is method level we do 
	  // plain method return. if block is around us we do block.return

	  if( nestedClosures->size() == 0 || updatingMethod ){   // method level

	    int numParameters = frame->ownerMethod->numParameters();

	    // we may return from function or procedure. in functions we actually return
	    // some value, whereas in procedures nothing is returned

	    if( numParameters == 0 ){   // no parameters means no arguments cleanup
	
	      
	      if(frame->ownerMethod->isFunction()){  // return from function
		// return top of stack	
		code = "Core.return_1 l1"; 
		mString dbgmsg = "return block result on TOS";
		code = debug(code, dbgmsg);
		code+='\n';
		
		currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_1_LOCAL, 1 );
		// skip_ifclear will skip 6 bytes which is the length of this instruction


	      }else{   // return from procedure
		
		code = "Core.return_1"; 
		mString dbgmsg = "return block result (nothing)";
		code = debug(code, dbgmsg);
		code+='\n';
		
		currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_1 );
		// we should skip 2 bytes
		out.setInstructionValue( sizeOffset, 2 );

	      }

	    }else{   // return with arguments cleanup
	      
	      // otherwise we call return_2 which also cleans pushed arguments	
	      
	      if(frame->ownerMethod->isFunction()){   // return from function

		code = "Core.return_2 l1, "; code += numParameters;
		mString dbgmsg = "return block result on TOS and cleanup"; 
		code = debug(code, dbgmsg);
		code+='\n';
		currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_2_LOCAL, 1,
							    numParameters );
		// we should skip 10 bytes
		out.setInstructionValue( sizeOffset, 10 );

	      }else{   // return from procedure
		
		code = "Core.return_2, "; code += numParameters; 
		mString dbgmsg = "return block result (nothing) and cleanup";
		code = debug(code, dbgmsg);
		code+='\n';
		
		currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_2, numParameters );
		// we should skip 6 bytes which is already set!

	      }
	    }   // if( numParameters != 0 )

	  }else{  // if( nestedClosures->size() == 0 )

	    // still within closure (this is nested block). we pass return result to our parent via 
	    // Block.return

	    int envSize = nestedClosures->top()->nParentVariables;

	    if(frame->ownerMethod->isFunction()){   // return from function

	      // return local variable from block
	      code = "Block.return l1"; 
	      code +=", "; code += envSize;
	      mString dbgmsg = "return block result on TOS and cleanup (in block)"; 
	      code = debug(code, dbgmsg);
	      code +='\n';
	      currentCodeOffset += out.appendInstruction( code, ClassBlock::RETURN_1_LOCAL, 1,
							  envSize );
	      // we should skip 10 bytes
	      out.setInstructionValue( sizeOffset, 10 );

	    }else{   // return from procedure
		
	      code = "Block.return "; code += envSize; 
	      mString dbgmsg = "return block result (nothing) and cleanup (in block)";
	      code = debug(code, dbgmsg);
	      code+='\n';
	      
	      currentCodeOffset += out.appendInstruction( code, ClassBlock::RETURN_1, envSize );
	      // we should skip 6 bytes which is already set!

	    }
	    
	  }   // if( nestedClosures->size() != 0 )
	  
	
      

      /////////////////////////////////////////////////////////////////////////////////
      // end of test for return from within if
      /////////////////////////////////////////////////////////////////////////////////


      // otherwise this if statement will be stored somewhere, so save its reference

      }else{

	// save reference
	frame->result = fvar;

      }
           


    }  // case crap
    break;
	  
	

    ////////////////////////////////////////////////////////////////////
    // for 
    ////////////////////////////////////////////////////////////////////
    case NODE_FOR: {

      // first is initialisation block, second is condition, third is expression, fourth is execution block

      // evaluate all blocks      
      evaluateNode( first, frame, env );      

      leftr = frame->result;

      requireStatus = true;
      evaluateNode( second, frame, env );
      requireStatus = false;

      rightr = frame->result;
 
      evaluateNode( third, frame, env );
      int thirdr = frame->result;

      evaluateNode( tree->children->at(3), frame, env );
      int fourthr = frame->result;




      // find free variable to store loop
      fvar = frame->getFreeVariable();

      code = 'l'; code+=fvar; code +=".new For ";
      code += "l"; code += leftr;  code += ", l"; code += rightr;
      code = debug(code, "new for loop");
      code +='\n';

      currentCodeOffset += out.appendInstruction( code, ClassFor::NEW, fvar, leftr, rightr );

      code = 'l'; code+=fvar; code +=".addblocks ";
      code += "l"; code += thirdr;  code += ", l"; code += fourthr;
      code = debug(code, "add blocks");
      code +='\n';
      currentCodeOffset += out.appendInstruction( code, ClassFor::ADDBLOCKS, fvar, thirdr, fourthr );


      // free
      frame->free( leftr );  frame->free( rightr );  frame->free( thirdr );  frame->free( fourthr );



      // if this is standalone for loop we execute it
      if( tree->parent->type == NODE_CLOSURE ){

          // direct parent is closure, so this is a standalone while

          // generate exec instruction
          code="l"; code+=fvar; code+= ".exec";
          mString dbg = "execute for";

          code = debug(code, dbg);
          code +='\n';
          currentCodeOffset += out.appendInstruction( code, ClassFor::EXEC,  fvar );

          // variable no longer needed because block has already been executed
          frame->free( fvar );


          /////////////////////////////////////////////////////////////////////////////////
          // test for return from within for
          /////////////////////////////////////////////////////////////////////////////////


	  // execution closure of while loop may have returned something which we check here. bit 1 is set
          // if return did
          // occur within block and result is on top of stack. if block just returned (from a procedure)
          // no value is passed on stack but bit will be set anyway

          // we don't know the number of bytes to skip until we have generated actual return instruction
          code="Core.skip_ifclear 0, ??";
          dbg = "test if block returned";

          code = debug(code, dbg);
          code +='\n';
          // make it 6 bytes
          currentCodeOffset += out.appendInstruction( code, ClassCore::SKIP_IFCLEAR,  0, 6 );

          // this is where number of bytes to skip is stored
          int sizeOffset = out.binaryLength - 4;


          // now we return value that was returned by block. if this is method level we do
          // plain method return. if block is around us we do block.return


	  if( nestedClosures->size() == 0 || updatingMethod ){   // method level
 
            int numParameters = frame->ownerMethod->numParameters();
 
            // we may return from function or procedure. in functions we actually return
            // some value, whereas in procedures nothing is returned
 
            if( numParameters == 0 ){   // no parameters means no arguments cleanup
         
               
              if(frame->ownerMethod->isFunction()){  // return from function
                // return top of stack
                code = "Core.return_1 l1";
                mString dbgmsg = "return block result on TOS";
                code = debug(code, dbgmsg);
                code+='\n';
                 
                currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_1_LOCAL, 1 );
                // skip_ifclear will skip 6 bytes which is the length of this instruction
 
 
              }else{   // return from procedure
                 
                code = "Core.return_1";
                mString dbgmsg = "return block result (nothing)";
                code = debug(code, dbgmsg);
                code+='\n';
                 
                currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_1 );
                // we should skip 2 bytes
                out.setInstructionValue( sizeOffset, 2 );
 
              }
 
            }else{   // return with arguments cleanup

              // otherwise we call return_2 which also cleans pushed arguments

              if(frame->ownerMethod->isFunction()){   // return from function

                code = "Core.return_2 l1, "; code += numParameters;
                mString dbgmsg = "return block result on TOS and cleanup";
                code = debug(code, dbgmsg);
                code+='\n';
                currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_2_LOCAL, 1,
                                                            numParameters );
                // we should skip 10 bytes
                out.setInstructionValue( sizeOffset, 10 );

              }else{   // return from procedure
                code = "Core.return_2, "; code += numParameters;
                mString dbgmsg = "return block result (nothing) and cleanup";
                code = debug(code, dbgmsg);
                code+='\n';

                currentCodeOffset += out.appendInstruction( code, ClassCore::RETURN_2, numParameters );
                // we should skip 6 bytes which is already set!

              }
            }   // if( numParameters != 0 )

          }else{  // if( nestedClosures->size() == 0 )

            // still within closure (this is nested block). we pass return result to our parent via
            // Block.return

            int envSize = nestedClosures->top()->nParentVariables;

            if(frame->ownerMethod->isFunction()){   // return from function

              // return local variable from block
              code = "Block.return l1";
              code +=", "; code += envSize;
              mString dbgmsg = "return block result on TOS and cleanup (in block)";
              code = debug(code, dbgmsg);
              code +='\n';
              currentCodeOffset += out.appendInstruction( code, ClassBlock::RETURN_1_LOCAL, 1,
                                                          envSize );
              // we should skip 10 bytes
              out.setInstructionValue( sizeOffset, 10 );

            }else{   // return from procedure

              code = "Block.return "; code += envSize;
              mString dbgmsg = "return block result (nothing) and cleanup (in block)";
              code = debug(code, dbgmsg);
              code+='\n';

              currentCodeOffset += out.appendInstruction( code, ClassBlock::RETURN_1, envSize );
              // we should skip 6 bytes which is already set!

            }

          }   // if( nestedClosures->size() != 0 )




      /////////////////////////////////////////////////////////////////////////////////
      // end of test for return from within for
      /////////////////////////////////////////////////////////////////////////////////


      // otherwise this loop will be stored somewhere, so save its reference

      }else{

        // save reference
        frame->result = fvar;

      }


      }  // case crap
    break;
     


    ////////////////////////////////////////////////////////////////////
    // method update
    ////////////////////////////////////////////////////////////////////
    case NODE_METHODUPDATE: {

      // NODE_METHODUPDATE may have 1 or 2 children. 1 means we have closure block in first
      // 2 children mean first is methodcall and second closure block
      

      // we need location of self and method index
      int selfLocation = env->getSelfLocation();

      mMethod* meth = (mMethod*)tree->symbol;

      receiver = this;


      if( second ){

	// we have to set this, because methodcall node invokes method on reference in frame->result
	frame->result = selfLocation;

	evaluateNode( first, frame, env );
	selfLocation = frame->result;  // object returned is having its method updated

	// note: this is neccesary because for some reason method call frees it
	frame->occupy( frame->result );


      
      }
	       

                      
        

      // evaluate method block
      updatingMethod++;


      // currentMethod moramo tukaj nastavit na to metodo. dejansko bomo metodo dali na sklad,
      // ker lahko imamo vgnezdene updejte currentMethods = new mStack<mMethod>(); .....

      ////////////////////////////////////////////////////////////////////
      // push parent variables
      ////////////////////////////////////////////////////////////////////

      // first is NODE_METHOD which can have 1 child (NODE_CLOSURE) or 2 (NODE_PARAMETERS and NODE_CLOSURE)
      mParseTree *body = first;

      if(second)
	body = second;
      


      mEnvironment* closureEnv = env->getChildOfClosure( body );  // closure environment
      
      mClosure *closure = 0;      
             

      // push all variables from parent which are used in this closure
      int stackDelta = 0;

      // 1st is current self
      
      int nParentVariables = 0;   // at least 1 variable - self

      // update locations of parent variables in this closure
      closureEnv->updateParentVariables();
	
      
      // for all from parent....
      int frameEntry=5;   // 1st parent var is at location 5 for functions
      if(!meth->isFunction()) frameEntry = 4;  // 4 for procedures


      for(mSymbol *s = closureEnv->parentSymbols->first(); s != 0; s = closureEnv->parentSymbols->next()){


	if( s->used ){
	  
	  int entry = 0;
	  if( s->symbolType == SYMBOL_VARIABLE ){ 
	    
	    entry = ((mVariable*)s)->entry;
	    
	    ((mVariable*)s)->entry = frameEntry;
	    
	    //cout<<"pushing environment variable "<<*s->name<<" at loc "<<entry<<endl;
	    
	  }else if( s->symbolType == SYMBOL_PARAMETER ){
	    
	    entry = ((mParameter*)s)->entry;
	    
	    //cout<<"pushing parameter "<<*s->name<<" at loc "<<entry<<endl;
	    
	    ((mParameter*)s)->entry = frameEntry;
	  }
	  
	  
	  
	  
	  
	  // primitive values are pushed as references
	  if( s->theClass->isPrimitive() || /*!closureEnv->declaredInParent(s->name)*/env->pushedDirectly(s->name) ){
	    
	    code = "Core.push l"; code += (entry + stackDelta);
	    mString dbg = "push environment var ";  dbg += *s->name;
	    code = debug(code, dbg);
	    code += '\n';
	    currentCodeOffset += out.appendInstruction( code, ClassCore::PUSH, entry + stackDelta  );
	    
	  }else{
	    
	    // all other values must firstly be dereferenced and then pushed on stack

	    // this is where we store actual reference
	    fvar = frame->getFreeVariable();
	  
	    code = "Core.deref l"; code += fvar;  code += ", l"; code += (entry + stackDelta);
	    mString dbgmsg = "dereference object";
	    code = debug(code, (const char*)dbgmsg);
	    code += '\n';
	    currentCodeOffset += out.appendInstruction( code, 
							ClassCore::DEREF, 
							fvar,
							entry + stackDelta );
	    
	    // if variable is declared in this environment then we push it as address
	    // however, if it was declared in one of the parent environments we push it as
	    // plain local because its address was already pushed
	    
	    code = "Core.push l"; code += fvar;
	    mString dbg = "push environment var ";  dbg += *s->name;
	    code = debug(code, dbg);
	    code += '\n';
	    currentCodeOffset += out.appendInstruction( code, ClassCore::PUSH, fvar  );

	    frame->free(fvar);  // free reference
	    
	  }
	  
	  stackDelta++;
	  nParentVariables++;
	  frameEntry++;
	  
	}  // if(s->used)
	
      }   // for( parentSymbols....
      


      ////////////////////////////////////////////////////////////////
      // setmethod
      ////////////////////////////////////////////////////////////////


      // object that owns the method.  if we're assigning method block then source is
      // our own object.
      int sourceSelf = env->getSelfLocation();
      

      //if(method_from_self){

      
      if(nParentVariables == 0){


	// method update without parent environment

	code = "Core.setmethod_1 l"; code += selfLocation; 
	code += ", l"; code += sourceSelf;

	int vindex = meth->theClass->methodTable->getMethodIndex(meth);
	int descriptor = 0;
	//if(meth->theClass != receiver->theClass)
	  descriptor = receiver->methodTable->getDescriptor(meth->theClass, 8 );

	code +=", #"; code += vindex; code +=", ??, "; 
	code += descriptor;
	mString dbgmsg = "method update (no environment)";
	code = debug(code, dbgmsg);
	code+='\n';
	
	// we set offset of closure to 0 because it's not yet known
	currentCodeOffset += out.appendInstruction( code, ClassCore::SETMETHOD_1, selfLocation, 
						    sourceSelf, vindex, 0, descriptor );
	
      }else{

	// method update WITH parent environment

	code = "Core.setmethod_2 l"; code += selfLocation + stackDelta; 
	code += ", l"; code += sourceSelf + stackDelta; 
	code +=", "; code += nParentVariables;
	int vindex = meth->theClass->methodTable->getMethodIndex(meth);
	int descriptor = 0;
	//if(meth->theClass != receiver->theClass)
	  descriptor = receiver->methodTable->getDescriptor(meth->theClass, 8 );


	code +=", #"; code += vindex; code +=", ??, ";
	code += descriptor;
	mString dbgmsg = "method update (with environment)";
	code = debug(code, dbgmsg);
	code+='\n';
	
	// we set offset of closure to 0 because it's not yet known
	currentCodeOffset += out.appendInstruction( code, ClassCore::SETMETHOD_2, selfLocation +stackDelta, 
						    sourceSelf+stackDelta, nParentVariables, 
						    vindex, 0, descriptor );


      }

      // free variable that hold receiver self
      frame->free( selfLocation );


      //} else  // we set foreign method (setmethod_2)

      // this is the location of SETMETHOD instruction where we expect closure address
      addRelocation( currentCodeOffset - 8 );

      // we need 'direct' reference to closure , that is without literals, etc.
      int closureRef = env->code->binaryLength - 8; 


      ////////////////////////////////////////////////////////////////
      // end of setmethod
      ////////////////////////////////////////////////////////////////
      
	
      // now shift all variables
      closureEnv->shiftVariables( nParentVariables/*-1*/ );


      

      // make room for parent locals
      closureEnv->currentVariableEntry += nParentVariables;//-1;


      

      // closure will store current code offset and relocation entry for closure address in
      // ClassBlock::NEW instruction
      closure = new mClosure( currentCodeOffset, closureRef, env->code );

      // number of parent variables without self
      closure->nParentVariables = nParentVariables ;//- 1 ;


      // create frame for this closure. owner method changes
      FrameRecord* closureFrame = new FrameRecord( closureEnv, meth );

	
      //cout<<"closureEnv->currentVariableEntry="<<closureEnv->currentVariableEntry<<endl;


      currentCodeOffset += 10;   // reserve for new_frame_copy instruction
	

      nestedClosures->push( closure );   // all relocation and link entries will be recorded in this closure

      closures->append( closure );  // also add it to list	
		
            

      // create code for this closure
      closureEnv->code = new mCode();

      partial = true;

      

      // evaluate all children (statements) of this closure
      for(mParseTree *c = body->children->first(); c != 0; c = body->children->next()){

	//cout<<"************* evaluating method closure "<<c->toString()<<endl;
	evaluateNode( c, closureFrame, closureEnv );
      }
            

      //////////////////////////////////////
      // remove pushed closure from stack
      //////////////////////////////////////
	
      mClosure* cl = nestedClosures->top();
      
      // calculate size
      cl->calculateSize( currentCodeOffset );
      
      
      
      // remove closure and add it to list of closures
      nestedClosures->pop();
      
      
      //////////////////////////////////////////////////
      // now add block prologue and epilogue
      //////////////////////////////////////////////////
      mCode* closureCode = closureEnv->code;
      
      
      // we must provide space for  3 implicitly pushed values (stack_ptr, ret address, self),
      // all variables from parent closures and all local variables
      int itemsToCopy = 3;  // at least 3 values will always be copied (stack, ret, self)
      
      if( meth->returnType != 0 && !meth->isConstructor() ) itemsToCopy++;  // +1 for functions

      // if method accepts arguments we must copy those too. we take arguments from method
      // note that 'method' is method that is being updated, method on the right side
      // is anonymous method, but parser ensures that the number of their arguments matches 
      itemsToCopy += meth->parameterList->size();

      
      itemsToCopy += cl->nParentVariables;
      
      int frame_size = closureFrame->calculateSize();
      //frame_size += cl->nParentVariables;
      
      mString frame_instruction="Core.new_frame ";
      frame_instruction += frame_size;    // total frame size
      frame_instruction += ", ";
      frame_instruction += itemsToCopy;  // number of pushed variables + stack + return addr. + self
      
      frame_instruction += '\n';
      
      
      // closure frame no longer needed
      delete closureFrame;
      
      closureCode->prependInstruction( frame_instruction, ClassCore::NEW_FRAME_COPY, 
				       frame_size, itemsToCopy );
      
      
      // restore old offset
      currentCodeOffset = cl->codeOffset;
            	

      updatingMethod--;



      break;
  }



    ////////////////////////////////////////////////////////////////////
    // anonymous method closure
    ////////////////////////////////////////////////////////////////////
    case NODE_ANONMETHOD: {

      // NODE_ANONMETHOD has 2 children. 1st is closure and 2nd is params
      
      // treba je ustvarit objekt tipa Method
      

      
      // we need location of self and method index
      //int selfLocation = env->getSelfLocation();

      // anonymous method
      mMethod* meth = (mMethod*)tree->symbol;

      receiver = this;


      /*if( second ){

	// we have to set this, because methodcall node invokes method on reference in frame->result
	frame->result = selfLocation;

	evaluateNode( first, frame, env );
	selfLocation = frame->result;  // object returned is having its method updated

	// note: this is neccesary because for some reason method call frees it
	frame->occupy( frame->result );


      
	}*/
	       

                      
        

      // evaluate method block
      updatingMethod++;


      // currentMethod moramo tukaj nastavit na to metodo. dejansko bomo metodo dali na sklad,
      // ker lahko imamo vgnezdene updejte currentMethods = new mStack<mMethod>(); .....

      ////////////////////////////////////////////////////////////////////
      // push parent variables
      ////////////////////////////////////////////////////////////////////

      // first is NODE_CLOSURE
      mParseTree *body = first;

      //if(second)
      //body = second;
      


      mEnvironment* closureEnv = env->getChildOfClosure( body );  // closure environment
      
      mClosure *closure = 0;      
             

      // push all variables from parent which are used in this closure
      int stackDelta = 0;

      // 1st is current self
      
      int nParentVariables = 0;   // at least 1 variable - self

      // update locations of parent variables in this closure
      closureEnv->updateParentVariables();
	
      
      // for all from parent....
      int frameEntry=5;   // 1st parent var is at location 5 for functions
      if(!meth->isFunction()) frameEntry = 4;  // 4 for procedures


      for(mSymbol *s = closureEnv->parentSymbols->first(); s != 0; s = closureEnv->parentSymbols->next()){


	if( s->used ){
	  
	  int entry = 0;
	  if( s->symbolType == SYMBOL_VARIABLE ){ 
	    
	    entry = ((mVariable*)s)->entry;
	    
	    ((mVariable*)s)->entry = frameEntry;
	    
	    //cout<<"pushing environment variable "<<*s->name<<" at loc "<<entry<<endl;
	    
	  }else if( s->symbolType == SYMBOL_PARAMETER ){
	    
	    entry = ((mParameter*)s)->entry;
	    
	    //cout<<"pushing parameter "<<*s->name<<" at loc "<<entry<<endl;
	    
	    ((mParameter*)s)->entry = frameEntry;
	  }
	  
	  
	  
	  
	  
	  // primitive values are pushed as references
	  if( s->theClass->isPrimitive() || env->pushedDirectly(s->name) ){
	    
	    code = "Core.push l"; code += (entry + stackDelta);
	    mString dbg = "push environment var ";  dbg += *s->name;
	    code = debug(code, dbg);
	    code += '\n';
	    currentCodeOffset += out.appendInstruction( code, ClassCore::PUSH, entry + stackDelta  );
	    
	  }else{
	    
	    // all other values must firstly be dereferenced and then pushed on stack

	    // this is where we store actual reference
	    fvar = frame->getFreeVariable();
	  
	    code = "Core.deref l"; code += fvar;  code += ", l"; code += (entry + stackDelta);
	    mString dbgmsg = "dereference object";
	    code = debug(code, (const char*)dbgmsg);
	    code += '\n';
	    currentCodeOffset += out.appendInstruction( code, 
							ClassCore::DEREF, 
							fvar,
							entry + stackDelta );
	    
	    // if variable is declared in this environment then we push it as address
	    // however, if it was declared in one of the parent environments we push it as
	    // plain local because its address was already pushed
	    
	    code = "Core.push l"; code += fvar;
	    mString dbg = "push environment var ";  dbg += *s->name;
	    code = debug(code, dbg);
	    code += '\n';
	    currentCodeOffset += out.appendInstruction( code, ClassCore::PUSH, fvar  );

	    frame->free(fvar);  // free reference
	    
	  }
	  
	  stackDelta++;
	  nParentVariables++;
	  frameEntry++;
	  
	}  // if(s->used)
	
      }   // for( parentSymbols....
      


      ////////////////////////////////////////////////////////////////
      // createmethod:
      ////////////////////////////////////////////////////////////////



      int nParameters = meth->parameterList->size();


      // now we push return type and class addresses for all parameters last to first

      // first we push return type class
      int classAddress = 0;  // assume 0 (for procedure)

      // and also return type which in case of procedure is 0
      if(meth->isFunction() )
	  
	  if(meth->returnType == this){

	    classAddress = thisClassAddress;
	    // relocate offset
	    addRelocation( currentCodeOffset+2 );
	    
	  }else{
	    
	    // otherwise we don't know the address so we have to add this offset to link table
	    addLinkEntry( meth->returnType, currentCodeOffset + 2 );
	  }
      

      code = "Core.pushd "; code += classAddress;
      mString dbg = "push return class address ";
      code = debug(code, dbg);
      code += '\n';
      currentCodeOffset += out.appendInstruction( code, ClassCore::PUSHD, classAddress  );

      stackDelta++;



      // and now parameter class addresses

      classAddress = 0;

      for(mParameter* p = meth->parameterList->last(); p != 0; p = meth->parameterList->prev()){

	classAddress = 0;

	if(p->theClass == this){

	    classAddress = thisClassAddress;
	    // relocate offset
	    addRelocation( currentCodeOffset+2 );
	    
	  }else{
	    
	    // otherwise we don't know the address so we have to add this offset to link table
	    addLinkEntry( p->theClass, currentCodeOffset + 2 );
	  }


	code = "Core.pushd "; code += classAddress;
	mString dbg = "push parameter class address ";  dbg += *p->theClass->name;
	code = debug(code, dbg);
	code += '\n';
	currentCodeOffset += out.appendInstruction( code, ClassCore::PUSHD, classAddress  );

	stackDelta++;

      }


      
      // object that owns the method.  if we're assigning method block then source is
      // our own object.
      int sourceSelf = env->getSelfLocation();
      

      //if(method_from_self){

      // this is where instruction will store reference to Method object
      int reference = frame->getFreeVariable();  // free variable as if no pushes were made
      frame->result = reference;

      
      if(nParentVariables == 0){


	// create method without parent environment


	//TUKAJ je treba ustvarit objekt tipa Method:
	// 1. createmethod
	// 2. instrukcija bo ustvarila objekt Method in ga settala up z vsem potrebnim iz argumentov 
	//    na skladu		   	         	


	code = "Core.createmethod_1 l"; code += (reference + stackDelta); 
	code += ", l"; code += (sourceSelf + stackDelta); code += ", "; code += nParameters;
	code += ", ??"; 


	mString dbgmsg = "create method  (no environment)";
	code = debug(code, dbgmsg);
	code+='\n';
	
	// we set offset of closure to 0 because it's not yet known
	currentCodeOffset += out.appendInstruction( code, ClassCore::CREATEMETHOD_1, reference+stackDelta, 
						    sourceSelf+stackDelta, nParameters, 0 );	



      }else{

	// method update WITH parent environment

	code = "Core.createmethod_2 l"; code += reference + stackDelta; 
	code += ", l"; code += sourceSelf + stackDelta; 
	code += ", "; code += nParameters; code +=", "; code += nParentVariables;
	code += ", ??"; 
	
	mString dbgmsg = "create update (with environment)";
	code = debug(code, dbgmsg);
	code+='\n';
	
	// we set offset of closure to 0 because it's not yet known
	currentCodeOffset += out.appendInstruction( code, ClassCore::CREATEMETHOD_2, reference+stackDelta, 
						    sourceSelf+stackDelta, nParameters, 
						    nParentVariables, 0 );


      }

      // free variable that hold receiver self
      //frame->free( selfLocation );  to je bilo pri updejtu, za levo stran updejta - owner metode


      //} else  // we set foreign method (setmethod_2)

      // this is the location of CREATEMETHOD instruction where we expect closure address (last 4 bytes)
      addRelocation( currentCodeOffset - 4 );

      // we need 'direct' reference to closure , that is without literals, etc.
      int closureRef = env->code->binaryLength - 4; 


      ////////////////////////////////////////////////////////////////
      // end of createmethod
      ////////////////////////////////////////////////////////////////
      
	
      // now shift all variables
      closureEnv->shiftVariables( nParentVariables );


      

      // make room for parent locals
      closureEnv->currentVariableEntry += nParentVariables;//-1;


      

      // closure will store current code offset and relocation entry for closure address in
      // ClassBlock::NEW instruction
      closure = new mClosure( currentCodeOffset, closureRef, env->code );

      // number of parent variables without self
      closure->nParentVariables = nParentVariables ;//- 1 ;


      // create frame for this closure. owner method changes
      FrameRecord* closureFrame = new FrameRecord( closureEnv, meth );

	
      //cout<<"closureEnv->currentVariableEntry="<<closureEnv->currentVariableEntry<<endl;


      currentCodeOffset += 10;   // reserve for new_frame_copy instruction
	

      nestedClosures->push( closure );   // all relocation and link entries will be recorded in this closure

      closures->append( closure );  // also add it to list	
		
            

      // create code for this closure
      closureEnv->code = new mCode();

      partial = true;

      

      // evaluate all children (statements) of this closure
      for(mParseTree *c = body->children->first(); c != 0; c = body->children->next()){

	//cout<<"************* evaluating method closure "<<c->toString()<<endl;
	evaluateNode( c, closureFrame, closureEnv );
      }
            

      //////////////////////////////////////
      // remove pushed closure from stack
      //////////////////////////////////////
	
      mClosure* cl = nestedClosures->top();
      
      // calculate size
      cl->calculateSize( currentCodeOffset );
      
      
      
      // remove closure and add it to list of closures
      nestedClosures->pop();
      
      
      //////////////////////////////////////////////////
      // now add block prologue and epilogue
      //////////////////////////////////////////////////
      mCode* closureCode = closureEnv->code;
      
      
      // we must provide space for  3 implicitly pushed values (stack_ptr, ret address, self),
      // all variables from parent closures and all local variables
      int itemsToCopy = 3;  // at least 3 values will always be copied (stack, ret, self)
      
      if( meth->returnType != 0 && !meth->isConstructor() ) itemsToCopy++;  // +1 for functions

      // if method accepts arguments we must copy those too. we take arguments from method
      // note that 'method' is method that is being updated, method on the right side
      // is anonymous method, but parser ensures that the number of their arguments matches 
      itemsToCopy += meth->parameterList->size();

      
      itemsToCopy += cl->nParentVariables;
      
      int frame_size = closureFrame->calculateSize();
      //frame_size += cl->nParentVariables;
      
      mString frame_instruction="Core.new_frame ";
      frame_instruction += frame_size;    // total frame size
      frame_instruction += ", ";
      frame_instruction += itemsToCopy;  // number of pushed variables + stack + return addr. + self
      
      frame_instruction += '\n';
      
      
      // closure frame no longer needed
      delete closureFrame;
      
      closureCode->prependInstruction( frame_instruction, ClassCore::NEW_FRAME_COPY, 
				       frame_size, itemsToCopy );
      
      
      // restore old offset
      currentCodeOffset = cl->codeOffset;
            	

      updatingMethod--;



      break;
  }



  default:
    cout<<"unhandled node type "<<tree->type<<endl;
      break;
      
              

  }  // switch  
        

  // newline only after complete instructions
  if( !partial ) out.appendText ( newline );
  
}


void mClass::linkSuperClasses(mClass *c, mMethodTable *table, int &address)
{
  // parent classes
  for(mMethodTable *tab = table->superTables->first(); tab != 0; tab = table->superTables->next())    
    linkSuperClasses(c, tab, address);
  

  // add link entry for this class
  // only one vtable should exist 
  for( virtualTable *vtab = table->virtualTables->first(); vtab !=0; vtab = table->virtualTables->next()){

    addLinkEntry(vtab->theClass, address );
    address += 12;   // 
    

    //cout<<"dodajam link entry za razred "<<*vtab->theClass->name<<", address="<<address-12<<endl;
  }

}



void mClass::addRelocation(int offset)
{  

  if(!relocTable){
    // 1st entry
    relocTable = new int[1];
    relocTable[0] = offset;

    relocEntries++;

  }else{
    // append to end
    int *newReloc = new int[relocEntries + 1];
    for(int i=0;i<relocEntries;i++) newReloc[i] = relocTable[i];
    
    newReloc[relocEntries] = offset;
    relocEntries++;
    delete[] relocTable;
    relocTable = newReloc;
  }

  // add this entry to currently nested closure
  if( !nestedClosures->empty() )
    nestedClosures->top()->addRelocationEntry( relocEntries-1 );
}



void mClass::addLinkEntry(mClass* c, int offset)
{

  // don't add Void, in case we want to...by some strange reason
  // and also don't add special Method type because generic Method is added by default
  if( *c->name == "Void" || c->isMethodType ) return;


  // check if we already have an entry for that class
  for(linkTableEntry* e = linkTable->first(); e != 0; e = linkTable->next() )
    if( e->theClass == c ){

      // offset -1 means we don't add offset, just class. and in this case the
      // class is found so we don't have to do anything
      if( offset == -1 ) return;


      e->addOffset( offset );
      
      // add this entry to currently nested closure
      if( !nestedClosures->empty() ){
	
	nestedClosures->top()->addLinkEntry( e, e->nOffsets - 1 );

      }
      
      return;
    }


  // no entry with that class has been found so we create a new one
  linkTableEntry *e = new linkTableEntry( c );

  // and add it to linktable
  linkTable->append( e );

  // if offset is -1 we don't add it to the list
  if( offset == -1 ) return;

  e->addOffset( offset );
    


  // add this entry to currently nested closure
  if( !nestedClosures->empty() )
    nestedClosures->top()->addLinkEntry( e, e->nOffsets - 1 );
}



int mClass::getLinkEntry(mClass* c)
{
  // check if we already have an entry for that class
  int i=0;
  for(linkTableEntry* e = linkTable->first(); e != 0; e = linkTable->next(), i++ )
    if( e->theClass == c )
      return i;
  
      
  // not found
  return -1;
}



void mClass::dumpClosures(mEnvironment *env)
{
  
  mCode* code = env->code;
  
  if(!code) { cout<<"no code"<<endl; return; }

  cout<<*code->text<<endl;
  cout<<"---------------------"<<endl;      
  cout<<*code->bytes<<endl;
  cout<<"---------------------"<<endl;      

  for(mEnvironment* child = env->children->first(); child != 0; child = env->children->next())    
    dumpClosures( child );

}


///////////////////////////////////////////////////////////////////////////

FrameRecord::FrameRecord(mEnvironment *env, mMethod *owner)
{
  // create variable list
  variables = new mList<FrameVariable>();
  variables->setAutoDelete( true );

  permanentLiterals = env->permanentLiterals;

  constant = false;

  ownerMethod = owner;

 

  // frame record holds ONLY temporary vars (locals already have assigned
  // location)

  indexOfFirstTemp = env->currentVariableEntry;  

  // initialise flags
  flag_call_via_var = false;

}

FrameRecord::~FrameRecord()
{
  // delete temporary variables list
  delete variables;
}


int FrameRecord::getFreeVariable(/* mClass *type*/)
{
  FrameVariable *var = 0;
  // start searching at 1st temporary index
  for(unsigned i = 0; i < variables->size(); i++ ){
    var = variables->at( i );
    if( var->free ){ var->free = false; return var->location; }
  }

  // if no free variable was found we add a new one
  FrameVariable *next = new FrameVariable;
  if( var )
     next->location = var->location + 1;
  else 
    next->location = indexOfFirstTemp;  // 1st variable

  next->free = false;
  next->literal = 0;

  variables->append( next );

  return next->location;
}

void FrameRecord::free(int location)
{

  // find temporary variable with this location and free it
  for(FrameVariable* v= variables->first(); v != 0; v = variables->next() )
    if( v->location == location ){ 
      // if this is permanent literal we cannot free it
      if( v->literal ){
	if(!isPermanentLiteral(v->literal) ) v->free = true;
	return;
      }else
	{ v->free = true; return; }
    }
}


void FrameRecord::occupy(int location)
{

  // find temporary variable with this location and mark it used
  for(FrameVariable* v= variables->first(); v != 0; v = variables->next() )
    if( v->location == location ){ 
      // if this is permanent literal we cannot free it
      if( v->literal ){
	if(!isPermanentLiteral(v->literal) ) v->free = true;
	return;
      }else
	{ v->free = false; return; }
    }
}




bool FrameRecord::isTemp(int i)
{   
  if( i < indexOfFirstTemp ) return false;  // local variable is never temporary
  for(unsigned j = 0; j < variables->size(); j++ ){
    // otherwise it is temporary if it doesn't hold permanent literal    
    FrameVariable* v = variables->at( j );
    if(v->location == i )
      if( v->literal && isPermanentLiteral(v->literal) ) return false; 
  }

  return true;   
}


FrameVariable* FrameRecord::findLiteral(mLiteral *l)
{
  for(FrameVariable* v= variables->first(); v != 0; v = variables->next() )
    if( v->literal == l ) return v;
  return 0;
}

void FrameRecord::bindPermanentLiteral(mLiteral* l, int location)
{
  if( !isPermanentLiteral( l )) return;

  for(FrameVariable* v= variables->first(); v != 0; v = variables->next() )
    if( v->location == location ) v->literal = l;
}

bool FrameRecord::isPermanentLiteral(mLiteral* lit)
{
  for(mSymbol *s = permanentLiterals->first(); s != 0; s = permanentLiterals->next()){ 
     mLiteral *literal = (mLiteral*)s;
     if(literal == lit) return true;
  }

  return false;
}


int FrameRecord::calculateSize()
{
  // frame holds entries for named local variables and temporaries
  return indexOfFirstTemp -1 + variables->size();  // -1 because temporaries start at location 1
}
