#ifndef MCLOSURE_H
#define MCLOSURE_H

#include "mString.h"
#include "mSymbol.h"

#include "mClass.h"

// compile time closure holds relocation and linkage entries
class mClosure{
      public:
             // args: current code offset, parent frame, closureAddress in parent code, parent code
             mClosure(int/*, FrameRecord**/, int, mCode*);

             virtual ~mClosure();

	     // add entry into relocation table. this is used so that relocatable offsets which occured
	     // inside this closure can be adjusted
	     void addRelocationEntry(int);

	     // same as above only for link addresses
	     // args: linktable entry, index into address table
	     void addLinkEntry(linkTableEntry*, int);
	     

	     // updates offsets relocation table
	     // args: reloc table, offset of closure start
	     void updateRelocationTable(int*, int);

	     // updates offsets in link table
	     // args: offset of closure start
	     void updateLinkTable(int);

	     // updates parent closure instantiation address of this closure
	     // args: address of closure start, delta into code (prologue length)
	     void fixParentCode(int, int);

	     // calculates size of this closure. this is called after closure has been compiled	     
	     // arg: current offset
	     void calculateSize(int);
             
	     // returns codelength
	     int getLength(){ return codeLength; }

	     // offsets into relocation table for relocatible items in this closure
	     int *relocEntries;	     	     

	     // code offset immediately before this closure
	     int codeOffset;	     

	     // parent's frame is stored to restore it after this closure's evaluation
	     //FrameRecord* parentFrame;

	     // number of used parent variables (without self)
	     int nParentVariables;

      private:
	     

	     struct link{
	       linkTableEntry* entry;
	       int *entries;
	       int nEntries;
	     };

	     // link entries that should be relocated after method
             link* linkEntries;

	     // paren't closure's code
	     mCode* parentCode;

	     // reference to closure address in code 
	     int closureReference;

	     // length of code in this closure
	     int codeLength;	     

	     int nRelocEntries;
	     int nLinkEntries;
	     
	     

            
};


#endif

