#ifndef MDEFS_H
#define MDEFS_H

#include "mString.h"
#include "mList.h"

// returns string representation of access argument
extern mString access2String( short );

// transforms a list of strings into one string
extern mString list2String( mList<mString>* );
// transforms a fully qualified indentifier into list of strings
extern mList<mString> string2List( mString* );


// class header
struct classHeader{
  char signature[6];  // the signature
  int  totalSize;     // total length of class in bytes (without header
                      // linkage and relocation tables, so this is pure data )
  int  nMethods;      // number of methods in this class (without
                      // superclasses)
  int  vtableEntries; // number of entries in virtual table
  int  mtableEntries; // number of entries in method table
  int  literalLength; // length of literals in bytes
  int  metaLength;    // length of metadata in bytes  
  int  relocLength;   // number of relocation entries
  int  nExternalClasses;  // number of external classes
  int  nClosures;     // number of closures in this class
  int  mainIndex;     // index of method main in method table
};

// link table entry
struct linkEntry{
    short classNameLength;  // length of classname
    char* className;        // name of class
    int nEntries;           // number of offsets in this entry
};


// entry into metadata table for methods
struct MetaMethodEntry{
  
  unsigned char flags;       // method flags
  char*         methodName;  // ptr to methodname
  int           vtableIndex; // index of method in virtual table
  int           descriptor;  // descriptor of object this method belongs to (same as in invocation)
  unsigned char nParameters; // # of parameters
  int*          argClassAddresses;   // list of classaddresses for arguments
};

// metadata struct for anonymous methods. 
struct MetaAnonymousEntry{ 
  unsigned char flags;       // method flags identical meaning as for ordinary methods
  //char*         methodName;  // ptr to methodname
  //int           vtableIndex; // index of method in virtual table
  //int           descriptor;  // descriptor of object this method belongs to (same as in invocation)
  int           closure;       // address of closure code
  unsigned char nParameters; // # of parameters
  int*          argClassAddresses;   // list of classaddresses for arguments
  int*          environment;   // parent environment references
};



// access modifiers
#define ACCESS_PUBLIC      0
#define ACCESS_PRIVATE     1
#define ACCESS_PROTECTED   2
#define ACCESS_RESTRICTED  3
#define ACCESS_NATIVE      6

// bind modifiers
#define BIND_STATIC        4
#define BIND_NONSTATIC     5


// symbol types
#define SYMBOL_VARIABLE    0
#define SYMBOL_PARAMETER   1
#define SYMBOL_LITERAL     2
#define SYMBOL_METHOD      3
#define SYMBOL_CLASS      4


// environment types
#define ENVIRONMENT_CLASS  0
#define ENVIRONMENT_METHOD 1
#define ENVIRONMENT_BLOCK  2
#define ENVIRONMENT_NULL   3     // null environment
#define ENVIRONMENT_ROOT   4     // root env
#define ENVIRONMENT_UPDATE 5     // for method update


// code generating
#define CODE_PARTIAL       0
#define CODE_COMPLETE      1

// parse tree node type
#define NODE_LITERAL      1
#define NODE_PLUS         2
#define NODE_MINUS        3
#define NODE_ASSIGN       4
#define NODE_MULTIPLY     5
#define NODE_DIVIDE       6
#define NODE_MOD          7
#define NODE_VARIABLE     8
#define NODE_QUALIFIEDID  9
#define NODE_METHODCALL   10
#define NODE_UNRESOLVED   11
#define NODE_ARRAYACCESS  12
#define NODE_ARGUMENTS    13
#define NODE_METHODUPDATE 14
#define NODE_METHODREF    15
#define NODE_LESS         16
#define NODE_GREATER      17
#define NODE_CLOSURE      18
#define NODE_NEW          19
#define NODE_EQUALS       20
#define NODE_NOT_EQUALS   21
#define NODE_GREATER_OR_EQUALS       22
#define NODE_LESS_OR_EQUALS          23
#define NODE_PRIMITIVE_CAST          24
#define NODE_CONSTRUCTION 25
#define NODE_RETURN       26
#define NODE_MODULE       27
#define NODE_BLOCK_EXIT   28
#define NODE_WHILE        29
#define NODE_DOWHILE      30
#define NODE_IF           31
#define NODE_FOR          32
#define NODE_DELIMITER    33
#define NODE_METHOD       34
#define NODE_PARAMETER    35
#define NODE_COMPLEX_CAST 36
#define NODE_CLASSREF     37
#define NODE_SUPERCONSTRUCTOR 38
#define NODE_NATIVECALL   39
#define NODE_AND          40
#define NODE_OR           41
#define NODE_NOT          42
#define NODE_TYPE         43
#define NODE_PARENTHESES  44
#define NODE_ANONMETHOD   45
#define NODE_INCREMENT    46
#define NODE_DECREMENT    47


// when adding expression node here, add it to isStandaloneMEthodCall in mParseTree as well


#endif
