#include "mEnvironment.h"
#include "mBuiltinClasses.h"
#include <iostream>

using namespace std;

mEnvironment::mEnvironment(short t, mString *Name, mEnvironment *p, mClass *cl)
{

 // create new list for symbols
 symbols = new mList<mSymbol>();

 parentSymbols = new mList<mSymbol>();
 
 // automatically delete them
 symbols->setAutoDelete( true );

 // automatically delete them
 parentSymbols->setAutoDelete( true );

 // permanent literals
 permanentLiterals = new mList<mSymbol>();
 permanentLiterals->setAutoDelete( false );


 // no children at this point
 children = new mList<mEnvironment>();
 children->setAutoDelete( true );
 
 ownerClass = cl;

 parent = p;
 type = t;
 name = Name;

 // this environment becomes a child of parent
 if(p /*&& type != ENVIRONMENT_UPDATE*/ ) p->addChild ( this );
 

 // variables start at location 1
 currentVariableEntry = 1; 


 
 methodClosure = false;

 if(parent && parent->type == ENVIRONMENT_METHOD) methodClosure = true;


 // but if we have any parameters then we put variables AFTER parameters
 // this is for 1st level method blocks, because ENVIRONMENT_METHOD only holds parameters
 if( p && p->type == ENVIRONMENT_METHOD ){
   currentVariableEntry = p->currentVariableEntry;

   addSelf( p->getSelfLocation() );

   // method body is always standalone
   standalone = true;

 }

 // this is for nested blocks (ENVIRONMENT_BLOCK where parent is also ENVIRONMENT_BLOCK)
 if( (p && p->type == ENVIRONMENT_BLOCK && type == ENVIRONMENT_BLOCK) ||
     (p && p->type == ENVIRONMENT_UPDATE && type == ENVIRONMENT_BLOCK) ){


   addSelf( 4 );  // self is at location 4 for closure

   // 1st local in block starts at location 5 (after stack_ptr, return_address, block ref. and self)
   currentVariableEntry = 5;
   
   standalone = p->standalone;
 }
 


 // if this is environment for method update we add self explicitly
 if( type == ENVIRONMENT_UPDATE ){

   //addSelf( 3 );  // self is at location 3 for method

   // 1st local in block starts at location 5 (after stack_ptr, return_address, block ref. and self)
   //currentVariableEntry = 4;

   // method body is always standalone
   standalone = true;

 }


 // copy parent environment into this environment
 if( type == ENVIRONMENT_BLOCK || type == ENVIRONMENT_UPDATE ){


   /////////////////////////////////////////////////
   // first copy variables defined in parent   
   /////////////////////////////////////////////////

   for(mSymbol *sym = p->symbols->first(); sym != 0; sym = p->symbols->next() ){
     

     // we're only interested in variables and parameters
     if( sym->symbolType == SYMBOL_VARIABLE ){ 

       mVariable* var = (mVariable*)sym;
       parentSymbols->append( new mVariable( new mString(*var->name), var->theClass, var->bind, 
					     var->entry ) );

     }else if( sym->symbolType == SYMBOL_PARAMETER ){

       mParameter* par = (mParameter*)sym;

 
       mParameter* newp = new mParameter( new mString(*par->name), par->theClass );
       newp->entry = par->entry;
       parentSymbols->append( newp );
       
     }
     
     
   }  // for
   



   /////////////////////////////////////////////////
   // and now parent variables of parent
   /////////////////////////////////////////////////



   // first copy variables defined in parent   
   for(mSymbol *sym = p->parentSymbols->first(); sym != 0; sym = p->parentSymbols->next() ){

     // only add if it doesnt exist
     if(getParentSymbol(sym->name, SYMBOL_VARIABLE ) != 0) continue;


     // we're only interested in variables and parameters
     if( sym->symbolType == SYMBOL_VARIABLE ){ 
       

       mVariable* var = (mVariable*)sym;
       parentSymbols->append( new mVariable( new mString(*var->name), var->theClass, var->bind, 
					     var->entry ) );

     }else if( sym->symbolType == SYMBOL_PARAMETER ){

       mParameter* par = (mParameter*)sym;


       mParameter* newp = new mParameter( new mString(*par->name), par->theClass );
       newp->entry = par->entry;
       parentSymbols->append( newp );
       
     }
     
     
   }  // for



 }

 

 closure = 0;    // no closure and no code at this point
 code = 0;

 hasReturn = hasParentAssignment = false;
}

mEnvironment::~mEnvironment()
{
 delete symbols;

 delete parentSymbols;

 //cout<<"done deleting symbols "<<endl;
 delete permanentLiterals;

 if(code) delete code;

 //delete children;
 
}

void mEnvironment::append(mSymbol *sym)
{
  // variables and parameters both have locations
  // variables however are fixed, parameters will be shifted once the total number
  // of variables is known
  if( sym->symbolType == SYMBOL_VARIABLE || sym->symbolType == SYMBOL_PARAMETER ) 
    currentVariableEntry++;


 // if this symbol is a literal, check its existence 
 if( sym->symbolType == SYMBOL_LITERAL ){

   mLiteral *lit = findLiteral( sym->name );
   if( !lit ){
     symbols->append( sym );   // add it if it doesn't already exist
   }else{
     // if it does exist than this literal is used multiple times, so add it
     // to permanentLiterals
     mLiteral *l = findPermanentLiteral( sym->name );
     if( !l )
       permanentLiterals->append( sym );
   }
 }else{
   // ordinary symbol is simply added
   
   symbols->append( sym ); 
 }
}


void mEnvironment::appendFromParent(mSymbol *sym)
{
  
  // only variables and parameters can be transferred from parent environment
  if( sym->symbolType != SYMBOL_VARIABLE || sym->symbolType != SYMBOL_PARAMETER ) return;

  // check if it already exists
  for( mSymbol *s = parentSymbols->first(); s != 0; s = parentSymbols->next() ){

    if( *s->name == *sym->name ) return;  // symbol found

  }
  


  // add symbol to parent symbols
  if( sym->symbolType == SYMBOL_VARIABLE ){ 

    mVariable* var = (mVariable*)sym;
    parentSymbols->append( new mVariable( new mString(*var->name), var->theClass, var->bind, var->entry ) );

  }else{

    mParameter* par = (mParameter*)sym;

    mParameter* newp = new mParameter( new mString(*par->name), par->theClass );
    newp->entry = par->entry;
    parentSymbols->append( newp );
  
  }

}



void mEnvironment::addChild(mEnvironment *child)
{
 children->append( child );
}



mSymbol* mEnvironment::getLocalSymbol(mString *name, int symType)
{
  // start in this environment
	
  // if symbol type is not explicitly given search for variables, parameters and parameterless methods
  if( symType == -1 ){
    for(mSymbol *s = symbols->first(); s != 0; s = symbols->next()){
      if( s->symbolType == SYMBOL_METHOD ){
	// if it's method it must not have arguments
	if( *s->name == *name && ((mMethod*)s)->parameterList->size() == 0) return s;
      }else if( s->symbolType == SYMBOL_VARIABLE || s->symbolType == SYMBOL_PARAMETER 
		|| s->symbolType == SYMBOL_CLASS )
	if( *s->name == *name ) return s;

      // if this is CLASS environment we check parameterless methods from superclasses
      if(type == ENVIRONMENT_CLASS) return ownerClass->getMethod( name );
      
    }

  // variables and parameters only
  }else if( symType == SYMBOL_VARIABLE || symType == SYMBOL_PARAMETER ){

    for(mSymbol *s = symbols->first(); s != 0; s = symbols->next())
      if( s->symbolType == SYMBOL_VARIABLE && *s->name == *name )       // found?
	return s;

  // methods only
  }else if( symType == SYMBOL_METHOD ){

    for(mSymbol *s = symbols->first(); s != 0; s = symbols->next())
      if( s->symbolType == SYMBOL_METHOD && *s->name == *name 
	  && ((mMethod*)s)->parameterList->size() == 0 )       // found?
	return s;

  // classes only
  }else if( symType == SYMBOL_CLASS ){
    for(mSymbol *s = symbols->first(); s != 0; s = symbols->next())
      if( s->symbolType == SYMBOL_CLASS && *s->name == *name ){       // found?
	return s;
      }
  }

  return 0;   // symbol not found
}


mSymbol* mEnvironment::getSymbol(mString *name, int symType)
{

  symbolIsLocal = false;

  // 1st we check local environment (this)
  mSymbol *sym = getLocalSymbol( name, symType );
 
  if( sym != 0 ){ symbolIsLocal = true; return sym; }

  // now we check parent symbols
  sym = getParentSymbol( name, symType );

  if( sym != 0 ){

    // we found parent symbol (which is either variable or parameter). this means this symbol will
    // be pushed on stack when this environment's closure is created. we mark it 'used'

    markSymbolUsed( name );
    
    // move up and mark this symbol in parent environment

    return sym;

  }

 

  // search up
  if( parent != 0 )
    return parent->getSymbol( name, symType );
 
  return 0;   // symbol not found
}





mSymbol* mEnvironment::getParentSymbol(mString *name, int symType)
{
  // only variables and parameters can be transferred from parent environment
  //if( symType != SYMBOL_VARIABLE || symType != SYMBOL_PARAMETER ) return 0;

  for( mSymbol *s = parentSymbols->first(); s != 0; s = parentSymbols->next() ){

    if( *s->name == *name ){
      
      return s;

    }

  }

  return 0;

}



mClass* mEnvironment::getFirstClass()
{
  if(type == ENVIRONMENT_CLASS )
     return ownerClass;

  return parent->getFirstClass();
}





mClass* mEnvironment::getDeclaredClass(mString* name)
{
  mSymbol *sym = getLocalSymbol( name, SYMBOL_CLASS );

  if( sym ) return (mClass*)sym;

  return 0;
}


mEnvironment* mEnvironment::getClassEnvironment(mString* name)
{
  for(mEnvironment *env = children->first(); env != 0; env = children->next()){
    
    if( env->type == ENVIRONMENT_CLASS && *env->name == *name ) return env;

  }

  return 0;
}


mString mEnvironment::toString()
{
 mString str( toStringThis(1) );
 return str; 
}


mString mEnvironment::toStringThis(int level)
{
 mString str;
 for(int i=0;i<level;i++) str += "  ";

 // make env type
 str += "-environment type = ";

 switch(type){
        case ENVIRONMENT_CLASS:  str += "CLASS";
                                 break;
        case ENVIRONMENT_METHOD: str += "METHOD";
                                 break;
        case ENVIRONMENT_BLOCK:  str += "BLOCK";
                                 break;
        case ENVIRONMENT_NULL:   str += "NULL";
                                 break;
        case ENVIRONMENT_ROOT:   str += "ROOT";
                                 break;                         
 }

 if(name != 0){
    str +="  ( '";
    str += *name;
    str +="' )";
 }

 
 str += "\n";


 for(mSymbol *s = symbols->first(); s != 0; s = symbols->next()){ 
   //if(s->symbolType == SYMBOL_VARIABLE || s->symbolType == SYMBOL_PARAMETER){
     for(int i=0;i<level;i++) str += "  ";
     str += " -- SYM: ";
     str += s->toString();
     str += "\n";
   //}
 }


 // all statements and expressions are stored in closure
 if(closure != 0){
    for(int i=0;i<level;i++) str += "  ";
    str += "(closure\n";
    // closure's children are statements and expressions
    for(mParseTree *child = closure->children->first(); child != 0;
        child = closure->children->next() ){
           for(int i=0;i<level;i++) str += "  ";
           //str += "(";
           str += child->toString();
           //str += ") ";
           str += "\n";
    }
    for(int i=0;i<level;i++) str += "  ";
    str +=")";
 }
 

 // get environment for children
 for(mEnvironment *env = children->first(); env != 0; env = children->next()){
     for(int i=0;i<level+1;i++) str += " ";
     str+="\n";
     str += env->toStringThis( level + 2 );
 }
    
 return str;
}



void mEnvironment::addSelf(int location)
{
 
  // ownerClass is this environment's enclosing class
  mVariable *self = new mVariable(new mString("self"), ownerClass, BIND_NONSTATIC, location);

  // add it to symbol list
  symbols->append( self );

}


// this is meant to be called in root environment to add all builtin
// classes 
void mEnvironment::createBuiltinClasses()
{

  ////////////////////////////////////////////////////////////////
  // Void
  ////////////////////////////////////////////////////////////////
  mClass *v = new mClass( new mString("Void"), ACCESS_PUBLIC );
  // add Void to this environment
  append( v );


  ////////////////////////////////////////////////////////////////
  // Object
  ////////////////////////////////////////////////////////////////
  mClass *o = new mClass( new mString("Object"), ACCESS_PUBLIC );

  // methods for Object
  mMethod *Object_clone=new mMethod( new mString("clone"), o, ACCESS_PUBLIC, 
				   BIND_NONSTATIC );
  o->addMethod( Object_clone );

  mMethod *Object_print=new mMethod( new mString("print"), v, ACCESS_PUBLIC, 
				   BIND_NONSTATIC );
  o->addMethod( Object_print );


  // getClass returns Class, but we don't have it at this point yet so we set
  // its return type not to Void but to 0
  mMethod *Object_getclass=new mMethod( new mString("getClass"), 0, ACCESS_PUBLIC, 
					BIND_NONSTATIC, ClassObject::GETCLASS );
  // and we add 'late' return type with name 'Class' which will be inserted later when it  becomes
  // available
  Object_getclass->lateReturnType = new mString("Class");

  o->addMethod( Object_getclass );
  
    
  // add Object to this environment
  append( o );
  


  ////////////////////////////////////////////////////////////////
  // Integer
  ////////////////////////////////////////////////////////////////
  mClass *i = new mClass( new mString("Integer"), ACCESS_PUBLIC );
  
  i->addSuperClass( o );

  mMethod *Integer_clone=new mMethod( new mString("clone"), i, ACCESS_PUBLIC, 
				    BIND_NONSTATIC, ClassInteger::NEW_COPY );
  i->addMethod( Integer_clone );

  mMethod *Integer_print=new mMethod( new mString("print"), v, ACCESS_PUBLIC, 
				    BIND_NONSTATIC, ClassInteger::PRINT );
  i->addMethod( Integer_print );
  
  // add Integer to this environment
  append( i );


  ////////////////////////////////////////////////////////////////
  // Character
  ////////////////////////////////////////////////////////////////
  mClass *c = new mClass( new mString("Character"), ACCESS_PUBLIC );
  
  c->addSuperClass( o );

  mMethod *Character_clone=new mMethod( new mString("clone"), i, ACCESS_PUBLIC, 
				    BIND_NONSTATIC, ClassCharacter::NEW_COPY );
  c->addMethod( Character_clone );

  mMethod *Character_print=new mMethod( new mString("print"), v, ACCESS_PUBLIC, 
				    BIND_NONSTATIC, ClassCharacter::PRINT );
  c->addMethod( Character_print );
  
  // add Character to this environment
  append( c );


  ////////////////////////////////////////////////////////////////
  // Double
  ////////////////////////////////////////////////////////////////
  mClass *d = new mClass( new mString("Double"), ACCESS_PUBLIC );
  
  d->addSuperClass( o );

  mMethod *Double_print=new mMethod( new mString("print"), v, ACCESS_PUBLIC, 
				   BIND_NONSTATIC, ClassDouble::PRINT );
  d->addMethod( Double_print );

  
  // add Double to this environment
  append( d );


  ////////////////////////////////////////////////////////////////
  // String
  ////////////////////////////////////////////////////////////////
  mClass *s = new mClass( new mString("String"), ACCESS_PUBLIC );

  s->addSuperClass( o );

  // methods for String
  mMethod *String_print=new mMethod( new mString("print"), v, ACCESS_PUBLIC, 
				   BIND_NONSTATIC, ClassString::PRINT );
  s->addMethod( String_print );
  

  
  // add String to this environment
  append( s );


  ////////////////////////////////////////////////////////////////
  // Boolean
  ////////////////////////////////////////////////////////////////
  mClass *bl = new mClass( new mString("Boolean"), ACCESS_PUBLIC );

  bl->addSuperClass( o );

  // methods for Boolean
  mMethod *Boolean_print=new mMethod( new mString("print"), v, ACCESS_PUBLIC, 
				   BIND_NONSTATIC, ClassBoolean::PRINT );
  bl->addMethod( Boolean_print );
  

  
  // add Boolean to this environment
  append( bl );

  
  ////////////////////////////////////////////////////////////////
  // Block
  ////////////////////////////////////////////////////////////////
  mClass *b = new mClass( new mString("Block"), ACCESS_PUBLIC );

  b->addSuperClass( o );

  // methods for Block
  //mMethod *Block_print=new mMethod( new mString("print"), 0, ACCESS_PUBLIC, 
  //			   BIND_NONSTATIC, ClassString::PRINT );
  //s->addMethod( String_print );

  mMethod *Block_exec=new mMethod( new mString("exec"), v, ACCESS_PUBLIC, 
				   BIND_NONSTATIC, ClassBlock::EXEC );
  b->addMethod( Block_exec );


 

  // isTopLevel : Boolean
  mMethod *Block_isTopLevel=new mMethod( new mString("isTopLevel"), bl, ACCESS_PUBLIC, 
					 BIND_NONSTATIC );
  b->addMethod( Block_isTopLevel );
  

  
  // add Block to this environment
  append( b );

  
  ////////////////////////////////////////////////////////////////
  // While
  ////////////////////////////////////////////////////////////////
  mClass *wh = new mClass( new mString("While"), ACCESS_PUBLIC );

  wh->addSuperClass( o );

  // methods for While
  //mMethod *Block_print=new mMethod( new mString("print"), 0, ACCESS_PUBLIC, 
  //			   BIND_NONSTATIC, ClassString::PRINT );
  //s->addMethod( String_print );

  mMethod *While_exec=new mMethod( new mString("exec"), v, ACCESS_PUBLIC, 
				   BIND_NONSTATIC, ClassWhile::EXEC );
  wh->addMethod( While_exec );
  

  
  // add While to this environment
  append( wh );


  ////////////////////////////////////////////////////////////////
  // DoWhile
  ////////////////////////////////////////////////////////////////
  mClass *dowh = new mClass( new mString("DoWhile"), ACCESS_PUBLIC );

  dowh->addSuperClass( o );

  // methods for While
  //mMethod *Block_print=new mMethod( new mString("print"), 0, ACCESS_PUBLIC, 
  //			   BIND_NONSTATIC, ClassString::PRINT );
  //s->addMethod( String_print );

  mMethod *DoWhile_exec=new mMethod( new mString("exec"), v, ACCESS_PUBLIC, 
				   BIND_NONSTATIC, ClassDoWhile::EXEC );
  dowh->addMethod( DoWhile_exec );
  

  
  // add DoWhile to this environment
  append( dowh );



  ////////////////////////////////////////////////////////////////
  // If
  ////////////////////////////////////////////////////////////////
  mClass *If = new mClass( new mString("If"), ACCESS_PUBLIC );

  If->addSuperClass( o );

  // methods for While
  //mMethod *Block_print=new mMethod( new mString("print"), 0, ACCESS_PUBLIC, 
  //			   BIND_NONSTATIC, ClassString::PRINT );
  //s->addMethod( String_print );

  mMethod *If_exec=new mMethod( new mString("exec"), v, ACCESS_PUBLIC, 
				   BIND_NONSTATIC, ClassIf::EXEC );
  If->addMethod( If_exec );
  

  
  // add If to this environment
  append( If );


  ////////////////////////////////////////////////////////////////
  // For
  ////////////////////////////////////////////////////////////////
  mClass *For = new mClass( new mString("For"), ACCESS_PUBLIC );

  For->addSuperClass( o );

  // methods for For
  //mMethod *Block_print=new mMethod( new mString("print"), 0, ACCESS_PUBLIC, 
  //			   BIND_NONSTATIC, ClassString::PRINT );
  //s->addMethod( String_print );

  mMethod *For_exec=new mMethod( new mString("exec"), v, ACCESS_PUBLIC, 
				   BIND_NONSTATIC, ClassFor::EXEC );
  For->addMethod( For_exec );
  

  
  // add For to this environment
  append( For );

}


void mEnvironment::updateBuiltinClasses()
{

  // check return types of methods for all builtin classes
  for(mSymbol *s = symbols->first(); s != 0; s = symbols->next())

   if( s->symbolType == SYMBOL_CLASS ){
     
     mClass *c = (mClass*)s;

     for(mMethod *m =  c->methodList->first(); m != 0; m = c->methodList->next()){

       if(m->lateReturnType != 0){

	 // update return type of this method

	 // find class with specified name in this environment
	 mSymbol *sym = getLocalSymbol( m->lateReturnType );
	 if(sym != 0){
	   m->returnType = (mClass*)sym;
	   //cout<<"Setting return type of "<<*m->name<<" to "<<*sym->name<<endl;
	 }
       }

     }

   }

}



mLiteral* mEnvironment::findLiteral(mString *name)
{
  for(mSymbol *s = symbols->first(); s != 0; s = symbols->next()){ 
   if( s->symbolType == SYMBOL_LITERAL ){
     mLiteral *literal = (mLiteral*)s;
     if(*name == *literal->name ) return literal;
   }
 }
  return 0;
}

mLiteral* mEnvironment::findPermanentLiteral(mString *name)
{
  for(mSymbol *s = permanentLiterals->first(); s != 0; s = permanentLiterals->next()){ 
     mLiteral *literal = (mLiteral*)s;
     if(*name == *literal->name ) return literal;
  }
  return 0;
}


void mEnvironment::removeLiteral(mLiteral* lit)
{
  // remove literal from both lists
  if(permanentLiterals->exists(lit)) permanentLiterals->remove( lit );
  // symbols will delete it completely
  symbols->remove( lit );  
}


mEnvironment* mEnvironment::getChildOfClosure(mParseTree* closure)
{

  for(mEnvironment *env = children->first(); env != 0; env = children->next())
    if( env->closure == closure ) return env;
 
  return 0;
}

int mEnvironment::getSelfLocation()
{

  mString selfName("self");
  mSymbol *sym = getLocalSymbol( &selfName, SYMBOL_VARIABLE );
  
  if(!sym) return -1;

  return ((mVariable*)sym)->entry;
}


void mEnvironment::shiftVariables(int offset)
{
  
  for(mSymbol *sym = symbols->first(); sym != 0; sym = symbols->next() ){
    
    if(*sym->name == "self") continue;

    // only variables and parameters
     if( sym->symbolType == SYMBOL_VARIABLE ){ 

       mVariable* var = (mVariable*)sym;
       var->entry += offset;

     }else if( sym->symbolType == SYMBOL_PARAMETER ){

       mParameter* par = (mParameter*)sym;
       par->entry += offset;       
       
     }         
     
  }  // for

  
}


void mEnvironment::markSymbolUsed(mString* name)
{
  if( type != ENVIRONMENT_BLOCK && type != ENVIRONMENT_UPDATE ) return;

  mSymbol* s = getParentSymbol( name );

  if(!s) return;

  s->used = true;

  // search up if in block environment
  if( parent->type == ENVIRONMENT_BLOCK || type == ENVIRONMENT_UPDATE )
    parent->markSymbolUsed(name);
  
}



void mEnvironment::updateParentVariables()
{

  for( mSymbol *s = parentSymbols->first(); s != 0; s = parentSymbols->next() ){

    mSymbol *orig = findInParent( s->name );

    if(!orig) continue;

    if( s->symbolType == SYMBOL_VARIABLE ){

      ((mVariable*)s)->entry = ((mVariable*)orig)->entry;

    }else if( s->symbolType == SYMBOL_PARAMETER ){

     

      ((mParameter*)s)->entry = ((mParameter*)orig)->entry;
      

    }
    
  }

  
}


mSymbol* mEnvironment::findInParent(mString* name)
{

  // check symbols defined in parent
  for(mSymbol *sym = parent->symbols->first(); sym != 0; sym = parent->symbols->next() ){

     // we're only interested in variables and parameters
     if( *sym->name == *name ){
       
       if( sym->symbolType == SYMBOL_VARIABLE || sym->symbolType == SYMBOL_PARAMETER ){
	 
	 return sym;
       }
     }
     
  }  // for
  

  

  // check symbols defined in parent  
   for(mSymbol *sym = parent->parentSymbols->first(); sym != 0; sym = parent->parentSymbols->next() ){

     // we're only interested in variables and parameters
     if( *sym->name == *name ){
       
       if( sym->symbolType == SYMBOL_VARIABLE || sym->symbolType == SYMBOL_PARAMETER ){

	

	 return sym;     
       }

     }
     
   }  // for

   return 0;
}


bool mEnvironment::declaredInParent(mString* name)
{

  if(!parent) return false;

  // check symbols defined in parent
  for(mSymbol *sym = parent->symbols->first(); sym != 0; sym = parent->symbols->next() ){

    // we're only interested in variables and parameters
    if( *sym->name == *name ){
      
      if( sym->symbolType == SYMBOL_VARIABLE || sym->symbolType == SYMBOL_PARAMETER ){

	return true;
      }

    }

  }  // for
  
  return false;
}


bool mEnvironment::declaredHere(mString* name)
{


  // check symbols defined in parent
  for(mSymbol *sym = symbols->first(); sym != 0; sym = symbols->next() ){

    // we're only interested in variables and parameters
    if( *sym->name == *name ){
      
      if( sym->symbolType == SYMBOL_VARIABLE || sym->symbolType == SYMBOL_PARAMETER ){

	return true;
      }

    }

  }  // for
  
  return false;
}


bool mEnvironment::pushedDirectly(mString* name)
{

  // if symbol is declared here it is pushed as direct reference
  if( declaredHere(name) ) return true;

  // if it's not declared here, then we have 2 options:
  // 1. the variable may be declared in parent block in which case it was pushed as address
  // 2. the variable may be declared as an argument in which case it was pushed as direct reference

  // we check for 2nd option
  if( parent->type == ENVIRONMENT_METHOD )
    if(parent->declaredHere(name)) return true;


  // false
  return false;
  
}






void mEnvironment::setReturn(bool ret, int line)
{
  hasReturn = ret;
}


void mEnvironment::setParentAssignment(mClass* c, bool assign, int line)
{
  // primitive values can be assigned
  if(c->isPrimitive()) return;


  hasParentAssignment = assign;
}


bool mEnvironment::containsReturn()
{
 if(hasReturn) return true;

 // check all children
 for(mEnvironment *env = children->first(); env != 0; env = children->next())
    if( env->containsReturn() ) return true;

 return false;
}


bool mEnvironment::containsParentAssignment()
{
  if(hasParentAssignment) return true;

 // check all children
 for(mEnvironment *env = children->first(); env != 0; env = children->next())
    if( env->containsParentAssignment() ) return true;

 return false;
}






