#ifndef __MLIST__H
#define __MLIST__H


template<class T>
class mList{
      public:
             /*!Constructs new (empty) list of size 0.*/
             mList(){ First=Last=Current=0; Size = MaxSize = 0; autoDel = false; }
             /*!Constructs new list with specified size.*/
             mList(unsigned s) { First=Last=Current=0; Size=0; MaxSize=s; autoDel = false; }
             /*!Copy constructor.*/
             mList(mList const &s){
               First=Last=Current=0;
               Size = 0;
               for(unsigned i=0;i<s.size();i++) append(s.at(i));
	       Size=s.size();
	       MaxSize=s.maxSize();
	       autoDel=s.autoDeletion();
             }
             /*!Assignment operator.*/
             mList& operator=(mList const &s){
               if(this != &s){
                  clear();
                  MaxSize = s.maxSize();
                  autoDel = s.autoDeletion();
                  for(unsigned i=0;i<s.size();i++) append(s.at(i));
               }
               return (*this);
             }
             /*!Returns class name. */
             virtual char const* className() const { return "mObject.mAbstractCollection.mList"; }
             /*!Destructor.*/
             virtual ~mList(){ clear(); }
             /*!Creates new instance of this object. */
             virtual mList& clone() { return (*(new mList(*this))); }
             /*!Sets whether objects should be deleted when removed from collection.
             Default is false.*/
             void setAutoDelete(bool on) { autoDel = on; }
             /*!Returns true if auto deletion enabled.*/
             bool autoDeletion() const { return autoDel; }
             /*!Returns size (# elements) of this collection.*/
             unsigned size() const { return Size; }
             /*!Returns max size or 0 if unlimited.*/
             unsigned maxSize() const { return MaxSize; }
             /*!Returns true if collection is empty.*/
             bool empty() const { return Size == 0; }             
             
             /*!Deletes all elements.*/
             virtual void clear(){
               item *it = First;
               while(it){
                     item *next = it->next;
                     removeItem(it);
                     it = next;
               }
               Size = 0;
               First=Last=Current=0;
             }
             /*!Adds element to the end of list. Returns true if successful.*/
             inline bool append(T*);
             /*!Adds element to the beginning of list. Returns true if successful.*/
             inline bool prepend(T *e){ return insert(0,e); }
             /*!Adds element at the specified location. Returns true if successful.
              Throws mOutOfBoundsException if index is out of range.*/
             inline bool insert(unsigned,T*);
             /*!Removes element and returns true if successful. Throws mEmptyCollectionException if list empty.*/
             inline bool remove(T*);

	     /*!Replaces element at specified position with new one. Returns true if successful.*/
             inline bool replace(unsigned,T*);

             /*!Removes element at specified position and returns true if successful.
               Throws mEmptyCollectionException if list empty, mOutOfBoundsException if
               index out of range. Index can range from 0 to size()-1 inclusive.*/
             inline bool remove(unsigned);
             /*!Returns true if element exists.*/
             bool exists(T const*);
             /*!Returns index of element or -1 if it doesn't exist.*/
             int find(T const*);
             /*!Returns element at specified position. Throws mEmptyCollectionException if list empty,
               mOutOfBoundsException if index is out of range. Index can range from 0 to size()-1 inclusive.*/
             T* at(unsigned) const;
             /*!Returns current item.*/
             T* current() { return Current; }
             /*!Returns 1st element and makes it the current item.*/
             T* first() { if(!First) return 0; Current=First; return First->i; }
             /*!Returns last element and makes it the current item.*/
             T* last() { if(!Last) return 0; Current=Last; return Last->i; }
             /*!Returns next element and makes it the current item.*/
             T* next(){
                if(!Current) return 0;
                item *i = Current->next;
                Current = i;
                if(!i) return 0;
                return i->i;
             }
             /*!Returns previous element and makes it the current item.*/
             T* prev(){
                if(!Current) return 0;
                item *i = Current->prev;
                Current = i;
                if(!i) return 0;
                return i->i;
             }
      private:
             struct item{
                    T    *i;
                    item *next,*prev;
             };
             
             item *First;          // 1st item
             item *Last;           // last
             item *Current;
             
             void removeItem(item*);

	     unsigned Size;
             unsigned MaxSize;
             bool     autoDel;
             
};

template<class T>
bool mList<T>::append(T *e)
{
 if(MaxSize > 0 && Size >= MaxSize) return false;

 if(Size == 0){     // 1st element
    First = new item;
    First->i = e; First->next = First->prev = 0;
    Last = Current = First;
 }else{
   item *it = new item;
   it->i = e; it->next = 0; it->prev = Last;
   Last->next = it;
   Last = it;
 }
 Size++;
 return true;
}

template<class T>
bool mList<T>::insert(unsigned index,T *e)
{
 if(MaxSize > 0 && Size >= MaxSize) return false;
 
 item *it = new item;
 it->i = e;

 if(Size==0 || index==Size){
    return append(e);
 }else{
    unsigned cur=0;
    item *f = First;
    while(cur++ < index) f = f->next;
    it->next = f;
    it->prev = f->prev;
    if(f->prev) f->prev->next=it;
    f->prev = it;
    if(index==0) First=it;
    if(index==1) First->next=it;
    if(index==Size-1) Last->prev=it;
 }
 
 Size++;
 return true;
}


template<class T>
inline bool mList<T>::replace(unsigned index,T* e)
{
  unsigned cur=0;
  item *f = First;
  while(cur++ < index) f = f->next;
 
  f->i = e;
  return true;
}

template<class T>
bool mList<T>::remove(T *e)
{
 item *f = First;
 while(f->i != e)
       if(f->next) f=f->next;
       else return false;

 // reassign prev,next
 if(f->prev) f->prev->next = f->next;
 else First = f->next;  // deleting First
 if(f->next) f->next->prev = f->prev;
 else Last = f->prev;  // deleting Last

 removeItem(f);
 Size--;
 
 return true; 
}

template<class T>
bool mList<T>::remove(unsigned i)
{
 
 item *f = First;
 unsigned cur=0;
 while(cur++ != i)
       if(f->next) f=f->next;
       else return false;

 // reassign prev,next
 if(f->prev) f->prev->next = f->next;
 else First = f->next;  // deleting First
 if(f->next) f->next->prev = f->prev;
 else Last = f->prev;  // deleting Last

 removeItem(f);
 Size--;
 
 return true; 
}

template<class T>
void mList<T>::removeItem(item *it)
{
 if(!it) return;
 if(autoDel) delete it->i;
 delete it;
}

template<class T>
bool mList<T>::exists(T const *i)
{
 if(Size == 0) return false;
 item *it = First; 
 while(it){
       item *next = it->next;
       if(it->i == i) return true;
       it = next;
 }
 return false;
}

template<class T>
int mList<T>::find(T const *e)
{
 if(Size == 0) return -1;
 item *it = First;
 int cur=0;
 while(it){
       item *next = it->next;
       if(it->i == e) return cur;
       it = next; cur++;
 }
 return -1;
}

template<class T>
T* mList<T>::at(unsigned i) const
{
 unsigned cur=0;
 item *f = First;
 while(cur++ < i) f = f->next;
 
 return f->i;
}

#endif
