#include "mLiteral.h"
#include "mEnvironment.h"
#include "mDefs.h"
#include "mClass.h"
#include <iostream>

using namespace std;

mLiteral::mLiteral(int value, mClass *Type)
{
 theClass = Type;
 multiple = false;

 symbolType = SYMBOL_LITERAL;   // this is literal

 special = false;

 // generate name from type and value by checking all integer types
 name = new mString("const_");
 
 
 if(*Type->name == "Integer"){    // Integer

    *name += "i_";
    intValue = value;
    mString val((int)value);
    *name += val;  

 }else if(*Type->name == "Boolean"){  // Boolean
    *name += "b_";
    intValue = value;
    mString val((int)value);
    *name += val;  
 }


 // char, short, byte ,... and possibly more

 stringValue = 0;
}


mLiteral::mLiteral(double value, mClass *Type)
{
 theClass = Type;
 multiple = false;

 symbolType = SYMBOL_LITERAL;   // this is literal

 special = false;

 // generate name from type and value by checking all built-in types
 name = new mString("const_");
 
 if(*Type->name == "Float"){
    *name += "f_";
    floatValue = value;
    mString val((float)value);
    *name += val;

 }else if(*Type->name == "Double"){
    *name += "d_";
    doubleValue = value;
    mString val((double)value);
    *name += val;
 }

 stringValue = 0;
 
}

mLiteral::mLiteral(mString *value, mClass *Type, bool spec)
{
 theClass = Type;
 multiple = false;

 symbolType = SYMBOL_LITERAL;   // this is literal

 special = spec;

 // generate string name from type and value
 name = new mString("const_");
 
 if( special ) *name += "special_";
 *name += "s_";
 stringValue = value;
 *name += *value; 
}


mLiteral::mLiteral(mMethod *method, mClass* Type, mClass *ownerClass)
{

  theClass = Type;
  multiple = false;
  special = false;

  symbolType = SYMBOL_LITERAL;   // this is literal
  
  
  // generate string name from method name, its class and arguments
  name = new mString("Z__");

  // add class name (once fully qualified class name)
  *name += *ownerClass->name;

  *name += "__";

  // add method name
  *name += *method->name;

  *name += "__";

  // add number of arguments
  *name += method->numParameters();

  // add types of all arguments
  for(mSymbol *l = method->parameterList->first(); l != 0; l = method->parameterList->next()){

    *name += "__";
    *name += *l->theClass->name;    
    
  }

  // add '0' to reserve space for NULL (method name literal is stored as null terminated string)
  *name += "0";
    
  stringValue = new mString( *name );
}



mLiteral::~mLiteral()
{
 delete name;
 //delete type;

 if(stringValue) delete stringValue;
}

mString mLiteral::toString()
{
 mString s;
 s += *theClass->name;
 s += " literal '";
 s += *name;
 s += "' ";
 s += "(offset ";
 s += offset;

 if( multiple ){
   s += ", multi ";
 }

 s += ")";

 
 
 return s;
}


void mLiteral::calculateLiteralAddresses(mList<mSymbol>* list)
{
  int currentOffset = 0;

  for(mSymbol *l = list->first(); l != 0; l = list->next()){
      mLiteral *lit = (mLiteral*)l;

      lit->offset = currentOffset;
      mString *name = lit->theClass->name;

      if( *name == "Integer" || *name == "Float" ) currentOffset += 4;
      else if( *name == "Double" || *name == "Long" ) currentOffset += 8;
      else if( *name == "Boolean" || *name == "Character" || *name == "Byte" ) currentOffset += 1;
      else if( *name == "Short" ) currentOffset += 2;
      else if( *name == "String" ) currentOffset += lit->length();

  }

}

mLiteral* mLiteral::find(mString* lname, mList<mSymbol>* list)
{
 for(mSymbol *l = list->first(); l != 0; l = list->next())
     if(*l->name == *lname) return (mLiteral*)l;

 return 0;
}


mLiteral* mLiteral::calculate(mLiteral* l1, mLiteral* l2, int operation, 
			      mList<mSymbol>* list, mEnvironment* env, mClass* returnType )
{

  mLiteral *result = 0;


  // check single literal case
  if( l2 == 0 ){

    if( *l1->theClass->name == "Integer" ){

      int value1 = l1->intValue;    // int value of 1st literal    
      
      switch(operation){
      case NODE_INCREMENT:  
	result = new mLiteral( value1+1, l1->theClass  );
	break;
      case NODE_DECREMENT:  
	result = new mLiteral( value1-1, l1->theClass  );
	break;	        
      }  // switch
      
    }else if( *l1->theClass->name == "Double" ){ 
      double value1 = l1->doubleValue;
      
      switch(operation){
      case NODE_INCREMENT:  
	result = new mLiteral( (double)value1+1.0, l1->theClass  );
	break;
      case NODE_DECREMENT:  
	result = new mLiteral( (double)value1-1.0, l1->theClass  );
	break;    
	
      }  // switch
      
      
    }

  }else{

  /////////////////////////////////////////////////////////
  // 1st literal is Integer
  /////////////////////////////////////////////////////////
  if( *l1->theClass->name == "Integer" ){

    int value1 = l1->intValue;    // int value of 1st literal

    // second can be integer, double, float, ...

    if( *l2->theClass->name == "Integer" ){ // Integer op Integer => Integer
      int value2 = l2->intValue;

      switch(operation){
        case NODE_PLUS:  
	  result = new mLiteral( value1+value2, l1->theClass  );
	  break;
	case NODE_MINUS:  
	  result = new mLiteral( value1-value2, l1->theClass  );
	  break;
	case NODE_MULTIPLY:  
	  result = new mLiteral( value1*value2, l1->theClass  );
	  break;  
        case NODE_DIVIDE:
	  if( value2 == 0 ) value1 = 0;
	  result = new mLiteral( value1/value2, l1->theClass  );
	  break;
        case NODE_MODULE:
	  if( value2 == 0 ) value1 = 0;
	  result = new mLiteral( value1 % value2, l1->theClass  );
	  break;
        case NODE_LESS:
	  result = new mLiteral( value1 < value2 ? 1 : 0, returnType  );
	  break;
	case NODE_GREATER:
	  result = new mLiteral( value1 > value2 ? 1 : 0, returnType  );
	  break;
        case NODE_EQUALS:
	  result = new mLiteral( value1 == value2 ? 1 : 0, returnType  );
	  break;
        case NODE_NOT_EQUALS:
	  result = new mLiteral( value1 != value2 ? 1 : 0, returnType  );
	  break;
	case NODE_GREATER_OR_EQUALS:
	  result = new mLiteral( value1 >= value2 ? 1 : 0, returnType  );
	  break;
	case NODE_LESS_OR_EQUALS:
	  result = new mLiteral( value1 <= value2 ? 1 : 0, returnType  );
	  break;
        

      }  // switch

    }else if( *l2->theClass->name == "Double" ){ // Integer op Double => Double
      double value2 = l2->doubleValue;

      switch(operation){
        case NODE_PLUS:  
	  result = new mLiteral( (double)value1+value2, l2->theClass  );
	  break;
	case NODE_MINUS:  
	  result = new mLiteral( (double)value1-value2, l2->theClass  );
	  break;
	case NODE_MULTIPLY:  
	  result = new mLiteral( (double)value1*value2, l2->theClass  );
	  break;  
        case NODE_DIVIDE:
	  if( value2 == 0 ) value1 = 0;
	  result = new mLiteral( (double)value1/value2, l2->theClass  );
	  break;
        case NODE_LESS:
	  result = new mLiteral( value1 < value2 ? 1 : 0, returnType  );
	  break;
	case NODE_GREATER:
	  result = new mLiteral( value1 > value2 ? 1 : 0, returnType  );
	  break;
        case NODE_EQUALS:
	  result = new mLiteral( value1 == value2 ? 1 : 0, returnType  );
	  break;
        case NODE_NOT_EQUALS:
	  result = new mLiteral( value1 != value2 ? 1 : 0, returnType  );
	  break;
	case NODE_GREATER_OR_EQUALS:
	  result = new mLiteral( value1 >= value2 ? 1 : 0, returnType  );
	  break;
	case NODE_LESS_OR_EQUALS:
	  result = new mLiteral( value1 <= value2 ? 1 : 0, returnType  );
	  break;      

      }  // switch

    }else if( *l2->theClass->name == "String" ){ // Integer op String => String
      mString* value2 = l2->stringValue;

      switch(operation){
        case NODE_PLUS: {
	  mString* str = new mString(value1); *str += *value2;
	  result = new mLiteral( str, l2->theClass  );
        }
	  break;

	  //default:
	  // strings can only be added, everything else is an error
	  //return 0;
	  //break;

      }  // switch
    }

  /////////////////////////////////////////////////////////
  // 1st literal is Double
  /////////////////////////////////////////////////////////
  }else if( *l1->theClass->name == "Double" ){

    double value1 = l1->doubleValue;    // double value of 1st literal

    // second can be integer, double, float, ...

    if( *l2->theClass->name == "Integer" ){ // Double op Integer => Double
      int value2 = l2->intValue;

      switch(operation){
        case NODE_PLUS:  
	  result = new mLiteral( value1+value2, l1->theClass  );
	  break;
	case NODE_MINUS:  
	  result = new mLiteral( value1-value2, l1->theClass  );
	  break;
	case NODE_MULTIPLY:  
	  result = new mLiteral( value1*value2, l1->theClass  );
	  break;  
        case NODE_DIVIDE:
	  if( value2 == 0 ) value1 = 0;
	  result = new mLiteral( value1/value2, l1->theClass  );
	  break;
        case NODE_LESS:
	  result = new mLiteral( value1 < value2 ? 1 : 0, returnType  );
	  break;
	case NODE_GREATER:
	  result = new mLiteral( value1 > value2 ? 1 : 0, returnType  );
	  break;
        case NODE_EQUALS:
	  result = new mLiteral( value1 == value2 ? 1 : 0, returnType  );
	  break;
        case NODE_NOT_EQUALS:
	  result = new mLiteral( value1 != value2 ? 1 : 0, returnType  );
	  break;
	case NODE_GREATER_OR_EQUALS:
	  result = new mLiteral( value1 >= value2 ? 1 : 0, returnType  );
	  break;
	case NODE_LESS_OR_EQUALS:
	  result = new mLiteral( value1 <= value2 ? 1 : 0, returnType  );
	  break;

      }  // switch

    }else if( *l2->theClass->name == "Double" ){ // Double op Double => Double
      double value2 = l2->doubleValue;

      switch(operation){
        case NODE_PLUS:  
	  result = new mLiteral( (double)value1+value2, l2->theClass  );
	  break;
	case NODE_MINUS:  
	  result = new mLiteral( (double)value1-value2, l2->theClass  );
	  break;
	case NODE_MULTIPLY:  
	  result = new mLiteral( (double)value1*value2, l2->theClass  );
	  break;  
        case NODE_DIVIDE:
	  if( value2 == 0 ) value1 = 0;
	  result = new mLiteral( (double)value1/value2, l2->theClass  );
	  break;
        case NODE_LESS:
	  result = new mLiteral( value1 < value2 ? 1 : 0, returnType  );
	  break;
	case NODE_GREATER:
	  result = new mLiteral( value1 > value2 ? 1 : 0, returnType  );
	  break;
        case NODE_EQUALS:
	  result = new mLiteral( value1 == value2 ? 1 : 0, returnType  );
	  break;
        case NODE_NOT_EQUALS:
	  result = new mLiteral( value1 != value2 ? 1 : 0, returnType  );
	  break;
	case NODE_GREATER_OR_EQUALS:
	  result = new mLiteral( value1 >= value2 ? 1 : 0, returnType  );
	  break;
	case NODE_LESS_OR_EQUALS:
	  result = new mLiteral( value1 <= value2 ? 1 : 0, returnType  );
	  break;

      }  // switch

    }else if( *l2->theClass->name == "String" ){ // Double op String => String
      mString* value2 = l2->stringValue;

      switch(operation){
        case NODE_PLUS: {
	  mString* str = new mString(value1); *str += *value2;
	  result = new mLiteral( str, l2->theClass  );
        }
	  break;

	default:
	  // strings can only be added, everything else is an error
	  return 0;
	  break;

      }  // switch
    }  

  /////////////////////////////////////////////////////////
  // 1st literal is String
  /////////////////////////////////////////////////////////
  }else if( *l1->theClass->name == "String" ){

    mString *value1 = l1->stringValue;    // string value of 1st literal

    // second can be integer, double, float, ...

    if( *l2->theClass->name == "Integer" ){ // String op Integer => String
      int value2 = l2->intValue;

      switch(operation){
        case NODE_PLUS: {
	  mString* str = new mString(*value1); *str += value2;
	  result = new mLiteral( str, l1->theClass  );
	}
	  break;

	default:
	  // strings can only be added, everything else is an error
	  return 0;
	  break;

      }  // switch

    }else if( *l2->theClass->name == "Double" ){ // String op Double => String
      double value2 = l2->doubleValue;

      switch(operation){
        case NODE_PLUS: {
	  mString* str = new mString(*value1); *str += value2;
	  result = new mLiteral( str, l1->theClass  );
        }
	  break;

      default:
	  // strings can only be added, everything else is an error
	  return 0;
	  break;

      }  // switch

    }else if( *l2->theClass->name == "String" ){ // String op String => String
      mString* value2 = l2->stringValue;

      switch(operation){
        case NODE_PLUS: {
	  mString* str = new mString(*value1); *str += *value2;
	  result = new mLiteral( str, l1->theClass  );
        }
	  break;

      default:
	  // strings can only be added, everything else is an error
	  return 0;
	  break;

      }  // switch

    }else if( *l2->theClass->name == "Boolean" ){ // String op Boolean => String
      int value2 = l2->intValue;

      switch(operation){
        case NODE_PLUS: {
	  mString bval("true");
	  if(value2 == 0) bval="false";
	  mString* str = new mString(*value1); *str += bval;
	  result = new mLiteral( str, l1->theClass  );
        }
	  break;

      default:
	  // strings can only be added, everything else is an error
	  return 0;
	  break;

      }  // switch

    }


  /////////////////////////////////////////////////////////
  // 1st literal is Boolean
  /////////////////////////////////////////////////////////
  }if( *l1->theClass->name == "Boolean" ){

    int value1 = l1->intValue;    // int value of 1st literal

    // second can only be Boolean or String

    // if there is no second literal then we have a single boolean expression
    if( l2 == 0){ // single Boolean


      // !boolean expression
      if(operation == NODE_NOT)
	  result = new mLiteral( !value1, returnType  );
      

    } else if( *l2->theClass->name == "Boolean" ){ // Boolean op Boolean => Boolean

      int value2 = l2->intValue;

      switch(operation){
	/*
        case NODE_PLUS: {
	  mString bval("true");
	  if(value2 == 0) bval="false";
	  mString* str = new mString(*value1); *str += bval;
	  result = new mLiteral( str, l1->theClass  );
        }
	break;*/

	case NODE_AND: {
	  // when ANDing both values must be true if the result should be set to true
	  result = new mLiteral( (value1 == 1) && (value2 == 1) ? 1 : 0, returnType  );
        }
	break;

	case NODE_OR: {
	  // when ORing at least one of the values must be true if the result should be set to true
	  result = new mLiteral( (value1 == 1) || (value2 == 1) ? 1 : 0, returnType  );
        }
	break;	
	

      default:
	  // Booleans can only be logically compared, everything else is an error
	  return 0;
	  break;

      }  // switch      



    }else if( *l2->theClass->name == "String" ){ // Boolean op String => String
      mString* value2 = l2->stringValue;

      switch(operation){
	
        case NODE_PLUS: {
	  mString bval("true");
	  if(value1 == 0) bval="false";
	  mString* str = new mString(bval); *str += *value2;
	  result = new mLiteral( str, l2->theClass  );
        }
	break;

	

      default:
	  // Booleans and String can only be added, everything else is an error
	  return 0;
	  break;

      }  // switch      

    }

  }

  }
  // if no appropriate pairs have been found then types don't match and we have
  // an error
  if( !result ) return 0;


  // now that we have a new literal, we must remove old ones if they're not used
  // more than once
  mLiteral *f1 = find( l1->name, list );
  
  if( !f1->multiple ){ list->remove( l1 ); env->removeLiteral( l1 );  }


  if(l2 != 0){
    mLiteral *f2 = find( l2->name, list );
    if( !f2->multiple ){ list->remove( l2 ); env->removeLiteral( l2 );  }
  }


  // store result to literal list and environment
  mLiteral* lit = mLiteral::find( result->name, list );

  if( lit != 0 ){  // already exists?
    delete result; result = lit; result->multiple = true;
  }else
    // append it to list
    list->append( result );

  env->append( result );

  return result;
}


int mLiteral::length()
{
  if(*theClass->name == "Integer" || *theClass->name == "Float")
    return 4;
  else if(*theClass->name == "Double" || *theClass->name == "Long")
    return 8;
  else if(*theClass->name == "Short")
    return 2;
  else if(*theClass->name == "Byte" || *theClass->name == "Character" || *theClass->name == "Boolean")
    return 1;
  else if(*theClass->name == "String")
    return stringValue->length();

  return 0;
}
