#ifndef MLITERAL_H
#define MLITERAL_H

#include "mString.h"
#include "mSymbol.h"

class mEnvironment;

class mClass;
class mMethod;

// literal
class mLiteral : public mSymbol{
      public:
             // int value, desired type, offset
             mLiteral(int, mClass*);
             // double value, desired type, offset
             mLiteral(double, mClass*);
             // string value, desired type, special flag
	     // note: special flag is used for 'special' literals such as
	     // the nil
             mLiteral(mString*, mClass*, bool=false);

	     // generates literal from method (this is used for native methods)	     
	     // method, desired type, owner class of method
             mLiteral(mMethod*, mClass*, mClass*);

             virtual ~mLiteral();

             // returns string representation of this literal
             virtual mString toString();

	     // calculates literal addresses in constant pool
	     // this should be called after parse and before code generation
	     // begins
	     static void calculateLiteralAddresses(mList<mSymbol>*);

             // returns literal from list with the matching name or 0 if not
             // found
             static mLiteral* find(mString*,mList<mSymbol>*);

	     // calculates the result of two literals (primitive types) and returns new
	     // literal of appropriate type by using simple integral promotion
	     // 0 is returned if calculation cannot be performed (type mismatch)
	     static mLiteral* calculate(mLiteral*, mLiteral*, int, mList<mSymbol>*,
					mEnvironment*, mClass* = 0);

	     // returns true if this literal is nil
	     bool isNil(){ return *name == "const_special_s_nil"; }

	     // returns length of this literal in bytes
	     // i.e. integer,float 4, double 8, string length
	     int length();

             // change to mClass later!!!
             //mString *type;

             // offset in constant pool
             int offset;

	     // true if used multiple times
	     bool multiple;

	     // true if this is special literal (nil,...)
	     bool special;
             
             // possible values (char and short could be taken by int, but
             // we declare them separately for clarity
             char    charValue;
             short   shortValue;
             int     intValue;
             float   floatValue;
             double  doubleValue;
             mString *stringValue;

};

#endif

