#include "mMethod.h"
#include <iostream>
#include "mClass.h"


mMethod::mMethod(mString *Name, mClass *Return, short Access, short Bind, short op)
{
     name = Name;
     returnType = Return;
     access = Access;
     bind = Bind;
     builtinOpcode = op;
     symbolType = SYMBOL_METHOD;   // this is method
             
     // create parameter list
     parameterList = new mList<mParameter>();
     // environment will delete parameters
     parameterList->setAutoDelete( false );
     environment = 0;

     // default: no late return type
     lateReturnType = 0;

     native = false;   // false by default

     returnsSelf = false;
           
}

mMethod::~mMethod()
{
 delete parameterList;
 delete lateReturnType;

}


void mMethod::addParameter(mParameter *p)
{
 parameterList->append ( p );
}

mString mMethod::toString()
{
 mString s;
 s += access2String( access );
 // constructors and destructors have no explicit return types, so
 // we don't include them
 if( returnType ){
     s += " ";
     s += *returnType->name;
 }else{
     s += " ";
     s += "void";
 }
 
 s += " method ";
 s += *name;
 s += " ( ";
 
 // parameters
 for(mParameter *p = parameterList->first(); p != 0; p = parameterList->next()){
     s += p->toString();
     s += ",";
 }
 
 s = s.truncate( s.length() - 1 );

 s += ")";
 
 return s;
}


mString mMethod::toStringDeclaration()
{
 mString s;
 // constructors and destructors have no explicit return types, so
 // we don't include them
 if( returnType ){
     s += " ";
     s += *returnType->name;
 }
 
 s += *name;
 
 if(parameterList->size() > 0){
   s += " ( ";
 
   // parameters
   for(mParameter *p = parameterList->first(); p != 0; p = parameterList->next()){
     s += *p->theClass->name;
     s += ",";
   }
 
   s = s.truncate( s.length() - 1 );

   s += ")";
 }
 
 return s;
}



bool mMethod::isDefaultConstructor()
{
  return ( *name == *theClass->name && parameterList->size() == 0 );
} 

bool mMethod::isConstructor()
{ 
  return *name == *theClass->name; 
}


bool mMethod::matchesSignature(mMethod* other)
{
  // names and number of parameters must match
  if( *name != *other->name || (int)parameterList->size() != other->numParameters() ) return false;


  // parameters too
  for( unsigned i=0; i < parameterList->size(); i++ )

      if( *parameterList->at( i )->theClass->name != 
	  *other->parameterList->at( i )->theClass->name ) return false;

  return true;

}


bool mMethod::mayOverride(mMethod* other)
{
  // names and number of parameters must match
  return  *name == *other->name && (int)parameterList->size() == other->numParameters();
}


bool mMethod::isSubtypeOf(mMethod* other)
{

  // number of parameters are already equal

  // first we check if return types are compatible: return of this method must be a subtype of other
 if( (other->isFunction() && !this->isFunction()) || (!other->isFunction() && this->isFunction()))
    return false;


  if( other->isFunction() ){

    if( !returnType->isSubtype(other->returnType) ) return false;

  }
      

  // now we check types of parameters which may vary contravariantly
  for( unsigned i=0; i < parameterList->size(); i++ )

      if( !other->parameterList->at( i )->theClass->isSubtype( 
	  parameterList->at( i )->theClass) )

	return false;


  return true;

}


int mMethod::getSelfLocation()
{
  if( *returnType->name == "Void" )
    return 3;   // for procedures
  else 
    return 4;   // for functions

}

bool mMethod::isFunction()
{
  // primitive 
  if(returnType == 0) return false;
 
  // constructor has no-void return type but is not considered a function
  return *returnType->name != "Void" && !isConstructor(); 
}


bool mMethod::isAccessibleWithin(mClass* cl)
{
  // method is always accessible in its own class regardless of access modifier
  // the same goes if it was declared with public or restricted modifier. it it available in ANY class
  // note: restricted is the same as public for accessing, it only has meaning in method update
  if( theClass == cl || access == ACCESS_PUBLIC || access == ACCESS_RESTRICTED ) return true;


  // if method was declared private it is only available in its own class
  if( access == ACCESS_PRIVATE && theClass == cl ) return true;
  
  
  // if it was declared protected, it is only accessible in its own class and all subclasses

  // check if in its own class
  if( access == ACCESS_PROTECTED && theClass == cl ) return true;

  // check if cl is subclass of method's class
  if( access == ACCESS_PROTECTED && cl->isSubclass( theClass ) ) return true;


  // otherwise the method is not accessible in class cl
  return false;
  
}


bool mMethod::isAccessibleOutside(mClass* cl)
{
  // method is accessible only if it is declared public
  return  access == ACCESS_PUBLIC; 
  
}


bool mMethod::parameterTypesMatch(mList<mParameter>* plist)
{

  if(!plist && (int)parameterList->size() > 0) return false;
  if(!plist && (int)parameterList->size() == 0) return true;

  // number of parameters must match
  if( (int)parameterList->size() != (int)plist->size() ) return false;


  for( unsigned i=0; i < parameterList->size(); i++ )

      if( parameterList->at( i )->theClass != 
	  plist->at( i )->theClass ) return false;

  return true;

}
