#ifndef M_METHOD__H
#define M_METHOD__H

#include "mString.h"
#include "mParameter.h"
#include "mDefs.h"
#include "mInstruction.h"
#include "mSymbol.h"
#include "mEnvironment.h"

class mClass;
class mEnvironment;

// method
class mMethod : public mSymbol{
      public:
      // name, return type, access, binding, opcode
      // if opcode is present then this is a builtin type method such as
      // Integer.print, String.print, etc.
      mMethod(mString*, mClass*, short, short, short=-1);
      
      virtual ~mMethod();

      // adds formal parameter
      void addParameter(mParameter*);

      // returns string representation of this method
      mString toString();

      // returns string representation of declaration form
      mString toStringDeclaration();
      
      // returns true if this is a builtin method f a primitive type
      bool isBuiltin(){ return builtinOpcode != -1; }

      // returns true if this method is a default (without parameters)
      // constructor
      bool isDefaultConstructor();

      // return true if this method is a constructor
      bool isConstructor();


      // returns true if this method is a function (i.e. has returnType)
      bool isFunction();

      // returns number of parameters
      int numParameters(){ return parameterList->size(); }

      // returns true if this method's signature matches other's
      bool matchesSignature(mMethod*);

      // returns true if this method matches other method in name and number of parameters 
      bool mayOverride(mMethod*);

      // returns true if this method is a submethod of other. return types may vary covariantly,
      // parameter types contravariantly, or invariantly with Self
      bool isSubtypeOf(mMethod*);

      // returns frame location of self
      int getSelfLocation();

      // returns true if method is accessible within specified class. 
      // method can have any access modifier if it's defined in this class, otherwise it must not
      // be private (in one of superclasses)
      bool isAccessibleWithin(mClass*);


      // returns true if method is accessible from outside of specified class.       
      bool isAccessibleOutside(mClass*);


      // returns true if this method is static
      bool isStatic(){ return bind == BIND_STATIC; }

      // returns true if this method is native
      bool isNative(){ return native; }

      // sets native to true
      void setNative(){ native = true; }

      
      // returns true if types of parameters of this method match with specified ones
      bool parameterTypesMatch(mList<mParameter>*);


      // adds the name of return type which will become available somewhere later
      // after the creation of method
      //void addLateReturnType(mString*);


      mClass *returnType;       // its returns type
      
      // note: owner class of this method is in 'theClass'
      

      short access;              // access modifier can be private, protected,
                                 // public and restricted

      short bind;                // binding modifier is static or nonstatic

      // list of formal parameters
      mList<mParameter> *parameterList;

      // environment of this method's definitions and statements
      mEnvironment *environment;

      int vtableIndex;          // index of this method in virtual table

      short builtinOpcode;      // meaningful only for built in methods of
                                // primitive types

      bool native;              // true if native

      // 'late' return type is the name of class of return type
      mString* lateReturnType;

      // true if function returns Self
      bool returnsSelf;

};


#endif

