#include "mMethodTable.h"
#include "mMethod.h"


mMethodTable::mMethodTable()
{
  // create virtual tables
  virtualTables = new mList<virtualTable>();
  virtualTables->setAutoDelete( true );

  // create method table
  methodTable = new mList<methodTableEntry>();
  methodTable->setAutoDelete( true );

  // create super list
  superTables = new mList<mMethodTable>();
  superTables->setAutoDelete( false );

  // we begin at 0
  startingIndex = 0;

}


mMethodTable::mMethodTable(mMethodTable* other)
{
  // create virtual tables
  virtualTables = new mList<virtualTable>();
  virtualTables->setAutoDelete( true );

  // create method table
  methodTable = new mList<methodTableEntry>();
  methodTable->setAutoDelete( true );

  // create super list
  superTables = new mList<mMethodTable>();
  superTables->setAutoDelete( false );

  startingIndex = other->startingIndex;

  // copy virtual tables
  for(virtualTable *vtab=other->virtualTables->first();vtab!=0;vtab=other->virtualTables->next())
    virtualTables->append( new virtualTable(vtab) );

  // copy method table
  for( methodTableEntry *entry=other->methodTable->first(); entry!=0; entry = other->methodTable->next())
    methodTable->append( new methodTableEntry(entry) );

  // super tables
  for(mMethodTable *vtab=other->superTables->first();vtab!=0;vtab=other->superTables->next())
    superTables->append( new mMethodTable(vtab) );
  

}

mMethodTable::~mMethodTable()
{
  delete virtualTables;
  delete methodTable;
  delete superTables;
}


mString mMethodTable::toString()
{
  mString s;
   
  for(mMethodTable *vtab=superTables->first();vtab!=0;vtab=superTables->next()){

    s += vtab->toString();

  }
  

  for(virtualTable *vtab=virtualTables->first();vtab!=0;vtab=virtualTables->next()){
    // classname
    s += " -- vtable for class "; s += *vtab->theClass->name;
    s += '\n';


    // the methods
    for(virtualTableEntry* entry = vtab->entries->first(); entry != 0;
	entry = vtab->entries->next()){

      s += "  --- index "; 
      s += entry->methodTableindex; s+= ", ";
      s += entry->method->toString();      
      s += '\n';
    }
  }

  // method table
  s += "\n  -- method table --\n";
  for( methodTableEntry *entry=methodTable->first(); entry!=0;
       entry = methodTable->next()){

    // zraven izpisemo se razred
    s += "    --- address ";
    s += entry->address;
    s += ", ";
    s += entry->method->toString();
    s += '\n';
  }
  s += "------------- end -------------\n";

  //}

  return s;
}


/******************************************************************************************/

void mMethodTable::addClass(mClass *c)
{
  // create new virtual table for this class
  virtualTable *vtable = new virtualTable( c, startingIndex );
  virtualTables->append( vtable );
}


/******************************************************************************************/


void mMethodTable::addMethod(mMethod *m)
{

  // create new method table entry for this method
  methodTableEntry *entry = new methodTableEntry;
  entry->method = m;
  entry->address = -1;   // address is unknown at this point

  // method must be added to method table
  methodTable->append( entry );

  // as well as virtual table for current class
  // (vtable for current class is last in list)
  virtualTables->last()->addMethod( m );

  for(mMethodTable *table = superTables->first(); table != 0; table = superTables->next())

    table->updateMethodIndex(m);
}



void mMethodTable::updateMethodIndex(mMethod *m)
{
    
  // find method in virtual table and change its index to index of m
  for(virtualTable *vtab=virtualTables->first();vtab!=0;vtab=virtualTables->next())

    vtab->updateMethodIndex(m);

  // now update in all supertables
  for(mMethodTable *table = superTables->first(); table != 0; table = superTables->next())

    table->updateMethodIndex(m);
}

/******************************************************************************************/


void mMethodTable::addSuper(mMethodTable *t)
{
  // table must be cloned because it may be modified
  mMethodTable *clone = new mMethodTable(t);


  // refresh indices in supertable 
  clone->updateIndices( numMethods() );

  superTables->append( clone );

  // increase starting index by method table size of super table
  startingIndex += t->numMethods();
  virtualTables->last()->startingIndex = startingIndex;
}

/******************************************************************************************/


void mMethodTable::updateIndices(int newStart)
{

  for(mMethodTable *tab = superTables->first(); tab != 0; tab = superTables->next())
    
      tab->updateIndices(newStart);    

  // now for this table
  for( virtualTable *vtab = virtualTables->first(); vtab !=0; vtab = virtualTables->next())
    
    for(virtualTableEntry* entry= vtab->entries->first(); entry != 0; entry = vtab->entries->next())
      entry->methodTableindex += newStart;

}

int mMethodTable::methodTableSizeOnly()
{

  int len = 0;  

  // + method table size
  len += methodTable->size() * 12;

  // + all supertables
  for(mMethodTable *tab = superTables->first(); tab != 0; tab = superTables->next())
    len += tab->methodTableSizeOnly();

  return len;

}


/******************************************************************************************/

int mMethodTable::size()
{

  // number of entries in virtual table * 4 +
  // number of entries in method table * 12 (we have 3 values for every method)

  int len = 0;
  for(virtualTable *vtab=virtualTables->first();vtab!=0;vtab=virtualTables->next())
    // number of entries in this vtable 
    len += vtab->entries->size() * 4;

  // + method table size
  len += methodTable->size() * 12;

  // + all supertables
  for(mMethodTable *tab = superTables->first(); tab != 0; tab = superTables->next())
    len += tab->size();

  return len + 12;   // 12 bytes is for entry in object table (4 bytes for object address, 
                    // 4 for selfish offset, 4 for mtable offset)
}



int mMethodTable::getMethodIndex(mMethod* method)
{
  // first check in this virtual table
  int count=0;
  for(virtualTable *vtab=virtualTables->first();vtab!=0;vtab=virtualTables->next())
    for(virtualTableEntry* entry= vtab->entries->first(); entry != 0; entry = vtab->entries->next(),count++)
      if(entry->method == method)
	
	return count;//entry->methodTableindex;

      

  // if no more supertables we return -1
  if(superTables->size() == 0) return -1;

  int index=0;
  // check supertables
  for(mMethodTable *tab = superTables->first(); tab != 0; tab = superTables->next()){
    index = tab->getMethodIndex(method);
    if(index != -1)
      return index;

  }

  return -1;
    
}


int mMethodTable::getStaticMethodIndex(mMethod* method, int& count)
{
  // check supertables
  for(mMethodTable *tab = superTables->first(); tab != 0; tab = superTables->next()){
    int index = tab->getStaticMethodIndex(method, count);
    if(index != -1)
      return index;
  }

  // search this method table
  for( methodTableEntry *entry = methodTable->first(); entry !=0;
       entry = methodTable->next(), count++)
    if( entry->method == method ) return count;
        

  return -1;
    
}



int mMethodTable::numMethods()
{

  // all parents
  int size = 0;
  for(mMethodTable *tab = superTables->first(); tab != 0; tab = superTables->next())
    size += tab->numMethods();

  // and our own
  return size + methodTable->size(); 

}


void mMethodTable::setMethodAddress(mMethod* m, int address)
{
  // find method in method table
  for( methodTableEntry *entry = methodTable->first(); entry !=0;
       entry = methodTable->next())
    if( entry->method == m ){ entry->address = address; break; }
}

int mMethodTable::getIndexOfMain()
{

  int startingIndex = methodTableSizeOnly()/12 - methodTableEntries();  

  int i=0;
  for( methodTableEntry *entry = methodTable->first(); entry !=0;
       entry = methodTable->next(), i++)
    if( *entry->method->name == "main" ) return i + startingIndex;

  return -1;
}


int mMethodTable::getDescriptor(mClass* c,int start)
{
          
  int offset = start;
  findDescriptor(c, offset);
  
  return offset;

}


bool mMethodTable::findDescriptor(mClass* c, int& start)
{

  // check supertables
  for(mMethodTable *tab = superTables->first(); tab != 0; tab = superTables->next()){
    if(tab->findDescriptor(c, start))
      return true;
  }


  for( virtualTable *vtab = virtualTables->first(); vtab !=0; vtab = virtualTables->next()){
    
    if(vtab->theClass == c) return true;    
    
    start += 8;  // descriptor is 8 bytes   
  }

  return false;

}



int mMethodTable::calculateVirtualTableOffsets(int start)
{

  // first come parent tables
  int count=0, currentStart = 0;  
  for(mMethodTable *tab = superTables->first(); tab != 0; tab = superTables->next(), count++){    

    //cout<<"# virtual table entries="<<tab->virtualTableEntries()<<", count="<<count<<endl;
    //currentStart = tab->virtualTableEntries() * 4;

    //if(count == 0)   // first stays the same
      start = tab->calculateVirtualTableOffsets(start);
    //else{
      //cout<<"calculating vtable offsets for table > 1 : starting at "<<start + currentStart
      //  <<", start="<<start<<", currentStart="<<currentStart<<endl;
      //start = tab->calculateVirtualTableOffsets(start + currentStart);    

    //}

  }

  // now for this table
  for( virtualTable *vtab = virtualTables->first(); vtab !=0; vtab = virtualTables->next()){
    
    vtab->selfishOffset = start;

    //if(advance)
    start += vtab->entries->size() * 4;
      
  }

  return start;

}



int mMethodTable::calculateMethodTableOffsets(int start)
{

  // first come parent tables
  for(mMethodTable *tab = superTables->first(); tab != 0; tab = superTables->next())    
    start = tab->calculateMethodTableOffsets(start);

  // now for this table
  for( virtualTable *vtab = virtualTables->first(); vtab !=0; vtab = virtualTables->next()){
    
    vtab->mtableOffset = start;

    // 12 bytes per entry in method table
    start += vtab->entries->size() * 12;
      
  }

  return start;

}


void mMethodTable::setObjectTableIndex(int& index)
{

  for(mMethodTable *tab = superTables->first(); tab != 0; tab = superTables->next())    
    tab->setObjectTableIndex(index);

  for( virtualTable *vtab = virtualTables->first(); vtab !=0; vtab = virtualTables->next()){

    vtab->theClass->cTableIndex = index;
    index++;
  }

}


/***********************************************************************************
 ***********************************************************************************
*/

void virtualTable::addMethod(mMethod *m)
{  
  virtualTableEntry *entry = new virtualTableEntry;
  entry->method = m;
  entry->methodTableindex = startingIndex + entries->size();
  // update index in method
  m->vtableIndex = entry->methodTableindex;  

  entries->append( entry );
}


void virtualTable::updateMethodIndex(mMethod *m)
{
       
  for(virtualTableEntry* entry= entries->first(); entry != 0; entry = entries->next()){

    if(m->mayOverride( entry->method)){      

      entry->methodTableindex = m->vtableIndex;
      entry->method->vtableIndex = m->vtableIndex;
    }


  }

}


