#ifndef M_METHOD_TABLE_H
#define M_METHOD_TABLE_H

#include "mString.h"
#include "mList.h"

class mMethod;
class mClass;


class mEnvironment;

// entry in method table
class methodTableEntry{
 public:
  methodTableEntry(){}
  methodTableEntry(methodTableEntry* other){

    method = other->method;
    address = other->address;
  }

  //int index;       // index of method
  mMethod* method;
  int address;       // offset in code

};


// each entry in vtable consists of method and index into method table
class virtualTableEntry{
 public:
  virtualTableEntry(){}
  virtualTableEntry(virtualTableEntry* other){
    method = other->method;
    methodTableindex = other->methodTableindex;    
  }

  mMethod* method;             // the method
  int methodTableindex;        // index of this method in method table
};


// virtual table consists of virtual table entries..
class virtualTable{
 public:
  virtualTable(mClass *c, int start){
    theClass = c;
    startingIndex = start;
    entries = new mList<virtualTableEntry>();
    entries->setAutoDelete( true );
  }

  virtualTable(virtualTable* other){

    theClass = other->theClass;
    startingIndex = other->startingIndex;
    selfishOffset = other->selfishOffset; 
    mtableOffset = other->mtableOffset; 

    entries = new mList<virtualTableEntry>();
    entries->setAutoDelete( true );
    
    for(virtualTableEntry* e = other->entries->first(); e != 0; e = other->entries->next())

      entries->append( new virtualTableEntry(e) );

  }
  
  // adds method
  void addMethod(mMethod*);

  void updateMethodIndex(mMethod*);

  virtual ~virtualTable(){
    delete entries;
  }
  mClass* theClass;     // the class this vtable belongs to
  mList<virtualTableEntry> *entries;

  // offset into common virtual table. this is written into object table
  int selfishOffset;

  // offset into common method table. this is also written into object table
  int mtableOffset;

  int startingIndex;
};



// this is the method table. method table contains addresses of all class methods.
class mMethodTable{
      public:
             // construts empty method table
             mMethodTable();
             
	     mMethodTable(mMethodTable*);

             virtual ~mMethodTable();

             // returns string representation of this table
             virtual mString toString();

	     // adds new class and creates virtual table for it
	     void addClass(mClass*);

	     // adds new class method
	     void addMethod(mMethod*);

	     // adds method table to this one (used for adding method tables of superclasses)
	     void addSuper(mMethodTable*);	     

	     // returns size in bytes. this includes object table, all virtual tables and
	     // method table
	     int size();

	     // returns index of specified method
	     int getMethodIndex(mMethod*);

	     // returns index of specified static method
	     int getStaticMethodIndex(mMethod*, int&);

	     // sets index into class(object) table for this class
	     void setObjectTableIndex(int&); 

	     // returns size of combined method tables
	     int methodTableSizeOnly();

	     // returns number of method entries including all super tables
	     int numMethods();

	     // sets the address of a specified method in method table
	     void setMethodAddress(mMethod*, int);

             // returns literal from list with the matching name or 0 if not
             // found
             //static mLiteral* find(mString*,mList<mSymbol>*);

	     // returns index of method main in method table. -1 if it doesn't exist
	     int getIndexOfMain();

	     // returns number of method table entries
	     int methodTableEntries(){
	       return methodTable->size();
	     }

	     // updates method index in supertables
	     void updateMethodIndex(mMethod*);

	     // returns offset of descriptor in object for specified class
	     int getDescriptor(mClass*,int);

	     // updates virtual table indicies in this methodtable's virtual table
	     // this is used to recalculate indices of supertable in 2nd and further places
	     // when inheriting multiple classes
	     void updateIndices(int);

	     // returns number of vtable entries
	     int virtualTableEntries(){
	       int size=0;
	       for( virtualTable* v=virtualTables->first(); v!=0; 
		   v=virtualTables->next() )

		 size += v->entries->size();

	       // super tables
	       for(mMethodTable *tab = superTables->first(); tab != 0; tab = superTables->next())
		 size += tab->virtualTableEntries();

	       return size;
	     }

	     // returns number of vtable entries for this class only
	     int thisVirtualTableEntries(){
	       int size=0;
	       for( virtualTable* v=virtualTables->first(); v!=0; 
		   v=virtualTables->next() )

		 size += v->entries->size();	       

	       return size;
	     }


	     // returns size of object table
	     int objectTableSize(){
	       int size=0;
	       // object table has the same number of entries as there are virtual tables in a class
	       for( virtualTable* v=virtualTables->first(); v!=0; 
		   v=virtualTables->next() )

		 size += 12;   // 12 bytes per entry in object table (class addx, vtab offs, mtab offs)

	       // super tables
	       for(mMethodTable *tab = superTables->first(); tab != 0; tab = superTables->next())
		 size += tab->objectTableSize();

	       return size;
	     }

	     // calculates offsets into virtual table for every class in hierarchy (selfish offsets)
	     // this method should be called after method table has been constructed
	     // argument is starting offset (for first table this is after class_data+object_table)
	     int calculateVirtualTableOffsets(int);

	     // calculates offsets into method table for every class in hierarchy
	     // this method should be called after method table has been constructed
	     // argument is starting offset (for first table this is after class_data+object_table+vtable)
	     int calculateMethodTableOffsets(int);

	     // for simplicity we separate vtables for each class in this hierarchy
	     mList<virtualTable> *virtualTables;
	     
	     mList<methodTableEntry> *methodTable;

	     // method tables for superclasses
	     mList<mMethodTable>* superTables;


	     // index at which methods for this portion of method table start
	     // i.e. if there exists a supertable with 3 methods for this table, this
	     // index will be 3, because methods of superclass are indexed 0,1,2.
	     int startingIndex;

 private:

	     bool findDescriptor(mClass*, int&);
		     

};

#endif

