#include "mParseTree.h"
#include "mBuiltinClasses.h"

#include <iostream>

using namespace std;

mParseTree::mParseTree(short Type, mClass* ret)
{
 type = Type;
 returnType = ret;
 
 children = new mList<mParseTree>();
 children->setAutoDelete( true );
 
 symbol = 0; name = 0;
 parent = 0;

 implicitReference = false;

}

mParseTree::mParseTree(short Type, mSymbol *Sym, mClass* ret)
{
 type = Type;

 // return type of this tree becomes return type of symbol
 if( ret == 0 )
   returnType = Sym->theClass;
 else 
   returnType = ret;
 
 children = new mList<mParseTree>();
 children->setAutoDelete( true );
 
 symbol = Sym;
 name = 0;

 parent = 0;

 implicitReference = false;

}

mParseTree::mParseTree(short Type,mString *id, mClass* ret)
{
 type = Type;
 returnType = ret;
 
 children = new mList<mParseTree>();
 children->setAutoDelete( true );
 
 symbol = 0;
 name = id; 

 parent = 0;

 implicitReference = false;
}

mParseTree::~mParseTree()
{ 
 delete children; 
}

void mParseTree::append(mParseTree *tree)
{
 tree->parent = this;
 children->append( tree );
}


mString mParseTree::getType()
{
  if(returnType) return *returnType->name;
  else return "void";
}


mString mParseTree::toString()
{
 mString str("(");
  

 mParseTree* inv = 0;  // invocation child

 switch(type){

     case NODE_LITERAL: 

       str += *symbol->name;       
       // check method call
       inv = getInvocationChild();
       if( inv != 0){
	 str += '.';
	 str += inv->toString();
	 str += ' ';
       }
       
       break;

     case NODE_VARIABLE: 

       if(symbol) 
	 str += *symbol->name;
       else 
	 str += *name;

       // check method call
       inv = getInvocationChild();
       if( inv != 0){
	 str += '.';
	 str += inv->toString();
	 str += ' ';
       }
 
       break;

     case NODE_NEW: str += "new ";
                     str += children->at(0)->toString();
                     if(children->size() > 1){
                        str += ' ';
                        str += children->at(1)->toString();
                     }
                     break;

     case NODE_CONSTRUCTION: 
       str += "construct ";
       
       // check method call
       inv = getInvocationChild();
       if( inv != 0){
	 str += '.';
	 str += inv->toString();
	 str += ' ';
       }

       if( children->size() > 1 ){
	 mParseTree* child = children->at(1);
	 if( child->type == NODE_METHODCALL ){    // yet additional methodcall
	   str += '.';
	   str += child->toString();
	 }
       }
       break;
                                                                   

     case NODE_QUALIFIEDID: str = *name;
       //return str;
                        break;                   


     case NODE_ASSIGN: str += "= ";
                     str += children->at(0)->toString();
                     str += ' ';
                     str += children->at(1)->toString();
                     break;    

     case NODE_INCREMENT: str += "++";
                     str += children->at(0)->toString();                     
                     break; 

     case NODE_DECREMENT: str += "--";
                     str += children->at(0)->toString();                     
                     break;                             
             

     case NODE_PLUS: str += "+ ";
                     str += children->at(0)->toString();
                     str += ' ';
                     str += children->at(1)->toString();

		     if( (inv=getInvocationChild(2)) != 0){
			 str += '.';
			 str += inv->toString();
			 str += ' ';
		     }
		     
                     break;
                     
     case NODE_MINUS: str += "- ";
                      str += children->at(0)->toString();
                      str += ' ';
                      str += children->at(1)->toString();

		      if( (inv=getInvocationChild(2)) != 0){
			 str += '.';
			 str += inv->toString();
			 str += ' ';
		      }

                      break;


     case NODE_MULTIPLY: str += "* ";
                      str += children->at(0)->toString();
                      str += ' ';
                      str += children->at(1)->toString();

		      if( (inv=getInvocationChild(2)) != 0){
			 str += '.';
			 str += inv->toString();
			 str += ' ';
		      }

                      break;

     case NODE_DIVIDE: str += "/ ";
                      str += children->at(0)->toString();
                      str += ' ';
                      str += children->at(1)->toString();

		      if( (inv=getInvocationChild(2)) != 0){
			 str += '.';
			 str += inv->toString();
			 str += ' ';
		      }

                      break;

     case NODE_MODULE: str += "% ";
                      str += children->at(0)->toString();
                      str += ' ';
                      str += children->at(1)->toString();

		      if( (inv=getInvocationChild(2)) != 0){
			 str += '.';
			 str += inv->toString();
			 str += ' ';
		      }

                      break;

     case NODE_LESS: str += "< ";
                     str += children->at(0)->toString();
                     str += ' ';
                     str += children->at(1)->toString();

		     if( (inv=getInvocationChild(2)) != 0){
			 str += '.';
			 str += inv->toString();
			 str += ' ';
		      }

                     break;

     case NODE_GREATER: str += "> ";
                     str += children->at(0)->toString();
                     str += ' ';
                     str += children->at(1)->toString();

		     if( (inv=getInvocationChild(2)) != 0){
			 str += '.';
			 str += inv->toString();
			 str += ' ';
		      }

                     break;

     case NODE_EQUALS: str += "== ";
                     str += children->at(0)->toString();
                     str += ' ';
                     str += children->at(1)->toString();
		     
		     if( (inv=getInvocationChild(2)) != 0){
			 str += '.';
			 str += inv->toString();
			 str += ' ';
		      }

                     break;

     case NODE_NOT_EQUALS: str += "!= ";
                     str += children->at(0)->toString();
                     str += ' ';
                     str += children->at(1)->toString();

		     if( (inv=getInvocationChild(2)) != 0){
			 str += '.';
			 str += inv->toString();
			 str += ' ';
		      }

                     break;

     case NODE_GREATER_OR_EQUALS: str += ">= ";
                     str += children->at(0)->toString();
                     str += ' ';
                     str += children->at(1)->toString();

		     if( (inv=getInvocationChild(2)) != 0){
			 str += '.';
			 str += inv->toString();
			 str += ' ';
		      }

                     break;

     case NODE_LESS_OR_EQUALS: str += "<= ";
                     str += children->at(0)->toString();
                     str += ' ';
                     str += children->at(1)->toString();

		     if( (inv=getInvocationChild(2)) != 0){
			 str += '.';
			 str += inv->toString();
			 str += ' ';
		      }

                     break; 


     case NODE_AND: str += "&& ";
                      str += children->at(0)->toString();
                      str += ' ';
                      str += children->at(1)->toString();

		      if( (inv=getInvocationChild(2)) != 0){
			 str += '.';
			 str += inv->toString();
			 str += ' ';
		      }

                      break;

     case NODE_OR: str += "|| ";
                      str += children->at(0)->toString();
                      str += ' ';
                      str += children->at(1)->toString();

		      if( (inv=getInvocationChild(2)) != 0){
			 str += '.';
			 str += inv->toString();
			 str += ' ';
		      }

                      break;    


     case NODE_NOT: str += "! ";
                      str += children->at(0)->toString();                      

		      if( (inv=getInvocationChild(2)) != 0){
			 str += '.';
			 str += inv->toString();
			 str += ' ';
		      }

                      break;           
                     
     case NODE_CLOSURE: str += "closure ";
                       // closure's children are statements and expressions
                        for(unsigned i=0;i<children->size(); i++){
			  mParseTree* child=children->at(i);
			  str += "(";                               
			  str += child->toString();
			  str += ") ";
                        }
                     /*str += children->at(0)->toString();
                     str += ' ';
                     str += children->at(1)->toString();*/
                     break;


     case NODE_WHILE: str += "while ";
                      // while's children are condition and execution statements                        
                      str += "(";                               
		      str += children->at(0)->toString();
		      str += ") do ";

		      str += "(";                               
		      str += children->at(1)->toString();
		      str += ") ";
                        		     
                     break;

     case NODE_DOWHILE: str += "do ";
                      // while's children are condition and execution statements                        
                      str += "(";                               
		      str += children->at(0)->toString();
		      str += ") while ";

		      str += "(";                               
		      str += children->at(1)->toString();
		      str += ") ";
                        		     
                     break;
                     
                              
                      
     case NODE_METHODCALL: str += "call ";                          
                           str += *symbol->name;			    
                           

			   if(children->size() == 1 ){  
			     // 1 means additional method call or  arguments	   
			     if(children->at(0)->type == NODE_ARGUMENTS ){
                                str += ' ';
			        str += children->at(0)->toString();
				str += ' ';
                             }else{
			       str += '.';
			       str += children->at(0)->toString();
			       str += ' ';
			     }
			   }else if(children->size() == 2 ){
                              str += ' ';
			      str += children->at(0)->toString(); // args
			      str += ".";
			      str += children->at(1)->toString(); // method call
			   }
                     break;


     case NODE_PARENTHESES: str += "{ ";
                     str += children->at(0)->toString();
                     str += "}.";
                     str += children->at(1)->toString();

		     /*if( (inv=getInvocationChild(2)) != 0){
			 str += '.';
			 str += inv->toString();
			 str += ' ';
			 }*/
		     
                     break;

     case NODE_NATIVECALL: str += "nativecall ";                          
                           str += *symbol->name;
			   str += ' ';
			   str += children->at(0)->toString(); // args
                           			   
                     break;

     case NODE_SUPERCONSTRUCTOR: str += "superconst ";                          
                           str += *symbol->name;			    
                           
			   if(children->size() == 1 ){  
			     // 1 means additional method call or  arguments	   
			     if(children->at(0)->type == NODE_ARGUMENTS ){
                                str += ' ';
			        str += children->at(0)->toString();
				str += ' ';
                             }else{
			       str += '.';
			       str += children->at(0)->toString();
			       str += ' ';
			     }
			   }else if(children->size() == 2 ){
                              str += ' ';
			      str += children->at(0)->toString(); // args
			      str += ".";
			      str += children->at(1)->toString(); // method call
			   }
                     break;

     case NODE_ARGUMENTS: 
                          str += "args ";
                          for(mParseTree *arg = children->first(); arg != 0; 
			      arg = children->next() ){
			         str += "#(";
			         str += arg->toString();
				 str += ") "; 
			  }
                        
                     break;

     case NODE_METHODUPDATE:
                          str += "-> ";
                          str += children->at(0)->toString();
			  if(children->size() > 1){
			    str += ' ';
			    str += children->at(1)->toString();
			  }
                          break;

     case NODE_ANONMETHOD:
                          str += "method (";
                          str += children->at(1)->toString();
			  str += ") { ";
			  str += children->at(0)->toString();
			  str += " } ";
                          break;

     case NODE_METHODREF:
                          //str += " ";
                          str += *symbol->name;
                          break;

     case NODE_CLASSREF:                         
                          str += *symbol->name;
                          break;


     case NODE_ARRAYACCESS:                      
                     str += "[]";

                     // arguments are always present      
                     str += ' ';
                     str += children->at(0)->toString();
                     str += ' ';       

		     if( (inv=getInvocationChild()) != 0){
			 str += '.';
			 str += inv->toString();
			 str += ' ';
		      }
                                                                   
                     break;		     


     case NODE_UNRESOLVED: str += "call ";
                           str += *name;
                           str += "??";
                     if(children->size() > 0 ){
                        str += children->at(0)->toString();
                        str += ' ';
                     }
                     if(children->size() > 1 )
                        str += children->at(1)->toString();
                     break;


     case NODE_PRIMITIVE_CAST: str += "pcast ";
                           str += *children->at(0)->returnType->name;
                           str += "->";
                           str += *returnType->name;
                           str += ' ';
			   if(children->size() > 0 ){
			     str += children->at(0)->toString();
			     str += ' ';
			   }
                     
			   break;

     case NODE_TYPE: str += "type ";
                           str += *returnType->name;
                           
                     
			   break;

     case NODE_COMPLEX_CAST: 
       
                           str += "ccast ";
                           str += *children->at(0)->returnType->name;
                           str += "->";
                           str += children->at(1)->toString();
                           str += ' ';
			   if(children->size() > 0 ){
			     str += children->at(0)->toString();
			     str += ' ';
			   }


			   if( (inv=getInvocationChild(2)) != 0){
			     str += '.';
			     str += inv->toString();
			     str += ' ';
			   }
                     
			   break;
			   
     case NODE_RETURN: 

       str += "return";
       if(children->size() > 0 && children->at(0)->type != NODE_METHODCALL ){
	 str += " ";
	 str += children->at(0)->toString();
       }

       break;

    
     case NODE_BLOCK_EXIT: 

       str += "exit";
       if(children->size() > 0 && children->at(0)->type != NODE_METHODCALL ){
	 str += " ";
	 str += children->at(0)->toString();
       }

       break;


      case NODE_METHOD: // method is built of params and block
        str += children->at(0)->toString();
	if(children->size() == 2){   // parameters + block
          str += '(';
	  str += children->at(1)->toString();
          str += ')';
	  
	}
                 
	break;


                      
 }  // switch


 // check method call
 /*inv = getInvocationChild();
 if( inv != 0){
   str += '.';
   str += inv->toString();
   str += ' ';
   }*/


 str += ')';

 return str;
}


mString mParseTree::toStringArguments()
{
  mString str;
  for(unsigned i = 0; i < children->size(); i++){
    str += *children->at(i)->returnType->name;
    if( i + 1 < children->size() )
    str += ','; 
  }

  return str;
}


mParseTree* mParseTree::getInvocationChild(int pos)
{
  int cpos=0;
  for(mParseTree *c = children->first(); c != 0; c = children->next(), cpos++ )
    
    if( pos != -1 ){   // must be at specified position

      if( pos == cpos && c->type == NODE_METHODCALL ) return c;

    }else
      if( c->type == NODE_METHODCALL ) return c;   // any position
      

  return 0;
}



mParseTree* mParseTree::getLastMethodNode()
{
  int size = children->size();
  if( size == 0 ){
    // if no children then this must be the last method node
      return this;
  }


  // otherwise search children that represent methodcalls
  if( size == 1 ){
      // 1 child can be either parameters or additional methodcall
      mParseTree *child = children->at(0);

      if( child->type == NODE_ARGUMENTS )
	  return this;

      if( child->type == NODE_METHODCALL || child->type == NODE_ARRAYACCESS )
          return children->at(0)->getLastMethodNode();      
  }else if( size == 2){
      // 2 children: 2nd is always additional methodcall
    //cout<<"seeking "<<children->at(1)->type<<endl;
      return children->at(1)->getLastMethodNode();
  }
 return 0;
}




// *********************************************************************************************

bool mParseTree::endsWithReturn()
{
  int last = children->size();

  if( last == 0 )   // empty closure
      return false;
 
  return children->at( last-1 )->type == NODE_RETURN;  
}


// *********************************************************************************************

void mParseTree::evaluate()
{

}


void mParseTree::addSuperConstructorCall(mMethod* method)
{

  // generate methodcall node
  mParseTree *call = new mParseTree( NODE_SUPERCONSTRUCTOR, method );
  call->implicitReference = true;
  call->parent = this;
  children->prepend(  call );

}


// *********************************************************************************************

bool mParseTree::promotePrimitivesViaCast(mParseTree* &left, mParseTree* &right, int op, mClass* retType)
{

  // check if void
  if( !left->returnType && right->returnType || left->returnType && !right->returnType ) return false;
  
  if(retType != 0) returnType = retType;

  if( !left->returnType && !right->returnType ) return true;

  mString &ltype = *left->returnType->name;
  mString &rtype = *right->returnType->name;


  // don't do anything if types are equal
  if(ltype == rtype ){ if(retType == 0) returnType=left->returnType;  return true; }

  /////////////////////////////////////////////////////////////
  // left is Integer
  /////////////////////////////////////////////////////////////
  if( ltype == "Integer" ){

    // Integer op Double => Double
    if( rtype == "Double" ){

      // mod defined only on integers
      if( op == NODE_MODULE ) return false;

      returnType = right->returnType;   // returnType of this parsetree

      // cast Integer to Double
      mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
      //cast->fromType = left->returnType;    // fromType is Integer
      cast->returnType = right->returnType; // returnType is Double

      cast->append( left );
      left = cast;
    }
 
    // Integer op String => String
    else if( rtype == "String" ){      

      // strings can only be added
      if( op != NODE_PLUS ) return false;

      returnType = right->returnType;   // returnType of this parsetree

      // cast Integer to String
      mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
      cast->returnType = right->returnType; // returnType is String

      cast->append( left );
      left = cast;
    }else
      return false;

    /////////////////////////////////////////////////////////////
    // left is Double
    /////////////////////////////////////////////////////////////
  } else if( ltype == "Double" ){

    // mod defined only on integers
    if( op == NODE_MODULE ) return false;

    // Double op Integer => Double
    if( rtype == "Integer" ){

      returnType = left->returnType;

      // cast Integer to Double
      mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
      cast->returnType = left->returnType;   // returnType is Double
      cast->append( right );
      right = cast;
    }
    // Double op String => String
    else if( rtype == "String" ){

      // strings can only be added
      if( op != NODE_PLUS ) return false;

      returnType = right->returnType;

      // cast Double to String
      mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
      cast->returnType = right->returnType;   // returnType is String
      cast->append( left );
      left = cast;
    }else
      return false;

    /////////////////////////////////////////////////////////////
    // left is String
    /////////////////////////////////////////////////////////////
  }else if( ltype == "String" ){

    // strings can only be added
    if( op != NODE_PLUS ) return false;

    // String op Integer => String
    if( rtype == "Integer" ){

      returnType = left->returnType;

      // cast Integer to String
      mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
      cast->returnType = left->returnType;   // returnType is String
      cast->append( right );
      right = cast;
    }

    // String op Double => String
    else if( rtype == "Double" ){

      returnType = left->returnType;

      // cast Double to String
      mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
      cast->returnType = left->returnType;   // returnType is String
      cast->append( right );
      right = cast;
      // String op Boolean => String
    }else if( rtype == "Boolean" ){

      returnType = left->returnType;

      // cast Boolean to String
      mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
      cast->returnType = left->returnType;   // returnType is String
      cast->append( right );
      right = cast;
    }else
      return false;

    /////////////////////////////////////////////////////////////
    // left is Boolean
    /////////////////////////////////////////////////////////////
  } else if( ltype == "Boolean" ){
    

    // Booleans can only be added to String
    if( op != NODE_PLUS ) return false;

    
    // Boolean op String => String
    if( rtype == "String" ){      

      returnType = right->returnType;

      // cast Boolean to String
      mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
      cast->returnType = right->returnType;   // returnType is String
      cast->append( left );
      left = cast;
    }else
      return false;

  }else{

    // no suitable types we found ! type mismatch
    return false;

  }

  if(retType != 0) returnType = retType;

  return true;
}



// *********************************************************************************************

bool mParseTree::promoteSinglePrimitiveViaCast(mParseTree* &left, int op, mClass* retType)
{

  // check if void
  if( !left->returnType ) return false;
  
  if(retType != 0) returnType = retType;


  mString &ltype = *left->returnType->name;


  // don't do anything if types are equal
  if(ltype == *retType->name )  return true;

  

  /////////////////////////////////////////////////////////////
  // left is Boolean
  /////////////////////////////////////////////////////////////
  if( ltype == "Boolean" ){
    

    // Booleans can only be added to String
    if( op != NODE_NOT ) return false;    


  }else{

    // no suitable types we found ! type mismatch
    return false;

  }

  if(retType != 0) returnType = retType;

  return true;
}


// *********************************************************************************************


bool mParseTree::promotePrimitiveViaCast(mParseTree* &ret, mClass* type)
{

  // check if void
  if( !returnType && type || returnType && !type ) return false;
  if( !returnType && !type ) return true;

  // don't do anything if types are equal
  if(*returnType->name == *type->name ) return true;

  // otherwise a cast is necessary

  /////////////////////////////////////////////////////////////
  // current returnType is Integer
  /////////////////////////////////////////////////////////////
  if(  *returnType->name == "Integer" ){

      // desired type is Double: Integer to Double
      if( *type->name == "Double" ){
      

	// cast Integer to Double
	mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
	cast->returnType = type;  // returnType is Double

	cast->append( this );
	ret = cast;
      }
      // desired type is String: Integer to String
      else if( *type->name == "String" ){
      

	// cast Integer to String
	mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
	cast->returnType = type;  // returnType is String

	cast->append( this );
	ret = cast;	
      } 

      // desired type is Object: Integer to Object
      else if( *type->name == "Object" ){
      
	// cast Integer to Object
	mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
	cast->returnType = type;  // returnType is Object

	cast->append( this );
	ret = cast;
      }
  /////////////////////////////////////////////////////////////
  // current returnType is Double
  /////////////////////////////////////////////////////////////
  }else if(  *returnType->name == "Double" ){

      // desired type is Integer: Double to Integer
      if( *type->name == "Integer" ){
      

	// cast Double to Integer
	mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
	cast->returnType = type;  // returnType is Integer

	cast->append( this );
	ret = cast;
      }
      // desired type is String: Double to String
      else if( *type->name == "String" ){
      
	// cast Double to String
	mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
	cast->returnType = type;  // returnType is String

	cast->append( this );
	ret = cast;
      }

      // desired type is Object: Double to Object
      else if( *type->name == "Object" ){
      

	// cast Integer to Object
	mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
	cast->returnType = type;  // returnType is Object

	cast->append( this );
	ret = cast;
      }

  /////////////////////////////////////////////////////////////
  // current returnType is String
  /////////////////////////////////////////////////////////////
  }else if(  *returnType->name == "String" ){    

      // desired type is Integer: String to Integer
    
      if( *type->name == "Integer" ){
      
	/*
	// cast String to Integer
	mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
	cast->returnType = type;  // returnType is Integer

	cast->append( this );
	ret = cast;
	*/
	return false;
      }
      // desired type is Double: String to Double
      else if( *type->name == "Double" ){

	/*
	// cast String to Double
	mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
	cast->returnType = type;  // returnType is Double

	cast->append( this );
	ret = cast;
	*/
	return false;
      }

      // desired type is Object: String to Object
      else if( *type->name == "Object" ){
      

	// cast Integer to Object
	mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
	cast->returnType = type;  // returnType is Object

	cast->append( this );
	ret = cast;
      }

  /////////////////////////////////////////////////////////////
  // current returnType is Boolean
  /////////////////////////////////////////////////////////////
  }else if(  *returnType->name == "Boolean" ){

      
      // desired type is String: Boolean to String
      if( *type->name == "String" ){
      
	// cast Boolean to String
	mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
	cast->returnType = type;  // returnType is String

	cast->append( this );
	ret = cast;
      }

      // desired type is Object: Boolean to Object
      else if( *type->name == "Object" ){
      

	// cast Boolean to Object
	mParseTree *cast = new mParseTree( NODE_PRIMITIVE_CAST );
	cast->returnType = type;  // returnType is Object

	cast->append( this );
	ret = cast;
      }

  }else
    // no suitable type was found
    return false;

  return true;

}


// *********************************************************************************************

bool mParseTree::performTypeCast(mParseTree*& tree, mClass* type)
{

  // check if void
  if( !returnType && type || returnType && !type ) return false;
  
  if( !returnType && !type ) return true;

  // don't do anything if types are already equal
  if(*returnType->name == *type->name ) return true;


  // 1st we check if we're dealing with primitive types, because they
  // have builtin methods for conversion between them. that is all primitive types
  if( returnType->isPrimitiveButNoObject() && type->isPrimitiveButNoObject() ){

    // cast primitive type    
    return promotePrimitiveViaCast( tree, type );

  }else{

    // we have complex (object) type
    
    // 1st check if we're casting from nil, in which case we don't do any
    // casting at all
    if( tree->type == NODE_LITERAL ){
      // check nil
      if( *tree->symbol->name == "const_special_s_nil" ) return true;
    }

    // otherwise we have a cast from one class to another       
    // classes must be related

    if(tree->returnType->isSubclass( type ) || type->isSubclass( tree->returnType )){

      mParseTree *cast = new mParseTree( NODE_COMPLEX_CAST );
      cast->returnType = type;

      cast->append( this );   // append expression to be casted
      cast->append( new mParseTree( NODE_TYPE, type ) );


      tree = cast;

      return true;      

    }


    // error for now!
    return false;

  }

  return true;
}


// *********************************************************************************************


bool mParseTree::performArgumentsCast(mParseTree*& args, mMethod *method)
{

  // if this is not NODE_ARGUMENTS return false
  if( type != NODE_ARGUMENTS )
    return false;


  // formal parameters of this method
  mList<mParameter> *params = method->parameterList;


  // now try to cast every argument in this parse tree to type of formal parameter
  for(unsigned i = 0; i< params->size(); i++){    

    // length of both lists is guaranteed to be the same

    mClass *paramType = params->at(i)->theClass;   // type of formal parameter

    mParseTree* argument = args->children->at(i);        // argument expression
    //mClass *argumentType = argument->returnType;   // type of actual argument
    

    //cout<<"---> performing cast from "<<*argument->returnType->name<<" to "<<*paramType->name<<endl;

    if( !argument->performTypeCast(argument , paramType ) ) return false;
    // update if it was casted
    args->children->replace(i, argument);
  }

  return true;

}


bool mParseTree::closureReturns()
{
  if( type == NODE_CLOSURE ){
    
    return env->containsReturn();

  }else if( type == NODE_WHILE ){

    // while has environment in second (execution) block
    return children->at(1)->env->containsReturn();

  }else if( type == NODE_DOWHILE ){

    // dowhile has environment in first (execution) block
    return children->at(0)->env->containsReturn();

  }else if( type == NODE_IF ){

    // if has environment in second block
    if( children->at(1)->env->containsReturn() ) return true;

    // but we must check third block also. this may be an else or another if
    if( children->size() == 3 )
      if( children->at(2)->type == NODE_CLOSURE ) return children->at(2)->env->containsReturn();
      else 
	return children->at(2)->closureReturns();

  }else if( type == NODE_FOR ){

    // for has environment in first, third and fourth block
    if( children->at(0)->env->containsReturn() ) return true;
    if( children->at(2)->env->containsReturn() ) return true;
    if( children->at(3)->env->containsReturn() ) return true;
    

  }

  return false;
}


bool mParseTree::closureAssignsToParent()
{
  if( type == NODE_CLOSURE ){
    
    return env->containsParentAssignment();

  }else if( type == NODE_WHILE ){

    // while has environment in second (execution) block
    return children->at(1)->env->containsParentAssignment();

  }else if( type == NODE_DOWHILE ){

    // dowhile has environment in first (execution) block
    return children->at(0)->env->containsParentAssignment();

  }else if( type == NODE_IF ){

    // if has environment in second block
    if( children->at(1)->env->containsParentAssignment() ) return true;

    // but we must check third block also. this may be an else or another if
    if( children->size() == 3 )
      if( children->at(2)->type == NODE_CLOSURE ) return children->at(2)->env->containsParentAssignment();
      else 
	return children->at(2)->closureAssignsToParent();

  }else if( type == NODE_FOR ){

    // for has environment in first, third and fourth block
    if( children->at(0)->env->containsParentAssignment() ) return true;
    if( children->at(2)->env->containsParentAssignment() ) return true;
    if( children->at(3)->env->containsParentAssignment() ) return true;
    

  }

  return false;
}


bool mParseTree::isBooleanClosure(mClass* bclass)
{
  if( type != NODE_CLOSURE )
    return false;
  

  // check if more than 1 child  ( 2nd child may be exit or explicit return)
  if(children->size() > 2) return false;

  // check if child is boolean expression
  return children->at(0)->returnType == bclass;  

}




int mParseTree::numIfConditions()
{

  if( type != NODE_IF ) return 0;
    
  // only one condition/block pair
  if(children->size() == 2) return 1;
  
  // otherwise we must have an additional block which is either ELSE or IF
  if(children->at(2)->type == NODE_CLOSURE) return 2;

  // if it's if then we move down recursovely
  return 1 + children->at(2)->numIfConditions();

}


bool mParseTree::removeNode(short type)
{

  for(mParseTree *c = children->first(); c != 0; c = children->next() )
    
    if( c->type == type ){

      children->remove( c );
      return true;

    }

  return false;

}


bool mParseTree::isVariableReference()
{
  
  // there must be no children and type must be node variable
  return type == NODE_VARIABLE && children->size() == 0; 

}


bool mParseTree::isSimpleMethodReference()
{
  
  // there must be no children and type must be node methodcall
  if( type != NODE_METHODCALL ) return false;

  if(children->size() == 0) return true;

  if(children->size() == 1 && children->at(0)->type == NODE_ARGUMENTS) return true;

  return false;
}


bool mParseTree::argumentsMatchExactly()
{
  if( type != NODE_ARGUMENTS ) return false;

  for(mParseTree *c = children->first(); c != 0; c = children->next() ){

    // 
    if( c->type == NODE_PRIMITIVE_CAST || c->type == NODE_COMPLEX_CAST ) return false;

  }

  return true;
}


bool mParseTree::isMultiMethodCall()
{

  if( type != NODE_METHODCALL && type != NODE_VARIABLE ) return false;


  // if we have 2 children we have additional call
  if(children->size() == 2) return true;

  // no children
  if(children->size() == 0) return false;

    
  // if we have 1 child it may be arguments or additional method call
  return children->at(0)->type == NODE_METHODCALL; 
  
}




bool mParseTree::isStandaloneClosureMethodCall(mParseTree* last)
{

  mParseTree* tree = this;
  while(tree->type == NODE_METHODCALL || tree->type == NODE_VARIABLE ){

    tree = tree->parent;
    
  }

  // we found a node that is not methodcall or variable, if it's a closure, then
  // this is standalone call otherwise not
  if(tree->type == NODE_CLOSURE) return true;

  // node is not a closure. it's probably an expression which means we must check if this 
  // expression has invocation child which must equal previous tree
  
  //cout<<"---> tree type= "<<tree->type<<endl;

  
  // first check binary operations: invocation child here is at third position (+ left right).invocation
  if(tree->type == NODE_PLUS || tree->type == NODE_MINUS || tree->type == NODE_MULTIPLY || tree->type == NODE_DIVIDE ||
     tree->type == NODE_MODULE || tree->type == NODE_LESS || tree->type == NODE_GREATER || tree->type == NODE_NEW ||
     tree->type == NODE_EQUALS || tree->type == NODE_NOT_EQUALS || tree->type == NODE_GREATER_OR_EQUALS || 
     tree->type == NODE_LESS_OR_EQUALS || tree->type == NODE_AND || tree->type == NODE_OR ||
     tree->type == NODE_COMPLEX_CAST ){

    // invocation child must be at position 2
    // if no child is found at position then this is not standalone call
    mParseTree* invocation = tree->getInvocationChild( 2 );
    if( invocation  == 0) return false;

    
    // node has invocation child at position 2, it must be equal to the last invocation for which we're 
    // checking if it's standalone
    if(invocation != last) return false;
    

    // otherwise we may continue up the tree
    return tree->parent->isStandaloneClosureMethodCall(last);
    

  }else if( tree->type == NODE_PRIMITIVE_CAST || tree->type == NODE_CONSTRUCTION  || tree->type == NODE_NOT){

    // if invocation child exists it's on first position
    // if no child is found at position then this is not standalone call
    mParseTree* invocation = tree->getInvocationChild( 1 );
    if( invocation  == 0) return false;

    
    // node has invocation child at position 1, it must be equal to the last invocation for which we're 
    // checking if it's standalone
    if(invocation != last) return false;
    

    // otherwise we may continue up the tree
    return tree->parent->isStandaloneClosureMethodCall(last);

  }


  return false;
  
}


void mParseTree::removeLastMethodNode()
{

  int size = children->size();
  if( size == 0 ){
    // if no children then this must be the last method node
    parent->children->remove( this );
    
  }


  // otherwise search children that represent methodcalls
  if( size == 1 ){
      // 1 child can be either parameters or additional methodcall
      mParseTree *child = children->at(0);

      if( child->type == NODE_METHODCALL || child->type == NODE_ARRAYACCESS )
          children->at(0)->removeLastMethodNode();      
  }else if( size == 2){
      // 2 children: 2nd is always additional methodcall
      children->at(1)->removeLastMethodNode();
  }
 
}


bool mParseTree::bothPrimitive(mParseTree* first, mParseTree* second)
{
  return first->returnType->isPrimitive() && second->returnType->isPrimitive();
}
