// mString class definition

#ifndef __MSTRING__H
#define __MSTRING__H

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>

class mString /*: public mObject*/{
      public:
            //!Constructor.
            /*!Constructs a null string (one whose length is said to be 0). */
            mString();
            //!Constructor.
            /*!Constructs a string from array of chars, ie. from classic c string. */
            mString(char const*);
            //!Copy Constructor.
            /*!Constructs a copy of another string.*/
            mString(mString const&);
            //!Constructor.
            /*!Constructs a string containing one character. */
            mString(char);
            //!Constructor.
            /*!Constructs string representation of a signed integer. */
            mString(int);
            //!Constructor.
            /*!Constructs string representation of an unsigned integer. */
            mString(unsigned int);
            //!Constructor.
            /*!Constructs a string representation of a double & possibly float. */
            mString(double);
            //!Destructor.
            virtual ~mString();
            //!Clone
            /*!Creates new instance of this object. */
            virtual mString& clone() { return (*(new mString(*this))); }
            //!className
            /*!Returns class name. */
            virtual char const* className() const { return "mObject.mString"; }	    
	    //!toString
            /*!Returns string representation. */
            virtual mString toString() { return *this; }
            //!Overloaded operator.
            /*!Assigns another string and returns reference to updated string. */
            mString& operator=(mString const&);
            //!Overloaded operator.
            /*!Assigns one character and returns reference to updated string. */
            mString& operator=(char);
            //!Overloaded operator.
            /*!Assigns a signed integer and returns reference to updated string. */
            mString& operator=(int);
            //!Overloaded operator.
            /*!Assigns an unsigned integer and returns reference to updated string. */
            mString& operator=(unsigned int);
            //!Overloaded operator.
            /*!Assigns a double or float and returns reference to updated string. */
            mString& operator=(double);
            //!Overloaded operator.
            /*!Adds string and returns reference to updated string. */
            mString& operator+=(mString const&);
	    //!Overloaded operator.
            /*!Adds string and returns reference to updated string. */
            mString& operator+=(char const*);
            //!Overloaded operator.
            /*!Adds single character and returns reference to updated string. */
            mString& operator+=(char);
            //!Overloaded operator.
            /*!Adds signed int and returns reference to updated string. */
            mString& operator+=(int);
            //!Overloaded operator.
            /*!Adds unsigned int and returns reference to updated string. */
            mString& operator+=(unsigned int);
            //!Overloaded operator.
            /*!Adds double and returns reference to updated string. */
            mString& operator+=(double);
            //!Overloaded operator.
            /*!Returns true if strings are equal. */
            bool operator==(mString const&);
            //!Overloaded operator.
            /*!Returns true if strings are equal. 2nd string is a c-style string. */
            bool operator==(char const*);
            //!Overloaded operator.
            /*!Returns true if strings are not equal. */
            bool operator!=(mString const&);
            //!Overloaded operator.
            /*!Returns true if strings are not equal. 2nd string is a c-style string. */
            bool operator!=(char const*);
            //!Overloaded operator.
            /*!Returns true if string 1 is greater than string 2. */
            bool operator>(mString const&);
            //!Overloaded operator.
            /*!Returns true if string 1 is greater than string 2.
            2nd string is a c-style string. */
            bool operator>(char const*);
            //!Overloaded operator.
            /*!Returns true if string 1 is greater or equal than string 2. */
            bool operator>=(mString const&);
            //!Overloaded operator.
            /*! Returns true if string 1 is greater or equal than string 2.
            2nd string is a c-style string. */
            bool operator>=(char const*);
            //!Overloaded operator.
            /*!Returns true if string 1 is lower than string 2. */
            bool operator<(mString const&);
            //!Overloaded operator.
            /*! Returns true if string 1 is lower than string 2.
            2nd string is a c-style string. */
            bool operator<(char const*);
            //!Overloaded operator.
            /*!Returns true if string 1 is lower or equal than string 2. */
            bool operator<=(mString const&);
            //!Overloaded operator.
            /*! Returns true if string 1 is lower or equal than string 2.
            2nd string is a c-style string. */
            bool operator<=(char const*);
            //!Conversion operator.
            /*!Converts mString to c-style char* string.*/
            operator char const*() const { return this->string; }
            //!Conversion operator.
            /*!Returns i-th character in string. */
            char operator[](int i) const { return string[i]; }
            //!At.
            /*!Returns i-th character in string. */
            char at(unsigned i) const { return string[i]; }
            /*!Converts mString to its signed integer representation.*/
            int Int() const { return strtol(string,NULL,Base); }
            //!Conversion function.
            /*!Converts mString to its unsigned integer representation.*/
            unsigned int Uint() const { return strtoul(string,NULL,Base); }
            //!Conversion function.
            /*!Converts mString to its double representation.*/
            double Double() const { return strtod(string,NULL); }
            //! Length.
            /*!Returns length of string. */
            unsigned length() const { return strlen(string); }
            //! Null.
            /*!Returns true if string is a null (of length 0) string. */
            bool null() const { return (strlen(string)==0); }
            //!Base.
            /*!Sets base for converting from int and unsigned int.
            Base can be 8(octal), 10(decimal) or 16(hexadecimal).
            Returns true if base was successfully set, false otherwise. */
            bool setBase(int);
            //!Base.
            /*!Returns current base used for converting values to string. */
            int base() const { return Base; }
            //!Fill.
            /*! Fills string with character given as the 1st parameter.
            2nd parameter is optional length; if omitted, the entire string will be filled. */
            void fill(char,int=-1);
            //!isSpace.
            /*! Returns true character at position given as parameter is a space. */
            bool isSpace(unsigned) const;
            //!Upper.
            /*! Converts string to its uppercase representation and returns reference
            to updated string. */
            mString& upper();
            //!Lower.
            /*! Converts string to its lowercase representation and returns reference
            to updated string. */
            mString& lower();
            //!Replace.
            /*! Replaces character at postion given with 1st parameter
            with char given as the 2nd parameter. Return reference to
            updated string. */
            mString& replace(unsigned,char);
            //!Append.
            /*! Appends string and returns reference to updated string.
            Same can be achieved with malevolent operator +=. */
            mString& append(mString const&);
            //!Prepend.
            /*!  Prepends string and returns reference to updated string. */
            mString& prepend(mString const&);
            //!Truncate.
            /*!Truncates string at specified position and return reference to updated string.
            If position is out of range, nothing is done. */
            mString& truncate(unsigned);
            //!Remove.
            /*!Removes length characters from specified position
            and return reference to updated string.
            Position is specified with 1st parameter, length with 2nd.
            If position is out of range, nothing is done. */
            mString& remove(unsigned,unsigned);
            //!Insert.
            /*! Inserts substring into strings at position specified by 1st parameter.
            Return reference to updated string. */
            mString& insert(unsigned,mString const&);
            //!Find.
            /*! Finds first occurrence of specified char in the string. Char
            is given as the 1st parameter. The search starts at pos, which
            is optionally given as the 2nd parameter. The search is case sensitive
            if cs, given optionally as 3rd parameter, equals true. Returns
            position of the char or -1 if not found. */
            int find(char,unsigned=0,bool=true) const;
            /*!Finds first occurrence of string given as 1st parameter. */
            int find(mString const&,unsigned=0,bool=true) const;
            //!Find.
            /*! Same as above,only search starts at the end and searches
            backwards. */
            int findRev(char,unsigned=0,bool=true) const;
            /*!Returns number of time a char occurs in string. */
            unsigned contains(char,unsigned=0,bool=true) const;
            /*!Returns number of time a string given as 1st parameter
            occurs in string. */
            unsigned contains(mString const&,unsigned=0,bool=true) const;
            //!Left.
            /*! Returns left part of the string from the specified position. */
            mString left(unsigned);
            //!Right.
            /*! Returns right part of the string from the specified position. */
            mString right(unsigned);
            //!Middle.
            /*! Returns middle part of the string from the specified position.
             2nd parameter is length. */
            mString middle(unsigned,unsigned);
            //!RemoveWhiteSpace.
            /*! Removes predceding & succeding whitespace from string. Returns reference to
             updated string. */
            mString& removeWhiteSpace();
            //!Sprintf.
            /*! Makes formatted string. Standard c flags (%) are supported. */
            void sprintf(char const*,...);
            //!Overloaded operator.
            /*! Returns new string, which is a sum of string1 & string2. */
            friend mString operator+(mString const&,mString const&);
            //!Overloaded operator.
            /*! Returns new string, which is a sum of string1 & string2.
             String1 is an int. */
            friend mString operator+(int,mString const&);
            //!Overloaded operator.
            /*! Returns new string, which is a sum of string1 & string2.
             String2 is an int. */
            friend mString operator+(mString const&,int);
            friend mString operator+(mString const&,char);

            friend bool operator==(mString const&,mString const&);
      private:
            // copies characters to _temp_str
            // params: chars, length
            /*inline*/ void _store(char*,unsigned);

            int      Base;
            char     *string;
            char     *_temp_str;
};

#endif
