hodAccessModifier IDENTIFIER MethodParametersOpt ':' Type
  {    
    // method must be added as a symbol to class environment before its body
    // because it may reference itself in a recursive manner


    if( !mSymbol::isValid( $2 ) ){
      myerror();
      cout<<"method '"<<*$2<<"' cannot be declared"<<endl;
      break;
    }


    // check if this is 'main' method. 'main' is always a procedure
    if( *$2 == "main"){
       myerror();
       cout<<"method 'main' doesn't return"<<endl;
       break;
    }



    mClass* ownerClass = envStack->top()->ownerClass;



    // check if this is a constructor. if it is it's an error because constructors
    // don't return.
    if(*$2 == *ownerClass->name ){
       myerror();
       cout<<"constructors don't return"<<endl;
       break;
    }

   


    if($5 == 0){
       myerror();
       cout<<"method '"<<*$2<<"' returns unknown type"<<endl;
       break;
    }


    // check if method with this signature has already been defined
    if( importing )
      if( ownerClass->methodDefinedHere( $2, $3 ) ){      
	
	myerror();
	cout<<"method '"<<*$2<<"' already defined"<<endl;
	break;	
      }        

        
    mEnvironment *env = 0;
    mMethod *meth = 0; 

    if( !importing ){


      // create method environment
      env = new mEnvironment(ENVIRONMENT_METHOD, $2, envStack->top(),
			     ownerClass );

      // find method in our main class
      meth = ownerClass->getDeclaredMethod( $2, $3 );      

    }else{

      // create new method      
      // ( name, return type, access, binding )
      meth = new mMethod ( $2, $5 ,$1, BIND_NONSTATIC );
      
      // add it to class
      ownerClass->addMethod( meth );

      // now add method to class environment. note that method is appended to
      // class env. only as a SYMBOL here so that we can search class
      // environments simply for symbols.
      // class environment also has a child environment for every method
      // which is assigned lower    
    
      envStack->top()->append( meth );
      

    }        
    
    
    // method environment onto stack
    if( !importing ){

      envStack->push( env );
      meth->environment = env;
      currentMethod = meth;   // this is the method we are processing at the moment      

    }


    // finally we add formal parameters to method environment if there are any
    // note: method needs parameters for checking when searching for appropriate method,
    // and environment needs them in searching when a parameter is referenced in code    

    int currentEntry = 5;  // default location for 1st parameter on stack is 5 for function
                           // except for static methods which don't expect anything on stack except
                           // parameters
                           // 1 is stack_ptr, 2 is result destination, 3 is return address, 4 is self, 
                           // 5 is 1st param


    if( !importing )
      // change to static later (the only static method for now is 'main')
      if( *meth->name == "main" ){

	currentEntry = 1;

      }else{

	// we add self to all nonstatic methods
	env->addSelf( 4 );  // for functions it resides at location 4

      }


    if($3){      

      for( mParameter* param = $3->first(); param != 0; param = $3->next() ){
	if( !importing ){
	  param->entry = currentEntry++;
	  env->append( param );
	}else
	  meth->addParameter( param );
      }
      
      delete $3;
    }

    if( !importing )
      env->currentVariableEntry = currentEntry;     // set current entry of method env so that child block
                                                    // will know where its own variables start    

  } 

  MethodBody {
  

     // if no explicit return was given we signal error because functions MUST return excplicitly 

     if( !importing ){

       mParseTree* closure = envStack->top()->children->at(0)->closure;

       if( !closure->endsWithReturn() ){

	 myerrorBefore();
	 cout<<"method '"<<*currentMethod->name<<"' must return '"<<*currentMethod->returnType->name
	     <<"' at end"<<endl;      
       }


       // remove this method's environment
       if($5 != 0)
	 envStack->pop();

     }

  }
  ;


// no return type for procedures
ProcedureDeclaration: MethodAccessModifier IDENTIFIER MethodParametersOpt
  {
    // method must be added as a symbol to class environment before its body
    // because it may reference itself in a recursive manner


    

    if( !mSymbol::isValid( $2 ) ){
      myerror();
      cout<<"method '"<<*$2<<"' cannot be declared"<<endl;
      break;
    }

    // returnType of procedure is 0 except for constructors which have a returnType of
    // their own class (this is implicit of course)
    mClass *retType = 0;

    mClass* ownerClass = envStack->top()->ownerClass;

    if(*$2 == *ownerClass->name )
      retType = ownerClass;
    else{
      // if it's not constructor, we set returnType to Void
      mString Void("Void");
      retType = (mClass*)root->getSymbol( &Void, SYMBOL_CLASS );
    }
    

    // check if method with this signature has already been defined
    if( importing )
      if( ownerClass->methodDefinedHere( $2, $3 ) ){      

	myerror();
	cout<<"method '"<<*$2<<"' already defined"<<endl;
	break;

      }

    
    mEnvironment *env = 0;
    mMethod *meth = 0; 

    if( !importing ){


      // create method environment
      env = new mEnvironment(ENVIRONMENT_METHOD, $2, envStack->top(),
			     ownerClass );

      // find method in our main class
      meth = ownerClass->getDeclaredMethod( $2, $3 );
      

    }else{

      // create new method

      // check if it's 'main'
      short binding = BIND_NONSTATIC;
      if( *$2 == "main") binding = BIND_STATIC;


      // ( name, return type, access, binding )
      meth = new mMethod ( $2, retType ,$1, binding );    
      
      // add it to class
      ownerClass->addMethod( meth );

      // now add method to class environment. note that method is appended to
      // class env. only as a SYMBOL here so that we can search class
      // environments simply for symbols.
      // class environment also has a child environment for every method
      // which is assigned lower    
    
      envStack->top()->append( meth );
      

    }
    
    // method environment onto stack
    if( !importing ){

      envStack->push( env );
      meth->environment = env;
      currentMethod = meth;   // this is the method we are processing at the moment
    }




    int currentEntry = 4;  // default location for 1st parameter on stack is 4
                           // except for static methods which don't expect anything on stack except
                           // parameters

    // change to static later (the only static method for now is 'main')
    if( !importing )
      if( *meth->name == "main" ){
	currentEntry = 1;
      }else{
	
	// we add self to all nonstatic methods
	env->addSelf( 3 );  // for procedures it resides at location 3
	
      }

    // finally we add formal parameters to method environment if there are any    
    if($3){        

      for( mParameter* param = $3->first(); param != 0; param = $3->next() ){
	if( !importing ){
	  param->entry = currentEntry++;
	  env->append( param );
	}else
	  meth->addParameter( param );
      }      
      delete $3;
    }


    if( !importing ){
      env->currentVariableEntry = currentEntry;   // set current entry of method env so that child block
                                                  // will know where its own variables start

      //envStack->push( env );

      //meth->environment = env;

      //currentMethod = meth;   // this is the method we are processing at the moment
    }
  }

  MethodBody {


  // if no explicit return was given we add it here (because return is mapped directly to RETURN 
  // instruction)

  if( !importing ){

    mParseTree* closure = envStack->top()->children->at(0)->closure;

    if( !closure->endsWithReturn() ){

      if(*currentMethod->name != "main")  // method main doesn't return
	closure->append( new mParseTree( NODE_RETURN ) );

    }
   

    // remove this method's environment
    envStack->pop();

  }


  }
  ;




// optional parameters for methods
MethodParametersOpt: { $$ = 0; }  // empty
                     | '(' MethodParameterList ')' {
		          $$ = $2;
                       }
                     ;

// classic type id, type id, ...
MethodParameterList:   Type IDENTIFIER {

		       if( !mSymbol::isValid( $2 ) ){
			 myerror();
			 cout<<"parameter '"<<*$2<<"' cannot be declared"<<endl;
			 $$ = 0;
			 break;
		       }

		       // create parameter and add it to list
		       // which is on top of stack
		       if($1){			
			 mList<mParameter> *list = new mList<mParameter>();
			 list->append ( new mParameter ( $2, $1 ) );
			 $$ = list;
		       }else $$ = 0;
                     }
                     | MethodParameterList ',' Type IDENTIFIER {
		         if($3){
			   if(!$1) $1 = new mList<mParameter>();
			   $1->append ( new mParameter ( $4, $3 ) );			   
			 }
			 $$ = $1;
                       }
                     ;

// possible access modifiers for methods...
MethodAccessModifier:        PUBLIC { $$ = ACCESS_PUBLIC; }
                           | PROTECTED { $$ = ACCESS_PROTECTED; }
                           | PRIVATE { $$ = ACCESS_PRIVATE; }
                           | RESTRICTED { $$ = ACCESS_RESTRICTED; }
                           | error {
                             cout<<"invalid access modifier"<<endl;
                             //yyerrok;
                           }
                           ;



// method body is a block
MethodBody: Block {

            }
            ;


// block is empty or built of local variables and statements
Block:   '{' LocalsOrStatementsOpt '}' {

            // do nothing if just importing
            if( importing ) break;

	    mParseTree* closure = envStack->top()->closure;

	    // every block except that of top-level method must exit with implicit NODE_BLOCK_EXIT
	    // or explicit return 
	    if( envStack->top()->parent->type != ENVIRONMENT_METHOD ){

	      // we have nested closure

	      if( !closure->endsWithReturn() ){

		// append exit
		closure->append( new mParseTree( NODE_BLOCK_EXIT ) );

	      }
	    }

            // block now holds a closure
            $$ = closure;
            
            // remove environment for local definitions in this block
            envStack->pop();
           }
         | '{' LocalsOrStatementsOpt '}' '.' MethodAccess Semicolon {

            // do nothing if just importing
	    if( importing ) break;

            envStack->pop();
               //$$ = $2;
         }
         /*| '{' LocalsOrStatementsOpt '}' '.' FieldAccess Semicolon {
            envStack->pop();
               //$$ = $2;
         }*/
         ;

LocalsOrStatementsOpt:    {


	                    // do nothing if just importing
	                    if( importing ) break;

                            // create empty environment
                            mEnvironment *env = new mEnvironment(ENVIRONMENT_BLOCK, 
						new mString("#empty locals"), 
						envStack->top(), 
						envStack->top()->ownerClass );

                            mParseTree *closure = new mParseTree( NODE_CLOSURE );
			    // return type of closure is Block
			    mString block("Block");
			    mSymbol *classOfType = root->getSymbol( &block, SYMBOL_CLASS );		    
			    closure->returnType = (mClass*)classOfType;


                            env->closure = closure;
                            
                            // push on top
                            envStack->push( env );
                          }

                        | LocalsOrStatements {
	                   // environment and its closure are created below
                        }
                        ;


LocalsOrStatements:    {

                          // do nothing if just importing
			  if( importing ) break;

                          // this is where we create new environment for
                          // current local definitions
                          mEnvironment *env = new mEnvironment( ENVIRONMENT_BLOCK,
								new mString("#locals"),
								envStack->top(),
								envStack->top()->ownerClass );


                          // since locals and statements are always within
                          // a block, we also create a closure for this block
                          mParseTree *closure = new mParseTree( NODE_CLOSURE );

			  // return type of closure is Block
			  mString block("Block");
			  mSymbol *classOfType = root->getSymbol( &block, SYMBOL_CLASS );		    
			  closure->returnType = (mClass*)classOfType;

                          env->closure = closure;

                          // push on top
                          envStack->push( env );

                          
                       }

                       // one statement
                       LocalOrStatement {
                       }
                      // or multiple 
                      | LocalsOrStatements LocalOrStatement {
                      }
                     ;


// local variable definition or statement
LocalOrStatement: Type { currentType=$1; } LocalVariableDeclarations Semicolon {
                    //$$ = $1;
                  }
                  | Statement { /*$$ = 0;*/ }
                  ;


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//
// VARIABLE DECLARATION
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

// variable declarations always have a preceeding TYPE, that's how they
// differ from common assignment expression
LocalVariableDeclarations:   LocalVariableDeclaration
                           | LocalVariableDeclarations ',' LocalVariableDeclaration
                           ;


// declaration of a local variable without initialisation
LocalVariableDeclaration:    IDENTIFIER {

                                // do nothing if just importing
                                if( importing ) break;

                                if( !mSymbol::isValid( $1 ) ){
                                    myerror();
                                    cout<<"variable '"<<*$1<<"' cannot be declared"<<endl;
                                    break;
                                }

                                // get active environment from top of stack
                                mEnvironment *env = envStack->top();

                                // first we check if variable with this name
                                // already exists in current environment
                                if( env->getLocalSymbol( $1 ) != 0 ){
                                    myerror();
                                    cout<<"variable '"<<*$1<<"' already defined"<<endl;
                                    break;
                                }

                                // bail out (don't add this variable) if type
                                // is erroneous
                                if(currentType == 0) break;
                                
                                // create new variable
                                // access is public , binding nonstatic
                                mVariable *v = new mVariable( $1, currentType, 
							      BIND_NONSTATIC, 
							      env->currentVariableEntry );

                                // ... and add it to active environment
                                env->append( v );

                                // variable is created with NODE_NEW
                                mParseTree *tree = new mParseTree( NODE_NEW );

                                tree->append( new mParseTree( NODE_VARIABLE, v ) );

				// add this variable declaration to closure
                                env->closure->append( tree );
                                
                                $$ = tree;

                             }
                             
                           // declaration with explicit initialisation expression
                           | IDENTIFIER '=' Expression {

			        // do nothing if just importing
			        if( importing ) break;


			        if( !mSymbol::isValid( $1 ) ){
                                    myerror();
                                    cout<<"variable '"<<*$1<<"' cannot be declared"<<endl;
                                    break;
                                }


			        // bail out (don't create this variable) if type
                                // or expression are erroneous
                                if(currentType == 0 || !$3 ) break;

                                // get active environment from top of stack
                                mEnvironment *env = envStack->top();

                                // check if variable with this name already exists
                                if( env->getLocalSymbol( $1 ) != 0){
                                    myerror();
                                    cout<<"symbol '"<<*$1<<"' already defined"<<endl;
                                    break;
                                }
			

				// perform implicit cast to currentType if necessary
                                if( !$3->performTypeCast( $3, currentType ) ){				  
                                  myerror();
                                  cout<<"cannot perform cast from '"
                                      <<$3->getType()<<"' to '"
                                      <<*currentType->name<<"'"<<endl;
                                  break;
                                }
								

                                // create new initialised variable
                                mVariable *v = new mVariable( $1, currentType, 
					  BIND_NONSTATIC, env->currentVariableEntry );

                                //cout<<"variable "<<*$1<<"'s code: "<<*$3->text<<endl;

                                // and add it to environment
                                env->append( v );

                                // variable is created with NODE_NEW
                                mParseTree *tree = new mParseTree( NODE_NEW );
                                
                                tree->append( new mParseTree( NODE_VARIABLE, v ) );

				// append expression
                                tree->append( $3 );                               

				// add this variable declaration to closure
                                env->closure->append( tree );
                                
                                $$ = tree;

                           }
                           // this one takes all statements except expressions
                           // in initialisation. used to create variable and
                           // assign statement to it (blocks, loops, ...)
                           | IDENTIFIER '=' OtherStatement {


			     // do nothing if just importing
			     if( importing ) break;

			     if( !mSymbol::isValid( $1 ) ){
			       myerror();
			       cout<<"variable '"<<*$1<<"' cannot be declared"<<endl;
			       break;
			     }

			     // bail out (don't create this variable) if type
			     // or statement are erroneous
			     if(currentType == 0 || !$3 ) break;

			     // get active environment from top of stack
			     mEnvironment *env = envStack->top();

			     // check if variable with this name already exists
			     if( env->getLocalSymbol( $1 ) != 0){
			       myerror();
			       cout<<"symbol '"<<*$1<<"' already defined"<<endl;
			       break;
			     }

			     // create new initialised variable
			     mVariable *v = new mVariable( $1, currentType, 
							   BIND_NONSTATIC, 
							   env->currentVariableEntry );
			     
			     //cout<<"variable "<<*$1<<"'s code: "<<*$3->text<<endl;
			     
			     // and add it to environment
			     env->append( v );
			     
			     
			     mParseTree *tree = new mParseTree( NODE_NEW );
			     
			     tree->append( new mParseTree( NODE_VARIABLE, v ) );
			     
			     // append statement
			     tree->append( $3 );                                
			     
			     env->closure->append( tree );
			     
			     $$ = tree;

                           }
                           ;


// we distinguish between statements and expressions
Statement:  OtherStatement  {

              // do nothing if just importing
              if( importing ) break;

              if( $1 ){
                 mEnvironment *env = envStack->top();

                 // append to current closure
                 env->closure->append( $1 );
              }
	    }

            | Expression Semicolon {

              // do nothing if just importing
              if( importing ) break;

              // get active environment from top of stack
              // this should be closure env
              if( $1 ){
                 mEnvironment *env = envStack->top();

                 // append to current closure not environment
                 env->closure->append( $1 );

              }
              //$$ = $1;
            }

            | ReturnStatement {

              // do nothing if just importing
              if( importing ) break;

              if( $1 ){
                 mEnvironment *env = envStack->top();

                 // append to current closure not environment
                 env->closure->append( $1 );
              }
	    }
            ;



// the return statement

                  // with expression (functions only)
ReturnStatement:  RETURN Expression Semicolon {
           
                    // do nothing if just importing
	            if( importing ) break;


                    if(!$2){ $$=0; break; }    // error in expression

                    // returning expression from procedure is error
		    if(!currentMethod->isFunction()){
		      myerror();
		      cout<<"method '"<<*currentMethod->name<<"'"<<" does not return '"
			  <<$2->getType()<<"'"<<endl;
		      $$ = 0;
		      break;

		    }

		    // returning expression from constructor is error
		    if( currentMethod->isConstructor() ){
		      myerror();
		      cout<<"constructor '"<<*currentMethod->name<<"'"<<" does not return '"
			  <<$2->getType()<<"'"<<endl;
		      $$ = 0;
		      break;

		    }
                   
		    // perform implicit cast to returnType if necessary
		    if( !$2->performTypeCast( $2, currentMethod->returnType ) ){
		      myerror();
		      cout<<"cannot perform cast from '"
			  <<$2->getType()<<"' to return type '"
			  <<*currentMethod->returnType->name<<"'"<<endl;
		      $$ = 0;
		      break;
		    }


		    mParseTree *tree = new mParseTree( NODE_RETURN );

		    // append expression		    
		    tree->append( $2 );
		    $$ = tree;

                                       

                  }


                  ////////////////////////////////////////////
                  // return statement
                  ////////////////////////////////////////////
                  | RETURN OtherStatement Semicolon {
           
                    // do nothing if just importing
	            if( importing ) break;


                    if(!$2){ $$=0; break; }    // error in expression

                    // returning from procedure is error
		    if(!currentMethod->isFunction()){
		      myerror();
		      cout<<"method '"<<*currentMethod->name<<"'"<<" does not return '"
			  <<$2->getType()<<"'"<<endl;
		      $$ = 0;
		      break;

		    }

		    // returning from constructor is error
		    if( currentMethod->isConstructor() ){
		      myerror();
		      cout<<"constructor '"<<*currentMethod->name<<"'"<<" does not return '"
			  <<$2->getType()<<"'"<<endl;
		      $$ = 0;
		      break;

		    }
                   
		    // perform implicit cast to returnType if necessary
		    if( !$2->performTypeCast( $2, currentMethod->returnType ) ){
		      myerror();
		      cout<<"cannot perform cast from '"
			  <<$2->getType()<<"' to return type '"
			  <<*currentMethod->returnType->name<<"'"<<endl;
		      $$ = 0;
		      break;
		    }


		    mParseTree *tree = new mParseTree( NODE_RETURN );

		    // append statement
		    tree->append( $2 );
		    $$ = tree;

                                       

                  }

                  // without expression (procedures only)
                  | RETURN Semicolon {


                     // do nothing if just importing
                     if( importing ) break;

                     // returning nothing from function is error (except for constructors)
		    if(currentMethod->isFunction() && !currentMethod->isConstructor() ){
		      myerror();
		      cout<<"method '"<<*currentMethod->name<<"' must return '"
			  <<*currentMethod->returnType->name<<"'"<<endl;
		      $$ = 0;
		      break;

		    }


		    // method main doesn't return
		    if(*currentMethod->name == "main"){
		      $$ = 0;
		      break;
		    }


		    // create return node
		    mParseTree *tree = new mParseTree( NODE_RETURN );
		    $$ = tree;
		     
                  }
                  ;

// all possible statements except expression and return
// these are statements that can be assigned to lvalue
OtherStatement:   //EmptyStatement { $$ = 0; }
                   WhileStatement { $$ = 0; }
                  | DoWhileStatement { $$ = 0; }
                  | Block {
                    $$ = $1;
                  }
                  /*| error ';' {
                             cout<<"bad statement"<<endl;
                             yyerrok;
                           }*/
                  ;


// nothing
//EmptyStatement:   ';' // Semicolon
//                 ;


// while loop
WhileStatement: W