#include "mTestInterpreter.h"
#include "mBuiltinClasses.h"
#include "mVirtualMachine.h"
#include <dlfcn.h>
#include <stdio.h>
#include "ObjectOffsets.h"




inline void mTestInterpreter::dumpLocalVariable(int local, int *stack_ptr)
{
  cout<<"l"<<local<<" = "<<*(stack_ptr - local)<<endl;
}



inline bool mTestInterpreter::testnil(int** object)
{
  if(object == 0){
    cout<<"invocation on nil"<<endl;
    return true;   
  }

  return false;
}




mTestInterpreter::mTestInterpreter(mVirtualMachine* vm) 
  : mAbstractInterpreter( vm )
{
  
  environment = vm->getEnvironment();




  //////////////////////////////////////////////
  // setup flag and system variables here
  //////////////////////////////////////////////

  terminated=false;

  debug = false;
  

  // setup stack ptr
  stack_origin = vm->getStack();
  stack_ptr = stack_origin; /// -> 1st free item
  

  // system flags register
  // bits: 
  //     0 - block_return ( set to 0 if block didn't return, to 1 on return )
  
  flags = 0;


  // number of implicitly pushed parent references inside invocation.
  nPushed = 0;


  

}

mTestInterpreter::~mTestInterpreter()
{

}


int mTestInterpreter::createDescriptor(int selfish, int self)
{
  int* descriptor = (int*)vm->allocate( 8 );

  *descriptor = selfish;
  *(descriptor+1) = self;
  return (int)descriptor;    
}






void mTestInterpreter::execute(mLoadedClass* theClass)
{
  // we get index of main method from method table
  int* methodTable =  (int*) ( theClass->address + theClass->head.literalLength + theClass->head.metaLength 
			       + *(int*)(theClass->address + theClass->head.literalLength + 
					 theClass->head.metaLength + 4) );

  int addressOfMain = methodTable[ 3 * theClass->head.mainIndex ];

  executeMethod( addressOfMain );
  
  
}



void mTestInterpreter::executeMethod(int address)
{
  // structure of class data in memory:

  // |  literals   |   <-- starting address
  // | class info  |
  // | object table|
  // |  vtable(s)  |
  // |  mtable     |
  

  // we get index of main method from method table
  //int* methodTable =  (int*) ( theClass->address + theClass->head.literalLength + theClass->head.metaLength + 
  //		       *(int*)(theClass->address + theClass->head.literalLength + theClass->head.metaLength + 4) ); 




  char* addressOfMain = (char*)address; //methodTable[ 3 * theClass->head.mainIndex ];  


  


  short opcode=0;

  

  // system flags register
  // bits: 
  //     0 - block_return ( set to 0 if block didn't return, to 1 on return )
  
  //flags = 0;


  // number of implicitly pushed parent references inside invocation.
  //nPushed = 0;


  if(debug) cout<<"starting execution of method at "<<(int)addressOfMain<<endl;


  if( debug ) cout<<"stack_ptr="<<(int)stack_ptr<<endl;
  
  //int dest=0, src=0;

  // main loop
  while( !terminated ){

    // fetch opcode
    opcode = *(short*)addressOfMain;

    switch( opcode ){

    ///////////////////////////////////////////////////////////////////
    // Integer 
    ///////////////////////////////////////////////////////////////////

    // new Integer with default value
    case ClassInteger::NEW: {

      if( debug )
	cout<<"new Integer with default value"<<endl;


      int* classData = (int*)vm->getIntegerClass();

      int* object = vm->create_object(classData);

      // integer value comes last
      *(object + INTEGER_OFFSET) = 0;      

      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);

      addressOfMain+=6;
    }
      break;



    // new Integer from literal
    case ClassInteger::NEW_FROM_LITERAL: {

      if( debug )
	cout<<"new Integer from literal "<<**(int**)(addressOfMain + 6)<<endl;

      int* classData = (int*)vm->getIntegerClass();

      int* object =  vm->create_object(classData);

      
      

      // integer value comes last
      *(object + INTEGER_OFFSET) = **(int**)(addressOfMain + 6);      

      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);



      addressOfMain+=10;
    }
      break;




    // new Integer from Integer
    case ClassInteger::NEW_COPY: {

      if( debug ) cout<<"new Integer from copy"<<endl;

      int* classData = (int*)vm->getIntegerClass();

      int* object =  vm->create_object(classData);

      // integer value from copy      

      *(object + INTEGER_OFFSET) = *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +INTEGER_OFFSET);
      
      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);          


      addressOfMain+=10;
      break;
    }


    // new Integer from Double
    case ClassInteger::NEW_FROM_DOUBLE: {

      if( debug ) cout<<"new Integer from Double"<<endl;

      int* classData = (int*)vm->getIntegerClass();

      int* object =  vm->create_object(classData);

      // integer value from copy      

      *(object + INTEGER_OFFSET) = 
	(int)*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +DOUBLE_OFFSET);
      
      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);    

      addressOfMain+=10;
      break;
    }


    // add two Integers
    case ClassInteger::ADD: {

      if( debug ) cout<<"add 2 Integers"<<endl;    

      // source value is in local variable of type Integer
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +INTEGER_OFFSET) += 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +INTEGER_OFFSET);

      addressOfMain+=10;
      break;

    }

    // sub two Integers
    case ClassInteger::SUB: {

      if( debug ) cout<<"sub 2 Integers"<<endl;

      // source value is in local variable of type Integer
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +INTEGER_OFFSET) -= 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +INTEGER_OFFSET);

      addressOfMain+=10;
      break;
    }


    // mul two Integers
    case ClassInteger::MUL: {

      if( debug ) cout<<"mul 2 Integers"<<endl;

      // source value is in local variable of type Integer
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +INTEGER_OFFSET) *= 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +INTEGER_OFFSET);

      addressOfMain+=10;
      break;   

    }


    // div two Integers
    case ClassInteger::DIV: 

      if( debug ) cout<<"div 2 Integers"<<endl;

      // source value is in local variable of type Integer
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +INTEGER_OFFSET) /= 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +INTEGER_OFFSET);

      addressOfMain+=10;
      break;


    // mod Integer
    case ClassInteger::MOD: 

      if( debug ) cout<<"mod Integer"<<endl;

      // source value is in local variable of type Integer
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +INTEGER_OFFSET) %= 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +INTEGER_OFFSET);

      addressOfMain+=10;
      break;


    // set if greater
    case ClassInteger::IFGT: 

      if( debug ) cout<<"set if greater than Integer"<<endl;

      // set boolean variable
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Integer
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +INTEGER_OFFSET) > 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +INTEGER_OFFSET) ? 1 : 0;

      addressOfMain+=14;
      break;


    // set if greater or equal
    case ClassInteger::IFGTEQ: 

      if( debug ) cout<<"set if greater or equal than Integer"<<endl;

      // set boolean variable
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Integer
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +INTEGER_OFFSET) >= 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +INTEGER_OFFSET) ? 1 : 0;

      addressOfMain+=14;
      break;

    // set if less
    case ClassInteger::IFLS: 

      if( debug ) cout<<"set if less than Integer"<<endl;

      // set boolean variable
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Integer
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +INTEGER_OFFSET) < 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +INTEGER_OFFSET) ? 1 : 0;

      addressOfMain+=14;
      break;

    // set if less or equal
    case ClassInteger::IFLSEQ: 

      if( debug ) cout<<"set if less or equal than Integer"<<endl;

      // set boolean variable
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Integer
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +INTEGER_OFFSET) <= 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +INTEGER_OFFSET) ? 1 : 0;

      addressOfMain+=14;
      break;


    // set if equal
    case ClassInteger::IFEQ: 

      if( debug ) cout<<"set if equal"<<endl;


      // set boolean variable
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Integer
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +INTEGER_OFFSET) == 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +INTEGER_OFFSET) ? 1 : 0;


      addressOfMain+=14;
      break;

    // set if not equal
    case ClassInteger::IFNEQ: 

      if( debug ) cout<<"set if not equal"<<endl;

      // set boolean variable
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Integer
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +INTEGER_OFFSET) != 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +INTEGER_OFFSET) ? 1 : 0;

      addressOfMain+=14;
      break;


    // set value
    case ClassInteger::SET: {

      if( debug )
	cout<<"set Integer value"<<endl;

      // store value to our object
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +INTEGER_OFFSET) = 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +INTEGER_OFFSET);
      
      
      addressOfMain+=10;
    }
      break;


    // print Integer
    case ClassInteger::PRINT: 
            

      printf("Integer=%d\n", *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +INTEGER_OFFSET));
      addressOfMain+=6;
      break;



    // increment
    case ClassInteger::INC: {

      if( debug ) cout<<"inc Integer"<<endl;

      // source value is in local variable of type Integer
      (*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +INTEGER_OFFSET))++;

      addressOfMain+=6;
      break;
    }


    // decrement
    case ClassInteger::DEC: {

      if( debug ) cout<<"dec Integer"<<endl;

      // source value is in local variable of type Integer
      (*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +INTEGER_OFFSET))--;

      addressOfMain+=6;
      break;
    }


    ///////////////////////////////////////////////////////////////////
    // Double 
    ///////////////////////////////////////////////////////////////////

    // new Double with default value
    case ClassDouble::NEW: {
     

      if( debug ) cout<<"new Double with default value"<<endl;

      int* classData = (int*)vm->getDoubleClass();

      int* object =  vm->create_object(classData);

      // integer value comes last
      *(double*)(object + DOUBLE_OFFSET) = 0.0;      

      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);
      

      addressOfMain+=6;
      break;
    }


    // new Double from literal
    case ClassDouble::NEW_FROM_LITERAL: {

      if( debug ) cout<<"new Double from literal"<<endl;

      int* classData = (int*)vm->getDoubleClass();

      int* object =  vm->create_object(classData);

      // double value comes last
      *(double*)(object + DOUBLE_OFFSET) = **(double**)(addressOfMain + 6);      

      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);


      addressOfMain+=10;
      break;
    }


    // new Double from Double
    case ClassDouble::NEW_COPY: {

      if( debug ) cout<<"new Double from copy"<<endl;

      int* classData = (int*)vm->getDoubleClass();

      int* object = vm->create_object(classData);

      // value from copy      
      *(double*)(object + DOUBLE_OFFSET) = 
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +DOUBLE_OFFSET);
      
      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);

      addressOfMain+=10;
      break;

    }


    // new Double from Integer
    case ClassDouble::NEW_FROM_INTEGER: {

      if( debug ) cout<<"new Double from Integer"<<endl;

      int* classData = (int*)vm->getDoubleClass();

      int* object = vm->create_object(classData);

      // value from copy      
      *(double*)(object + DOUBLE_OFFSET) =
	(double)*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +DOUBLE_OFFSET);
      
      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);    

      addressOfMain+=10;
      break;
    }


    // add two Doubles
    case ClassDouble::ADD: {

      if( debug ) cout<<"add 2 Doubles"<<endl;

      // source value is in local variable of type Double

      *(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +DOUBLE_OFFSET) +=
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +DOUBLE_OFFSET);

      addressOfMain+=10;
      break;

    }

    // sub two Doubles
    case ClassDouble::SUB: {

      if( debug ) cout<<"sub 2 Doubles"<<endl;

      // source value is in local variable of type Double
      
      *(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +DOUBLE_OFFSET) -=
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +DOUBLE_OFFSET);
      

      addressOfMain+=10;
      break;
    }

    // mul two Doubles
    case ClassDouble::MUL: {

      if( debug ) cout<<"mul 2 Doubles"<<endl;

      // source value is in local variable of type Double
      *(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +DOUBLE_OFFSET) *=
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +DOUBLE_OFFSET);

      addressOfMain+=10;
      break;   
    }

    // div two Doubles
    case ClassDouble::DIV: {

      if( debug ) cout<<"div 2 Doubles"<<endl;

      // source value is in local variable of type Double
      *(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +DOUBLE_OFFSET) /=
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +DOUBLE_OFFSET);

      addressOfMain+=10;
      break;
    }


    // set if greater
    case ClassDouble::IFGT: 

      if( debug ) cout<<"set if greater than Double"<<endl;

      // set boolean variable
      *(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Double
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +DOUBLE_OFFSET) > 
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +DOUBLE_OFFSET) ? 1 : 0;

      addressOfMain+=14;
      break;


    // set if greater or equal
    case ClassDouble::IFGTEQ: 

      if( debug ) cout<<"set if greater or equal than Double"<<endl;

      // set boolean variable
      *(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Integer
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +DOUBLE_OFFSET) >= 
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +DOUBLE_OFFSET) ? 1 : 0;

      addressOfMain+=14;
      break;

    // set if less
    case ClassDouble::IFLS: 

      if( debug ) cout<<"set if less than Double"<<endl;

      // set boolean variable
      *(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Integer
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +DOUBLE_OFFSET) < 
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +DOUBLE_OFFSET) ? 1 : 0;

      addressOfMain+=14;
      break;

    // set if less or equal
    case ClassDouble::IFLSEQ: 

      if( debug ) cout<<"set if less or equal than Double"<<endl;

      // set boolean variable
      *(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Integer
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +DOUBLE_OFFSET) <= 
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +DOUBLE_OFFSET) ? 1 : 0;

      addressOfMain+=14;
      break;


    // set if equal
    case ClassDouble::IFEQ: 

      if( debug ) cout<<"set if equal"<<endl;

      // set boolean variable
      *(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Integer
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +DOUBLE_OFFSET) == 
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +DOUBLE_OFFSET) ? 1 : 0;

      addressOfMain+=14;
      break;

    // set if not equal
    case ClassDouble::IFNEQ: 

      if( debug ) cout<<"set if not equal"<<endl;

      // set boolean variable
      *(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Integer
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +DOUBLE_OFFSET) != 
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +DOUBLE_OFFSET) ? 1 : 0;

      addressOfMain+=14;
      break;


    // set value
    case ClassDouble::SET: {

      if( debug )
	cout<<"set Double value"<<endl;

      // store value to our object
      *(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +DOUBLE_OFFSET) = 
	*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +DOUBLE_OFFSET);
      
      addressOfMain+=10;
    }
      break;


    // print Double
    case ClassDouble::PRINT: 

      printf("Double=%f\n",*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain +2))+1)+DOUBLE_OFFSET));
      addressOfMain+=6;
      break;


    // increment
    case ClassDouble::INC: {

      if( debug ) cout<<"inc Double"<<endl;

      // source value is in local variable of type Integer
      (*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +DOUBLE_OFFSET))++;

      addressOfMain+=6;
      break;
    }


    // decrement
    case ClassDouble::DEC: {

      if( debug ) cout<<"dec Double"<<endl;

      // source value is in local variable of type Integer
      (*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +DOUBLE_OFFSET))--;

      addressOfMain+=6;
      break;
    }


    ///////////////////////////////////////////////////////////////////
    // String
    ///////////////////////////////////////////////////////////////////

    // new String with default value
    case ClassString::NEW: {

      if( debug )
	cout<<"new String with default value"<<endl;      

      int* classData = (int*)vm->getStringClass();

      int* object = vm->create_object(classData);

      char* chars = (char*)vm->allocate( 1 );
      *chars = '\0';

      // first is length
      *(object + STRING_OFFSET) = 0;
      // second is ptr to chars
      *(object + STRING_OFFSET+1) = (int)chars;

      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);
                

      addressOfMain+=6;
    }
      break;


    // new String from literal
    case ClassString::NEW_FROM_LITERAL: {

      if( debug ){	
	cout<<"new String from literal ";
	for(int i=0;i<*(int*)(addressOfMain + 10);i++)
	  cout<<*((char*)*(int*)(addressOfMain + 6) + i);
	cout<<endl;

      }
            
      int* classData = (int*)vm->getStringClass();

      int* object = vm->create_object(classData);


      int length = *(int*)(addressOfMain + 10);  // length of string literal

      
      char* chars = (char*)vm->allocate( length + 1 );

      memcpy( chars, *(int**)(addressOfMain + 6), length );
      *(chars + length) = '\0';  // add null

      // first is length
      *(object + STRING_OFFSET) = length;
      // second is ptr to chars
      *(object + STRING_OFFSET+1) = (int)chars;
            

      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);         

      addressOfMain+=14;
    }
      break;



    // new String from String
    case ClassString::NEW_COPY: {

      if( debug ) cout<<"new String from copy"<<endl;

      int* classData = (int*)vm->getStringClass();

      int* object = vm->create_object(classData);


      char* str = *(char**)(*((int**)*(stack_ptr - *(int*)(addressOfMain +6))+1)+STRING_OFFSET+1);      

      int length = strlen(str); 

      char* chars = (char*)vm->allocate( length + 1 );
      memcpy( chars, str, length );
      *(chars + length) = '\0';  // add null

      // first is length
      *(object + STRING_OFFSET) = length;
      // second is ptr to chars
      *(object + STRING_OFFSET+1) = (int)chars;
      

      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);             

      addressOfMain+=10;
      break;
    }


    // new String from Double
    case ClassString::NEW_FROM_DOUBLE: {

      if( debug ) cout<<"new String from Double"<<endl;

      int* classData = (int*)vm->getStringClass();

      int* object = vm->create_object(classData);


      char* chars = (char*)vm->allocate( 40 );
      sprintf(chars, "%f",*(double*)(*((int**)*(stack_ptr - *(int*)(addressOfMain +6))+1)+DOUBLE_OFFSET) );

      // first is length
      *(object + STRING_OFFSET) = strlen(chars);
      // second is ptr to chars
      *(object + STRING_OFFSET+1) = (int)chars;
      

      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);
            

      addressOfMain+=10;
      break;
    }

    // new String from Integer
    case ClassString::NEW_FROM_INTEGER: {

      if( debug ) cout<<"new String from Integer"<<endl;

      int* classData = (int*)vm->getStringClass();
      
      int* object = vm->create_object(classData);
    

      char* chars = (char*)vm->allocate( 40 );
      sprintf(chars, "%d",*(*((int**)*(stack_ptr - *(int*)(addressOfMain +6))+1)+INTEGER_OFFSET) );



      // first is length
      *(object + STRING_OFFSET) = strlen(chars);
      // second is ptr to chars
      *(object + STRING_OFFSET+1) = (int)chars;
      

      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);


      addressOfMain+=10;
      break;
    }

      
   // new String from Boolean
    case ClassString::NEW_FROM_BOOLEAN: {

      if( debug ) cout<<"new String from Boolean"<<endl;

      int* classData = (int*)vm->getStringClass();
      
      int* object = vm->create_object(classData);
    

      char* chars = (char*)vm->allocate( 40 );

      int bvalue = *(*((int**)*(stack_ptr - *(int*)(addressOfMain +6))+1)+BOOLEAN_OFFSET);

      if(bvalue==1)
	sprintf(chars, "%s","true" );
      else
	sprintf(chars, "%s","false" );
      


      // first is length
      *(object + STRING_OFFSET) = strlen(chars);
      // second is ptr to chars
      *(object + STRING_OFFSET+1) = (int)chars;
      

      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);


      addressOfMain+=10;
      break;
    }



    // add two Strings
    case ClassString::ADD: {

      if( debug ) cout<<"add 2 Strings"<<endl;

      int* classData = (int*)vm->getStringClass();

      int* object = vm->create_object(classData);
      
      char* head = *(char**)(*((int**)*(stack_ptr - *(int*)(addressOfMain +2))+1)+STRING_OFFSET+1);
      char* tail = *(char**)(*((int**)*(stack_ptr - *(int*)(addressOfMain +6))+1)+STRING_OFFSET+1);      

      int hlen = strlen(head);  int tlen = strlen(tail);

      char* chars = (char*)vm->allocate( hlen + + tlen + 1 );

      memcpy( chars, head, hlen ); memcpy( chars + hlen, tail, tlen );
      *(chars + hlen + tlen ) = '\0';  // add null
     

      // first is length
      *(object + STRING_OFFSET) = hlen + tlen;
      // second is ptr to chars
      *(object + STRING_OFFSET+1) = (int)chars;
      

      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);

      addressOfMain+=10;
      break;

    }


    // set value
    case ClassString::SET: {

      if( debug ) cout<<"set String value"<<endl;      

      char* str = *(char**)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +STRING_OFFSET+1);
      int len = *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +STRING_OFFSET);
      
           

      // alloc space for String object
      char* chars = (char*)vm->allocate( len + 1 );

      memcpy( chars, str, len );
      *(chars + len) = '\0';  // add null

      // store length and chars to object
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain +2))+1)+STRING_OFFSET) = len;
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain +2))+1)+STRING_OFFSET +1) = (int)chars;      
	
      addressOfMain+=10;
      break;
    }

    
      

   // set if equal
    case ClassString::IFEQ: 

      if( debug ) cout<<"set if equal"<<endl;

      // set boolean variable
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Integer
	strcmp( *(char**)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +STRING_OFFSET+1), 
	*(char**)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +STRING_OFFSET+1)) == 0 ? 1 : 0;

      addressOfMain+=14;
      break;


    // set if not equal
    case ClassString::IFNEQ: 

      if( debug ) cout<<"set if not equal"<<endl;

      // set boolean variable
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Integer
	strcmp( *(char**)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +STRING_OFFSET+1), 
	*(char**)(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +STRING_OFFSET+1)) != 0 ? 1 : 0;

      addressOfMain+=14;
      break;
      


    // print String
    case ClassString::PRINT: 

      printf("%s\n",*(char**)(*((int**)*(stack_ptr - *(int*)(addressOfMain +2))+1)+STRING_OFFSET+1));

      addressOfMain+=6;
      break;


    ///////////////////////////////////////////////////////////////////
    // Boolean
    ///////////////////////////////////////////////////////////////////

    // new Boolean with default value true
    case ClassBoolean::NEW: {

      if( debug ) cout<<"new Boolean with default value"<<endl;


      int* classData = (int*)vm->getBooleanClass();

      int* object = vm->create_object(classData);
     

      // boolean value comes last
      *(object + BOOLEAN_OFFSET) = 1;      // default is true

      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);
      

      addressOfMain+=6;
      break;
    }


    // new Boolean from literal
    case ClassBoolean::NEW_FROM_LITERAL: {

      if( debug ) cout<<"new Boolean from literal"<<endl;

      int* classData = (int*)vm->getBooleanClass();

      int* object = vm->create_object(classData);

      
      // boolean value comes last
      *(object + BOOLEAN_OFFSET) = (int)**(char**)(addressOfMain + 6);

      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);

      addressOfMain+=10;
      break;
    }


    // new Boolean from Boolean
    case ClassBoolean::NEW_COPY: {

      if( debug ) cout<<"new Boolean from copy"<<endl;

      int* classData = (int*)vm->getBooleanClass();
      int* object = vm->create_object(classData);


      // boolean value
      *(object + BOOLEAN_OFFSET) = *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +BOOLEAN_OFFSET);
      
      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);
      

      addressOfMain+=10;
      break;

    }

  
    // set value
    case ClassBoolean::SET: {

      if( debug )
	cout<<"set Boolean value"<<endl;

      // store value to our object
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +BOOLEAN_OFFSET) = 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +BOOLEAN_OFFSET);      
      
      addressOfMain+=10;
    }
      break;


    // set boolean if both booleans are true
    case ClassBoolean::AND: 

      if( debug ) cout<<"boolean and"<<endl;

      // set boolean variable
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Boolean
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +BOOLEAN_OFFSET) == 1 && 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +BOOLEAN_OFFSET) == 1 ? 1 : 0;

      addressOfMain+=14;
      break;

     
    // set boolean if at least one of the booleans are true
    case ClassBoolean::OR: 

      if( debug ) cout<<"boolean or"<<endl;

      // set boolean variable
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Boolean
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +BOOLEAN_OFFSET) == 1 || 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +BOOLEAN_OFFSET) == 1 ? 1 : 0;

      addressOfMain+=14;
      break;


    // set boolean to notboolean 
    case ClassBoolean::NOT: 

      if( debug ) cout<<"boolean not"<<endl;

      // set boolean variable
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Boolean
	! *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +BOOLEAN_OFFSET);

      addressOfMain+=10;
      break;

      

    // set if equal
    case ClassBoolean::IFEQ: 

      if( debug ) cout<<"set if equal"<<endl;

      // set boolean variable
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Boolean
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +BOOLEAN_OFFSET) == 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +BOOLEAN_OFFSET) ? 1 : 0;

      addressOfMain+=14;
      break;


      
    // set if not equal
    case ClassBoolean::IFNEQ: 

      if( debug ) cout<<"set if not equal"<<endl;

      // set boolean variable
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) = 
      
	// source value is in local variable of type Boolean
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +BOOLEAN_OFFSET) != 
	*(*((int**)*(stack_ptr - *(int*)(addressOfMain + 6))+1) +BOOLEAN_OFFSET) ? 1 : 0;

      addressOfMain+=14;
      break;
      
      

    // print Boolean
    case ClassBoolean::PRINT: {

      if( debug ) cout<<"print Boolean"<<endl;

      int value = *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +BOOLEAN_OFFSET);
      

      if(value==0)	
	printf("false\n");
      else
	printf("true\n");

      addressOfMain+=6;
      
      break;

    }



    ///////////////////////////////////////////////////////////////////
    // Object
    ///////////////////////////////////////////////////////////////////

    // new Object
    case ClassObject::NEW: {

      if( debug ) cout<<"new Object"<<endl;


      // class data for class Object
      //int* classData = (int*)vm->getObjectClass();

      //int* object = vm->create_object(classData);
        
      cout<<"ClassObject::NEW NOT IMPLEMENTED"<<endl;

      // store reference to dest variable
      //*(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+2);
     

      addressOfMain+=6;
      break;
    }

    


    // new Object from Object
    case ClassObject::NEW_COPY: {

      if( debug ) cout<<"new Object from copy"<<endl;

      int* classData = (int*)vm->getObjectClass();

      int* object = vm->create_object(classData);

      
      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)(object+4);

      addressOfMain+=10;
      break;

    }
    



    // print Object
    case ClassObject::PRINT: {

      if( debug ) cout<<"print Object"<<endl;      
	
      printf("Object@%d\n", (int)((int*)*(stack_ptr - *(int*)(addressOfMain + 2 )) +1));     

      addressOfMain+=6;
      
      break;

    }


    // getClass: creates instance of Class and stores it to dest variable
    case ClassObject::GETCLASS: {

      if( debug ) cout<<"getClass from Object"<<endl;      
	
      mString className("Class");
      mLoadedClass *theClass = vm->findAndLoadClass( &className );
      
      int* classData = (int*)theClass->dataStartAddress;
          

      // creates object already pointing to correct descriptor
      int* object = vm->create_user_object( classData );            


      int *originalObject = (int*) *(object + 1);            

      int nDescriptors = *(classData+3);


      // Class object includes address of classdata
      // we need classdata of receiver object
      int* self = (int*)*(stack_ptr - *(int*)(addressOfMain + 2));
      int otherClassData = *(int*)*(self + 1);
      *(originalObject + CLASS_OFFSET) = otherClassData;


      *(stack_ptr - *(int*)(addressOfMain + 6))=(int)object;


      // reserve stack and push current execution address and reference
      stack_ptr+=2;
      *(stack_ptr-2) = (int)object;  // reference
      *(stack_ptr-1) = (int)addressOfMain + 10;  // address of next instruction
                     

      // zigzag: index of constructor method in Class is the first method (0) in vtable for Class
      int constructorIndex = 0;
      addressOfMain = (char*)*( originalObject + 2 + nDescriptors*2 + 3*  *(*(int**)object + 
									    constructorIndex) );
            
      
      break;

    }
    


    ///////////////////////////////////////////////////////////////////
    // Core 
    ///////////////////////////////////////////////////////////////////


    // termination
    case ClassCore::TERMINATE:

      if( debug ) cout<<"terminating"<<endl;

      terminated = true;
      addressOfMain+=2;
      break;


    // allocate stack frame
    case ClassCore::NEW_FRAME:

      if( debug )
	cout<<"allocating stack frame of "<<*(int*)(addressOfMain+2)<<endl;

      stack_ptr += *(int*)(addressOfMain+2);

      addressOfMain+=6;
      break;

    // allocate stack frame and copies locals into it
    // static methods have different frame allocation (no self on stack which means old stack_ptr
    // starts 4 bytes lower )
    case ClassCore::NEW_FRAME_COPY: {

      if( debug )
	cout<<"allocating copy stack frame of "<<*(int*)(addressOfMain+2)<<endl;

      int nItems = *(int*)(addressOfMain+6);  // # of locations to copy into local frame

      int size = *(int*)(addressOfMain+2);      


      // this is important: old stack_ptr is as it was before invocation i.e. before implicit
      // push of self and return address and possible parent variables
      int old_stack_ptr = (int)(stack_ptr - 2 - nPushed);

      nPushed=0;

      // allocate
      stack_ptr += size;

      // push old stack_ptr
      *(stack_ptr-1) = old_stack_ptr;

      // now copy items from old stack into this frame
      for(int i=1;i<=nItems;i++) *(stack_ptr-i-1) = *(stack_ptr-size-i);


      addressOfMain+=10;
      break;

    }

    // free stack frame
    case ClassCore::FREE_FRAME:

      if( debug )
	cout<<"freeing stack frame of "<<*(int*)(addressOfMain+2)<<endl;

      stack_ptr -= *(int*)(addressOfMain+2);

      addressOfMain+=6;
      break;

    // push local onto frame
    case ClassCore::PUSH: {

      if( debug )
          cout<<"pushing local l"<<*(int*)(addressOfMain+2)<<endl;

      int reference = *(stack_ptr - *(int*)(addressOfMain + 2));
      
      *stack_ptr = reference;
      stack_ptr++;

      addressOfMain+=6;
      break;
    }


    // push local address onto frame
    case ClassCore::PUSHREF: {

      if( debug )
          cout<<"pushing local address l"<<*(int*)(addressOfMain+2)<<endl;

      int reference = (int)(stack_ptr - *(int*)(addressOfMain + 2));
      
      //cout<<"pushed reference is "<<reference<<endl;

      *stack_ptr = reference;
      stack_ptr++;

      addressOfMain+=6;
      break;
    }
    

    // push direct value onto frame
    case ClassCore::PUSHD: {

      if( debug )
          cout<<"pushing direct value "<<*(int*)(addressOfMain+2)<<endl;

      
      *stack_ptr = *(int*)(addressOfMain + 2);
      stack_ptr++;

      addressOfMain+=6;
      break;
    }


    // pop local from frame into local
    case ClassCore::POP:

      if( debug )
          cout<<"popping local into l"<<*(int*)(addressOfMain+2)<<endl;

      *(stack_ptr - *(int*)(addressOfMain + 2)) = *(--stack_ptr);
      

      addressOfMain+=6;
      break;  

    // copy reference from one location to another
    case ClassCore::COPY:

      if( debug )
	cout<<"copying reference "<<endl;

      *(stack_ptr - *(int*)(addressOfMain + 2)) = 
	*(stack_ptr - *(int*)(addressOfMain + 6));

      addressOfMain+=10;
      break;


    // copy reference from one location to another which is itself a refrence
    case ClassCore::COPYREF: {

      if( debug )
	cout<<"copying reference to reference"<<endl;

      // stack address of variable
      int* reference = (int*)*(stack_ptr - *(int*)(addressOfMain + 2));

      *reference = *(stack_ptr - *(int*)(addressOfMain + 6));

      addressOfMain+=10;
      break;
    }


    // dereference local address onto frame
    case ClassCore::DEREF: {

      if( debug )
          cout<<"dereferencing local variable l"<<*(int*)(addressOfMain+6)<<endl;

      int* reference = (int*)*(stack_ptr - *(int*)(addressOfMain + 6));
      
      *(stack_ptr - *(int*)(addressOfMain + 2)) = *reference;

      addressOfMain+=10;
      break;
    }


    // sets local variable reference to nil
    case ClassCore::NILIFY:

      if( debug )
	cout<<"nilifying reference "<<endl;

      // set to 0
      *(stack_ptr - *(int*)(addressOfMain + 2)) = 0;

      addressOfMain+=6;
      break;



    // creates new object
    case ClassCore::NEW_OBJECT: {

      if( debug )
          cout<<"creating new object ";


      // get object size
      int* classData = *(int**)(addressOfMain + 6);      

      // returns object already pointing to correct descriptor
      int object = (int)vm->create_user_object( classData );

      int *originalObject = (int*) *((int*)object + 1);            

      int nDescriptors = *(classData+3);

      // store reference to dest variable
       *(stack_ptr - *(int*)(addressOfMain + 2)) = object;

      // reserve stack and push current execution address and reference
      stack_ptr+=2;
      *(stack_ptr-2) = object;  // reference
      *(stack_ptr-1) = (int)addressOfMain + 14;  // address of next instruction
                     

      addressOfMain = (char*)*( originalObject + 2 + nDescriptors*2 + 3*  *(*(int**)object + *(int*)(addressOfMain + 10)) );
      

      break;
    }



    // returns from method (returns nothing, doesn't clean up)
    case ClassCore::RETURN_1: {

      if( debug )
          cout<<"returning from method"<<endl;

      // restore old return address and old stack_ptr
      // 
      // | old_stack_ptr| <-- stack_ptr 
      // | ret addr     |
      // | self         |
      // | param1,2,... |
      // | local1,2,... |
      

      // clear block_return flag (it may have been set by a preceeding Block.return)
      if( (flags & 1) != 0){
	flags &= 0xfffffffe;
      }


      addressOfMain = (char*)*(stack_ptr - 2);  // continue at old address
      stack_ptr = (int*)*(stack_ptr - 1);             // restore old stack_ptr

      // check if interpreting single method from native

      if( addressOfMain == 0)
	terminated = true;
      

      break;  
    }


    // returns local variable from method
    case ClassCore::RETURN_1_LOCAL: {

      if( debug )
          cout<<"returning local l"<<*(int*)(addressOfMain + 2) << " from method"<<endl;

      // restore old return address and old stack_ptr
      //                 <-- stack_ptr
      // | old_stack_ptr|  -1
      // | result local |  -2
      // | ret addr     |  -3
      // | self         |
      // | param1,2,... |
      // | local1,2,... |
      

      int ref = *(stack_ptr - *(int*)(addressOfMain + 2));           // reference in local variable
      
      // clear block_return flag (it may have been set by a preceeding Block.return)
      if( (flags & 1) != 0){
	flags &= 0xfffffffe;
	stack_ptr--;  // remove result pushed by Block
      }

      int dest = *(stack_ptr - 2);    // this is where we store our result - ref
      

      addressOfMain = (char*)*(stack_ptr - 3);        // continue at old address
      stack_ptr = (int*)*(stack_ptr - 1);             // restore old stack_ptr


      stack_ptr--;           // remove result destination which was pushed implicitly in invoke_1_f !!!
      

      // store reference to result destination
      //stack_ptr++;
      *(stack_ptr - dest) = ref;

      // check if interpreting single method from native
      if( addressOfMain == 0)
	terminated = true;

      break;  
    }

  
    // returns from method (returns nothing, cleans up arguments)
    case ClassCore::RETURN_2: {

      if( debug )
          cout<<"returning from method and cleanup arguments"<<endl;

      // restore old return address and old stack_ptr
      // 
      // | old_stack_ptr| <-- stack_ptr 
      // | ret addr     |
      // | self         |
      // | param1,2,... |
      // | local1,2,... |

      // clear block_return flag (it may have been set by a preceeding Block.return)
      if( (flags & 1) != 0){
	flags &= 0xfffffffe;
      }

      int nArgs = *(int*)(addressOfMain + 2);

      addressOfMain = (char*)*(stack_ptr - 2);        // continue at old address
      stack_ptr = (int*)*(stack_ptr - 1);             // restore old stack_ptr

      // clean up args
      stack_ptr -= nArgs;

      // check if interpreting single method from native
      if( addressOfMain == 0)
	terminated = true;

      break;  
    }

    // returns local variable from method and cleans up arguments
    case ClassCore::RETURN_2_LOCAL: {

      if( debug )
          cout<<"returning local from method and cleanup arguments"<<endl;

      // restore old return address and old stack_ptr
      //                 <-- stack_ptr
      // | old_stack_ptr|  -1
      // | result local |  -2
      // | ret addr     |  -3
      // | self         |
      // | param1,2,... |
      // | local1,2,... |
      

      int ref = *(stack_ptr - *(int*)(addressOfMain + 2));           // reference in local variable

      // clear block_return flag (it may have been set by a preceeding Block.return)
      if( (flags & 1) != 0){
	flags &= 0xfffffffe;
	stack_ptr--;   // remove result pushed by Block
      }

      int nArgs = *(int*)(addressOfMain + 6);
      int dest = *(stack_ptr - 2);    // this is where we store our result - ref

      addressOfMain = (char*)*(stack_ptr - 3);        // continue at old address
      stack_ptr = (int*)*(stack_ptr - 1);             // restore old stack_ptr
      stack_ptr--;           // remove result destination which was pushed implicitly in invoke_1_f !!!

      // clean up args
      stack_ptr -= nArgs;

      // store reference to result destination     
      *(stack_ptr - dest) = ref;
      
      // check if interpreting single method from native
      if( addressOfMain == 0)
	terminated = true;

      break;  
    }


    // invokes virtual procedure method
    case ClassCore::INVOKE_1_P: {

      if( debug )
	cout<<"invoking virtual procedure method #"<<*(int*)(addressOfMain + 6)<<endl;
      
      int** object = (int**) *(stack_ptr - *(int*)(addressOfMain + 2));

      if(terminated = testnil(object)) break;

      int index = *(int*)(addressOfMain + 6);
      int offset = *(int*)(addressOfMain + 10);

      
      if(offset != 0)
	object = (int**)((int)*(object + 1) + offset);  // select descriptor

     
      // method ptr -> self                +  3*  index in vtable
      int *method = *(object+1) + 2 + *(int*)(**(object+1) +12) * 2 +  3*  *( *object + index);

       
      //cout<<"# descriptors="<<*(int*)(**(object+1) +12)<<endl;

      // if environment exist we push its values
      int *env = (int*)*(method+2);
      

      if(env){

	nPushed= *env;
	stack_ptr += *env;
	for(int i=1;i<=*env;i++) *(stack_ptr-i) = *(env+i);

      }

      // if object addresses of actual self and this object are identical then self is set to this object
      // otherwise we take it from method table
      //int self = (int)object;     
      

      /*if((int)*(object+1) != (int)*((int*)*(method+1) +1)){
	self = *(method+1);
	cout<<"taking self from method table"<<endl;

	}*/


      // push return address and reference
      stack_ptr+=2;
      *(stack_ptr-2) = /*self;*/ *(method+1);   // actual self
      *(stack_ptr-1) = (int)addressOfMain + 14;  // address of next instruction
      
      addressOfMain = (char*)*method;     


      break;  
    }

    
    // invokes virtual function method
    case ClassCore::INVOKE_1_F: {

      if( debug )
	cout<<"invoking virtual function method #"<<*(int*)(addressOfMain + 6)<<", result to "
	    <<*(int*)(addressOfMain + 10)<<endl;
      
      int** object = (int**) *(stack_ptr - *(int*)(addressOfMain + 2));


      if(terminated = testnil(object)) break;

      int index = *(int*)(addressOfMain + 6);
      int offset = *(int*)(addressOfMain + 14);

      if(offset != 0)
	object = (int**)((int)*(object + 1) + offset);  // select descriptor

      // method ptr ->
      int *method = *(object+1) + 2 + *(int*)(**(object+1) +12) * 2 +  3*  *( *object + index);      


      
      // if environment exist we push its values
      int *env = (int*)*(method+2);
      

      if(env){

	// push all references
	//cout<<"invoking function with environment"<<endl;
	
	nPushed= *env;
	stack_ptr += *env;
	for(int i=1;i<=*env;i++) *(stack_ptr-i) = *(env+i);

      }

      //int self = (int)object;     

      /*if((int)*(object+1) != (int)*((int*)*(method+1) +1)){
	self = *(method+1);
	cout<<"taking self from method table"<<endl;

	}*/


      // push result destination, return address and reference
      stack_ptr+=3;
      *(stack_ptr-1) = *(int*)(addressOfMain + 10);  // location of result
      *(stack_ptr-2) = (int)addressOfMain + 18;  // address of next instruction
      *(stack_ptr-3) = /*self;*/ *(method+1);   // actual self;
      
      

      addressOfMain = (char*)*method;
      break;  
    }


    // invokes super constructor
    case ClassCore::INVOKE_S: {

      if( debug )
	cout<<"invoking super constructor #"<<*(int*)(addressOfMain + 6)<<endl;
      
      int** object = (int**) *(stack_ptr - *(int*)(addressOfMain + 2));
      int index = *(int*)(addressOfMain + 6);     
      int offset = *(int*)(addressOfMain + 10);

      if(offset != 0)
	object = (int**)((int)*(object + 1) + offset);  // select descriptor

      int nDescriptors = *(int*)(**(object+1) + 12);


      // method ptr -> self                +  3*  index in vtable
      int *method = *(object+1) + 2 + nDescriptors*2 + 3*  index;
                             

      // push return address and reference
      stack_ptr+=2;
      // TUKAJ je treba potisnit deskriptor objekta tega konstruktorja

      *(stack_ptr-2) = *(method+1);   // actual self
      *(stack_ptr-1) = (int)addressOfMain + 14;  // address of next instruction
      
      addressOfMain = (char*)*method;


      break;  
    }

      

    // invokes native procedure
    case ClassCore::INVOKE_N_P: {

      if( debug )
	cout<<"invoking native method "<<(char*)*(int*)(addressOfMain + 2)<<endl;
            

      void** handles = vm->getNativeHandles();

      void(*function)(int*&, mVMEnvironment* );

      for(int i = 0; i < vm->numNativeHandles(); i++){	

	function = (void (*)(int*&,mVMEnvironment* )) dlsym( handles[i], (char*)*(int*)(addressOfMain + 2) );

	if( dlerror() == NULL){	  

	  function( stack_ptr, environment );

	  // replace instruction with INVOKE_N_P_S

	  //tukaj...

	  break;

	}

      }

      

      int nArgs = *(int*)(addressOfMain + 6) - 3;   // 3 are implicitly pushed      
     

      addressOfMain = (char*)*(stack_ptr - 2);        // continue at old address
      stack_ptr = (int*)*(stack_ptr - 1);             // restore old stack_ptr

      // clean up args
      stack_ptr -= nArgs;


      break;  
    }


    // invokes native function
    case ClassCore::INVOKE_N_F: {

      if( debug )
	cout<<"invoking native function "<<(char*)*(int*)(addressOfMain + 2)<<endl;
      
      void** handles = vm->getNativeHandles();

      void(*function)(int*&, mVMEnvironment*);

      for(int i = 0; i < vm->numNativeHandles(); i++){	

	function = (void (*)(int*&, mVMEnvironment*)) dlsym( handles[i], (char*)*(int*)(addressOfMain + 2) );

	if( dlerror() == NULL){	  

	  function( stack_ptr, environment );
	  // replace instruction with INVOKE_N_P_S

	  //tukaj...

	  break;

	}else{

	  //cout<<"NATRIVE function NOT FOUND"<<endl;

	}

      }

            
      


      // get result from stack which was pushed inside native method 
      int result = *(stack_ptr - 1);
      stack_ptr--;

      int dest = *(stack_ptr - 2);  // this is where to store result


      int nArgs = *(int*)(addressOfMain + 6) - 4;   // 4 are implicitly pushed      

      addressOfMain = (char*)*(stack_ptr - 3);        // continue at old address
      stack_ptr = (int*)*(stack_ptr - 1);             // restore old stack_ptr

      stack_ptr--;           // remove result destination which was pushed implicitly in invoke_1_f !!!

      // clean up args
      stack_ptr -= nArgs;

      


      // store reference to result destination
      *(stack_ptr - dest) = result;
      
      break;  
    }




    // skips n bytes if bit in flags is set
    case ClassCore::SKIP_IFSET: {

      if( debug )
	cout<<"skipping on flags bit "<<*(int*)(addressOfMain + 2)<<endl;

      unsigned bit = 1 << *(int*)(addressOfMain + 2);


      if( (flags & bit) != 0 ){  // bit set ?
	addressOfMain += 10 + *(int*)(addressOfMain + 6);
      }else {
	addressOfMain+=10;
      }

      //cout<< *(int*)(addressOfMain + 2) 

      break;
    }


    // skips n bytes if bit in flags is NOT set (just the opposite of above)
    case ClassCore::SKIP_IFCLEAR: {

      if( debug )
	cout<<"not skipping on flags bit "<<*(int*)(addressOfMain + 2)<<endl;

      unsigned bit = 1 << *(int*)(addressOfMain + 2);


      if( (flags & bit) == 0 ){  // bit set ?
	//cout<<" skipping next instruction"<<endl;
	addressOfMain += 10 + *(int*)(addressOfMain + 6);
      }else {
	//cout<<" not skipping next instruction"<<endl;
	addressOfMain+=10;
      }


      break;
    }



    // set method without environment
    case ClassCore::SETMETHOD_1: {

      if( debug )
	cout<<"updating method (no environment)"<<*(int*)(addressOfMain + 10)<<endl;
      
     
      int **object = (int**)*(stack_ptr - *(int*)(addressOfMain + 2)); 

      int **source = (int**)*(stack_ptr - *(int*)(addressOfMain + 6)); 

      int offset = *(int*)(addressOfMain + 18);


      if(offset != 0)
	object = (int**)((int)*(object + 1) + offset);  // select descriptor


      // method ptr ->
      int *method = *(object+1) + 2 + *(int*)(**(object+1) +12) * 2 +  3*  *( *object + *(int*)(addressOfMain + 10));

     
      // now we go into method table with this index and set new address




      *method = *(int*)(addressOfMain + 14);   // address
      *(method+1) = (int)source;                 // actual self
      *(method+2) = 0;                           // no environment

      addressOfMain+=22;


      break;
    }



    // set method with environment
    case ClassCore::SETMETHOD_2: {

      if( debug )
	cout<<"updating method with environment"<<endl;
      

      int **object = (int**)*(stack_ptr - *(int*)(addressOfMain + 2)); 

      int **source = (int**)*(stack_ptr - *(int*)(addressOfMain + 6));
      

      int nVariables = *(int*)(addressOfMain + 10); 


      int offset = *(int*)(addressOfMain + 22);
     

      if(offset != 0)
	object = (int**)((int)*(object + 1) + offset);  // select descriptor

      
      int *method = *(object+1) + 2 + *(int*)(**(object+1) +12) * 2 +  3*  *( *object + *(int*)(addressOfMain + 14));
      
      
      // alloc space for environment
      int* env = (int*)vm->allocate( 4 * nVariables + 4);
       

      int *topEnv = env + nVariables+1;

      // copy parent references passed on stack into environment
      for(int i=1;i<=nVariables;i++){
	
	*(topEnv-i) = *(stack_ptr-i);

      }

      *env = nVariables;

      // now we go into method table with this index and set new address

      *method = *(int*)(addressOfMain + 18);   // address
      *(method+1) = (int)source;                 // actual self
      *(method+2) = (int)env;                    // environment
           
     
      // remove references from stack
      stack_ptr -= nVariables;

      addressOfMain+=26;


      break;
    }


    // upcast
    case ClassCore::UPCAST: {

      if( debug )
	cout<<"upcasting "<<endl;
      
      // source, dest, class address

      int **object = (int**)*(stack_ptr - *(int*)(addressOfMain + 2));      


      // the class that we are casting to
      int classAddress = *(int*)(addressOfMain + 10);

      // now stroll through class table for this object and check if it contains class address
      // class table starts after classinfo
      int *ctable = (int*)(**(object+1) + 20);   // --> class table

      int nEntries = *(int*)(**(object+1) + 12);      
      

      int newAddress = (int)*(object+1) + 8;  // --> start of descriptors in object

      bool found=false;
      for(int i=0;i<nEntries;i++){
	//cout<<"offset into vtable is "<<*(ctable + i*2 +1)<<", classaddress="<<*(ctable + i*2)<<endl;
	if( *(ctable + i*3) == classAddress){
	  found = true;
	  break;
	}
	newAddress += 8;
      }

      if(!found){

	cout<<"Invalid cast"<<endl;
	terminated = true;
	break;
      }



      //cout<<"new address="<<newAddress<<endl;

      *(stack_ptr - *(int*)(addressOfMain + 6)) = newAddress;//(int)*(object+1) + newAddress;
	

      addressOfMain+=14;


      break;
    }


    // downcast
    case ClassCore::DOWNCAST: {

      if( debug )
	cout<<"downsting "<<endl;
      
	addressOfMain+=14;


      break;
    }



    // check method signature
    case ClassCore::MCHECK: {

      if( debug )
          cout<<"checking method type "<<endl;

      
      //*stack_ptr = *(int*)(addressOfMain + 2);
      //stack_ptr++;

      cout<<"MCHECK NOT IMPLEMENTED"<<endl;


      // original Method object
      int *object = *((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1);

      // # parameters of method
      int nParams = *(int*)(addressOfMain + 6);


      // index of this method is after classdata in Method object
      int metaIndex = *(object + METHOD_OFFSET + 1);


      // anonymous method?
      if( metaIndex == -1 ){

	cout<<"methodcast: checking anon method!"<<endl;


	MetaAnonymousEntry* anonmeta = (MetaAnonymousEntry*) *(object + METHOD_OFFSET + 3);


	if( anonmeta->nParameters != nParams ){

	  cout<<"methodcast: # of parameters doesn't match"<<endl;
	  terminated = true;
	  break;

	}


	// check all parameter types and return type
	for(int i=1; i <= anonmeta->nParameters + 1; i++){
    
	  
	  // later: use subtype relation function !!!!!!!!!!!!!!!!!!!!!!!!!!!!!1
	  if( anonmeta->argClassAddresses[ i-1 ] != *(stack_ptr-i) ){

	    cout<<"methodcast: type of parameter " <<i <<" doesn't match"<<endl;
	    terminated = true;
	    break;
	    
	  }
	}	    		

      }else{

	// normal method

	// meta size is 4th field of classinfo in classdata	
	int metaSize = *( *(int**)object + 4);
	int* metaInfo = (int*)( (char*)*object - metaSize);  // ptr to start of metadata
	

	// meta method entries is 5th field of meta info in classdata
	MetaMethodEntry* entry =  (MetaMethodEntry*) *( (int*)*(metaInfo + 4) + metaIndex); // -> method entry
  

	if( entry->nParameters != nParams ){

	  cout<<"methodcast: # of parameters doesn't match"<<endl;
	  terminated = true;
	  break;

	}


	for(int i=1; i<=entry->nParameters; i++){
    
	  // later: use subtype relation function !!!!!!!!!!!!!!!!!!!!!!!!!!!!!1
	  if( entry->argClassAddresses[ i-1 ] != *(stack_ptr-i) ){

	    cout<<"methodcast: type of parameter " <<i <<" doesn't match"<<endl;
	    terminated = true;
	    break;
	    
	  }
	}
	  
	// and return type. return type is not always present in class methods. only for functions

	// TUKAJ 

      }


      // important: clear stack of all class address values that were pushed (+1 for return class)
      stack_ptr -= (nParams + 1);

      addressOfMain+=10;
      break;
    }




  
    // test references for equivalence
    case ClassCore::IFREFEQ: {

      if( debug )
	cout<<"comparing two references "<<endl;
      
      // source, dest, class address

      int **object1 = (int**)*(stack_ptr - *(int*)(addressOfMain + 2));
      int **object2 = (int**)*(stack_ptr - *(int*)(addressOfMain + 6));

      // set boolean variable
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) =
	object1 == object2 ? 1 : 0;

     
      addressOfMain+=14;


      break;
    }


    // test if references differ
    case ClassCore::IFREFNEQ: {

      if( debug )
	cout<<"comparing two references "<<endl;
      
      // source, dest, class address

      int **object1 = (int**)*(stack_ptr - *(int*)(addressOfMain + 2));
      int **object2 = (int**)*(stack_ptr - *(int*)(addressOfMain + 6));

      // set boolean variable
      *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 10))+1) +BOOLEAN_OFFSET) =
	object1 != object2 ? 1 : 0;

     
      addressOfMain+=14;


      break;
    }


      
    // load nil
    case ClassCore::NEW_NIL: {

      if( debug )
	cout<<"new nil "<<endl;
      
      // source, dest, class address

      *(stack_ptr - *(int*)(addressOfMain + 2)) = 0;      
     
      addressOfMain+=6;


      break;
    }

      

    // create anonymous method without environment
    case ClassCore::CREATEMETHOD_1: {
      

      if( debug )
	cout<<"creating method (no environment)"<<endl;
  

      // create object Method
      mString className("Method");
      mLoadedClass *theClass = vm->findAndLoadClass( &className );
      
      int* classData = (int*)theClass->dataStartAddress;
          

      // object already pointing to correct descriptor
      int* object = vm->create_user_object( classData );            


      int *originalObject = (int*) *(object + 1);            

      int nDescriptors = *(classData+3);


      // owner object of Method
      int* owner = (int*)*(stack_ptr - *(int*)(addressOfMain + 6));
      int ownerClassData = *(int*)*(owner + 1);

      // # parameters of method
      int nParams = *(int*)(addressOfMain + 10);




      // now we set up Method object for anonymous method which includes the following:
      // - classdata of owner
      // - meta index: set to -1 because this is anonymous method
      // - owner set to this method's owner
      // - anonymous_env_ptr  ->
            

      // address of classdata of owner of this method
      *(originalObject + METHOD_OFFSET) = ownerClassData;
      // meta index is for class methods, here we set it to -1 which identifies that this
      // is an anonymous method
      *(originalObject + METHOD_OFFSET + 1) = -1;

      // owner object of method 
      *(originalObject + METHOD_OFFSET + 2) = (int)owner;


      // set up meta data for anonymous method. because anonymous methods
      // have no place in metadata for class, we have to store it in
      // a special location - the meta environment

      // alloc space for meta data
      MetaAnonymousEntry* metaenv = (MetaAnonymousEntry*)vm->allocate( sizeof( MetaAnonymousEntry ) );

      // # of parameters
      metaenv->nParameters = nParams;
      // allocate space for parameter class addresses + 1 for return class (which may be 0...)
      metaenv->argClassAddresses = (int*)vm->allocate( 4 * nParams + 4 ); 

      // copy parameter class addresses and return class
      for(int i=1; i <= nParams + 1; i++){
	//cout<<"copying class address "<<*(stack_ptr-i)<<endl;
	metaenv->argClassAddresses[ i-1 ] = *(stack_ptr-i);
      }

      // setup flags 
      metaenv->flags = 0;

      // bit 1 set for functions
      if( metaenv->argClassAddresses[ nParams ] != 0 )
	metaenv->flags |= 1; 

      // access is PUBLIC
      metaenv->flags |= 0x30;


      // address of method code
      metaenv->closure = *(int*)(addressOfMain + 14);

      // no parent environment
      metaenv->environment = 0;      


      // ptr to meta environment for anonymous method 
      *(originalObject + METHOD_OFFSET + 3) = (int)metaenv;
      
            
      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)object;


      // important: clear stack of all class address values that were pushed (+1 for return class)
      stack_ptr -= (nParams + 1);

      
      // reserve stack and push current execution address and reference
      stack_ptr+=2;
      *(stack_ptr-2) = (int)object;  // reference
      *(stack_ptr-1) = (int)addressOfMain + 18;  // address of next instruction
                     

      // zigzag: index of constructor method in Method is the first method (0) in vtable for Method
      // if Method has default constructor at position 0...
      int constructorIndex = 0;
      addressOfMain = (char*)*( originalObject + 2 + nDescriptors*2 + 3*  *(*(int**)object + 
									    constructorIndex) );


      


      
      //addressOfMain+=14;


      break;
    }




    // create anonymous method with environment
    case ClassCore::CREATEMETHOD_2: {
      

      if( debug )
	cout<<"creating method (with environment)"<<endl;
  

      // create object Method
      mString className("Method");
      mLoadedClass *theClass = vm->findAndLoadClass( &className );
      
      int* classData = (int*)theClass->dataStartAddress;
          

      // object already pointing to correct descriptor
      int* object = vm->create_user_object( classData );            


      int *originalObject = (int*) *(object + 1);            

      int nDescriptors = *(classData+3);


      // owner object of Method
      int* owner = (int*)*(stack_ptr - *(int*)(addressOfMain + 6));
      int ownerClassData = *(int*)*(owner + 1);

      // # parameters of method
      int nParams = *(int*)(addressOfMain + 10);

      // # parent variables
      int nParentVariables = *(int*)(addressOfMain + 14);



      // now we set up Method object for anonymous method which includes the following:
      // - classdata of owner
      // - meta index: set to -1 because this is anonymous method
      // - owner set to this method's owner
      // - anonymous_env_ptr  ->
            

      // address of classdata of owner of this method
      *(originalObject + METHOD_OFFSET) = ownerClassData;
      // meta index is for class methods, here we set it to -1 which identifies that this
      // is an anonymous method
      *(originalObject + METHOD_OFFSET + 1) = -1;

      // owner object of method 
      *(originalObject + METHOD_OFFSET + 2) = (int)owner;


      // set up meta data for anonymous method. because anonymous methods
      // have no place in metadata for class, we have to store it in
      // a special location - the meta environment

      // alloc space for meta data
      MetaAnonymousEntry* metaenv = (MetaAnonymousEntry*)vm->allocate( sizeof( MetaAnonymousEntry ) );

      // # of parameters
      metaenv->nParameters = nParams;
      // allocate space for parameter class addresses + 1 for return class (which may be 0...)
      metaenv->argClassAddresses = (int*)vm->allocate( 4 * nParams + 4 ); 

      // copy parameter class addresses and return class
      for(int i=1; i <= nParams + 1; i++){
	//cout<<"copying class address "<<*(stack_ptr-i)<<endl;
	metaenv->argClassAddresses[ i-1 ] = *(stack_ptr-i);
      }

      // setup flags 
      metaenv->flags = 0;

      // bit 1 set for functions
      if( metaenv->argClassAddresses[ nParams ] != 0 )
	metaenv->flags |= 1; 

      // access is PUBLIC
      metaenv->flags |= 0x30;


      // address of method code
      metaenv->closure = *(int*)(addressOfMain + 18);

      // copy environment references ( +4 for size )
      metaenv->environment = (int*)vm->allocate( 4 * nParentVariables +4  );
      metaenv->environment[ 0 ] = nParentVariables;
      
 
      // parent variables start after parameters on stack
      for(int i=1; i <=nParentVariables; i++){
	//cout<<"copying parent variable "<<*(stack_ptr - nParams - 1 - i)<<endl;
	metaenv->environment[ i ] = *(stack_ptr - nParams - 1 - i);
      }



      // ptr to meta environment for anonymous method 
      *(originalObject + METHOD_OFFSET + 3) = (int)metaenv;
      
            
      // store reference to dest variable
      *(stack_ptr - *(int*)(addressOfMain + 2))=(int)object;


      // important: clear stack of all class address values that were pushed (+1 for return class)
      stack_ptr -= (nParams + 1 + nParentVariables);

      
      // reserve stack and push current execution address and reference
      stack_ptr+=2;
      *(stack_ptr-2) = (int)object;  // reference
      *(stack_ptr-1) = (int)addressOfMain + 22;  // address of next instruction
                           

      // zigzag: index of constructor method in Method is the first method (0) in vtable for Method
      // if Method has default constructor at position 0...
      int constructorIndex = 0;
      addressOfMain = (char*)*( originalObject + 2 + nDescriptors*2 + 3*  *(*(int**)object + 
									    constructorIndex) );


      


      
      //addressOfMain+=14;


      break;
    }



    ///////////////////////////////////////////////////////////////////
    // Block
    ///////////////////////////////////////////////////////////////////

      // block structure:
      // -----------------------
      // | -> closure address  |
      // |---------------------|
      // | env. length         |
      // |---------------------|
      // | -> env. address     |
      // |---------------------|
      // | block exit code     |
      // |---------------------|
      // | block flags         |
      // |---------------------|


    // new block
    case ClassBlock::NEW:{

      if( debug ) cout<<"creating new block"<<endl;


      int* classData = (int*)vm->getBlockClass();


      int* object = vm->create_object(classData);      

     
      int nVariables = *(int*)(addressOfMain + 10);  // number of variables from parent closure
      

      // alloc space for environment
      int* env = (int*)vm->allocate( 4 * nVariables );
       
      int *topEnv = env + nVariables;

      // copy parent references passed on stack into environment
      for(int i=1;i<=nVariables;i++) *(topEnv-i) = *(stack_ptr-i);


      if( debug ){
	cout<<"allocated for "<<nVariables<<" parental local(s)"<<endl;
	
      }
      
     
      // store address of closure to our object
      *(object + BLOCK_OFFSET) = *(int*)(addressOfMain + 6);
      // store environment length
      *(object + BLOCK_OFFSET+1) = nVariables;
      // store environment address
      *(object + BLOCK_OFFSET+2) = (int)env;
      // set initial status to 0
      *(object + BLOCK_OFFSET+3) = 0;
      // set flags to 0
      *(object + BLOCK_OFFSET+4) = 0;

      // store reference to local variable
      *(stack_ptr - *(int*)(addressOfMain + 2)) = (int)(object+4); //(int)descriptor;
    

      // cleanup pushed environment on stack
      stack_ptr -= nVariables;

      //cout<<"block address="<<*object<<endl;

      addressOfMain+=14;

      break;
    }


    // execute block
    case ClassBlock::EXEC: {

      if( debug ) cout<<"executing block"<<endl;


      // block object
      int* object = *((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +BLOCK_OFFSET;

      int descriptor = *(stack_ptr - *(int*)(addressOfMain + 2));

      int nVariables = *(object+1);

      int* env = (int*)*(object+2);

      // push all references in this block's environment, and then return address
      stack_ptr += nVariables;  // # locals
            
      for(int i=1;i<=nVariables;i++) *(stack_ptr-i) = *(env+i-1);


      stack_ptr++;   // return address

      *(stack_ptr-1) = (int)addressOfMain + 6;  // push address of next instruction

      
      stack_ptr++;
      *(stack_ptr-1) = descriptor;  // push block reference
      
      
      // jump to start of block
      addressOfMain = (char*)*object;

      // set block_return flag to 0
      flags &= 0xfffffffe;


      break;

    }


    // set block status
    case ClassBlock::STATUS: {

      if( debug ) cout<<"setting block status"<<endl;
      

      // 
      // | old_stack_ptr| <-- stack_ptr
      // | block ref    |
      // | ret addr     |
      // | pushed locals|

      // block object
      int* object = (*((int**)*(stack_ptr - 2)+1) +BLOCK_OFFSET);

      // check boolean value
      *(object + 3) = *(*((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +BOOLEAN_OFFSET) == 1 ?  
	*(int*)(addressOfMain + 6) : *(int*)(addressOfMain + 10);


      if( debug) cout<<"status set to "<<*(object + 3)<<endl;

      
      addressOfMain+=14;
      break;

    }

    // exit block
    case ClassBlock::EXIT_1: {

      if( debug ) cout<<"exiting block"<<endl;
      

      // restore old return address and old stack_ptr
      // 
      // | old_stack_ptr| <-- stack_ptr
      // | block ref    |
      // | ret addr     |
      // | pushed locals|

      // block object
      int* object = (*((int**)*(stack_ptr - 2)+1) +BLOCK_OFFSET);


     

      addressOfMain = (char*)*(stack_ptr - 3);  // continue at old address

      stack_ptr = (int*)*(stack_ptr - 1);             // restore old stack_ptr


      // stack_ptr is now as it was before frame allocation inside block
      //stack_ptr--;   // remove block reference ( new_frame didn't take it into account)


      // pushed references however are still on stack. remove them.
      // note: this removes all parent variables as well as block reference!
      stack_ptr -= *(object+1);   // number of variables pushed


      //addressOfMain+=6;
      break;

    }


    // return local from block
    case ClassBlock::RETURN_1_LOCAL: {

      if( debug ) cout<<"returning local from block"<<endl;


      

      // block object
      //int* object = (int*)*(stack_ptr - *(int*)(addressOfMain + 2));

      // restore old return address and old stack_ptr
      // 
      // | old_stack_ptr| <-- stack_ptr
      // | block ref    |
      // | ret addr     |
      // | pushed locals|

      
      int* ref = (int*)*(stack_ptr - *(int*)(addressOfMain + 2));

      // clear block_return flag (it may have been set by a preceeding Block.return)
      if( (flags & 1) != 0){
	flags &= 0xfffffffe;
	stack_ptr--;   // remove result pushed by Block
      }

      // # of parent variables to clean from stack. important: this does NOT include pushed self
      // because frame_copy saves old stack_ptr as it was before implicit push of self and ret. address
      //int nParentVariables = *(int*)(addressOfMain + 6); //*(int*)(object + 4);


      // block object
      int* object = (*((int**)*(stack_ptr - 2)+1) +BLOCK_OFFSET); //(int*)*(stack_ptr - 2);

      addressOfMain = (char*)*(stack_ptr - 3);  // continue at old address
      stack_ptr = (int*)*(stack_ptr - 1);             // restore old stack_ptr


      // stack_ptr is now as it was before frame allocation inside block

      //stack_ptr--;

      //cout<<"--- removing "<<nParentVariables<<" variables from stack"<<endl;
      // pushed references however are still on stack. remove them
      stack_ptr -= *(object+1);   // number of variables pushed

      // now push result on top
      stack_ptr++;
      *(stack_ptr-1) = (int)ref;

      // set block_return flag to 1
      flags |= 0x00000001;

      //addressOfMain+=10;
      break;

    }



    // return nothing from block
    case ClassBlock::RETURN_1: {

      if( debug ) cout<<"returning nothing from block"<<endl;


      

      // block object
      //int* object = (int*)*(stack_ptr - *(int*)(addressOfMain + 2));

      // restore old return address and old stack_ptr
      // 
      // | old_stack_ptr| <-- stack_ptr 
      // | ret addr     |
      // | pushed locals|

      
      // clear block_return flag (it may have been set by a preceeding Block.return)
      if( (flags & 1) != 0){
	flags &= 0xfffffffe;
      }

      // # of parent variables to clean from stack. important: this does NOT include pushed self
      // because frame_copy saves old stack_ptr as it was before implicit push of self and ret. address
      //int nParentVariables = *(int*)(addressOfMain + 2); //*(int*)(object + 4);

      // block object
      int* object = (*((int**)*(stack_ptr - 2)+1) +BLOCK_OFFSET);//(int*)*(stack_ptr - 2);

      addressOfMain = (char*)*(stack_ptr - 3);  // continue at old address
      stack_ptr = (int*)*(stack_ptr - 1);             // restore old stack_ptr


      // stack_ptr is now as it was before frame allocation inside block

      //stack_ptr--;

      //cout<<"--- removing "<<nParentVariables<<" variables from stack"<<endl;
      // pushed references however are still on stack. remove them
      stack_ptr -= *(object+1);   // number of variables pushed
      

      // set block_return flag to 1
      flags |= 0x00000001;

      //addressOfMain+=10;
      break;

    }    



    ///////////////////////////////////////////////////////////////////
    // While
    ///////////////////////////////////////////////////////////////////

    // while structure:
    // -----------------------
    // | -> cond block ref   |
    // |----------------------
    // | -> exec block ref   |
    // |----------------------
    // | -> while status     |
    // -----------------------

    // status is 0 if conditional block has not been executed, 1 if it has

    // new While
    case ClassWhile::NEW:{

      if( debug ) cout<<"creating new while"<<endl;
      

      int* classData = (int*)vm->getWhileClass();

      int* object = vm->create_object(classData);
      

      // store conditional closure reference
      *(object+WHILE_OFFSET) = *(stack_ptr - *(int*)(addressOfMain + 6));

      // store conditional closure reference
      *(object+WHILE_OFFSET+1) = *(stack_ptr - *(int*)(addressOfMain + 10));

      // initial status is 0
      *(object+WHILE_OFFSET+2) = 0;
            
      // store reference to local variable
      *(stack_ptr - *(int*)(addressOfMain + 2)) = (int)(object+4);     
      

      addressOfMain+=14;

      break;
    }


    // execute while
    case ClassWhile::EXEC: {

      if( debug ) cout<<"executing while"<<endl;      

      // if Block.return_flag is set then this while's main block returned. exit loop 
      if( (flags & 1) != 0){

	if( debug ) cout<<"exiting while loop because of return"<<endl;

	//flags &= 0xfffffffe;   // clear flag
	//stack_ptr--;  // remove result pushed by Block
	addressOfMain+=6;
	break;
      }
      



      // while object
      int* object = *((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +WHILE_OFFSET;

      int whileStatus = *(object+2);  

      int* cond =  // condition block
	*((int**)*object +1) +BLOCK_OFFSET;

      int descriptor = (int)*object;  // descriptor of condition

      int status = *(cond+3);     // status of conditional block

      // condition bloc status will be set to 1 if condition is true, 2 if false, 0 initially

      // execute condition only if it hasn't been yet
      if(whileStatus == 0 && (status == 1 || status == 0)){	

	int nVariables = *(cond+1);
	
	int* env = (int*)*(cond+2);
	
	// push all references in this block's environment, and then return address
	stack_ptr += nVariables;  // # locals
	
	for(int i=1;i<=nVariables;i++) *(stack_ptr-i) = *(env+i-1);
	
	
	stack_ptr++;   // return address
	
	*(stack_ptr-1) = (int)addressOfMain;  // we must return HERE!
	
	
	stack_ptr++;
	*(stack_ptr-1) = descriptor;  // push block reference
	
	
	// set while status to 1 which means that condition has been executed
	*(object+2) = 1;
	

	if( debug ) cout<<"executing condition"<<endl;
	
	// jump to start of block
	addressOfMain = (char*)*cond;
	
	
      }else{  // whileStatus == 0


	///////////////////////////////////////
	// execute main block
	///////////////////////////////////////

	// terminate loop if status was set to 2
	if(status == 2){

	  addressOfMain+=6;
	  break;
	}
	

	// set block_return flag to 0
	flags &= 0xfffffffe;


	// condition has been executed, now execute main block
	int* block = //(int*)*(object + 1);  // main block
	  *((int**)*(object + 1) +1) +BLOCK_OFFSET;

	int descriptor = (int)*(object + 1);

      
	int nVariables = *(block+1);
	
	int* env = (int*)*(block+2);
	
	// push all references in this block's environment, and then return address
	stack_ptr += nVariables;  // # locals
	
	for(int i=1;i<=nVariables;i++) *(stack_ptr-i) = *(env+i-1);
	
	
	stack_ptr++;   // return address
	
	*(stack_ptr-1) = (int)addressOfMain;  // we must return HERE!
	
	
	stack_ptr++;
	*(stack_ptr-1) = descriptor;  // push block reference
	
	
	// set while status to 0 which means that condition must be executed again
	*(object+2) = 0;

	if( debug ) cout<<"executing main block"<<endl;
	
	// jump to start of block
	addressOfMain = (char*)*block;



      }

      
      break;

    }



    ///////////////////////////////////////////////////////////////////
    // DoWhile
    ///////////////////////////////////////////////////////////////////

    // dowhile structure:
    // -----------------------
    // | -> exec block ref   |
    // |----------------------
    // | -> cond block ref   |
    // |----------------------
    // | -> while status     |
    // -----------------------

    // new block
    case ClassDoWhile::NEW:{

      if( debug ) cout<<"creating new dowhile"<<endl;
      
      int* classData = (int*)vm->getDoWhileClass();

      int* object = vm->create_object(classData);

      
      // store execution closure reference
      *(object+DOWHILE_OFFSET) = *(stack_ptr - *(int*)(addressOfMain + 6));

      // store conditional closure reference
      *(object+DOWHILE_OFFSET+1) = *(stack_ptr - *(int*)(addressOfMain + 10));

      // initial status is 0
      *(object+DOWHILE_OFFSET+2) = 0;
            
      // store reference to local variable
      *(stack_ptr - *(int*)(addressOfMain + 2)) = (int)(object+4);            


      addressOfMain+=14;

      break;
    }


    // execute dowhile
    case ClassDoWhile::EXEC: {

      if( debug ) cout<<"executing dowhile"<<endl;      


      // if Block.return_flag is set then this while's main block returned. exit loop 
	if( (flags & 1) != 0){
	  
	  if( debug ) cout<<"exiting dowhile loop because of return"<<endl;
	  	 
	  addressOfMain+=6;
	  break;
	}


      // dowhile object
      int* object = *((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +DOWHILE_OFFSET;

      int whileStatus = *(object+2);  

      if( debug ) cout<<"dowhile status="<<whileStatus<<endl;      
            

      int descriptor_block = (int)*object;
      int descriptor_cond = (int)*(object+1);

      int* block = *((int**)*object +1) +BLOCK_OFFSET;  // condition block
      int* cond = *((int**)*(object+1) +1) +BLOCK_OFFSET;  // main execution block



      if( whileStatus == 0 ){     // status is 0 only the first time

	// 1st time execution is always performed

	// set block_return flag to 0
	flags &= 0xfffffffe;
	
	
	int nVariables = *(block+1);
	
	int* env = (int*)*(block+2);
	
	// push all references in this block's environment, and then return address
	stack_ptr += nVariables;  // # locals
	
	for(int i=1;i<=nVariables;i++) *(stack_ptr-i) = *(env+i-1);
	
	
	stack_ptr++;   // return address
	
	*(stack_ptr-1) = (int)addressOfMain;  // we must return HERE!
	
	
	stack_ptr++;
	*(stack_ptr-1) = descriptor_block;  // push block reference
      
	// set while status to 1 which means that main block has been executed
	*(object+2) = 1;
		
	if( debug ) cout<<"executing main block"<<endl;
	
	// jump to start of block
	addressOfMain = (char*)*block;
		
	 
      }   // whileStatus == 0
      

      // now we execute conditional block. status is either 1 because main block has executed 
      
      else if(whileStatus == 1){
	
	
	int nVariables = *(cond+1);
	
	int* env = (int*)*(cond+2);
	
	// push all references in this block's environment, and then return address
	stack_ptr += nVariables;  // # locals
	
	for(int i=1;i<=nVariables;i++) *(stack_ptr-i) = *(env+i-1);
	
	
	stack_ptr++;   // return address
	
	*(stack_ptr-1) = (int)addressOfMain;  // we must return HERE!
	
	
	stack_ptr++;
	*(stack_ptr-1) = descriptor_cond;  // push block reference
	
	// set while status to 2 which means that condition block has been executed and that we
	// must test the condition
	*(object+2) = 2;
	
	if( debug ) cout<<"executing condition block"<<endl;
	
	// set block_return flag to 0
	flags &= 0xfffffffe;

	// jump to start of block
	addressOfMain = (char*)*cond;
	
	
      }  // whileStatus == 1
      


      // test condition
      else if(whileStatus == 2){	
	
	
	int status = *(cond+3);
	
	// condition block status will be set to 1 if condition is true, 2 if false, 0 initially
	
	
	if(status == 1){
	  
	  // set while status to 1 which means that main block has been executed
	  *(object+2) = 1;
	  
	  if( debug ) cout<<"condition true"<<endl;

	  int nVariables = *(block+1);
	  
	  int* env = (int*)*(block+2);
	  
	  // push all references in this block's environment, and then return address
	  stack_ptr += nVariables;  // # locals
	
	  for(int i=1;i<=nVariables;i++) *(stack_ptr-i) = *(env+i-1);
	
	
	  stack_ptr++;   // return address
	  
	  *(stack_ptr-1) = (int)addressOfMain;  // we must return HERE!
	  
	  
	  stack_ptr++;
	  *(stack_ptr-1) = descriptor_block;  // push block reference
	  
	  
	  // jump to start of block
	  addressOfMain = (char*)*block;
	  
	}else{

	  if( debug ) cout<<"condition false"<<endl;

	  // condition is false - exit loop
	  addressOfMain+=6;
	  break;

	}

      }  // whileStatus == 2

      
      break;

    }




    ///////////////////////////////////////////////////////////////////
    // For
    ///////////////////////////////////////////////////////////////////

    // for structure:
    // -----------------------
    // | -> init block ref   |
    // |----------------------
    // | -> cond block ref   |
    // |----------------------
    // | -> next block ref   |
    // |----------------------
    // | -> main block ref   |
    // |----------------------
    // | -> for status       |
    // -----------------------

    // new for
    case ClassFor::NEW:{

      if( debug ) cout<<"creating new for"<<endl;
      

      int* classData = (int*)vm->getForClass();

      int* object = vm->create_object(classData);

      

      // store initialisation closure reference
      *(object+FOR_OFFSET) = *(stack_ptr - *(int*)(addressOfMain + 6));

      // store conditional closure reference
      *(object+FOR_OFFSET+1) = *(stack_ptr - *(int*)(addressOfMain + 10));

      // initial status is 0
      *(object+FOR_OFFSET+4) = 0;
            
      // store reference to local variable
      *(stack_ptr - *(int*)(addressOfMain + 2)) = (int)(object+4);           


      addressOfMain+=14;

      break;
    }



    // add both blocks
    case ClassFor::ADDBLOCKS:{

      if( debug ) cout<<"adding for blocks"<<endl;
           
      int* object = *((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +FOR_OFFSET; 

      // store initialisation closure reference
      *(object+2) = *(stack_ptr - *(int*)(addressOfMain + 6));

      // store conditional closure reference
      *(object+3) = *(stack_ptr - *(int*)(addressOfMain + 10));                

      addressOfMain+=14;

      break;
    }



    // execute for
    case ClassFor::EXEC: {

      if( debug ) cout<<"executing for"<<endl;      

      // if Block.return_flag is set then this while's main block returned. exit loop 
      if( (flags & 1) != 0){

	if( debug ) cout<<"exiting for loop because of return"<<endl;

	addressOfMain+=6;
	break;
      }
      



      // for object
      int* object = *((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +FOR_OFFSET;

      int status = *(object+4);  

      // if status is 0 we execute initialisation block

      if( status == 0){

	int descriptor = *object;

	int* init = //(int*)*object;  // initialisation block
	    *((int**)*object +1) +BLOCK_OFFSET;

	int nVariables = *(init+1);
	
	int* env = (int*)*(init+2);
	
	// push all references in this block's environment, and then return address
	stack_ptr += nVariables;  // # locals
	
	for(int i=1;i<=nVariables;i++) *(stack_ptr-i) = *(env+i-1);
	
	
	stack_ptr++;   // return address
	
	*(stack_ptr-1) = (int)addressOfMain;  // we must return HERE!
	
	
	stack_ptr++;
	*(stack_ptr-1) = descriptor;  // push block reference
	
	
	// set for status to 1 which means that initialisation has been executed
	*(object+4) = 1;
	

	if( debug ) cout<<"executing initialisation"<<endl;
	
	// set block_return flag to 0
	flags &= 0xfffffffe;

	// jump to start of block
	addressOfMain = (char*)*init;



      }else if( status == 1){


	// we must execute condition
	int descriptor = *(object+1);
	int* cond = *((int**)*(object+1) +1) +BLOCK_OFFSET;  // condition block

	int nVariables = *(cond+1);
	
	int* env = (int*)*(cond+2);
	
	// push all references in this block's environment, and then return address
	stack_ptr += nVariables;  // # locals
	
	for(int i=1;i<=nVariables;i++) *(stack_ptr-i) = *(env+i-1);
	
	
	stack_ptr++;   // return address
	
	*(stack_ptr-1) = (int)addressOfMain;  // we must return HERE!
	
	
	stack_ptr++;
	*(stack_ptr-1) = descriptor;  // push block reference
	
	
	// set while status to 2 which means that condition has been executed
	*(object+4) = 2;
	

	if( debug ) cout<<"executing condition"<<endl;
	
	// jump to start of block
	addressOfMain = (char*)*cond;



      }else if( status == 2 ){


	// now we execute main block

	int* cond = *((int**)*(object+1) +1) +BLOCK_OFFSET;  // condition block

	int bstatus = *(cond+3);     // status of conditional block

	// terminate loop if status was set to 2
	if(bstatus == 2){

	  addressOfMain+=6;
	  break;
	}
	

	// set block_return flag to 0
	flags &= 0xfffffffe;
	
	int descriptor = *(object+3);

	int* block = *((int**)*(object+3) +1) +BLOCK_OFFSET;  // main block

      
	int nVariables = *(block+1);
	
	int* env = (int*)*(block+2);
	
	// push all references in this block's environment, and then return address
	stack_ptr += nVariables;  // # locals
	
	for(int i=1;i<=nVariables;i++) *(stack_ptr-i) = *(env+i-1);
	
	
	stack_ptr++;   // return address
	
	*(stack_ptr-1) = (int)addressOfMain;  // we must return HERE!
	
	
	stack_ptr++;
	*(stack_ptr-1) = descriptor;  // push block reference
	
	
	// set while status to 3 which means that main has been executed
	*(object+4) = 3;

	if( debug ) cout<<"executing main block"<<endl;
	
	// jump to start of block
	addressOfMain = (char*)*block;


      } else if( status == 3){


	// execute next block
	int descriptor = *(object+2);
	int* next = *((int**)*(object+2) +1) +BLOCK_OFFSET;  // next block

	int nVariables = *(next+1);
	
	int* env = (int*)*(next+2);
	
	// push all references in this block's environment, and then return address
	stack_ptr += nVariables;  // # locals
	
	for(int i=1;i<=nVariables;i++) *(stack_ptr-i) = *(env+i-1);
	
	
	stack_ptr++;   // return address
	
	*(stack_ptr-1) = (int)addressOfMain;  // we must return HERE!
	
	
	stack_ptr++;
	*(stack_ptr-1) = descriptor;  // push block reference
	
	
	// set while status to 1 which means that condition must be executed again
	*(object+4) = 1;
	

	if( debug ) cout<<"executing next"<<endl;
	
	// jump to start of block
	addressOfMain = (char*)*next;


      }


    
      
      break;

    }




    ///////////////////////////////////////////////////////////////////
    // If 
    ///////////////////////////////////////////////////////////////////

    // if structure:
    // -----------------------
    // | # of conditions     |     --------   --------                --------
    // |----------------------     |cond1 |   |cond2 |                |condN |
    // | ->list of cond/block| --->|block1|-->|block2|---> ....... -->|blockN|  
    // |----------------------     --------   --------                --------
    // | status              |
    // |----------------------
    // | next condition      |
    // -----------------------

    // new block
    case ClassIf::NEW:{

      if( debug ) cout<<"creating new if with single condition/block"<<endl;
      
      int* classData = (int*)vm->getIfClass();

      int* object = vm->create_object(classData);


      
      // only 1 condition/block 
      *(object+IF_OFFSET) = 1;

      // allocate list (8 bytes for 1 pair) and store pair
      int* pairs = (int*)vm->allocate( 8 );

      *pairs = *(stack_ptr - *(int*)(addressOfMain + 6));
      *(pairs+1) = *(stack_ptr - *(int*)(addressOfMain + 10));

      

      // store list
      *(object+IF_OFFSET+1) = (int)pairs;

      // initial status is 0
      *(object+IF_OFFSET+2) = 0;

      // nextcondition to execute is 0
      *(object+IF_OFFSET+3) = 0;
            
      // store reference to local variable
      *(stack_ptr - *(int*)(addressOfMain + 2)) = (int)(object+4);


      addressOfMain+=14;

      break;
    }




    // new block
    case ClassIf::NEW_1:{

      if( debug ) cout<<"creating new if with "<<*(int*)(addressOfMain + 6)<<" condition/blocks"<<endl;
      
      int* classData = (int*)vm->getIfClass();

      int* object = vm->create_object(classData);
      

      // # of condition/block pairs
      *(object+IF_OFFSET) = *(int*)(addressOfMain + 6);

      // allocate list (8 bytes for 1 pair) and store pair
      int* pairs = (int*)vm->allocate( 8 * *(int*)(addressOfMain + 6) );     
      

      // store list
      *(object+IF_OFFSET+1) = (int)pairs;

      // initial status is 0
      *(object+IF_OFFSET+2) = 0;

      // nextcondition to execute is 0
      *(object+IF_OFFSET+3) = 0;
            
      // store reference to local variable
      *(stack_ptr - *(int*)(addressOfMain + 2)) = (int)(object+4); //descriptor;           


      addressOfMain+=10;

      break;
    }



    // add condition/block to if
    case ClassIf::ADDPAIR:{

      if( debug ) cout<<"adding pair"<<endl;
      
      // if object
      int* object = *((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +IF_OFFSET; 
              


      // we use nextcondition status of If object to find where we we place this one
      int next = *(object+3);
      int* pairs = (int*)*(object +1);   // pairs of condition/block

      *(pairs+next*2) = *(stack_ptr - *(int*)(addressOfMain + 6));      // condition block
      *(pairs+next*2 + 1) = *(stack_ptr - *(int*)(addressOfMain + 10)); // main block
      
      
      // if this was last block added we set next to 0 to set it up for execution
      if(next == *object - 1)
	*(object+3) = 0;
      else
	(*(object+3))++;

      addressOfMain+=14;

      break;
    }



    // add block to if
    case ClassIf::ADDBLOCK:{

      if( debug ) cout<<"adding block without condition"<<endl;
      
      // if object
      int* object = *((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +IF_OFFSET; 
              


      // we use nextcondition status of If object to find where we we place this one
      int next = *(object+3);
      int* pairs = (int*)*(object +1);   // pairs of condition/block

      *(pairs+next*2) = 0;    // no condition block !!!
      // store main block
      *(pairs+next*2 + 1) = *(stack_ptr - *(int*)(addressOfMain + 6));
      
      
      // this was last block added so we set next to 0 to set iz up for execution
      *(object+3) = 0;

      addressOfMain+=10;

      break;
    }



    // execute if
    case ClassIf::EXEC: {

      if( debug ) cout<<"executing if"<<endl;      


      
      // if Block.return_flag is set then this while's main block returned. exit if 
      if( (flags & 1) != 0){
	  
	if( debug ) cout<<"exiting if statement because of return"<<endl;
	  	 
	addressOfMain+=6;
	break;
      }


      // if object
      int* object = *((int**)*(stack_ptr - *(int*)(addressOfMain + 2))+1) +IF_OFFSET;       


      int* pairs = (int*)*(object +1);   // pairs of condition/block

      //if( debug ) cout<<"got object="<<endl;

      int ifStatus = *(object+2);  

      int next = *(object+3);     // condition to execute
      

      if( debug ) cout<<"if status="<<ifStatus<<endl;      
           
      
      // test whether we are done
	if(next == -1){
	  addressOfMain+=6;
	  break;
	  
	}


      // first we execute condition block


      if( ifStatus == 0 ){     // status 0 means we must execute condition block

	
	// test whether last condition has been executed
	if(next == *(object)){
	  addressOfMain+=6;
	  break;
	  
	}


	int descriptor = *(pairs+next*2);        // condition descriptor 

	// if condition is 0 then this is unconditional else block
	if(descriptor == 0){
	  *(object+2) = 1;
	  break;

	}
       
	int* cond = *((int**)*(pairs+next*2) + 1) +BLOCK_OFFSET;   // condition is now actual block


	// set block_return flag to 0
	//flags &= 0xfffffffe;
	
	
	int nVariables = *(cond+1);
	
	int* env = (int*)*(cond+2);
	
	// push all references in this block's environment, and then return address
	stack_ptr += nVariables;  // # locals
	
	for(int i=1;i<=nVariables;i++) *(stack_ptr-i) = *(env+i-1);
	
	
	stack_ptr++;   // return address
	
	*(stack_ptr-1) = (int)addressOfMain;  // we must return HERE!
	
	
	stack_ptr++;
	*(stack_ptr-1) = descriptor;  // push block reference
      
	// set while status to 1 which means that conditional block has been executed
	*(object+2) = 1;

	// increment next condition
	//(*(object+3))++;
		
	if( debug ) cout<<"executing conditional block"<<endl;
	
	// jump to start of block
	addressOfMain = (char*)*cond;
		
	 
      }   // ifStatus == 0
      

      // now we execute main block. 
      
      else if (ifStatus == 1){
	

	int descriptor = *(pairs+next*2+1);
	int conddescriptor = *(pairs+next*2);
	

	int* cond = 0;

	if(conddescriptor != 0) cond = *((int**)*(pairs+next*2) + 1) +BLOCK_OFFSET;       // condition block
	int* block =  *((int**)*(pairs+next*2+1) + 1) +BLOCK_OFFSET;  // main execution block

	

	// test if status of condition is true
	
	int status = 1;

	if(conddescriptor != 0) status = *(cond+3);


	if(status == 1){
	  


	  int nVariables = *(block+1);
	  
	  int* env = (int*)*(block+2);
	  
	  // push all references in this block's environment, and then return address
	  stack_ptr += nVariables;  // # locals
	  
	  for(int i=1;i<=nVariables;i++) *(stack_ptr-i) = *(env+i-1);
	  
	  
	  stack_ptr++;   // return address
	  
	  *(stack_ptr-1) = (int)addressOfMain;  // we must return HERE!
	  
	  
	  stack_ptr++;
	  *(stack_ptr-1) = descriptor;  // push block reference
	  
	  // set while status to 0 which means that main block has been executed and that we
	  // must execute next condition block (if it exists)
	  *(object+2) = 0;

	  // set next condition to indicate that we are done
	  (*(object+3)) = -1;
	  
	  if( debug ) cout<<"executing main block"<<endl;
	  



	  // set block_return flag to 0
	  flags &= 0xfffffffe;
	  
	  // jump to start of block
	  addressOfMain = (char*)*block;

	}else{

	  *(object+2) = 0;
	  // increment next condition
	  (*(object+3))++;
	  if( debug ) cout<<"going to next condition"<<endl;
	  break;


	}
	
	
      }  // ifStatus == 1
            

      
      break;

    }


    default:        
	  cout<<"invalid opcode "<<hex<<opcode<<endl;
	  terminated = true;
    }  // switch
    
    
  }  // while

    if( debug ) cout<<"stack_ptr="<<(int)stack_ptr<<endl;


    // if single method was executed from within native environment, terminated was set to true
    // but to keep interpreter going, we set it back to false here - after the end of method execution
    terminated = false;

}



void mTestInterpreter::printObject(int address)
{
  int* classData = *(int**)(address+4);
  cout<<"-- Object @ "<<address<<"--"<<endl;
  cout<<"-- size = "<<*classData<<endl;
  cout<<"-- mtable offset = "<<*(classData+1)<<endl;
  cout<<"-- initial selfish offset = "<<*(classData+2)<<endl;
  cout<<"-- # entries in class table = "<<*(classData+3)<<endl;

  //cout<<"   --- object table ---"<<endl;
  //cout<<"-- index 0 "<<*(classData+3)<<endl;

  
}
