#ifndef __MTESTINTERPRETER_H
#define __MTESTINTERPRETER_H

#include "mString.h"
#include "mAbstractInterpreter.h"
#include <iostream>
#include "mVMEnvironment.h"
#include "mLoadedClass.h"

using namespace std;

class mTestInterpreter : public mAbstractInterpreter{
 public:
  mTestInterpreter(mVirtualMachine*);
  virtual ~mTestInterpreter();


  // executes method main of a given class
  virtual void execute(mLoadedClass*);

  // executes method at a given address
  virtual void executeMethod(int);

  virtual void printObject(int);

  virtual int* getStack(){ return stack_ptr; }

  // pushes onto stack
  virtual void push(int a){ *stack_ptr = a; stack_ptr++;}

  // setsup environment
  // used to set internal variable (nPushed) to size of environment
  // this method is used in native code when invoking methods with environment
  virtual void setEnvironmentSize(int s){ nPushed = s; }

 private:


  // creates object descriptor (selfish,self pair)
  // self is previously allocated object
  inline int createDescriptor(int selfish,int self);  

  inline void dumpLocalVariable(int,int*);
  

  inline bool testnil(int**);
  
  // env of virtual machine
  mVMEnvironment *environment;


  // stack pointer
  int* stack_ptr;


  // flag variables inside interpretation

  // set to true in termination instruction
  bool terminated;


  // set to true when debugging instructions
  bool debug;

  // original (starting) address of stack. this is obtained from vm
  int* stack_origin;

  // system flags register
  // bits: 
  //     0 - block_return ( set to 0 if block didn't return, to 1 on return )  
  unsigned flags;


  // number of implicitly pushed parent references inside invocation.
  int nPushed;


  
  
  
};

#endif
