#include "mVMEnvironment.h"
#include "mVirtualMachine.h"
#include "ObjectOffsets.h"
#include "mString.h"



mVMEnvironment::mVMEnvironment(mVirtualMachine* vm)
{

  theMachine = vm;

}


int* mVMEnvironment::createInteger(int i)
{

  int* classData = (int*)theMachine->getIntegerClass();

  int* object =  theMachine->create_object(classData);     
  
  *(object + INTEGER_OFFSET) = i;

  return object;
}


int* mVMEnvironment::createFloat(float i)
{

  //int* classData = (int*)theMachine->getFloatClass();

  //int* object =  theMachine->create_object(classData);     
  
  //*(object + FLOAT_OFFSET) = i;

  return 0; //object;
}


int* mVMEnvironment::createDouble(double i)
{

  int* classData = (int*)theMachine->getDoubleClass();

  int* object =  theMachine->create_object(classData);     
  
  *(double*)(object + DOUBLE_OFFSET) = i;

  return object;
}


int* mVMEnvironment::createString(char* i)
{

  int* classData = (int*)theMachine->getStringClass();

  int* object =  theMachine->create_object(classData); 


  int length = strlen( i );  // length of string

      
  char* chars = (char*)theMachine->allocate( length + 1 );

  memcpy( chars, i, length );
  *(chars + length) = '\0';  // add null

  // first is length
  *(object + STRING_OFFSET) = length;
  // second is ptr to chars
  *(object + STRING_OFFSET+1) = (int)chars;
      

  return object;
}


int* mVMEnvironment::createBoolean(int i)
{

  int* classData = (int*)theMachine->getBooleanClass();

  int* object =  theMachine->create_object(classData);     
  

  
  if(i == 0)
    *(object + BOOLEAN_OFFSET) = 0;
  else
    *(object + BOOLEAN_OFFSET) = 1;


  return object;
}



int* mVMEnvironment::createWithDefaultConstructor(mString* classname, int size)
{

  // find and load class if it hasn't already been
  mLoadedClass* loaded = theMachine->findAndLoadClass( classname );

  // create object from classdata
  int* object =  theMachine->create_user_object( (int*)loaded->dataStartAddress, size );


  // find and execute default constructor
  //theMachine->getInterpreter()->executeMethod( 0 );   tukaj




  return object;

}



int* mVMEnvironment::createPrimitiveObject(mString* classname)
{

  // find and load class if it hasn't already been
  mLoadedClass* loaded = theMachine->findAndLoadClass( classname );

  // create object from classdata
  int* object =  theMachine->create_object( (int*)loaded->dataStartAddress );

  return object;

}


int* mVMEnvironment::createObjectFromClassData(int classdata, int size)
{

  // create object from classdata
  int* object =  theMachine->create_user_object( (int*)classdata, size);  

  return object;
}



int mVMEnvironment::getMethodIndex(char* className, char* methodName, char* arguments, int& mtableOffset,
				   int& classAddress)
{

  // find and load class if it hasn't already been
  mString cname(className);
  mLoadedClass* loaded = theMachine->findAndLoadClass( &cname );
  
  // find method entry in meta method info of metadata

  int* classData = (int*)loaded->dataStartAddress;  

  classAddress = (int)classData;


  int metaSize = *(classData + 4);

  int* metaInfo = (int*)( (char*)classData - metaSize);  // ptr to start of metadata

  // # of methods
  int nMethods = *(metaInfo + 3);
  

  // move to meta info for methods

  MetaMethodEntry* methodList = (MetaMethodEntry*) *(metaInfo + 4);


  char** args = new char*[100];   // 100 parameters max

  int nArgs = 0;
  char currentArg[100];  
  int len = 0;

  if( strlen(arguments) > 0 )
    for(unsigned i=0; i< strlen(arguments) +1; i++){

      // terminate it
      if( arguments[i] == ' ' || arguments[i] == '\0'){
	currentArg[len] = '\0';
	args[nArgs] = new char[ len ];
	strcpy( args[nArgs], currentArg);
	nArgs++; len=0;      
      }else
	currentArg[len++] = arguments[i];    
    }
  
  
  


  //for(int i=0;i<nArgs;i++)
  //cout<<"------- argument type = "<<args[i]<<endl;


  for(int i=0; i<nMethods; i++){

    MetaMethodEntry* entry = (MetaMethodEntry*) *((int*)methodList + i);

    // method name and # parameters must match
    if(strcmp(entry->methodName, methodName) == 0 && entry->nParameters == nArgs){
      

      // name and # params match, now check parameter types

      bool match = true;
      for(int j=0;j < entry->nParameters; j++){
	
	// classdata for parameter class
	int* classdata = (int*) *(entry->argClassAddresses + j);

	int metaSize = *(classdata + 4);

	int* paramMeta = (int*)( (char*)classdata - metaSize);  // ptr to start of metadata
	
	
	if( strcmp( (char*)*paramMeta, args[j] ) != 0 ){   // -> classname of param j doesn't match
	  match = false;
	  break;
	  
	}
	
      }

      if(match){

	if(mtableOffset == -1){

	  // mtableOffset == -1 means that we must search for it!

	  // match found! return index of this method and find mtable offset for this class

	  // get mtable offset for this class from classtable
	  int ctableEntries = *(classData + 3);
	  
	  //cout<<"--- object table  ( "<<ctableEntries<<" entries )" <<endl;
	  int *classTable = (classData + 5);   // --> class table
	  
	  for(int t=0; t < ctableEntries; t++, classTable+=3){
	    
	    int* cData = (int*)*classTable;
	    
	    // get classname of this class
	    int metaSize = *(cData + 4);
	    
	    int* cMeta = (int*)( (char*)cData - metaSize);  // ptr to start of metadata
	    if( strcmp( (char*)*cMeta, className ) == 0 ){
	      
	      mtableOffset = *(classTable + 2);
	      break;
	    }
	    
	    
	  }
	}
	
	return entry->vtableIndex;
      }
    

    }

  }
  
  cout<<"native: method "<<methodName<<" NOT found! "<<endl;
  return -1;
}




int* mVMEnvironment::getClassMethod(char* className, char* methodName, char* arguments, int* object)
{


  int mtabOffset = -1;  // means that getMethodIndex will search for this offset
                        // in classtable

  int* address = 0;


  int classAddress;
  

  // first check this class
  int vtableIndex = getMethodIndex( className, methodName, arguments, mtabOffset, classAddress );

  //int mtableIndex = -1;

  if(vtableIndex != -1){

    int** obj = (int**)object;   // just to convert to double ptr

    //

    // go into method table and and find method entry    
    int* methodEntry = *(obj+1) + 2 + *(int*)(**(obj+1) +12) * 2 +  3*  *( *obj + vtableIndex);


    
    address = methodEntry; 

    //cout<<"vtable index="<<vtableIndex<<", mtableOffset= "<<mtabOffset<<endl
    //<<"method address="<<*methodEntry<<", env ptr="<<*(methodEntry+2)<<endl;


  }


  return address;
  

}


void mVMEnvironment::invokeMethod(int* methodEntry, int* arg1, int* arg2, 
				  int* arg3, int* arg4)
{

  // invoke method at methodEntry position

  mAbstractInterpreter* interpreter = theMachine->getInterpreter();
  

  // push method arguments
  if(arg4!=0) interpreter->push( (int)arg4 );
  if(arg3!=0) interpreter->push( (int)arg3 );
  if(arg2!=0) interpreter->push( (int)arg2 );
  if(arg1!=0) interpreter->push( (int)arg1 );
    

  // push enviroment TUKAJ,ce obstaja
  // if environment exist we push its values
  int *env = (int*)*(methodEntry+2);
      

  if(env){

    interpreter->setEnvironmentSize( *env );

    // push all references (back to front)
    for(int i=*env; i > 0; i--)

      interpreter->push( *(env + i) );
    
  }

  

  // actual self
  interpreter->push( *(methodEntry+1) );


  // address of next instruction is 0!!! ( this means that after return instruction we finish 
  // interpretation)
  interpreter->push( 0 );

  //stack_ptr+=2;
  //*(stack_ptr-2) = *(methodEntry+1);   // actual self
  //*(stack_ptr-1) = 0;                  // address of next instruction is 0!!! ( this means 
                                       // that after return instruction we finish interpretation)


  // execute method 
  interpreter->executeMethod( *methodEntry );



  

}


void mVMEnvironment::invokeFunction(int* methodEntry, int* arg1, int* arg2, 
				  int* arg3, int* arg4)
{

  // invoke method at methodEntry position

  mAbstractInterpreter* interpreter = theMachine->getInterpreter();
  

  // push method arguments
  if(arg4!=0) interpreter->push( (int)arg4 );
  if(arg3!=0) interpreter->push( (int)arg3 );
  if(arg2!=0) interpreter->push( (int)arg2 );
  if(arg1!=0) interpreter->push( (int)arg1 );
    

  // if environment exist we push its values
  int *env = (int*)*(methodEntry+2);
      

  if(env){
    
    interpreter->setEnvironmentSize( *env );

    // push all references (back to front)
    for(int i=*env; i > 0; i--){

      interpreter->push( *(env + i) );

    }
    
  }



  // actual self
  interpreter->push( *(methodEntry+1) );


  // address of next instruction is 0!!! ( this means that after return instruction we finish 
  // interpretation)
  interpreter->push( 0 );


  // location of result - top of stack
  interpreter->push( -1 );
  

  // execute method  
  interpreter->executeMethod( *methodEntry );

  

}



void mVMEnvironment::invokeMethodOnObject(int* methodEntry, int* object, int* arg1, int* arg2, 
				  int* arg3, int* arg4)
{

  // invoke method at methodEntry position

  mAbstractInterpreter* interpreter = theMachine->getInterpreter();
  

  // push method arguments
  if(arg4!=0) interpreter->push( (int)arg4 );
  if(arg3!=0) interpreter->push( (int)arg3 );
  if(arg2!=0) interpreter->push( (int)arg2 );
  if(arg1!=0) interpreter->push( (int)arg1 );
    

  // if environment exist we push its values
  int *env = (int*)*(methodEntry+2);
      

  if(env){

    interpreter->setEnvironmentSize( *env );

    // push all references (back to front)
    for(int i=*env; i > 0; i--)

      interpreter->push( *(env + i) );
    
  }
  

  // object is original object, but we need specific descriptor
  // this we can get by adding default selfish from methodtable in classdata to
  // the start of our receiver
  interpreter->push( (int)object /*+ *(methodEntry+1)*/ );


  // address of next instruction is 0!!! ( this means that after return instruction we finish 
  // interpretation)
  interpreter->push( 0 );

  //stack_ptr+=2;
  //*(stack_ptr-2) = *(methodEntry+1);   // actual self
  //*(stack_ptr-1) = 0;                  // address of next instruction is 0!!! ( this means 
                                       // that after return instruction we finish interpretation)


  // execute method 
  interpreter->executeMethod( *methodEntry );



  

}


void mVMEnvironment::invokeFunctionOnObject(int* methodEntry, int* object, int* arg1, int* arg2, 
				  int* arg3, int* arg4)
{

  // invoke method at methodEntry position

  mAbstractInterpreter* interpreter = theMachine->getInterpreter();
  

  // push method arguments
  if(arg4!=0) interpreter->push( (int)arg4 );
  if(arg3!=0) interpreter->push( (int)arg3 );
  if(arg2!=0) interpreter->push( (int)arg2 );
  if(arg1!=0) interpreter->push( (int)arg1 );
    

  // if environment exist we push its values
  int *env = (int*)*(methodEntry+2);
      

  if(env){
    
    interpreter->setEnvironmentSize( *env );

    // push all references (back to front)
    for(int i=*env; i > 0; i--){

      interpreter->push( *(env + i) );

    }
    
  }



  // actual self
  // object is original object, but we need specific descriptor
  // this we can get by adding default selfish from methodtable in classdata to
  // the start of our receiver

  interpreter->push( (int)object /*+ *(methodEntry+1)*/ );


  // address of next instruction is 0!!! ( this means that after return instruction we finish 
  // interpretation)
  interpreter->push( 0 );


  // location of result - top of stack
  interpreter->push( -1 );
  

  // execute method  
  interpreter->executeMethod( *methodEntry );

  

}



void mVMEnvironment::invokeMethod(int methodAddress)
{

  
  mAbstractInterpreter* interpreter = theMachine->getInterpreter();
  

  
  // push enviroment TUKAJ,ce obstaja

  

  
  // execute method
  interpreter->executeMethod( methodAddress );

  

}



void mVMEnvironment::invokeAnonymousMethod(MetaAnonymousEntry* entry, int* object)
{
  mAbstractInterpreter* interpreter = theMachine->getInterpreter();
  

  // if environment exist we push its values
  int *env = entry->environment;
      

  if(env){
    
    interpreter->setEnvironmentSize( *env );    

    for(int i=1; i <= *env ; i++){

      interpreter->push( *(env + i) );

    }
    
  }

    
  interpreter->push( (int)object );


  // address of next instruction is 0!!! ( this means that after return instruction we finish 
  // interpretation)
  interpreter->push( 0 );
  
  
  // execute method
  interpreter->executeMethod( entry->closure );

} 



void mVMEnvironment::invokeAnonymousFunction(MetaAnonymousEntry* entry, int* object)
{
  mAbstractInterpreter* interpreter = theMachine->getInterpreter();
  

  // if environment exist we push its values
  int *env = entry->environment;

  if(env){
    
    interpreter->setEnvironmentSize( *env );
    
    for(int i=1; i <= *env ; i++){

      interpreter->push( *(env + i) );

    }

    
  }

   
  interpreter->push( (int)object );


  // address of next instruction is 0!!! ( this means that after return instruction we finish 
  // interpretation)
  interpreter->push( 0 );
  

  // location of result - top of stack
  interpreter->push( -1 );

  
  // execute method
  interpreter->executeMethod( entry->closure );

} 




void mVMEnvironment::pushObject(int* object)
{

  mAbstractInterpreter* interpreter = theMachine->getInterpreter();
  
  // push 
  interpreter->push( (int)object );

}
