#ifndef __MVMENVIRONMENT
#define __MVMENVIRONMENT


#include "../base_classes/mDefs.h"

class mString;
class mVirtualMachine;






// environment of virtual machine
// includes utility functions for creating objects of base classes and functionality to invoke methods

class mVMEnvironment{

 public:

  mVMEnvironment(mVirtualMachine*);  

  //////////////////////////////////////////////////////
  // functions for creating objects of primitve types
  // NOTE: objects returned point to 'original' objects: to Object
  // to make object point to correct descriptor add 4 to all primitive objects
  //////////////////////////////////////////////////////

  // creates an Integer object from value and returns a pointer to it
  int* createInteger(int i);

  // creates a Float object from value and returns a pointer to it
  int* createFloat(float i);

  // creates a Double object from value and returns a pointer to it
  int* createDouble(double i);

  // creates a String object from value and returns a pointer to it
  int* createString(char* i);

  // creates a Boolean object from value and returns a pointer to it
  // i=0 -> false, i=1 -> true
  int* createBoolean(int i);

  // creates other primitive object (Block, While,...)
  int* createPrimitiveObject(mString*);

  // pushes object onto stack
  void pushObject(int*);

  // creates an Object 
  // do we really need 'just' Objects from native world???
  //int* createObject();

  // creates a user object from class in argument with default constructor
  int* createWithDefaultConstructor(mString* classname, int size=0);


  // creates a user object from classdata
  int* createObjectFromClassData(int classdata, int size);  


  // returns address of method table entry for specified method
  // args: class name, method name, argument list (separated by space), instance upon which
  // to invoke
  int* getClassMethod(char* className, char* methodName, char* arguments, int* object);

  // method invocation
  // args: object reference, method entry returned by getClassMethodAddress,
  // method parameters
  void invokeMethod(int* methodEntry, int* arg1=0, int* arg2=0, int* arg3=0, int* arg4=0); 


  // function invocation
  // args: object reference, method entry returned by getClassMethodAddress,
  // method parameters
  // same as above only result is returned on top of stack
  void invokeFunction(int* methodEntry, int* arg1=0, int* arg2=0, int* arg3=0, int* arg4=0); 



  // these two are same as above, the only difference being that they are invoked
  // upon explicit receiver

  // function invocation
  // args: object reference, method entry returned by getClassMethodAddress,
  // method parameters
  // same as above only result is returned on top of stack
  void invokeFunctionOnObject(int* methodEntry, int* object, int* arg1=0, int* arg2=0, int* arg3=0, int* arg4=0); 


  // method invocation, same as above, only arguments should already have been pushed
  // param is method address
  void invokeMethodOnObject(int* methodEntry, int* object, int* arg1=0, int* arg2=0, int* arg3=0, int* arg4=0);



  // anonymous method invocation
  // args: ptr to anonymous meta environment, receiver
  void invokeAnonymousMethod(MetaAnonymousEntry* entry, int* object);


  // anonymous function invocation
  // args: ptr to anonymous meta environment, receiver
  void invokeAnonymousFunction(MetaAnonymousEntry* entry, int* object);



  // method invocation, same as above, only arguments should already have been pushed
  // param is method address
  void invokeMethod(int);
  

  //void pushArgument(int value);

  //int getStack(){ theMachine->getInterpreter()->getStack(); }


  


 private:

  // returns index of specified method in virtual table, and offset into mtable for specified class,
  // class address for that same class
  // args: class name, method name, argument list (separated by space)  
  int getMethodIndex(char* className, char* methodName, char* arguments, int& mtableOffset, 
		     int& classAddress);


  mVirtualMachine *theMachine;
  


};



#endif
