#include <iostream>
#include "nativeInterface.h"
#include "shared.h"
#include "../base_classes/mDefs.h"


/*

 Implements functions of Class in native.
 Class Class contains a 4-byte value, after method table, which is 
 the address of classdata for specified class. This address is used to gather
 all meta information about that class. Meta information is placed after literal
 pool and before class info fields.
 Meta information layout is as follows:
 
 char* className;            // ptr to classname with terminating '\0'
 int nDirectSupersclasses;   // # of direct superclasses of this class

*/


extern "C" {

  // Returns the name of the Class.
  void Z__Class__getName__0(functionFrame* &frame, mVMEnvironment* env);


  // Sets the name of the Class.
  void Z__Class__setName__1__String(procedureFrame* &frame, mVMEnvironment* env);


  // Returns vector of direct superclasses
  void Z__Class__getSuperClasses__0(functionFrame*& frame, mVMEnvironment *env );


  // Returns a vector of declared methods
  void Z__Class__getDeclaredMethods__0(functionFrame*& frame, mVMEnvironment *env );

  // Returns a specific method
  void Z__Class__getDeclaredMethod__2__String__Vector(functionFrame*& frame, mVMEnvironment *env );

  // Returns a vector of constructor methods
  void Z__Class__getDeclaredConstructors__0(functionFrame*& frame, mVMEnvironment *env );

  // Returns true if this class is a subclass of the one passed as argument
  void Z__Class__isSubclassOf__1__Class(functionFrame*& frame, mVMEnvironment *env );

  // Returns true if this class is a superclass of the one passed as argument
  void Z__Class__isSuperclassOf__1__Class(functionFrame*& frame, mVMEnvironment *env );

  // Returns true if object is an instance of this class
  void Z__Class__isClassOf__1__Object(functionFrame*& frame, mVMEnvironment *env );

  // Sets class method
  void Z__Class__setMethod__2__Method__Method(procedureFrame*& frame, mVMEnvironment *env );

}


using namespace std;




void Z__Class__getName__0(functionFrame*& frame, mVMEnvironment *env )
{

  // get Class object
  int* object = (int*)Utils::getSelf(frame);  

  int* metaInfo = getMetaAddress( object );

  // classname is the first meta info in classdata
  char* cnameptr = (char*)*metaInfo;  // -> classname chars  

  // create String object and push it as return value
  int* name = env->createString( cnameptr );

  
  Utils::pushObject( (int*)frame, name + 2 );
  
}




void Z__Class__setName__1__String(procedureFrame*& frame, mVMEnvironment *env )
{  

  // get Class object

  int* object = (int*)Utils::getSelf(frame);  

  int* metaInfo = getMetaAddress( object );

  // classname is the first meta info in classdata
  char* cnameptr = (char*)*metaInfo;  // -> classname chars  


  delete[] cnameptr;    // remove old string


  // get atring argument - new classname
  char* newString = Utils::getStringValue(frame, 0);

  // alloc new string and copy name into it
  cnameptr = new char[ strlen(newString) + 1 ];
  
  strcpy( cnameptr, newString );

  // store new string to 1st field of metainfo
  *metaInfo = (int)cnameptr;

}



void Z__Class__getSuperClasses__0(functionFrame*& frame, mVMEnvironment *env )
{

  // create Vector of Class objects


  // 1st create Vector and find method 'add' which will be used for adding Class'es
  mString vecName( "Vector" );
  int* vec = env->createWithDefaultConstructor( &vecName );


  int* addMethod = env->getClassMethod("Vector", "add", "Object", vec);  


  // get Class object to obtain superclasses
  int* object = (int*)Utils::getSelf(frame);  

  int* metaInfo = getMetaAddress( object );

  // number of direct superclasses is second field of meta info in classdata
  int nSuperClasses = *(metaInfo + 1);

  

  int* superList = (int*)*(metaInfo + 2);



  // now we'll create Class objects for all direct superclasses
  mString classObjectName( "Class" );


  int* obj = 0;

  for(int i=0;i<nSuperClasses;i++){    

    // obj already points to correct descriptor
    obj = env->createWithDefaultConstructor( &classObjectName );

    int *originalObject = (int*) *(obj + 1);

    // store classdata to classdata field of create Class object
    *(originalObject + CLASS_OFFSET) = *(superList + i);

    // store this object into vector by invoking 'Vector.add'
    env->invokeMethod( addMethod, obj );    


  }  

  
  // push Vector of Class'es as result
  Utils::pushObject( (int*)frame, vec );

}


void Z__Class__getDeclaredMethods__0(functionFrame*& frame, mVMEnvironment *env )
{


  // create Vector of Method objects


  // 1st create Vector and find method 'add' which will be used for adding Method's
  mString vecName( "Vector" );
  int* vec = env->createWithDefaultConstructor( &vecName );


  int* addMethod = env->getClassMethod("Vector", "add", "Object", vec);  


  // get Class object
  int* object = (int*)Utils::getSelf(frame);


  int classData = *(object + CLASS_OFFSET );

  int* metaInfo = getMetaAddress( object );

  // number of methods is fourth field of meta info in classdata
  
  int nMethods = *(metaInfo + 3);

  
  
  // now we'll create Method objects
  mString classMethodName( "Method" );

  int* obj = 0;

  MetaMethodEntry* methodList = (MetaMethodEntry*) *(metaInfo + 4);

  for(int i=0; i<nMethods; i++){

    MetaMethodEntry* entry = (MetaMethodEntry*) *((int*)methodList + i);    


    obj = env->createWithDefaultConstructor( &classMethodName );

    int *originalObject = (int*) *(obj + 1);

    // store classdata to classdata field of created Method object
    *(originalObject + METHOD_OFFSET) = classData;

    // and index ( this is entry->virtualTableIndex, but this actually equals i
    // because indices start at 0 in each virtual table)
    *(originalObject + METHOD_OFFSET + 1) = i;



    // TUKAJ!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    // add owner object

    // for this we need to find receiver object.....




    *(originalObject + METHOD_OFFSET + 2) = i;

    // store this object into vector by invoking 'Vector.add'
    env->invokeMethod( addMethod, obj );

  }   

 
  
  Utils::pushObject( (int*)frame, vec );
    

}


///////////////////////////////////////////////////////////////////////////////////////////

void Z__Class__getDeclaredMethod__2__String__Vector(functionFrame*& frame, mVMEnvironment *env )
{

  // get Class object
  int* object = (int*)Utils::getSelf(frame);


  int classData = *(object + CLASS_OFFSET );

  int* metaInfo = getMetaAddress( object );

  // number of methods is fourth field of meta info in classdata
  
  int nMethods = *(metaInfo + 3);

  char* argumentNames[64];   // argument names

  
  // # of arguments in Vector
  int nArguments = 0;

  if( Utils::isNil(frame, 0) ){ // name is nil, return nil
    
    Utils::pushObject( (int*)frame, 0 );
    return;

  }
  

  // get method name
  char* methodName = Utils::getStringValue( frame, 0 );



  // check if arguments are not nil
  if( !Utils::isNil(frame, 1) ){

    // vector of parameters
    int* vec = Utils::getReference(frame, 1);    

    // invoke Vector.size to see how may parameters do we have in Vector    
    int* sizeMethod = env->getClassMethod("Vector", "size", "", vec);    
    
    env->invokeFunction( sizeMethod );     

    int* intSize =  (int*)Utils::top( (int*)frame );   // Integer object on top of frame
    

    // # of pushed arguments
    nArguments = *( *(int**)(intSize+1) +INTEGER_OFFSET);


    // Vector.at will give us every argument    
    int* atMethod = env->getClassMethod("Vector", "at", "Integer", vec);

    // fill in argument names
    // this is to avoid repetitive invocation inside loop later
    for(int i=0; i<nArguments; i++){

      // create Integer for index for method at
      int* b = env->createInteger( i );

      env->invokeFunction( atMethod, b + 4 );

      int* argument =  (int*)Utils::top( (int*)frame );   // argument object on top of frame


      // check if it's String and get chars, if not, return error
      // because Vector accepts any kind of Objects, so it may not
      // neccesarily be a String
     
      argumentNames[i] = (char*)  *( *(int**)(argument + 1) + STRING_OFFSET + 1);
    }


  }
  
  

  MetaMethodEntry* methodList = (MetaMethodEntry*) *(metaInfo + 4);

  int* obj = 0;
  

  // check if method matches
  for(int i=0; i<nMethods; i++){

    MetaMethodEntry* entry = (MetaMethodEntry*) *((int*)methodList + i);    


    // name and # of arguments must match
    if( strcmp( entry->methodName, methodName ) == 0 && entry->nParameters == nArguments ){

      // name and argument count match, now check argument types
      bool match = true;
      for(int j=0; j<nArguments; j++){

	// classname of argument in classdata must match that in Vector at jth position
	int* metaArgument = getMetaAddressFromClassData( (int*)entry->argClassAddresses[j] );
	
	char* cnameptr = (char*)*metaArgument;  // name is 1st field of metadata

	// get name from Vector
	if( strcmp( cnameptr, argumentNames[j] ) != 0){  // they don't match

	  // this is not the method, break out
	  match = false;
	  break;

	}
	

      }

      if(!match) continue;  // arguments don't match, skip to next method


      mString classMethodName( "Method" );

      obj = env->createWithDefaultConstructor( &classMethodName );

      int *originalObject = (int*) *(obj + 1);

      // store classdata to classdata field of created Method object
      *(originalObject + METHOD_OFFSET) = classData;
      
      // and index ( this is entry->virtualTableIndex)
      *(originalObject + METHOD_OFFSET + 1) = entry->vtableIndex;      
      
      // TUKAJ!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      // add owner object
      
      // for this we need to find receiver object.....
          


      //*(originalObject + METHOD_OFFSET + 2) = i;


      // method found, so we break out
      break;

    }


  }

  // push created method
  Utils::pushObject( (int*)frame, obj );
  
}



///////////////////////////////////////////////////////////////////////////////////////////

void Z__Class__getDeclaredConstructors__0(functionFrame*& frame, mVMEnvironment *env )
{


  // create Vector of Method objects


  // 1st create Vector and find method 'add' which will be used for adding Method's
  mString vecName( "Vector" );
  int* vec = env->createWithDefaultConstructor( &vecName );


  int* addMethod = env->getClassMethod("Vector", "add", "Object", vec);  


  // get Class object
  int* object = (int*)Utils::getSelf(frame);


  int classData = *(object + CLASS_OFFSET );

  int* metaInfo = getMetaAddress( object );

  // number of methods is fourth field of meta info in classdata
  
  int nMethods = *(metaInfo + 3);

  
  
  // now we'll create Method objects
  mString classMethodName( "Method" );

  int* obj = 0;

  MetaMethodEntry* methodList = (MetaMethodEntry*) *(metaInfo + 4);

  for(int i=0; i<nMethods; i++){

    MetaMethodEntry* entry = (MetaMethodEntry*) *((int*)methodList + i);
    
    
    // only constructors  (4th bit set)
    if( (entry->flags & 8) == 0) continue;


    obj = env->createWithDefaultConstructor( &classMethodName );

    int *originalObject = (int*) *(obj + 1);

    // store classdata to classdata field of created Method object
    *(originalObject + METHOD_OFFSET) = classData;

    // and index
    *(originalObject + METHOD_OFFSET + 1) = i;

    // store this object into vector by invoking 'Vector.add'
    env->invokeMethod( addMethod, obj );

  }   

 
  
  Utils::pushObject( (int*)frame, vec );
    

}


///////////////////////////////////////////////////////////////////////////////////////////


// returns true if thisClassData is superClassData or found in one of its superclasses
bool isThisClass(int thisClassData, int superClassData)
{
  
  // classdata matches
  if( thisClassData == superClassData) return true;

  // now search for thisClassData in superClassData
  int* metaInfo = getMetaAddressFromClassData( (int*)superClassData );
  int* superList = (int*)*(metaInfo + 2);   // superlist of superclass
  int nSuperClasses = *(metaInfo + 1);      // # superclasses of superclass

  for(int i=0;i<nSuperClasses;i++)

    if(isThisClass( thisClassData, *(superList + i) ))
      return true;
      
  return false;
}



void Z__Class__isSubclassOf__1__Class(functionFrame*& frame, mVMEnvironment *env )
{

  // class passed as argument must be in our superlist, because then argument class is
  // our superclass and hence we're its subclass

  // get Class object to obtain superclasses
  int* object = (int*)Utils::getSelf(frame);  

  int* metaInfo = getMetaAddress( object );

  // number of direct superclasses is second field of meta info in classdata
  int nSuperClasses = *(metaInfo + 1);


  int* superList = (int*)*(metaInfo + 2);

  // get Class argument
  int* otherClass = (int*)Utils::getArgument(frame, 0);  


  // classaddress of other class must be in this list
  int found = 0;  // assume not found

  // zdaj za vsak nadrazred preverimo, ce je njegov classdata enak temu v argumentu

  for(int i=0;i<nSuperClasses;i++){ 


    if(isThisClass( *(otherClass + CLASS_OFFSET), *(superList + i) ) ){       

      found = 1;
      break;

    }
  }


  // create Boolean object and push it as return value
  int* exists = env->createBoolean( found );

  
  Utils::pushObject( (int*)frame, exists + 4 );

}



void Z__Class__isSuperclassOf__1__Class(functionFrame*& frame, mVMEnvironment *env )
{

  // our class must be in superlist of argument class

  // get Class object to obtain superclasses
  int* object = (int*)Utils::getArgument(frame, 0);  

  int* metaInfo = getMetaAddress( object );

  // number of direct superclasses is second field of meta info in classdata
  int nSuperClasses = *(metaInfo + 1);


  int* superList = (int*)*(metaInfo + 2);

  // get Class argument
  int* thisClass = (int*)Utils::getSelf(frame);  


  // classaddress of other class must be in this list
  int found = 0;  // assume not found

  // zdaj za vsak nadrazred preverimo, ce je njegov classdata enak nasemu razredu

  for(int i=0;i<nSuperClasses;i++){ 


    if(isThisClass( *(thisClass + CLASS_OFFSET), *(superList + i) ) ){       

      found = 1;
      break;

    }
  }


  // create Boolean object and push it as return value
  int* exists = env->createBoolean( found );

  
  Utils::pushObject( (int*)frame, exists + 4 );
}



void Z__Class__isClassOf__1__Object(functionFrame*& frame, mVMEnvironment *env )
{  

  // our class must be in classlist of argument object

  // get Object object to obtain its classdata
  int* object = (int*)Utils::getArgument(frame, 0);  

  // get our Class
  int* thisClass = (int*)Utils::getSelf(frame); 
  
  int classAddress = *(thisClass + CLASS_OFFSET);   // classdata of our class

  int *classTable = (int*)( *object + 20 );   // -> classTable of Object argument (20 bytes
                                                   // for classinfo, after which classtable starts)

  int ctableEntries = *(int*)( *object + 12); // 4th field of classinfo is # ctable entries


  int found = 0;
  // now check if classdata of our class is among those in classtable
  for(int i=0;i<ctableEntries;i++){

    if( *(classTable + i*3) == classAddress){
      found = 1;
      break;
    }
  }
  

  // create Boolean object and push it as return value
  int* exists = env->createBoolean( found );

  
  Utils::pushObject( (int*)frame, exists + 4 );
  
}



void Z__Class__setMethod__2__Method__Method(procedureFrame*& frame, mVMEnvironment *env )
{

  //TUKAJ

}
