#include <iostream>
#include "nativeInterface.h"
#include "shared.h"
#include "../base_classes/mDefs.h"


/*

 Implements functions of Method in native.
 Class Method contains a 4-byte value, after method table, which is 
 the address of classdata for class of this method. A second 4-byte
 value is the index of this method in table for meta data for methods 

*/


extern "C" {

  // Returns the name of method.
  void Z__Method__getName__0(functionFrame* &frame, mVMEnvironment* env);
  
  // Returns true if method is a constructor
  void Z__Method__isConstructor__0(functionFrame* &frame, mVMEnvironment* env);

  // Returns true if method is a function
  void Z__Method__isFunction__0(functionFrame* &frame, mVMEnvironment* env);

  // Returns true if method is static
  void Z__Method__isStatic__0(functionFrame* &frame, mVMEnvironment* env);

  // Returns true if method is native
  void Z__Method__isNative__0(functionFrame* &frame, mVMEnvironment* env);

  // Returns true if method is public
  void Z__Method__isPublic__0(functionFrame* &frame, mVMEnvironment* env);

  // Returns true if method is private
  void Z__Method__isPrivate__0(functionFrame* &frame, mVMEnvironment* env);

  // Returns true if method is protected
  void Z__Method__isProtected__0(functionFrame* &frame, mVMEnvironment* env);

  // Returns true if method is restricted
  void Z__Method__isRestricted__0(functionFrame* &frame, mVMEnvironment* env);

  // Returns true if method is anonymous
  void Z__Method__isAnonymous__0(functionFrame* &frame, mVMEnvironment* env);

  // Returns return type or nil if procedure
  void Z__Method__getReturnType__0(functionFrame* &frame, mVMEnvironment* env);

  // Returns types of parameters  
  void Z__Method__getParameterTypes__0(functionFrame* &frame, mVMEnvironment* env);

  // Returns method closure (outer block)  
  void Z__Method__getBlock__0(functionFrame* &frame, mVMEnvironment* env);

  // Invokes method with arguments upon object
  void Z__Method__invoke__2__Object__Vector(functionFrame* &frame, mVMEnvironment* env);

  // Invokes method with arguments upon object
  void Z__Method__invoke__1__Vector(functionFrame* &frame, mVMEnvironment* env);
}


using namespace std;





void Z__Method__getName__0(functionFrame*& frame, mVMEnvironment *env )
{

  // get Method object
  int* object = (int*)Utils::getSelf(frame);  

  // index of this method is after classdata in Method object
  int metaIndex = *(object + METHOD_OFFSET + 1);


  int* name;

  // anonymous method?
  if( metaIndex == -1 ){

    // has special name
    name = env->createString( "#anonymous#" );

  }else{

    // normal method

    int* metaInfo = getMetaAddressWithOffset( object, METHOD_OFFSET );

    // meta method entries is 5th field of meta info in classdata
    MetaMethodEntry* entry =  (MetaMethodEntry*) *( (int*)*(metaInfo + 4) + metaIndex); // -> method entry  

 
    // create String object and push it as return value
    name = env->createString( entry->methodName );

  }

  
  Utils::pushObject( (int*)frame, name + 4 );
  
}



void Z__Method__isConstructor__0(functionFrame*& frame, mVMEnvironment *env )
{

  // get Method object
  int* object = (int*)Utils::getSelf(frame);  

  // index of this method is after classdata in Method object
  int metaIndex = *(object + METHOD_OFFSET + 1);


  int* b;  

  // anonymous method?
  if( metaIndex == -1 ){

    // anonymous is never construnctor
    b = env->createBoolean( 0 );

  }else{

    // normal method


    int* metaInfo = getMetaAddressWithOffset( object, METHOD_OFFSET );

    // meta method entries is 5th field of meta info in classdata
    MetaMethodEntry* entry =  (MetaMethodEntry*) *( (int*)*(metaInfo + 4) + metaIndex);  // -> method entry


    // create Boolean object and push it as return value
    // 4th bit is set for constructors
    b = env->createBoolean( (entry->flags & 8) != 0 ? 1 : 0 );
  }
  
  Utils::pushObject( (int*)frame, b + 4 );
  
}




void Z__Method__isFunction__0(functionFrame* &frame, mVMEnvironment* env)
{

  // get Method object
  int* object = (int*)Utils::getSelf(frame);  

  // index of this method is after classdata in Method object
  int metaIndex = *(object + METHOD_OFFSET + 1);


  int* b;

  // anonymous method?
  if( metaIndex == -1 ){

    // get flags
    MetaAnonymousEntry* anonmeta = (MetaAnonymousEntry*) *(object + METHOD_OFFSET + 3);
    b = env->createBoolean( anonmeta->flags & 1 == 1 ? 1 : 0 );

  }else{

    // normal method


    int* metaInfo = getMetaAddressWithOffset( object, METHOD_OFFSET );

    // meta method entries is 5th field of meta info in classdata
    MetaMethodEntry* entry =  (MetaMethodEntry*) *( (int*)*(metaInfo + 4) + metaIndex);  // -> method entry


    // create Boolean object and push it as return value
    // 1st bit is set for functions
    b = env->createBoolean( (entry->flags & 1) != 0 ? 1 : 0 );

  }
  
  Utils::pushObject( (int*)frame, b + 4 );

}


void Z__Method__isStatic__0(functionFrame* &frame, mVMEnvironment* env)
{

  // get Method object
  int* object = (int*)Utils::getSelf(frame);  

  // index of this method is after classdata in Method object
  int metaIndex = *(object + METHOD_OFFSET + 1);

  int* b;

  // anonymous method?
  if( metaIndex == -1 ){

    // get flags
    MetaAnonymousEntry* anonmeta = (MetaAnonymousEntry*) *(object + METHOD_OFFSET + 3);
    b = env->createBoolean( anonmeta->flags & 2 != 0 ? 1 : 0 );

  }else{

    // normal method


    int* metaInfo = getMetaAddressWithOffset( object, METHOD_OFFSET );

    // meta method entries is 5th field of meta info in classdata
    MetaMethodEntry* entry =  (MetaMethodEntry*) *( (int*)*(metaInfo + 4) + metaIndex);  // -> method entry
  

    // create Boolean object and push it as return value
    // 2nd bit is set for functions
    b = env->createBoolean( (entry->flags & 2) != 0 ? 1 : 0 );

  }

  
  Utils::pushObject( (int*)frame, b + 4 );

}


void Z__Method__isNative__0(functionFrame* &frame, mVMEnvironment* env)
{

  // get Method object
  int* object = (int*)Utils::getSelf(frame);  

  // index of this method is after classdata in Method object
  int metaIndex = *(object + METHOD_OFFSET + 1);

  int* b;

  // anonymous method?
  if( metaIndex == -1 ){

    // get flags
    MetaAnonymousEntry* anonmeta = (MetaAnonymousEntry*) *(object + METHOD_OFFSET + 3);
    b = env->createBoolean( anonmeta->flags & 4 != 0 ? 1 : 0 );

  }else{

    // normal method


    int* metaInfo = getMetaAddressWithOffset( object, METHOD_OFFSET );

    // meta method entries is 5th field of meta info in classdata
    MetaMethodEntry* entry =  (MetaMethodEntry*) *( (int*)*(metaInfo + 4) + metaIndex);  // -> method entry


    // create Boolean object and push it as return value
    // 3rd bit is set for functions
    b = env->createBoolean( (entry->flags & 4) != 0 ? 1 : 0 );

  }
  
  Utils::pushObject( (int*)frame, b + 4 );

}


void Z__Method__isPrivate__0(functionFrame* &frame, mVMEnvironment* env)
{

  // get Method object
  int* object = (int*)Utils::getSelf(frame);  

  // index of this method is after classdata in Method object
  int metaIndex = *(object + METHOD_OFFSET + 1);

  int* b;

  // anonymous method?
  if( metaIndex == -1 ){

    // get flags
    MetaAnonymousEntry* anonmeta = (MetaAnonymousEntry*) *(object + METHOD_OFFSET + 3);
    b = env->createBoolean( (anonmeta->flags & 48) != 0 ? 0 : 1 );

  }else{

    // normal method


    int* metaInfo = getMetaAddressWithOffset( object, METHOD_OFFSET );
    
    // meta method entries is 5th field of meta info in classdata
    MetaMethodEntry* entry =  (MetaMethodEntry*) *( (int*)*(metaInfo + 4) + metaIndex);  // -> method entry
    
    
    // create Boolean object and push it as return value
    // bits 5,6: 00 for private
    b = env->createBoolean( (entry->flags & 48) != 0 ? 0 : 1 );
  }
  
  Utils::pushObject( (int*)frame, b + 4 );

}


void Z__Method__isPublic__0(functionFrame* &frame, mVMEnvironment* env)
{

  // get Method object
  int* object = (int*)Utils::getSelf(frame);  

  // index of this method is after classdata in Method object
  int metaIndex = *(object + METHOD_OFFSET + 1);

  int* b;

  // anonymous method?
  if( metaIndex == -1 ){

    // get flags
    MetaAnonymousEntry* anonmeta = (MetaAnonymousEntry*) *(object + METHOD_OFFSET + 3);
    b = env->createBoolean( (anonmeta->flags & 48) == 48 ? 1 : 0 );

  }else{

    // normal method


    int* metaInfo = getMetaAddressWithOffset( object, METHOD_OFFSET );

    // meta method entries is 5th field of meta info in classdata
    MetaMethodEntry* entry =  (MetaMethodEntry*) *( (int*)*(metaInfo + 4) + metaIndex);  // -> method entry


    // create Boolean object and push it as return value
    // bits 5,6: 11 for public
    b = env->createBoolean( (entry->flags & 48) == 48 ? 1 : 0 );
  }
  
  Utils::pushObject( (int*)frame, b + 4 );

}


void Z__Method__isProtected__0(functionFrame* &frame, mVMEnvironment* env)
{

  // get Method object
  int* object = (int*)Utils::getSelf(frame);  

  // index of this method is after classdata in Method object
  int metaIndex = *(object + METHOD_OFFSET + 1);

  int* b;

  // anonymous method?
  if( metaIndex == -1 ){

    // get flags
    MetaAnonymousEntry* anonmeta = (MetaAnonymousEntry*) *(object + METHOD_OFFSET + 3);
    b = env->createBoolean( (anonmeta->flags & 0x10) == 0x10 && (anonmeta->flags & 0x20) == 0 ? 1 : 0 );

  }else{

    // normal method


    int* metaInfo = getMetaAddressWithOffset( object, METHOD_OFFSET );

    // meta method entries is 5th field of meta info in classdata
    MetaMethodEntry* entry =  (MetaMethodEntry*) *( (int*)*(metaInfo + 4) + metaIndex);  // -> method entry


    // create Boolean object and push it as return value
    // bits 5,6: 01 for public
    
    b = env->createBoolean( (entry->flags & 0x10) == 0x10 && (entry->flags & 0x20) == 0 ? 1 : 0 );
  }
  
  Utils::pushObject( (int*)frame, b + 4 );

}



void Z__Method__isRestricted__0(functionFrame* &frame, mVMEnvironment* env)
{

  // get Method object
  int* object = (int*)Utils::getSelf(frame);  

  // index of this method is after classdata in Method object
  int metaIndex = *(object + METHOD_OFFSET + 1);

  int* b;

  // anonymous method?
  if( metaIndex == -1 ){

    // get flags
    MetaAnonymousEntry* anonmeta = (MetaAnonymousEntry*) *(object + METHOD_OFFSET + 3);
    b = env->createBoolean( (anonmeta->flags & 0x20) == 0x20 && (anonmeta->flags & 0x10) == 0 ? 1 : 0 );

  }else{

    // normal method


    int* metaInfo = getMetaAddressWithOffset( object, METHOD_OFFSET );

    // meta method entries is 5th field of meta info in classdata
    MetaMethodEntry* entry =  (MetaMethodEntry*) *( (int*)*(metaInfo + 4) + metaIndex);  // -> method entry


    // create Boolean object and push it as return value
    // bits 5,6: 10 for restricted
    b = env->createBoolean( (entry->flags & 0x20) == 0x20 && (entry->flags & 0x10) == 0 ? 1 : 0 );
  }
  
  Utils::pushObject( (int*)frame, b + 4 );

}


void Z__Method__isAnonymous__0(functionFrame* &frame, mVMEnvironment* env)
{

  // get Method object
  int* object = (int*)Utils::getSelf(frame);  

  // index of this method is after classdata in Method object
  int metaIndex = *(object + METHOD_OFFSET + 1);
  

  // create Boolean object and push it as return value
  // meta index is -1 for anonymous methods
  int* b = env->createBoolean( metaIndex == -1 ? 1 : 0 );

  
  Utils::pushObject( (int*)frame, b + 4 );

}



void Z__Method__getReturnType__0(functionFrame* &frame, mVMEnvironment* env)
{

  // get Method object
  int* object = (int*)Utils::getSelf(frame);  

  // index of this method is after classdata in Method object
  int metaIndex = *(object + METHOD_OFFSET + 1);


  bool isFunc = false;
  int returnClassAddress = 0;


  // anonymous method?
  if( metaIndex == -1 ){

    MetaAnonymousEntry* anonmeta = (MetaAnonymousEntry*) *(object + METHOD_OFFSET + 3);
    isFunc = anonmeta->flags & 1 == 1;
    // return class address is last in arguments list for anonymous functions
    returnClassAddress = anonmeta->argClassAddresses[ anonmeta->nParameters ];

  }else{

    // normal method


    int* metaInfo = getMetaAddressWithOffset( object, METHOD_OFFSET );

    // meta method entries is 5th field of meta info in classdata
    MetaMethodEntry* entry =  (MetaMethodEntry*) *( (int*)*(metaInfo + 4) + metaIndex);  // -> method entry
    isFunc = entry->flags & 1 == 1;
    // last in list
    returnClassAddress = entry->argClassAddresses[entry->nParameters];
  }

  int* obj = 0;

  if( isFunc ){

    mString classObjectName( "Class" );

    obj = env->createWithDefaultConstructor( &classObjectName );

    int *originalObject = (int*) *(obj + 1);

    // store classdata to classdata field of created Class object
    *(originalObject + CLASS_OFFSET) = returnClassAddress;    
    

  }else{

    // no return type for procedures, create nil
    // obj is already 0 which is nil

  }

  
  // push Class object
  Utils::pushObject( (int*)frame, obj );
}


void Z__Method__getParameterTypes__0(functionFrame* &frame, mVMEnvironment* env)
{

  // create Vector of Class objects

  // 1st create Vector and find method 'add' which will be used for adding Class'es
  mString vecName( "Vector" );
  int* vec = env->createWithDefaultConstructor( &vecName );


  int* addMethod = env->getClassMethod("Vector", "add", "Object", vec); 



  // get Method object
  int* object = (int*)Utils::getSelf(frame);  

  // index of this method is after classdata in Method object
  int metaIndex = *(object + METHOD_OFFSET + 1);


  mString classObjectName( "Class" );

  // anonymous method?
  if( metaIndex == -1 ){

    MetaAnonymousEntry* anonmeta = (MetaAnonymousEntry*) *(object + METHOD_OFFSET + 3);

    int* obj = 0;

    for(int i=0; i<anonmeta->nParameters; i++){
    

      obj = env->createWithDefaultConstructor( &classObjectName );

      int *originalObject = (int*) *(obj + 1);

      // store classdata to classdata field of created Class object
      *(originalObject + CLASS_OFFSET) = anonmeta->argClassAddresses[i];    
    
      // store this object into vector by invoking 'Vector.add'
      env->invokeMethod( addMethod, obj ); 

    }

  }else{

    // normal method


    int* metaInfo = getMetaAddressWithOffset( object, METHOD_OFFSET );

    // meta method entries is 5th field of meta info in classdata
    MetaMethodEntry* entry =  (MetaMethodEntry*) *( (int*)*(metaInfo + 4) + metaIndex);  // -> method entry
  



    int* obj = 0;

    for(int i=0; i<entry->nParameters; i++){
    

      obj = env->createWithDefaultConstructor( &classObjectName );

      int *originalObject = (int*) *(obj + 1);

      // store classdata to classdata field of created Class object
      *(originalObject + CLASS_OFFSET) = entry->argClassAddresses[i];    
    
      // store this object into vector by invoking 'Vector.add'
      env->invokeMethod( addMethod, obj ); 

    }

  }
  
  
  // push Class object
  Utils::pushObject( (int*)frame, vec );
}


void Z__Method__getBlock__0(functionFrame* &frame, mVMEnvironment* env)
{

  // first get Method object
  int* object = (int*)Utils::getSelf(frame);  

  // index of this method is after classdata in Method object
  int metaIndex = *(object + METHOD_OFFSET + 1);

  int* metaInfo = getMetaAddressWithOffset( object, METHOD_OFFSET );

  // meta method entries is 5th field of meta info in classdata
  MetaMethodEntry* entry =  (MetaMethodEntry*) *( (int*)*(metaInfo + 4) + metaIndex);  // -> method entry


  // code address of this method is in method table, AMPAK SE POTREBUJEMO OFFSET V VTABELO
  // TO je treba se dodati v MetaMethodEntry (8, 16,...)
  // v ta namen dodaj ownerja v Method.z0 in offset 8,16,... v MetaMethodEntry, da bo kot
  // pri invokaciji

  cout<<"Index in virtual table ="<<entry->vtableIndex<<endl;

  cout<<"descriptor offset ="<<entry->descriptor<<endl;



  // outer block of this method doesn't exist on it own, so we have to create it

  mString block( "Block" );
  int* blockObj = env->createPrimitiveObject( &block );

  // set values for this block

  // te vrednosti dobimo iz methodEntry iz tabele metod in iz okolja metode, ki je
  // pravtako v istem methodEntryju

  // store address of closure to our object
  //*(object + BLOCK_OFFSET) = *(int*)(addressOfMain + 6);
  // store environment length
  //*(object + BLOCK_OFFSET+1) = nVariables;
  // store environment address
  //*(object + BLOCK_OFFSET+2) = (int)env;
  // set initial status to 0
  *(blockObj + BLOCK_OFFSET+3) = 0;
  // set bit 1 of flags because this is top level block
  *(blockObj + BLOCK_OFFSET+4) = 0x00000001;

  // push created block
  Utils::pushObject( (int*)frame, blockObj + 4 );
  
  
}

 
void Z__Method__invoke__2__Object__Vector(functionFrame* &frame, mVMEnvironment* env)
{
  
  // idi v metamethodentry za to metodo in najdi njen vtable indeks in deskriptor
  // potem jo lahko invociramo nad objektom

  // get Method object
  int* object = (int*)Utils::getSelf(frame);  

  // index of this method is after classdata in Method object
  int metaIndex = *(object + METHOD_OFFSET + 1);  

  int* metaInfo = getMetaAddressWithOffset( object, METHOD_OFFSET );

  // meta method entries is 5th field of meta info in classdata
  MetaMethodEntry* entry =  (MetaMethodEntry*) *( (int*)*(metaInfo + 4) + metaIndex);  // -> method entry  


  // first we need to check whether owner class of method is included in receiver class
  // if it's not then method cannot be invoked because the classes are not related
  // example: invoking method Myclass.mymethod upon Object - class of Myclass is not
  // present in classtable of Object

  // classdata of class owner is 1st field in this object 
  int* classData = (int*) *(object + METHOD_OFFSET + 0);

  // get receiver argument
  int* receiver = (int*)Utils::getArgument(frame, 0); 

  // !!!!!!! PREVERI, CE JE RECEIVER NIL


  // classdata of receiver object
  int* receiverClassData = (int*)*receiver; 


  int* ctableEntry = 0;  // entry in classtable where owner class was found

  // find correct descriptor within receiver class
  int descriptor = Utils::getDescriptor( receiver, classData, ctableEntry );
  if(descriptor == -1){
    
    cout<<"native: wrong receiver object!"<<endl;
    Utils::pushObject( (int*)frame, 0 );  // push nil
      
    return;


  }

  
  // now we need address of this method, which we'll get from class method table at index which is
  // given in virtual table
                             
  //                                   selfish (vtable start) from ctable + vtable index
  int mtableIndex =  *( receiverClassData  + ( *(ctableEntry + 1) >> 2) + entry->vtableIndex);
  
  
  // now we go into method table with this index
  // NOTE: this is methodEntry in classdata not object !
  //                    classdata    + mtab offset (2nd info value)
  int* methodEntry  =  ( receiverClassData   + (*(receiverClassData + 1) >> 2 ) + 3*mtableIndex);



  ///////////////////////////////////////////////////////////////////////////////////////////
  // TUKAJ preveri, ce ima metoda ze dolocenega prejemenika, ce ima, potem je ne moremo
  // izvesti nad posebej podanim prejemnikom, vrni napako, ker je potemtakem metoda
  // vzeta iz objekta, te pa izvrsujejo v posebni metodi

  ////////////////////////////////////////////////////////////////////////////////////////////


  //cout<<"method address="<<*methodEntry<<endl;
     
      

  // check if receiver object is an instance of class that declared this method
  // this means that class of receiver object must be one of the classes in classtable
  // of class which declares this method

  // dobi pravilen deskriptor za invokacijo nad pravim objektom
  // ce vzamemo metodo iz razreda Object in jo potem invociramo nad objektom MyObject
  // bo razred Object vsebovan v classtabeli MyObject-a. SelfishOffset (deskriptor) za metodo
  // v tabeli metod v razredu Object bo drugacen kot za razred Object v razredu MyObject
  // pri prvem bo 8, pri drugem pa 16, in tako dalje ...

  // classdata od ownerja metode mora biti v classtabeli od receiverja
  

  /*
  int descriptor = Utils::getDescriptor(receiver, classData);
  if(descriptor == -1){
    
    cout<<"native: wrong receiver object!"<<endl;
    Utils::pushObject( (int*)frame, 0 );  // push nil
      
    return;


  }
  */

  // check if arguments are not nil
  if( !Utils::isNil(frame, 1) ){

    // vector of parameters
    int* vec = Utils::getReference(frame, 1);    

    // invoke Vector.size to see how may parameters do we have in Vector    
    int* sizeMethod = env->getClassMethod("Vector", "size", "", vec);    
    
    env->invokeFunction( sizeMethod );     

    int* intSize =  (int*)Utils::top( (int*)frame );   // Integer object on top of frame
    

    // # of pushed arguments
    int nArguments = *( *(int**)(intSize+1) +INTEGER_OFFSET);                              


    // check if # of arguments matches those in metadata
    if( entry->nParameters != nArguments ){

      cout<<"native: # of arguments doesn't match!"<<endl;
      Utils::pushObject( (int*)frame, 0 );  // push nil
      
      return;

    }
    

    // Vector.at will give us every argument    
    int* atMethod = env->getClassMethod("Vector", "at", "Integer", vec);    


    int nPushed=0;
    // push arguments and perform casts when necessary
    // last to first!
    for(int i=nArguments-1; i >=0; i--){

      // create Integer for index for method at
      int* b = env->createInteger( i );

      env->invokeFunction( atMethod, b + 4 );

      int* argument =  (int*)Utils::top( (int*)frame );   // argument object on top of frame
      


      // check type of argument
      int* castedArgument = Utils::performCast( entry->argClassAddresses[i], argument );

      if( castedArgument == 0 ){

      cout<<"native: argument "<<i<<" doesn't match expected type!"<<endl;
      (int*)frame -= nPushed;  // restore stack
      Utils::pushObject( (int*)frame, 0 );  // push nil
      return;

    }


      // push argument object
      
      Utils::pushObject( (int*)frame, castedArgument  );
      nPushed++;
    }



  }else{

    // arguments is nil, if method receives arguments, we have mismatch
    if( entry->nParameters != 0 ){

      cout<<"native: # of arguments doesn't match!"<<endl;
      Utils::pushObject( (int*)frame, 0 );  // push nil
      
      return;

    }

  }
    

  // and finally invoke on receiver

  // check if procedure or function
  if(entry->flags & 1 == 1){   // bit 1 of flafs is set for functions
    
    env->invokeFunctionOnObject( methodEntry, (int*) ((int)receiver + descriptor) );

    // get returned value and push it

    int* returnedObject =  (int*)Utils::top( (int*)frame );   // on top of frame
    Utils::pushObject( (int*)frame, returnedObject );    

  }else{
    
    env->invokeMethodOnObject( methodEntry, (int*) ((int)receiver + descriptor) );
    // push nil
    Utils::pushObject( (int*)frame, 0 );
  }
  


}


void Z__Method__invoke__1__Vector(functionFrame* &frame, mVMEnvironment* env)
{
  
  // idi v metamethodentry za to metodo in najdi njen vtable indeks in deskriptor
  // potem jo lahko invociramo nad objektom

  // get Method object
  int* object = (int*)Utils::getSelf(frame);  

  // index of this method is after classdata in Method object
  int metaIndex = *(object + METHOD_OFFSET + 1);  


  if( metaIndex == -1 ){
    

    // we have anonymous method

    MetaAnonymousEntry* anonmeta = (MetaAnonymousEntry*) *(object + METHOD_OFFSET + 3);

    // get receiver - owner object of this method
    int* receiver = (int*)*(object + METHOD_OFFSET + 2); 


    // check if arguments are not nil
    if( !Utils::isNil(frame, 0) ){

      // vector of parameters
      int* vec = Utils::getReference(frame, 0);    

      // invoke Vector.size to see how may parameters do we have in Vector    
      int* sizeMethod = env->getClassMethod("Vector", "size", "", vec);    
      
      env->invokeFunction( sizeMethod );     
      
      int* intSize =  (int*)Utils::top( (int*)frame );   // Integer object on top of frame
      
      
      // # of pushed arguments
      int nArguments = *( *(int**)(intSize+1) +INTEGER_OFFSET);                              
      
      
      // check if # of arguments matches those in metadata
      if( anonmeta->nParameters != nArguments ){
	
	cout<<"native: # of arguments doesn't match!"<<endl;
	Utils::pushObject( (int*)frame, 0 );  // push nil
	
	return;
	
      }
          

      // Vector.at will give us every argument    
      int* atMethod = env->getClassMethod("Vector", "at", "Integer", vec);    
      
      
      int nPushed=0;
      // push arguments and perform casts when necessary
      // last to first!
      for(int i=nArguments-1; i >=0; i--){
	
	// create Integer for index for method at
	int* b = env->createInteger( i );
	
	env->invokeFunction( atMethod, b + 4 );
	
	int* argument =  (int*)Utils::top( (int*)frame );   // argument object on top of frame
	
	
	
	// check type of argument
	int* castedArgument = Utils::performCast( anonmeta->argClassAddresses[i], argument );
	
	if( castedArgument == 0 ){
	  
	  cout<<"native: argument "<<i<<" doesn't match expected type!"<<endl;
	  (int*)frame -= nPushed;  // restore stack
	  Utils::pushObject( (int*)frame, 0 );  // push nil
	  return;
	  
	}
	
	
	// push argument object
	
	Utils::pushObject( (int*)frame, castedArgument  );
	nPushed++;
      }
      
      
      
    }else{
      
      // arguments is nil, if method receives arguments, we have mismatch
      if( anonmeta->nParameters != 0 ){
	
	cout<<"native: # of arguments doesn't match!"<<endl;
	Utils::pushObject( (int*)frame, 0 );  // push nil
	
	return;
	
      }
      
    }

    


    // check if procedure or function
    if(anonmeta->flags & 1 == 1){   // bit 1 of flags is set for functions
    
      env->invokeAnonymousFunction(anonmeta, receiver /* + descriptor*/ );      

      // get returned value and push it
      int* returnedObject =  (int*)Utils::top( (int*)frame );   // on top of frame
      Utils::pushObject( (int*)frame, returnedObject );    

    }else{
    

      //env->invokeMethodOnObject( methodEntry, (int*) ((int)receiver + descriptor) );
      env->invokeAnonymousMethod(anonmeta, receiver /* + descriptor*/ );

      // push nil
      Utils::pushObject( (int*)frame, 0 );
    }



  }


  /*



  int* metaInfo = getMetaAddressWithOffset( object, METHOD_OFFSET );

  // meta method entries is 5th field of meta info in classdata
  MetaMethodEntry* entry =  (MetaMethodEntry*) *( (int*)*(metaInfo + 4) + metaIndex);  // -> method entry  


  // first we need to check whether owner class of method is included in receiver class
  // if it's not then method cannot be invoked because the classes are not related
  // example: invoking method Myclass.mymethod upon Object - class of Myclass is not
  // present in classtable of Object

  // classdata of class owner is 1st field in this object 
  int* classData = (int*) *(object + METHOD_OFFSET + 0);

  // get receiver argument
  int* receiver = (int*)Utils::getArgument(frame, 0); 

  // !!!!!!! PREVERI, CE JE RECEIVER NIL


  // classdata of receiver object
  int* receiverClassData = (int*)*receiver; 


  int* ctableEntry = 0;  // entry in classtable where owner class was found

  // find correct descriptor within receiver class
  int descriptor = Utils::getDescriptor( receiver, classData, ctableEntry );
  if(descriptor == -1){
    
    cout<<"native: wrong receiver object!"<<endl;
    Utils::pushObject( (int*)frame, 0 );  // push nil
      
    return;


  }

  
  // now we need address of this method, which we'll get from class method table at index which is
  // given in virtual table
                             
  //                                   selfish (vtable start) from ctable + vtable index
  int mtableIndex =  *( receiverClassData  + ( *(ctableEntry + 1) >> 2) + entry->vtableIndex);
  
  
  // now we go into method table with this index
  // NOTE: this is methodEntry in classdata not object !
  //                    classdata    + mtab offset (2nd info value)
  int* methodEntry  =  ( receiverClassData   + (*(receiverClassData + 1) >> 2 ) + 3*mtableIndex);



  ///////////////////////////////////////////////////////////////////////////////////////////
  // TUKAJ preveri, ce ima metoda ze dolocenega prejemenika, ce ima, potem je ne moremo
  // izvesti nad posebej podanim prejemnikom, vrni napako, ker je potemtakem metoda
  // vzeta iz objekta, te pa izvrsujejo v posebni metodi

  ////////////////////////////////////////////////////////////////////////////////////////////


  //cout<<"method address="<<*methodEntry<<endl;
     
      

  // check if receiver object is an instance of class that declared this method
  // this means that class of receiver object must be one of the classes in classtable
  // of class which declares this method

  // dobi pravilen deskriptor za invokacijo nad pravim objektom
  // ce vzamemo metodo iz razreda Object in jo potem invociramo nad objektom MyObject
  // bo razred Object vsebovan v classtabeli MyObject-a. SelfishOffset (deskriptor) za metodo
  // v tabeli metod v razredu Object bo drugacen kot za razred Object v razredu MyObject
  // pri prvem bo 8, pri drugem pa 16, in tako dalje ...

  // classdata od ownerja metode mora biti v classtabeli od receiverja
  
  

  // check if arguments are not nil
  if( !Utils::isNil(frame, 1) ){

    // vector of parameters
    int* vec = Utils::getReference(frame, 1);    

    // invoke Vector.size to see how may parameters do we have in Vector    
    int* sizeMethod = env->getClassMethod("Vector", "size", "", vec);    
    
    env->invokeFunction( sizeMethod );     

    int* intSize =  (int*)Utils::top( (int*)frame );   // Integer object on top of frame
    

    // # of pushed arguments
    int nArguments = *( *(int**)(intSize+1) +INTEGER_OFFSET);                              


    // check if # of arguments matches those in metadata
    if( entry->nParameters != nArguments ){

      cout<<"native: # of arguments doesn't match!"<<endl;
      Utils::pushObject( (int*)frame, 0 );  // push nil
      
      return;

    }
    

    // Vector.at will give us every argument    
    int* atMethod = env->getClassMethod("Vector", "at", "Integer", vec);    


    int nPushed=0;
    // push arguments and perform casts when necessary
    // last to first!
    for(int i=nArguments-1; i >=0; i--){

      // create Integer for index for method at
      int* b = env->createInteger( i );

      env->invokeFunction( atMethod, b + 4 );

      int* argument =  (int*)Utils::top( (int*)frame );   // argument object on top of frame
      


      // check type of argument
      int* castedArgument = Utils::performCast( entry->argClassAddresses[i], argument );

      if( castedArgument == 0 ){

      cout<<"native: argument "<<i<<" doesn't match expected type!"<<endl;
      (int*)frame -= nPushed;  // restore stack
      Utils::pushObject( (int*)frame, 0 );  // push nil
      return;

    }


      // push argument object
      
      Utils::pushObject( (int*)frame, castedArgument  );
      nPushed++;
    }



  }else{

    // arguments is nil, if method receives arguments, we have mismatch
    if( entry->nParameters != 0 ){

      cout<<"native: # of arguments doesn't match!"<<endl;
      Utils::pushObject( (int*)frame, 0 );  // push nil
      
      return;

    }

  }
    

  // and finally invoke on receiver

  // check if procedure or function
  if(entry->flags & 1 == 1){   // bit 1 of flafs is set for functions
    
    env->invokeFunctionOnObject( methodEntry, (int*) ((int)receiver + descriptor) );

    // get returned value and push it

    int* returnedObject =  (int*)Utils::top( (int*)frame );   // on top of frame
    Utils::pushObject( (int*)frame, returnedObject );    

  }else{
    
    env->invokeMethodOnObject( methodEntry, (int*) ((int)receiver + descriptor) );
    // push nil
    Utils::pushObject( (int*)frame, 0 );
  }
  
  */

}
