#ifndef __NATIVE_INTERFACE
#define __NATIVE_INTERFACE


#include "../mach/mVMEnvironment.h"
#include "../mach/ObjectOffsets.h"
#include "../base_classes/mString.h"

#include <iostream>

using namespace std;


// frame layout for procedures
struct procedureFrame{

  int oldStackPtr;    // old stack pointer
  int returnAddress;
  int self;           // this object
  // parameter 1
  // parameter 2
};


// frame layout for functions
struct functionFrame{

  int oldStackPtr;    // old stack pointer
  int resultLocal;    // local variable where result should be stored
  int returnAddress;
  int self;           // this object
  // parameter 1
  // parameter 2
};







// utility functions

class Utils{

 public:

  // returns argument from procedure frame
  // arguments start at 0
  // NOTE: returns 'whole' object from start not pointing to descriptor
  static int getArgument(procedureFrame *frame, int arg){ 
    return *( *((int**)frame - 4 - arg) + 1); 
  }

  // returns argument from function frame
  // arguments start at 0
  static int getArgument(functionFrame *frame, int arg){ 
    return  *( *((int**)frame - 5 - arg) + 1); 
  }


  // returns argument from procedure frame
  // arguments start at 0
  // NOTE: returns object pointing to descriptor
  static int* getReference(procedureFrame *frame, int arg){ 
    return  *((int**)frame - 4 - arg); 
  }

  // returns argument from function frame
  // arguments start at 0
  static int* getReference(functionFrame *frame, int arg){ 
    return  *((int**)frame - 5 - arg); 
  }



  // returns int value from Integer object in procedure frame at position arg
  // arg starts with 0
  static int getIntegerValue(procedureFrame *frame,int arg ){ 
    return *(   *((int**) *((int*)frame - 4-arg)+1) +INTEGER_OFFSET); 
  }

  // returns int value from Integer object in function frame at position arg
  // arg starts with 0
  static int getIntegerValue(functionFrame *frame,int arg ){ 
    return *(   *((int**) *((int*)frame - 5-arg)+1) +INTEGER_OFFSET); 
  }

  // returns double value from Double object in procedure frame at position arg
  // arg starts with 0
  static double getDoubleValue(procedureFrame *frame,int arg ){ 
    return *(double*)(   *((int**) *((int*)frame - 4-arg)+1) +DOUBLE_OFFSET); 
  }

  // returns double value from Double object in function frame at position arg
  // arg starts with 0
  static double getDoubleValue(functionFrame *frame,int arg ){ 
    return *(double*)(   *((int**) *((int*)frame - 5-arg)+1) +DOUBLE_OFFSET); 
  }

  // returns string value from String object in procedure frame at position arg
  // arg starts with 0
  static char* getStringValue(procedureFrame *frame,int arg ){ 
    return (char*)*(   *((int**) *((int*)frame - 4-arg)+1) +STRING_OFFSET + 1); 
  }

  // returns string value from String object in function frame at position arg
  // arg starts with 0
  static char* getStringValue(functionFrame *frame,int arg ){ 
    return (char*)*(   *((int**) *((int*)frame - 5-arg)+1) +STRING_OFFSET + 1); 
  }


  // returns int value from Boolean object in procedure frame at position arg
  // arg starts with 0
  static int getBooleanValue(procedureFrame *frame,int arg ){ 
    return *(   *((int**) *((int*)frame - 4-arg)+1) +BOOLEAN_OFFSET); 
  }

  // returns int value from Integer object in function frame at position arg
  // arg starts with 0
  static int getBooleanValue(functionFrame *frame,int arg ){ 
    return *(   *((int**) *((int*)frame - 5-arg)+1) +BOOLEAN_OFFSET); 
  }


  // returns self in procedure frame
  static int getSelf(procedureFrame *frame){ 
    return *( *((int**)frame - 3) +1); 
  }

  // returns self in function frame
  static int getSelf(functionFrame *frame){ 
    return *( *((int**)frame - 4) +1); 
  }
  

  /*static void setIntegerValue(procedureFrame *frame,int arg, int value ){ 
    *(   *((int**) *((int*)frame - 4-arg)+1) +INTEGER_OFFSET) = value; 
    }*/

  static void pushObject(int*& frame, void* object){ 

    *frame = (int)object; frame++;  
  }
  

  // returns object from top of frame
  static void* popObject(int*& frame){ 

    frame--;
    return (void*)*(frame+2);  
  }

  // returns object from top of frame and keeps the frame unaltered
  // used for obtaining return value from functions
  static void* top(int*& frame){ 

    return (void*)*(frame + 1);  
  }


  // returns true if argument arg is nil in procedure frame
  static bool isNil(procedureFrame *frame, int arg){ 
    return *((int**)frame - 4 - arg) == 0; 
  }


  // returns true if argument arg is nil in function frame
  static bool isNil(functionFrame *frame, int arg){ 
    return *((int**)frame - 5 - arg) == 0; 
  }


  // performs cast ob object to toClass (address of classdata)
  // returns casted argument if cast was successful, 0 otherwise
  static int* performCast(int toClass, int* object);



  // returns the descriptor of an object in classdata of another object 
  // give descriptor of classdata as it is in object object
  // returns -1 if classdata is not found in classtable of object
  // ctable entry holds entry where class was found
  static int getDescriptor(int* object, int* classData, int* &ctableEntry);




 private:



};


#endif 
